/* asdasd2
**	FILE:			LgiRes_Menu.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			6/3/2000
**	DESCRIPTION:	Menu Resource Editor
**
**
**	Copyright (C) 1999, Matthew Allen
**		fret@memecode.com
*/

////////////////////////////////////////////////////////////////////
#include <stdlib.h>
#include "LgiResEdit.h"
#include "LgiRes_Menu.h"

#define IDC_TREE				100

//////////////////////////////////////////////////////////////////
#define VAL_Separator			"Sep"
#define VAL_Enabled				"Enabled"

ResMenuItem::ResMenuItem(ResMenu *menu) :
	Str((menu)?menu->Group:0)
{
	Sep = false;
	Enabled = true;
	Menu = menu;
}

void ResMenuItem::OnMouseClick(GMouse &m)
{
	if (m.Down() AND m.Right())
	{
		GSubMenu *RClick = NEW(GSubMenu);
		if (RClick)
		{
			bool PasteData = false;
			bool PasteTranslations = false;

			{
				GClipBoard c(Tree);
				char *Clip = c.Text();
				if (Clip)
				{
					PasteTranslations = strstr(Clip, TranslationStrMagic);

					char *p = Clip;
					while (*p AND strchr(" \r\n\t", *p)) p++;
					PasteData = *p == '<';
					DeleteArray(Clip);
				}
			}

			RClick->AppendItem("Copy Text", IDM_COPY_TEXT, true);
			RClick->AppendItem("Paste Text", IDM_PASTE_TEXT, PasteTranslations);

			if (Tree->GetMouse(m, true))
			{
				switch (RClick->Float(Tree, m.x, m.y))
				{
					case IDM_COPY_TEXT:
					{
						Str.CopyText();
						break;
					}
					case IDM_PASTE_TEXT:
					{
						Str.PasteText();
						break;
					}
				}
			}

			DeleteObj(RClick);
		}
	}
}

char *ResMenuItem::GetText(int i)
{
	if (Sep)
	{
		return "--------------";
	}

	char *s = Str.Get();
	return (s) ? s : (char*)"<null>";
}

bool ResMenuItem::GetFields(DataDlgFieldList &Fields, ObjProperties *Props)
{
	Str.GetFields(Fields);
	Fields.Insert(NEW(DataDlgField(DATA_BOOL, 300, VAL_Separator, "Separator")));
	Fields.Insert(NEW(DataDlgField(DATA_BOOL, 301, VAL_Enabled, "Enabled")));

	Serialize(true);
	return true;
}

void ResMenuItem::OnFieldChange()
{
	Serialize(false);
	Update();
}

bool ResMenuItem::Serialize(bool Write)
{
	Str.Serialize(Write, &Values);
	
	if (Write)
	{
		if (Str.Ref == 0)
		{
			Str.Ref = Str.GetGroup()->UniqueRef();
		}		
		
		if (ValidStr(Str.Define) AND Str.Id == 0)
		{
			Str.Id = Str.GetGroup()->UniqueId(Str.Define);
		}
	}

	if (Write)
	{
		Values.Set(VAL_Separator, Sep);
		Values.Set(VAL_Enabled, Enabled);
	}
	else
	{
		int n = false;
		if (Values.Get(VAL_Separator, n)) Sep = n;
		if (Values.Get(VAL_Enabled, n)) Enabled = n;
	}

	return true;
}

bool ResMenuItem::Read(GXmlTag *t, ResMenuItem *Parent)
{
	bool Status = false;
	if (t)
	{
		bool SubMenu = stricmp(t->Tag, "submenu") == 0;
		bool MenuItem = stricmp(t->Tag, "menuitem") == 0;

		if (SubMenu OR MenuItem)
		{
			// Read item
			char *n = 0;
			Sep = (n = t->GetAttr("sep")) ? atoi(n) : false;
			Enabled = (n = t->GetAttr("enabled")) ? atoi(n) : true;

			if ((n = t->GetAttr("ref")) AND
				Menu AND
				Menu->Group)
			{
				int Ref = atoi(n);
				ResString *s = Menu->Group->FindRef(Ref);

				if (s)
				{
					// Copy the data from the string resource
					// allocated by the string group into
					// our object, allocated as part of this
					// item
					Str = *s;
					
					// Delete the string group allocated version
					// leaving the group containing just our
					// copy of the string
					DeleteObj(s);
				}
			}

			// Attach item
			Status = MenuItem;
			if (Parent)
			{
				Parent->Insert(this);
			}
			else if (Menu)
			{
				Menu->Insert(this);
			}

			// Read sub items
			if (SubMenu)
			{
				for (GXmlTag *c = t->Children.First(); c; c = t->Children.Next())
				{
					ResMenuItem *i = NEW(ResMenuItem(Menu));
					if (i AND i->Read(c, this))
					{
						Status = true;
					}
					else
					{
						LgiAssert(0);
						DeleteObj(i);
					}
				}
			}
		}
	}

	return Status;
}

bool ResMenuItem::Write(GXmlTag *t, int Tabs)
{
	bool SubMenu = GetChild() != 0;
	t->Tag = NewStr(SubMenu ? (char*) "submenu" : (char*) "menuitem");
	if (Sep)
	{
		t->SetAttr("Sep", 1);
	}
	else
	{
		if (NOT Enabled) t->SetAttr("Enabled", 0);
		t->SetAttr("Ref", Str.Ref);
	}

	if (SubMenu)
	{
		for (ResMenuItem *i = dynamic_cast<ResMenuItem*>(GetChild()); i; i = dynamic_cast<ResMenuItem*>(i->GetNext()))
		{
			GXmlTag *c = NEW(GXmlTag);
			if (c AND i->Write(c, 0))
			{
				t->InsertTag(c);
			}
			else
			{
				DeleteObj(c);
			}
		}
	}

	return true;
}

//////////////////////////////////////////////////////////////////
ResMenu::ResMenu(AppWnd *w, int type) :
	Resource(w, type),
	GTree(IDC_TREE, 0, 0, 200, 200, "Menu resource")
{
	Sunken(false);
	Group = NEW(ResStringGroup(w));
}

ResMenu::~ResMenu()
{
}

void AddChildren(GTreeItem *i, List<ResMenuItem> &Items)
{
	for (GTreeItem *c = i->GetChild(); c; c = c->GetNext())
	{
		ResMenuItem *m = dynamic_cast<ResMenuItem*>(c);
		if (m)
		{
			Items.Insert(m);
		}
		AddChildren(c, Items);
	}
}

void ResMenu::EnumItems(List<ResMenuItem> &Items)
{
	AddChildren(this, Items);
}

void ResMenu::Create()
{
	Name("IDM_MENU");
	if (Resource::Item) Resource::Item->Update();
	Visible(true);
}

GView *ResMenu::CreateUI()
{
	return Ui = NEW(ResMenuUi(this));
}

void ResMenu::OnItemClick(GTreeItem *Item, GMouse &m)
{
	GTree::OnItemClick(Item, m);
}

void ResMenu::OnItemBeginDrag(GTreeItem *Item, int Flags)
{
	GTree::OnItemBeginDrag(Item, Flags);
}

void ResMenu::OnItemExpand(GTreeItem *Item, bool Expand)
{
	GTree::OnItemExpand(Item, Expand);
}

void ResMenu::OnItemSelect(GTreeItem *Item)
{
	GTree::OnItemSelect(Item);

	if (AppWindow)
	{
		ResMenuItem *Item = dynamic_cast<ResMenuItem*>(Selection());
		AppWindow->OnObjSelect(Item);
	}
}

void ResMenu::OnRightClick(GSubMenu *RClick)
{
}

void ResMenu::OnCommand(int Cmd)
{
}

int ResMenu::OnCommand(int Cmd, int Event, OsView hWnd)
{
	switch (Cmd)
	{
		case IDM_NEW_SUB:
		{
			ResMenuItem *Item = dynamic_cast<ResMenuItem*>(Selection());
			GTreeItem *New = 0;
			if (Item)
			{
				New = Item->Insert(NEW(ResMenuItem(this)));
			}
			else
			{
				New = Insert(NEW(ResMenuItem(this)));
			}

			if (New)
			{
				Select(New);
			}
			break;
		}
		case IDM_NEW_ITEM:
		{
			ResMenuItem *Item = dynamic_cast<ResMenuItem*>(Selection());
			if (Item)
			{
				int n = Item->IndexOf();
				if (n >= 0) n++;

				GTreeItem *Parent = Item->GetParent();
				GTreeItem *New = 0;
				if (Parent)
				{
					New = Parent->Insert(NEW(ResMenuItem(this)), n);
				}
				else
				{
					New = Insert(NEW(ResMenuItem(this)), n);
				}

				if (New)
				{
					Select(New);
				}
			}
			break;
		}
		case IDM_DELETE_ITEM:
		{
			ResMenuItem *Item = dynamic_cast<ResMenuItem*>(Selection());
			if (Item)
			{
				Remove(Item);
				DeleteObj(Item);
			}
			break;
		}
		case IDM_SET_LANG:
		{
			if (Group)
			{
				Group->SetCurrent();
				Invalidate();
			}
			break;
		}
		case IDM_NEW_LANG:
		{
			if (Group)
			{
				Group->OnCommand(IDM_NEW_LANG, 0, 0);
				Invalidate();
			}
			break;
		}
		case IDM_DELETE_LANG:
		{
			if (Group)
			{
				Group->OnCommand(IDM_DELETE_LANG, 0, 0);
				Invalidate();
			}
			break;
		}
	}

	return 0;
}

ResString *ResMenu::GetStringByRef(int Ref)
{
	return Group ? Group->FindRef(Ref) : 0;
}

bool ResMenu::Test(ErrorCollection *e)
{
	return true;
}

bool ResMenu::Read(GXmlTag *t, ResFileFormat Format)
{
	bool Status = true;

	char *p = 0;
	if (p = t->GetAttr("Name"))
	{
		Name(p);
	}

	for (GXmlTag *c = t->Children.First(); c; c = t->Children.Next())
	{
		if (stricmp(c->Tag, "string-group") == 0)
		{
			if (Group)
			{
				if (NOT Group->Read(c, Format))
				{
					Status = false;
					LgiAssert(0);
				}
			}
			else
			{
				Status = false;
				LgiAssert(0);
			}
		}
		else if (stricmp(c->Tag, "submenu") == 0)
		{
			ResMenuItem *i = NEW(ResMenuItem(this));
			if (i AND i->Read(c))
			{
			}
			else
			{
				Status = false;
				LgiAssert(0);
				DeleteObj(i);
			}
		}
		else break;
	}

	return Status;
}

bool ResMenu::Write(GXmlTag *t, ResFileFormat Format)
{
	bool Status = true;

	// Write start tag
	t->Tag = NewStr("menu");
	if (Name()) t->SetAttr("Name", Name());

	// Write group
	if (Group)
	{
		GXmlTag *g = NEW(GXmlTag);
		if (g AND Group->Write(g, Format))
		{
			t->InsertTag(g);
		}
		else
		{
			Status = false;
			DeleteObj(g);
		}
	}

	// Write submenus/items
	for (GTreeItem *i = Items.First(); i; i = Items.Next())
	{
		ResMenuItem *m = dynamic_cast<ResMenuItem*>(i);
		if (m)
		{
			GXmlTag *c = NEW(GXmlTag);
			if (c AND m->Write(c, 1))
			{
				t->InsertTag(c);
			}
			else
			{
				Status = false;
				DeleteObj(c);
			}
		}
	}

	return Status;
}

//////////////////////////////////////////////////////////////////////////
ResMenuUi::ResMenuUi(ResMenu *Res)
{
	Menu = Res;
	Tools = 0;
	Status = 0;
	StatusInfo = 0;
}

ResMenuUi::~ResMenuUi()
{
	if (Menu)
	{
		Menu->Ui = 0;
	}
}

void ResMenuUi::OnPaint(GSurface *pDC)
{
	GRegion Client(0, 0, X()-1, Y()-1);
	for (GView *w = Children.First(); w; w = Children.Next())
	{
		GRect r = w->GetPos();
		Client.Subtract(&r);
	}

	pDC->Colour(LC_MED, 24);
	for (GRect *r = Client.First(); r; r = Client.Next())
	{
		pDC->Rectangle(r);
	}
}

void ResMenuUi::Pour()
{
	GRegion Client(GetClient());
	GRegion Update;

	for (GView *v = Children.First(); v; v = Children.Next())
	{
		GRect OldPos = v->GetPos();
		Update.Union(&OldPos);

		if (v->Pour(Client))
		{
			if (NOT v->Visible())
			{
				v->Visible(true);
			}

			if (OldPos != v->GetPos())
			{
				// position has changed update...
				v->Invalidate();
			}

			Client.Subtract(&v->GetPos());
			Update.Subtract(&v->GetPos());
		}
		else
		{
			// make the view not visible
			v->Visible(false);
		}
	}

	for (int i=0; i<Update.Length(); i++)
	{
		Invalidate(Update[i]);
	}
}

void ResMenuUi::OnPosChange()
{
	Pour();
}

void ResMenuUi::OnCreate()
{
	Tools = NEW(GToolBar);
	if (Tools)
	{
		char *FileName = LgiFindFile("_MenuIcons.gif");
		if (FileName AND Tools->SetBitmap(FileName, 16, 16))
		{
			Tools->Attach(this);

			Tools->AppendButton("New child menu", IDM_NEW_SUB);
			Tools->AppendButton("New menu item", IDM_NEW_ITEM);
			Tools->AppendButton("Delete menu item", IDM_DELETE_ITEM);
			Tools->AppendSeparator();

			Tools->AppendButton("Change language", IDM_SET_LANG, TBT_PUSH);
			Tools->AppendButton("New language", IDM_NEW_LANG, TBT_PUSH);
			Tools->AppendButton("Delete language", IDM_DELETE_LANG, TBT_PUSH);
		}
		else
		{
			DeleteObj(Tools);
		}
		DeleteArray(FileName);
	}

	Status = NEW(GStatusBar);
	if (Status)
	{
		Status->Attach(this);
		StatusInfo = Status->AppendPane("", 2);
	}

	ResFrame *Frame = NEW(ResFrame(Menu));
	if (Frame)
	{
		Frame->Attach(this);
	}
}

int ResMenuUi::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case M_COMMAND:
		{
			Menu->OnCommand(MsgA(Msg)&0xffff, MsgA(Msg)>>16, (OsView) MsgB(Msg));
			break;
		}
		case M_DESCRIBE:
		{
			char *Text = (char*) MsgB(Msg);
			if (Text)
			{
				StatusInfo->Name(Text);
			}
			break;
		}
	}

	return GView::OnEvent(Msg);
}

int ResMenuUi::CurrentTool()
{
	return false;
}

void ResMenuUi::SelectTool(int i)
{
}


