/// \file
/// \author Matthew Allen
#ifndef __DATE_TIME_H
#define __DATE_TIME_H

#define GDTF_DEFAULT				0
	
#define GDTF_DAY_MONTH_YEAR			0x01
#define GDTF_MONTH_DAY_YEAR			0x02
#define GDTF_YEAR_DAY_MONTH			0x04
#define GDTF_DATE_MASK				0x0f

#define GDTF_12HOUR					0x10
#define GDTF_24HOUR					0x20
#define GDTF_TIME_MASK				0xf0

/// A date/time class
class LgiClass GDateTime
{
	int16 _Day;			// 1-DaysInMonth
	int16 _Year;		// ####
	int16 _Thousands;	// 0-999

	int16 _Month;		// 1-12
	int16 _Seconds;		// 0-59
	int16 _Minutes;		// 0-59
	int16 _Hours;		// 0-23		(24hr)

	uchar _Format;
	static uchar DefaultFormat;

public:
	GDateTime();
	~GDateTime();

	// Date access

	/// Returns the day
	int Day() { return _Day; }
	/// Sets the day
	void Day(int d) { _Day = d; }
	/// Returns the month
	int Month() { return _Month; }
	/// Sets the month
	void Month(int m) { _Month = m; }
	/// Returns the year
	int Year() { return _Year; }
	/// Sets the year
	void Year(int y) { _Year = y; }

	// Time access
	int Thousands() { return _Thousands; }
	void Thousands(int t) { _Thousands = t; }
	int Seconds() { return _Seconds; }
	void Seconds(int s) { _Seconds = s; }
	int Minutes() { return _Minutes; }
	void Minutes(int m) { _Minutes = m; }
	int Hours() { return _Hours; }
	void Hours(int h) { _Hours = h; }

	// Format
	uchar GetFormat() { return _Format; }
	void SetFormat(uchar f) { _Format = f; }
	static uchar GetDefaultFormat();
	static void SetDefaultFormat(uchar f) { DefaultFormat = f; }

	// Calculated properties
	int DayOfWeek();

	// Misc

	/// Gets the date and time as a string
	void Get(char *Str);
	/// Gets the data and time as a 64 bit int (os specific)
	bool Get(int64 &s);
	/// Gets just the date as a string
	void GetDate(char *Str);
	/// Gets just the time as a string
	void GetTime(char *Str);

	/// Sets the date and time to the system clock
	void SetNow();
	/// Parses a date time from a string
	bool Set(char *Str);
	/// Sets the date and time from a 64 bit int (os specific)
	bool Set(int64 s);
	/// Parses the date from a string
	bool SetDate(char *Str);
	/// Parses the time from a string
	bool SetTime(char *Str);

	// Methods
	bool IsLeapYear(int Year = -1);
	bool IsSameDay(GDateTime &d);
	int DaysInMonth();
	void AddMinutes(int Minutes);
	void AddHours(int Hours);
	void AddDays(int Days);
	void AddMonths(int Months);
	int TimeZone();			// -60 * timezone
	int TimeZoneOffset();	// e.g. daylight savings
	
	// File
	int Sizeof();
	bool Serialize(class GFile &f, bool Write);
	bool Serialize(class ObjProperties *Props, char *Name, bool Write);

	// operators
	bool operator <(GDateTime &dt);
	bool operator <=(GDateTime &dt);
	bool operator >(GDateTime &dt);
	bool operator >=(GDateTime &dt);
	bool operator ==(GDateTime &dt);
	bool operator !=(GDateTime &dt);
	int Compare(GDateTime *d);

	GDateTime operator -(GDateTime &dt);
	GDateTime operator +(GDateTime &dt);
	GDateTime DiffMonths(GDateTime &dt);
};

struct GTimeZone
{
public:
	float Offset;
	char *Text;
};

extern GTimeZone GTimeZones[];

#endif
