/// \file
/// \author Matthew Allen (fret@memecode.com)
/// \brief Extension to the GWindow class to create a document processing application.
#ifndef __GDOCAPP_H
#define __GDOCAPP_H

#include <stdio.h>

#include "Lgi.h"
#include "GXml.h"
#include "GMru.h"

////////////////////////////////////////////////////////////////////////////////////////////
// Defines

// window messages
#define		IDM_SAVE			15002
#define		IDM_CLOSE			15003
#define		IDM_EXIT			15004

// Misc
#define		MainWnd				((AppWnd*)GApp::ObjInstance()->AppWnd)

/// A base for a document processing application.
class GDocApp :
	public GWindow,
	public GMru
{
private:
	class GDocAppPrivate *d;
	
	void			_Close();

protected:
	/// Loads the given menu.
	bool			_LoadMenu(char *Resource = 0);
	/// Call this to create the window, usually in the constructor of your main window class.
	bool			_Create();
	/// Call this to destroy the window, usually in the destuctor of your main window class.
	bool			_Destroy();
	/// Use the open file dialog to select a docuement to open.
	bool			_OpenFile(char *File);
	/// Save the current document.
	bool			_SaveFile(char *File);

	bool			_DoSerialize(bool Write);
	bool			_SerializeFile(GFile &f, bool Write);

	// Data
	GSubMenu		*_FileMenu;

public:
	/// Creates the class.
	GDocApp
	(
		/// The name of the application
		char *appname = 0,
		/// [Optional] The icon resource for the window.
		char *icon = 0
	);
	~GDocApp();

	/// Sets the current file name.
	void SetCurFile(char *f);
	/// Gets the current file name.
	char *GetCurFile();
	/// Sets the dirty state. If the doc is clean you are not asked to save it when you close the window.
	bool SetDirty(bool Dirty);
	/// Gets the dirty state.
	bool GetDirty();
	/// Gets the options list.
	ObjProperties *GetOptions();
	/// Gets the application name.
	char *GetAppName();

	/// Implement this to clear your docment from memory, i.e. "Close".
	virtual void Empty() {}
	/// Implement this to read/write your application specific options.
	virtual bool SerializeOptions(ObjProperties *Options, bool Write) { return false; }
	/// This is called when the dirty state changes.
	virtual void OnDirty(bool NewValue) {}

	// Impl
	void OnReceiveFiles(List<char> &Files);
	bool OnRequestClose(bool OsShuttingDown);
	int OnCommand(int Cmd, int Event, OsView Window);
	int OnEvent(GMessage *m);
};

#endif
