#ifndef __GSEMAPHORE_H
#define __GSEMAPHORE_H

// Thread Type and Functions
#if defined WIN32

	typedef DWORD						OsThreadId;
	typedef HANDLE						OsSemaphore;
	#define LgiGetCurrentThread()		GetCurrentThreadId()

#elif defined BEOS

	typedef thread_id					OsThreadId;
	typedef sem_id						OsSemaphore;
	#define LgiGetCurrentThread()		find_thread(0)

#elif defined LINUX

	typedef pthread_t					OsThreadId;
	typedef pthread_mutex_t				OsSemaphore;
	#define LgiGetCurrentThread()		pthread_self()

#else

	typedef int							OsThreadId;
	typedef int							OsSemaphore;
	#define LgiGetCurrentThread()		0

#endif

// Resource Locking Class
class LgiClass GSemaphore
{
	OsThreadId _Thread;
	OsSemaphore _Sem;
	bool _Lock();
	void _Unlock();
	char *_Name;

protected:
	int _Count;

public:
	GSemaphore(char *name = 0);
	virtual ~GSemaphore();

	bool Lock();
	bool LockWithTimeout(int Timeout /* in ms */);
	void Unlock();

	char *GetName();
	void SetName(char *s);
	
	#ifdef _DEBUG
	int GetCount() { return _Count; }
	#endif
};

#endif
