/// \file
/// \author Matthew Allen
/// \brief A tab view

#ifndef _GTABVIEW_H_
#define _GTABVIEW_H_

// Tab control
class GTabPage;
typedef List<GTabPage> TabPageList;

/// A tab control that displays multiple pages of information in a small area.
class LgiClass GTabView :
	public GView,
	public ResObject
{
	friend class GTabPage;
	class GTabViewPrivate *d;
	int TabY();

protected:
	List<GTabPage> Tabs;

public:
	GTabView(int id, int x, int y, int cx, int cy, char *name, int Init = 0);
	~GTabView();

	// Methods
	int Value();
	void Value(int i);

	// List modification
	bool Append(GTabPage *Page, int Where = -1);
	GTabPage *Append(char *name, int Where = -1);
	bool Delete(GTabPage *Page);

	// List access
	GTabPage *TabAt(int i) { return Tabs.ItemAt(i); }
	GTabPage *GetCurrent();
	int GetTabs() { return Tabs.GetItems(); }
	
	// Impl
	bool Attach(GView *parent);
	int OnEvent(GMessage *Msg);
	GView *FindControl(int Id);
	int OnNotify(GView *Ctrl, int Flags);
	void OnChildrenChanged(GView *Wnd, bool Attaching);
	GRect &GetTabClient();

	#if defined WIN32
	GView *FindControl(HWND hCtrl);
	#endif

	void OnPosChange();
	void OnPaint(GSurface *pDC);
	void OnMouseClick(GMouse &m);
	bool OnKey(GKey &k);
	void OnFocus(bool f);
};

class LgiClass GTabPage :
	public GView,
	public GLgiRes,
	public ResObject
{
	friend class GTabView;

	// Vars
	GTabView *TabCtrl;
	GRect TabPos;

	// Methods
	void PaintTab(GSurface *pDC, bool Selected);
	bool Attach(GView *parent);

public:
	GTabPage(char *name);
	~GTabPage();

	char *Name();
	bool Name(char *n);
	GTabView *GetTabControl() { return TabCtrl; }

	int OnEvent(GMessage *Msg);
	void OnPaint(GSurface *pDC);
	bool OnKey(GKey &k);
	void OnFocus(bool b);

	void Append(GView *Wnd);
	bool Remove(GView *Wnd);
	bool AppendDialogFromResource(int Resource);
};

#endif
