#ifndef __GTHREAD_H
#define __GTHREAD_H

//////////////////////////////////////////////////////////////////////////
// Thread types are defined in GSemaphore.h

#if defined BEOS

	#define PRIORITY_IDLE			B_LOW_PRIORITY
	#define PRIORITY_LOW			B_LOW_PRIORITY
	#define PRIORITY_NORMAL			B_NORMAL_PRIORITY
	#define PRIORITY_HIGH			B_DISPLAY_PRIORITY

#elif defined WIN32

	#define PRIORITY_IDLE			THREAD_PRIORITY_IDLE
	#define PRIORITY_LOW			THREAD_PRIORITY_BELOW_NORMAL

#else

	#define PRIORITY_IDLE			1
	#define PRIORITY_LOW			2

#endif

enum GThreadState
{
	LGITHREAD_INIT = 1,
	LGITHREAD_RUNNING = 2,
	LGITHREAD_ASLEEP = 3,
	LGITHREAD_EXITED = 4,
	LGITHREAD_ERROR = 5
};

class LgiClass GThread
{
	GThreadState State; // LGITHREAD_???
	int ReturnValue;
	OsThread hThread;

	#if defined WIN32
	friend uint WINAPI ThreadEntryPoint(int i);
	uint ThreadId;
	#elif defined LINUX
	friend void *ThreadEntryPoint(void *i);
	#else
	friend int32 ThreadEntryPoint(void *i);
	#endif

protected:
	bool DeleteOnExit;

public:
	GThread();
	virtual ~GThread();

	// Properties
	OsThread Handle() { return hThread; }
	GThreadState GetState() { return State; } // Volitle as best... only use for 'debug'
	virtual int ExitCode();
	virtual bool IsExited();

	// Methods
	virtual void Run();
	virtual void Terminate();
	virtual int Priority();
	virtual void Priority(int p);

	// Override to do something
	virtual int Main();

	// Events
	virtual void OnBeforeMain() {}
	virtual void OnAfterMain() {}
};

#endif
