
#ifndef __QApplication_h
#define __QApplication_h

#include "qwidget.h"
#include "GContainers.h"
#include "GSemaphore.h"

#define XA_WmDeleteWindow	XAtom[0]
#define XA_WmProtocols		XAtom[1]
#define XA_Clipboard		XAtom[2]
#define XA_XdndEnter		XAtom[3]
#define XA_XdndPosition		XAtom[4]
#define XA_XdndStatus		XAtom[5]
#define XA_XdndLeave		XAtom[6]
#define XA_XdndDrop			XAtom[7]
#define XA_XdndFinished		XAtom[8]
#define XA_XdndSelection	XAtom[9]
#define XA_NetWmIcon		XAtom[10]
#define XA_Cardinal			XAtom[11]
#define XA_Max				20		// leave room for growth

class QApplication : public QEventSink, public GSemaphore
{
	friend class QObject;
	friend class QPopup;
	friend class QWidgetPrivate;
	friend class QEvent;
	friend class QWidget;
	friend class QWindow;

	static QApplication *_App;
	static QWidget *WantsFocus;
	class QApplicationPrivate *d;

	Display *Dsp;
	QList<QWidget> Delete;

	// Atoms
	Atom XAtom[XA_Max];
	
	// Clipboard data..
	char *ClipText;
	Pixmap ClipImage;
	QEventSink *SelectionSink;
	
	// Dnd
	List<char> DndTypes, DndAccepted;
	class GDragDropTarget *DndTarget;
	
	// Methods	
	void OnDeleteWidget(QWidget *w);
	void OnEvent(XEvent *Event);
	QWidget *GetKeyTarget();

public:
	QApplication(int args, char **arg);
	virtual ~QApplication();

	// Api
	bool IsOk();
	QWidget *desktop();
	int exec();
	void onEvent(XEvent &Event);
	void processEvents();
	void exit(int code);
	void enter_loop();
	void exit_loop();
	void OnTabKey(QWidget *w, int Dir);
	uint64 GetLastButtonPressTime(int Button, int x, int y);
	OsThreadId GetGuiThread();

	// Focus
	QWidget *GetFocus();
	QWidget *GetFocusWindow();
	void SetFocus(QWidget *q);

	// Clipboard
	Atom GetClipboard() { return XA_Clipboard; }
	void EmptyClip();
	void SetClipText(QWidget *w, char *s);
	void SetClipImage(QWidget *w, class GSurface *pDC);
	void SetSelectionSink(QEventSink *s);
	bool GetSelection(Window w, Atom Selection, Atom Format, uchar *&Data, ulong &Len);
	
	// Static members
	static void postEvent(QWidget *o, QEvent *e);
	static int wheelScrollLines();
	static QApplication *GetApp() { return _App; }
};

#endif
