
#ifndef __QWidget_h
#define __QWidget_h

#include "qevent.h"

class QWidget : public QObject
{
	friend class QApplication;
	friend class QPainter;
	friend class ImagePainter;
	friend class QWidgetPrivate;
	friend class QPopup;
	friend class GApp;
	friend class GMemDC;

	static QList<QWidget> AllWidgets;
	static QWidget *MouseGrabber;
	static QWidget *OwnsClipboard;

	void _paint(XExposeEvent *e, QEvent *q);
	void _resize(XConfigureEvent *e, QEvent *q);
	void _map(bool m);
	bool _wait_mapped(int timeout = -1);
	QEvent *_event;
	virtual void *_GetWndPtr() { return 0; }

protected:
	class QWidgetPrivate *d;
	void setParentWidget(QWidget *p);

public:
	enum Background
	{
		NoBackground
	};

	enum MouseButton
	{
		LeftButton = 0x0001,
		RightButton = 0x0002,
		MidButton = 0x0004
	};

	enum KeyButton
	{
		AltButton = 0x0008,
		ShiftButton = 0x0010,
		ControlButton = 0x0020
	};
	
	QWidget(char *name = 0, bool top = false);
	QWidget(Window Existing, bool AddToList = false);
	~QWidget();

	Window handle();
    void detachHandle();
	class QMainWindow *GetWindow();
	
	// Internal
	void _dump(int i=0);
	bool _isdebug();
	void _SetClipRgn(GRegion *c); // int Rects, XRectangle *Rect);
	GRegion *_GetClipRgn(); // int &Rects, XRectangle *&Rect);
	virtual void _SetDeleteMe();
	virtual void _SetWndPtr(void *p) {}
	static QWidget *GetMouseGrabber() { return MouseGrabber; }

	// window Api
	void setIcon(char *Path);
	bool IsMouseOver(QEvent *e);
	bool IsPopup();
	void GetDecorationSize(int &x, int &y);

	// Widget Api
	virtual int height();
	virtual int width();
	virtual void setGeometry(int x, int y, int w, int h);
	virtual GRect &geometry();
	virtual void setBackgroundMode(Background b);
	virtual void repaint();
	virtual void repaint(int x, int y, int w, int h);
	virtual void update();
	virtual void update(int x, int y, int w, int h);
	virtual bool isEnabled();
	virtual void setEnabled(bool i);
	virtual bool isVisible();
	virtual void show(bool Raise = false);
	virtual void hide();
	virtual bool hasFocus();
	virtual void setFocus();
	virtual void clearFocus();
	virtual OsPoint &mapFromGlobal(OsPoint &p);
	virtual OsPoint &mapToGlobal(OsPoint &p);
	virtual QWidget *parentWidget();
	virtual QList<QWidget> *childrenWidgets();
	virtual bool reparent(QWidget *parent, OsPoint &p, bool show);
	virtual void grabMouse();
	virtual void ungrabMouse();
	bool grabbedMouse() { return MouseGrabber == this; }

	virtual char *getText();
	virtual void setText(char *text);
	virtual int value();
	virtual void setValue(int i);
	virtual int wantKeys();
	virtual void wantKeys(int keys);
	virtual bool isTabStop();
	virtual void isTabStop(bool i);

	// Static Api
	static QWidget *Find(Window w);

	// Events
	virtual void resizeEvent(QEvent *e) {}
	virtual void paintEvent(QEvent *e) {}
	virtual void customEvent(QEvent *e) {}
	virtual void mousePressEvent(QEvent *e) {}
	virtual void mouseDoubleClickEvent(QEvent *e) {}
	virtual void mouseReleaseEvent(QEvent *e) {}
	virtual void mouseMoveEvent(QEvent *e) {}
	virtual void leaveEvent(QEvent *e) {}
	virtual void enterEvent(QEvent *e) {}
	virtual void wheelEvent(QEvent *e) {}
	virtual void focusInEvent(QEvent *e) {}
	virtual void focusOutEvent(QEvent *e) {}
	virtual bool keyPressEvent(QEvent *e) {}
	virtual bool keyReleaseEvent(QEvent *e) {}
	virtual void closeEvent(QEvent *e) {}
	virtual void notifyEvent(int i = 0) {}
	virtual void propertyEvent(QEvent *e) {}

	// popup
	bool GetPopup();
	void SetPopup(bool b);
};

typedef QWidget						*OsView;

#endif
