

#ifndef __LgiLinux_h
#define __LgiLinux_h

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>

#include "GMem.h"
#include "GToken.h"

#undef Status
#undef Success
#undef None
#undef Above
#undef Below

#define XStatus		int
#define XSuccess	0
#define XAbove		0
#define XBelow		1
#define XNone		0L

extern "C" uint64 LgiCurrentTime();
extern bool _GetKdePaths(GToken &t, char *Type);
extern bool _GetIniField(char *Grp, char *Field, char *In, char *Out, int OutSize);
extern bool _GetKdeFont(char *FontType, char *Font, int FontBufSize, int &PointSize);

XChar2b *ConvertToX(char16 *s, int len = -1);
extern char *XErr(int i);
extern char *XMessage(int i);

class QObject
{
public:
	virtual ~QObject() {}

	static Display *XDisplay();
	static class QApplication *QApp();
};

class QInput : public QObject
{
	XIM Method;
	XIC Context;

public:
	QInput();
	~QInput();
};

class QPoint
{
	int _x, _y;

public:
	QPoint()
	{
		_x = _y = 0;
	}

	QPoint(int X, int Y)
	{
		_x = X;
		_y = Y;
	}

	int x() { return _x; }
	int y() { return _y; }
	void set(int x, int y) { _x = x; _y = y; }
};

class QCursor : public QObject
{
public:
	QPoint &pos();
};

class QRect
{
	int x1, y1, x2, y2;

public:
	QRect()
	{
		x1 = y1 = x2 = y2 = 0;
	}

	QRect(int l, int t, int r, int b)
	{
		x1 = l; y1 = t; x2 = r; y2 = b;
	}

	int set(int x, int y, int w, int h)
	{
		x1 = x; y1 = y; x2 = x1 + w - 1; y2 = y1 + h - 1;
	}

	int left() { return x1; }
	int top() { return y1; }
	int right() { return x2; }
	int bottom() { return y2; }
	int x() { return x2 - x1 + 1; }
	int y() { return y2 - y1 + 1; }
};

class QEventSink : public QObject
{
public:
	virtual void OnEvent(XEvent *Event) = 0;
};


#include "qlist.h"

#endif
