/// \file
/// \author Matthew Allen

#ifndef __OS_CLASS_H
#define __OS_CLASS_H

class LgiClass OsApplication
{
protected:
	HCURSOR hNormalCursor;

public:
	OsApplication()
	{
		hNormalCursor = 0;
	}
};

// Win32 Memory Handler
class LgiClass GMem {

	int64 Size;
	HGLOBAL hMem;
	void *pMem;

public:
	GMem(int64 size)
	{
		Size = size;
		hMem = GlobalAlloc(GMEM_FIXED, (DWORD)Size);
		pMem = 0;
	}

	GMem(HGLOBAL hmem)
	{
		hMem = hmem;
		Size = GlobalSize(hMem);
		pMem = 0;
	}

	~GMem()
	{
		if (pMem)
		{
			GlobalUnlock(hMem);
			pMem = 0;
		}
		if (hMem)
		{
			GlobalFree(hMem);
			hMem = 0;
		}
	}

	void *Lock()
	{
		return pMem = GlobalLock(hMem);
	}

	void UnLock()
	{
		if (pMem)
		{
			GlobalUnlock(hMem);
			pMem = 0;
		}
	}

	int64 GetSize()
	{
		return Size;
	}

	void Detach()
	{
		if (pMem)
		{
			GlobalUnlock(hMem);
			pMem = 0;
		}
		hMem = 0;
		Size = 0;
	}

	HGLOBAL Handle() { return hMem; }
};

// Win32 window class
class LgiClass GWin32Class : public GObject
{
	friend class GControl;
	friend class GApp;

	static LRESULT CALLBACK Redir(OsView hWnd, UINT m, WPARAM a, LPARAM b);
	static LRESULT CALLBACK SubClassRedir(OsView hWnd, UINT m, WPARAM a, LPARAM b);

	WNDPROC ParentProc;

public:
	union
	{
		WNDCLASSA a;
		WNDCLASSW w;
	} Class;

	GWin32Class(char *Name);
	~GWin32Class();

	bool Register();
	bool SubClass(char *Parent);
	LRESULT CALLBACK CallParent(OsView hWnd, UINT m, WPARAM a, LPARAM b);

	static GWin32Class *Create(char *ClassName);
};

/// Registry access class
#include "GContainers.h"
class LgiClass GRegKey
{
	HKEY k, Root;
	char s[256];
	char *KeyName;

public:
	/// Constructor
	GRegKey
	(
		/// The key name: you can use printf style syntax and extra arguments
		char *Key,
		...
	);
	~GRegKey();

	/// Returns true if the key was openned
	bool IsOk();
	/// Creates the key if not present
	bool Create();
	/// Returns the key name
	char *Name();

	/// Return a string value
	char *GetStr
	(
		/// Name of the subkey or NULL for the default string.
		char *Name = 0
	);
	/// Sets a string value
	bool SetStr(char *Name, char *Value);
	/// Get an int value
	int GetInt(char *Name = 0);
	/// Set an int value
	bool SetInt(char *Name, int Value);
	/// Get a binary value
	bool GetBinary(char *Name, void *&Ptr, int &Len);
	/// Set a binary value
	bool SetBinary(char *Name, void *Ptr, int Len);
	/// Delete a value
	bool DeleteValue(char *Name = 0);
	/// Delete a key
	bool DeleteKey();

	/// List all the key names under this key
	bool GetKeyNames(List<char> &n);
	/// List all the key value name under this key
	bool GetValueNames(List<char> &n);
};

#endif
