/*hdr
**      FILE:           GDb-Dbf.h
**      AUTHOR:         Matthew Allen
**      DATE:           8/2/2000
**      DESCRIPTION:    Csv implementation of the GDb API
**
**      Copyright (C) 2003 Matthew Allen
**			fret@memecode.com
*/

#include <stdio.h>

#include "Lgi.h"
#include "GDb.h"
#include "GToken.h"

///////////////////////////////////////////////////////////////////
// Forward decl
class CsvField;
class CsvRecord;

class CsvRecordset : public GDbRecordset
{
	friend class CsvField;

	char *FileName;
	List<CsvField> F;
	List<CsvRecord> R;
	Iterator<CsvRecord> *Cur;
	CsvRecord *Temp;
	CsvRecord *New;
	bool Dirty;

	void Empty();
	void Read();
	void Write();
	CsvRecord *Record();

public:
	CsvRecordset(char *file);
	~CsvRecordset();

	char *Name();
	GDbField *operator [](int Index);
	GDbField *operator [](char *Name);
	GDbField *InsertField(char *Name, int Type, int Len, int Index);
	bool DeleteField(GDbField *Fld);
	int Fields();
	bool Bof();
	bool Eof();
	bool AddNew();
	bool Edit();
	bool Update();
	void Cancel();
	bool MoveFirst();
	bool MoveNext();
	bool MovePrev();
	bool MoveLast();
	int SeekRecord(int i);
	int RecordIndex();
	int Records();
	bool DeleteRecord();
};


class CsvRecord
{
	int RawLen;
	char *Raw;

public:
	CsvRecordset *Rs;
	int Fields;
	char **Data;

	void EmptyField(int i)
	{
		if (Data AND
			i < Fields)
		{
			char *End = Raw + RawLen;
			if (NOT Raw OR (Data[i] < Raw OR Data[i] > End))
			{
				DeleteArray(Data[i]);
			}
		}
	}

	void Empty()
	{
		if (Data)
		{
			char *End = Raw + RawLen;
			for (int i=0; i<Fields; i++)
			{
				if (NOT Raw OR (Data[i] < Raw OR Data[i] > End))
				{
					DeleteArray(Data[i]);
				}
			}
		}

		DeleteArray(Raw);
	}

	bool SetFields(int i)
	{
		char **b = NEW(char*[i]);
		if (b)
		{
			memset(b, 0, sizeof(char*)*i);
			if (Data AND Fields > 0)
			{
				int m = min(Fields, i);
				memcpy(b, Data, sizeof(char*)*m);
			}
			DeleteArray(Data);
			Data = b;
			Fields = i;

			return true;
		}

		return false;
	}

	CsvRecord(CsvRecordset *rs)
	{
		Rs = rs;
		RawLen = 0;
		Raw = 0;
		Fields = 0;
		Data = 0;
		SetFields(Rs->Fields());
	}

	CsvRecord(CsvRecordset *rs, CsvRecord *r)
	{
		Rs = rs;
		RawLen = 0;
		Raw = 0;
		Fields = 0;
		Data = 0;
		if (r) *this = *r;
	}

	CsvRecord(CsvRecordset *rs, char *&In)
	{
		Rs = rs;
		RawLen = 0;
		Raw = 0;
		Fields = 0;
		Data = 0;

		int Flds = 1;
		char *Start = In;
		for (RawLen=0; In[RawLen]; )
		{
			if (In[RawLen] == ',')
			{
				Flds++;
				RawLen++;
			}
			else if (In[RawLen] == '\"')
			{
				RawLen++;
				while (In[RawLen] AND In[RawLen] != '\"') RawLen++;
				RawLen++;
			}
			else if (In[RawLen] == '\n' OR In[RawLen] == '\r')
			{
				break;
			}
			else
			{
				RawLen++;
			}
		}

		LgiAssert(RawLen);

		In += RawLen;
		while (*In == '\r' OR *In == '\n') In++;
		Raw = NewStr(Start, RawLen);

		if (Raw AND SetFields(Flds))
		{
			char *r = Raw;
			for (int n=0; n<Fields AND r; n++)
			{
				char *e = r;
				if (*r == '\"')
				{
					e = ++r;
					while (*e AND *e != '\"') e++;
					if (*e) *e++ = 0;
				}
				else
				{
					while (*e AND *e != '\n' AND *e != ',') e++;
				}
				Data[n] = r;
				if (*e)
				{
					r = e + 1;
					*e = 0;
				}
				else
				{
					break;
				}
			}
		}
		else In = 0;
	}

	~CsvRecord()
	{
		DeleteArray(Data);
	}

	CsvRecord &operator =(CsvRecord &r)
	{
		Empty();
		SetFields(r.Fields);
		for (int i=0; i<Fields; i++)
		{
			Data[i] = NewStr(r.Data[i]);
		}

		return *this;
	}
};

class CsvField : public GDbField
{
	int Index;
	char *FldName;
	CsvRecordset *Rs;

public:
	CsvField(CsvRecordset *rs, int i, char *n)
	{
		Index = i;
		Rs = rs;
		FldName = NewStr(n);
	}

	~CsvField()
	{
		DeleteArray(FldName);
	}

	char *Name()
	{
		return FldName;
	}

	bool Name(char *Str)
	{
		DeleteArray(FldName);
		return (FldName = NewStr(Str)) != 0;
	}

	int Type()
	{
		return GV_STRING;
	}

	bool Type(int NewType)
	{
		return false;
	}

	int Length()
	{
		return 255;
	}

	bool Length(int NewLength)
	{
		return false;
	}

	char *Description()
	{
		return 0;
	}

	bool Description(char *NewDesc)
	{
		return false;
	}

	bool Set(GVariant &v)
	{
		CsvRecord *r = Rs->Record();
		if (r AND v.Str())
		{
			if (Index < r->Fields OR
				r->SetFields(Index + 1))
			{
				r->EmptyField(Index);
				r->Data[Index] = NewStr(v.Str());
				return true;
			}
		}

		return false;
	}

	bool Get(GVariant &v)
	{
		CsvRecord *r = Rs->Record();
		if (r)
		{
			if (Index < r->Fields)
			{
				v = r->Data[Index];
				return true;
			}
		}
		
		return false;
	}
};

/////////////////////////////////////////////////////////////////////
CsvRecordset::CsvRecordset(char *file)
{
	Cur = 0;
	Temp = 0;
	New = 0;
	Dirty = false;
	FileName = NewStr(file);
	Read();
}

CsvRecordset::~CsvRecordset()
{
	if (Dirty)
	{
		Write();
	}
	DeleteArray(FileName);
}

char *CsvRecordset::Name()
{
	char *d = strrchr(FileName, DIR_CHAR);
	return d ? d + 1 : 0;
}

void CsvRecordset::Empty()
{
	F.DeleteObjects();
	R.DeleteObjects();
}

void CsvRecordset::Read()
{
	Empty();

	char *Txt = ReadTextFile(FileName);
	if (Txt)
	{
		char *t = strchr(Txt, '\n');
		if (t)
		{
			*t++ = 0;

			// Headers...
			GToken Flds(Txt, ",");
			for (int n=0; n<Flds.Length(); n++)
			{
				char *Name = TrimStr(Flds[n], " \r\t\n\"");
				if (Name)
				{
					F.Insert(NEW(CsvField(this, n, Name)));
				}
			}

			while (*t)
			{
				R.Insert(NEW(CsvRecord(this, t)));
				if (NOT t)
				{
					break;
				}
				else if (*t == '\n')
				{
					t++;
				}
			}
		}
	}
}

void CsvRecordset::Write()
{
	GFile f;
	if (FileName AND f.Open(FileName, O_WRITE))
	{
		char s[256];
		int NewLine = strlen(EOL_SEQUENCE);
		f.SetSize(0);

		// Headers
		for (CsvField *Fld = F.First(); Fld; )
		{
			CsvField *Nxt = F.Next();
			char *Name = Fld->Name();
			f.Write(s, sprintf(s, "\"%s\"", Name));
			if (Nxt) f.Write((void*)",", 1);
			Fld = Nxt;
		}
		f.Write((void*)EOL_SEQUENCE, NewLine);
	
		// Records
		for (bool b = MoveFirst(); b; b = MoveNext())
		{
			GVariant v;
			for (int i=0; i<F.GetItems(); i++)
			{
				if (i)
				{
					f.Write((void*)",", 1);
				}

				if ((*this)[i]->Get(v))
				{
					switch (v.Type)
					{
						case GV_STRING:
						{
							f.Write(s, sprintf(s, "\"%s\"", v.Str()));
							break;
						}
					}
				}
			}
			f.Write((void*)EOL_SEQUENCE, NewLine);
		}
	}
}

CsvRecord *CsvRecordset::Record()
{
	if (Temp) return Temp;
	if (New) return New;
	if (Cur) return Cur->Current();
	return 0;
}

GDbField *CsvRecordset::operator [](int Index)
{
	return F[Index];
}

GDbField *CsvRecordset::operator [](char *Name)
{
	if (Name)
	{
		for (GDbField *f=F.First(); f; f=F.Next())
		{
			if (f->Name() AND stricmp(Name, f->Name()) == 0)
			{
				return f;
			}
		}
	}
	return 0;
}

GDbField *CsvRecordset::InsertField(char *Name, int Type, int Length, int Index)
{
	if (Index < 0)
	{
		Index = F.GetItems();
	}

	CsvField *f = NEW(CsvField(this, Index, Name));
	if (f)
	{
		F.Insert(f, Index);
	}
	return f;
}

bool CsvRecordset::DeleteField(GDbField *Fld)
{
	return 0;
}

int CsvRecordset::Fields()
{
	return F.GetItems();
}

bool CsvRecordset::Bof()
{
	return NOT Cur OR NOT Cur->Current();
}

bool CsvRecordset::Eof()
{
	return NOT Cur OR NOT Cur->Current();
}

bool CsvRecordset::AddNew()
{
	Update();
	New = NEW(CsvRecord(this));
	return New != 0;
}

bool CsvRecordset::Edit()
{
	Update();
	if (Cur AND Cur->Current())
	{
		Temp = NEW(CsvRecord(this, Cur->Current()));
	}
	return Temp != 0;
}

bool CsvRecordset::Update()
{
	if (Temp)
	{
		CsvRecord *c = Cur ? Cur->Current() : 0;
		if (c)
		{
			*c = *Temp;
			DeleteObj(Temp);
			Dirty = true;
			return true;
		}
	}
	else if (New)
	{
		R.Insert(New);
		New = 0;
		Dirty = true;
		return true;
	}

	return false;
}

void CsvRecordset::Cancel()
{
	DeleteObj(Temp);
	DeleteObj(New);
}

bool CsvRecordset::MoveFirst()
{
	Cancel();

	if (NOT Cur)
	{
		Cur = NEW(Iterator<CsvRecord>(&R));
	}

	if (Cur)
	{
		return Cur->First();
	}

	return 0;
}

bool CsvRecordset::MoveNext()
{
	Cancel();

	return Cur ? Cur->Next() : 0;
}

bool CsvRecordset::MovePrev()
{
	Cancel();

	return Cur ? Cur->Prev() : 0;
}

bool CsvRecordset::MoveLast()
{
	Cancel();

	if (NOT Cur)
	{
		Cur = NEW(Iterator<CsvRecord>(&R));
	}

	if (Cur)
	{
		return Cur->Last();
	}

	return 0;
}

int CsvRecordset::SeekRecord(int i)
{
	Cancel();

	if (NOT Cur)
	{
		Cur = NEW(Iterator<CsvRecord>(&R));
	}

	if (Cur AND (*Cur)[i])
	{
		return true;
	}
	
	return false;
}

int CsvRecordset::RecordIndex()
{
	return Cur ? R.IndexOf(Cur->Current()) : -1;
}

int CsvRecordset::Records()
{
	return R.GetItems();
}

bool CsvRecordset::DeleteRecord()
{
	return 0;
}

//////////////////////////////////////////////////////////////////////////
class CsvDb : public GDb
{
	List<CsvRecordset> Tables;

public:
	~CsvDb()
	{
		Disconnect();
	}

	bool Connect(char *Init)
	{
		Tables.Insert(NEW(CsvRecordset(Init)));
		return Tables.First() != 0;
	}

	bool Disconnect()
	{
		Tables.DeleteObjects();
		return true;
	}

	GDbRecordset *Open(char *Name)
	{
		return Tables.First();
	}

	// Design
	GDbRecordset *TableAt(int i)
	{
		return Tables[i];
	}
};

///////////////////////////////////////////////////////////////////
GDb *OpenCsvDatabase(char *Path)
{
	CsvDb *Db = NEW(CsvDb);
	if (Db AND Db->Connect(Path))
	{
		return Db;
	}
	DeleteObj(Db);
	return 0;
}

