#ifndef _OS_H_
#define _OS_H_

#pragma warning(disable:4355)

// Includes
#include <stdio.h>
#include <string.h>
#include <assert.h>

// Defines
#define __INTEL__				1
#define _EXPORT
#define USER

#define _S_IFMT					0170000         /* file type mask */
#define _S_IFDIR				0040000         /* directory */
#define _S_IFCHR				0020000         /* character special */
#define _S_IFIFO				0010000         /* pipe */
#define _S_IFREG				0100000         /* regular */
#define _S_IREAD				0000400         /* read permission, owner */
#define _S_IWRITE				0000200         /* write permission, owner */
#define _S_IEXEC				0000100         /* execute/search permission, owner */

#define S_IFMT					_S_IFMT
#define S_IFDIR					_S_IFDIR
#define S_IFCHR					_S_IFCHR
#define S_IFIFO					_S_IFIFO
#define S_IFREG					_S_IFREG
#define S_IREAD					_S_IREAD
#define S_IWRITE				_S_IWRITE
#define S_IEXEC					_S_IEXEC

#define S_IFLNK					0000040
#define S_IXUSR					0000020
#define S_IRWXO					0000010
#define S_IUMSK					0000004

#define S_ISLNK(i)				(((i)&S_IFLNK)!=0)
#define S_ISREG(i)				(((i)&S_IFREG)!=0)
	
#define S_STR_INDEX				0x001
#define S_INT_INDEX				0x002
#define S_UINT_INDEX			0x004
#define S_LONG_LONG_INDEX		0x008
#define S_ULONG_LONG_INDEX		0x010
#define S_FLOAT_INDEX			0x020
#define S_DOUBLE_INDEX			0x040
#define S_INDEX_DIR				0x100
#define S_ATTR_DIR				0x200
#define S_ATTR					0x400
#define S_ALLOW_DUPS			0x800

#define B_OK					0
#define B_NO_ERROR				0
#define B_ERROR					-1
#define B_NO_INIT				-2
#define B_BAD_VALUE				-3
#define B_BAD_TYPE				-4
#define B_IO_ERROR				-5
#define B_NO_MEMORY				-6
#define B_DEVICE_FULL			-7
#define B_ENTRY_NOT_FOUND		-8
#define B_NAME_IN_USE			-9
#define B_INTERRUPTED			-10
#define B_BAD_INDEX				-11
#define B_BUSY					-12
#define B_READ_ONLY_DEVICE		-13
#define B_NOT_ALLOWED			-14
#define B_FILE_ERROR			-15
#define B_NOT_A_DIRECTORY		-16
#define B_IS_A_DIRECTORY		-17
#define B_DIRECTORY_NOT_EMPTY	-18
#define B_FILE_EXISTS			-19
#define B_QUERY_NON_INDEXED		-20
#define B_TIMEOUT				-21

#define B_STRING_TYPE			1
#define B_INT32_TYPE			2
#define B_UINT32_TYPE			3
#define B_INT64_TYPE			4
#define B_UINT64_TYPE			5
#define B_FLOAT_TYPE			6
#define B_DOUBLE_TYPE			7

#define B_DO_NOT_RESCHEDULE		1
#define B_FLUSH_DRIVE_CACHE		2
#define EFBIG					3 // FIXME?
#define B_LOW_PRIORITY			4
#define B_QUERY_UPDATE			6

#define B_MOUNT_READ_ONLY		0x01
#define B_LIVE_QUERY			0x02

#define _O_RDONLY				0x0000
#define _O_WRONLY				0x0001
#define _O_RDWR					0x0002
#define _O_APPEND				0x0008
#define _O_CREAT				0x0100
#define _O_TRUNC				0x0200
#define _O_EXCL					0x0400

#define O_RDONLY				_O_RDONLY
#define O_WRONLY				_O_WRONLY
#define O_RDWR					_O_RDWR
#define O_APPEND				_O_APPEND
#define O_CREAT					_O_CREAT
#define O_TRUNC					_O_TRUNC
#define O_EXCL					_O_EXCL
#define O_TEXT					_O_TEXT
#define O_BINARY				_O_BINARY
#define O_RAW					_O_BINARY
#define O_TEMPORARY				_O_TEMPORARY
#define O_NOINHERIT				_O_NOINHERIT
#define O_SEQUENTIAL			_O_SEQUENTIAL
#define O_RANDOM				_O_RANDOM
#define O_RWMASK				(O_RDONLY|O_WRONLY)

#define R_OK					0x01
#define W_OK					0x02
#define X_OK					0x04

#define B_FILE_NAME_LENGTH		300

#define min_c(a, b)				((a)<(b) ? (a) : (b))
#define dprintf					printf
#define debugger
#define kernel_debugger

// Typedefs
typedef unsigned char			uint8;
typedef char					int8;
typedef unsigned short			uint16;
typedef short					int16;
typedef unsigned int			uint32;
typedef int						int32;
typedef unsigned __int64		uint64;
typedef __int64					int64;

typedef unsigned char			uchar;
typedef unsigned short			ushort;
typedef unsigned int			uint;
typedef unsigned long			ulong;

typedef int						vint32;

typedef int64					off_t;
typedef int64					ino_t;
typedef uint64					bigtime_t;
typedef uint					time_t;
typedef int						status_t;
typedef int						mode_t;
typedef int						iovec;
typedef int						selectsync;
typedef int						port_id;
typedef int						type_code;
typedef ulong					dev_t;
typedef int						uid_t;
typedef int						gid_t;

// Semaphores
#define B_SYSTEM_TEAM			1
typedef int						sem_id;
typedef int						team_id;

extern sem_id create_sem(long thread_count, const char *name);
extern status_t delete_sem(sem_id sem);
extern status_t acquire_sem(sem_id sem);
extern status_t acquire_sem_etc(sem_id sem, long count, long flags, double timeout);
extern status_t release_sem(sem_id sem);
extern status_t release_sem_etc(sem_id sem, long count, long flags);
extern status_t set_sem_owner(sem_id sem, team_id team);

// Threads
typedef int						thread_id;
typedef int						(*thread_func)(void);
typedef int						(*thread_entry)(void);

#define spawn_kernel_thread		spawn_thread

extern thread_id find_thread(const char *name);
extern status_t resume_thread(thread_id thread);
extern thread_id spawn_thread(thread_entry func, const char *name, long priority, void *data);
extern long kill_thread(thread_id thread);

// Time
extern status_t snooze(bigtime_t microseconds);
extern bigtime_t time(void *p);
extern bigtime_t system_time();

// File/Device
#define B_FS_IS_PERSISTENT		0x01
#define B_FS_HAS_ATTR			0x02
#define B_FS_HAS_MIME			0x04
#define B_FS_HAS_QUERY			0x08
#define B_FS_IS_READONLY		0x10

struct fs_info
{
	int flags;
	int io_size;
	int block_size;
	int total_blocks;
	int free_blocks;
	char volume_name[64];
	char fsh_name[64];
};

struct stat
{
	int st_dev;
	int st_ino;
	int st_nlink;
	int st_blksize;

	int st_uid;
	int st_gid;
	int st_mode;
	size_t st_size;

	bigtime_t st_atime;
	bigtime_t st_mtime;
	bigtime_t st_crtime;
	bigtime_t st_ctime;
};

struct dirent
{
	int d_dev;
	int64 d_ino;
	int d_reclen;
	int d_pdev;
	int d_pino;
	char d_name[256];
};

struct attr_info
{
	int type;
	int size;
};

struct index_info
{
	int type;
	int size;
	time_t modification_time;
	time_t creation_time;
	uid_t uid;
	gid_t gid;
};

extern int fstat(int fd, struct stat *s);
extern int open(const char *name, int flags);
extern void close(int fd);
extern size_t read_pos(int fd, off_t offset, uchar *buf, size_t size);
extern size_t write_pos(int fd, off_t offset, const uchar *buf, size_t size);
extern int ioctl(int fildes, int request, ...);

// User
extern uid_t geteuid();
extern gid_t getegid();

// Atomic ops
extern int32 atomic_add(vint32 *atomicVariable, int32 addValue);
extern int32 atomic_and(vint32 *atomicVariable, int32 andValue);
extern int32 atomic_or(vint32 *atomicVariable, int32 orValue);

// Debug
extern char *__PRETTY_FUNCTION__();

// Number
extern int64 strtoll(char *s, char **end, int base);
extern uint64 strtoull(char *s, char **end, int base);



#endif
