/*hdr
**	FILE:			GFont_W.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			19/6/2002
**	DESCRIPTION:	Unicode Font Support
**
**	Copyright (C) 2002, Matthew Allen
**		fret@memecode.com
*/

//////////////////////////////////////////////////////////////////////
// Includes
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "Lgi.h"
#include "GFontSelect.h"
#include "GdiLeak.h"

#ifdef FontChange
#undef FontChange
#endif

/////////////////////////////////////////////////////////////////////////////
void GFont::_Measure(int &x, int &y, OsChar *Str, int Len)
{
	#if defined WIN32

	LgiAssert(Handle());

	HDC hDC = GetParam() ? (HDC)GetParam() : CreateCompatibleDC(0);
	HFONT hOldFont = (HFONT) SelectObject(hDC, Handle());

	SIZE Size;
	if (GetTextExtentPoint32W(hDC, Str, Len, &Size))
	{
		x = Size.cx;
		y = Size.cy;
	}
	else
	{
		x = y = 0;
	}

	SelectObject(hDC, hOldFont);
	if (NOT GetParam())
		DeleteDC(hDC);

	#elif defined XWIN

	QFontMetrics m(Handle());
	x = m.width(Str, Len);
	y = m.height();

	#elif defined BEOS
	
	if (Handle())
	{
		font_height h;
		Handle()->GetHeight(&h);

		x = Handle()->StringWidth(Str, Len);
		y = ceil(h.ascent + h.descent);
	}
	
	#else
	
	#error "Implement me."

	#endif
}

int GFont::_CharAt(int x, OsChar *Str, int Len)
{
	#if defined WIN32

	LgiAssert(Handle());

	INT Fit = 0;
	HDC hDC = CreateCompatibleDC((HDC)GetParam());
	HFONT hOldFont = (HFONT) SelectObject(hDC, Handle());

	if (hOldFont)
	{
		SIZE Size = {0, 0};
		if (IsWin9x)
		{
			char *Buf = NEW(char[Len << 1]);
			if (Buf)
			{
				int Bytes = WideCharToMultiByte(CP_ACP, 0, Str, Len, Buf, Len<<1, 0, 0);
				Buf[Bytes] = 0;
				int Chars = _mbstrlen(Buf);
				GetTextExtentExPointA(hDC, Buf, Chars, x, &Fit, 0, &Size);
				DeleteArray(Buf);
			}
		}
		else
		{
			if (NOT GetTextExtentExPointW(hDC, Str, Len, x, &Fit, 0, &Size))
			{
				DWORD e = GetLastError();
				Fit = -1;
			}
		}
		
		SelectObject(hDC, hOldFont);
	}
	else
	{
		DWORD e = GetLastError();
		Fit = -1;
		LgiAssert(0);
	}

	DeleteDC(hDC);
	return Fit;

	#elif defined XWIN

	if (x > 0)
	{
		QFontMetrics m(Handle());
		int w = abs(x);
		int Best = 0;
		for (int i=1; i<Len; i++)
		{
			int n = m.width(Str, i);
			n = abs(n - x);
			
			if (n < w)
			{
				Best = i;
				w = n;
			}
			else
				break;
		}

		return Best;
	}
	return 0;

	#elif defined BEOS

	if (Handle())
	{
		BRect *r = NEW(BRect[Len]);
		escapement_delta *e = NEW(escapement_delta[Len]);
		Handle()->GetBoundingBoxesAsString(Str, Len, B_SCREEN_METRIC, e, r);

		int c = 0;
		for (int i=0; i<Len; i++)
		{
			c += r[i].Width();
			if (c > x)
			{
				return i;
			}
		}
		
		return Len;
	}
	else
	{
		printf("%s:%i - No font.\n", __FILE__, __LINE__);
	}

	#else
	
	#error "Implement me."

	#endif
}

void GFont::_Draw(GSurface *pDC, int x, int y, OsChar *Str, int Len, GRect *r)
{
	#if defined WIN32

	LgiAssert(Handle());

	HDC hDC = pDC->StartDC();
	HFONT hOldFont = (HFONT) SelectObject(hDC, Handle());
	if (hOldFont)
	{
		SetTextColor(hDC, Fore() & 0xFFFFFF);
		SetBkColor(hDC, Back() & 0xFFFFFF);
		SetBkMode(hDC, Transparent() ? TRANSPARENT : OPAQUE);

		SIZE Size;
		if ((NOT Transparent() AND NOT r)
			OR
			(GetOwnerUnderline()))
		{
			GetTextExtentPoint32W(hDC, Str, Len, &Size);
		}

		if (Transparent() AND NOT r)
		{
			TextOutW(hDC, x, y, Str, Len);
		}
		else
		{
			RECT rc;
			if (r)
				rc = *r;
			else
			{
				rc.left = x;
				rc.top = y;
				rc.right = x + Size.cx;
				rc.top = y + Size.cy;
			}

			ExtTextOutW(hDC, x, y, ETO_CLIPPED | (Transparent()?0:ETO_OPAQUE), &rc, Str, Len, 0);
		}

		if (GetOwnerUnderline())
		{
			pDC->Colour(Fore(), 24);
			pDC->Line(x, y + GetHeight() - 1, x + Size.cx + 1, y + GetHeight() - 1);
		}

		HANDLE h = SelectObject(hDC, hOldFont);
	}
	else
	{
		DWORD e = GetLastError();
		LgiAssert(0);
	}
	pDC->EndDC();

	#elif defined XWIN

	COLOUR Old = pDC->Colour();
	if (pDC->Handle())
	{
		int _Back = Back();
		QFontMetrics m(Handle());
		pDC->Colour(Fore(), 24);
		pDC->Handle()->setFont(*Handle());
		pDC->Handle()->drawText(x, y + m.ascent(),
								Str, Len,
								Transparent() ? 0 : &_Back,
								r);
	}

	pDC->Colour(Old);

	#elif defined BEOS
	
	if (Handle())
	{
		BView *v = pDC->Handle();
		if (v AND v->LockLooper())
		{
			font_height h;
			Handle()->GetHeight(&h);
			
			COLOUR f = Fore(), b = Back();

			if (NOT Transparent())
			{
				v->SetHighColor(R24(b), G24(b), B24(b));
				if (r)
				{
					BRect Rc = *r;
					v->FillRect(Rc);
				}
			}

			v->SetHighColor(R24(f), G24(f), B24(f));
			v->SetLowColor(R24(b), G24(b), B24(b));
			
			v->SetFont(Handle());
			BPoint p(x, y + h.ascent);
			v->DrawString(Str, Len, p);
			
			if (Underline())
			{
				v->SetHighColor(R24(f), G24(f), B24(f));
				v->StrokeLine(BPoint(x, y + h.ascent + 1), BPoint(x + r->X(), y + h.ascent + 1));
			}
			
			if (NOT pDC->IsScreen())
			{
				v->Sync();
			}

			v->UnlockLooper();
		}
	}

	#else
	
	#error "Implement me."

	#endif
}

/////////////////////////////////////////////////////////////////////////////
void GFont::TextW(	GSurface *pDC,
					int x, int y,
					char16 *Str,
					int Len,
					GRect *r,
					int TabOrigin)
{
	if (pDC AND IsValid())
	{
		if (Str AND Len < 0)
		{
			Len = StrlenW(Str);
		}
		
		if (Len > 0)
		{
			GDisplayString Ds(this, Str, Len, TabOrigin);
			Ds.Draw(pDC, x, y, r);
		}
		else if (r AND NOT Transparent())
		{
			pDC->Colour(Back(), 24);
			pDC->Rectangle(r);
		}
	}
}

void GFont::SizeW(int *x, int *y, char16 *Str, int Len, int Flags)
{
	if (x) *x = 0;
	if (y) *y = GetHeight();
	
	if (Str AND IsValid())
	{
		if (x)
		{
			if (Len < 0)
			{
				Len = StrlenW(Str);
			}
			GDisplayString Ds(this, Str, Len);
			*x = Ds.X();
		}
	}
	else
	{
		printf("%s:%i - SizeW parameters invalid.\n", __FILE__, __LINE__);
	}
}

int GFont::CharAtW(int x, char16 *Str, int Len, int TabOffset)
{
	if (NOT Str OR x <= 0)
	{
		return 0;
	}

	if (IsValid())
	{
		GDisplayString s(this, Str, Len, TabOffset);
		return s.CharAt(x);

		/*
		int Chars = 0;
		int X = 0;

		if (Len < 0)
		{
			Len = StrlenW(Str);
		}

		for (int i=0; Str[i] AND i<Len; i++)
		{
			if (Str[i] == '\t')
			{
				X = ((X - TabOffset + TabSize()) / TabSize()) * TabSize() + TabOffset;
			}
			else
			{
				int CharX;
				SizeW(&CharX, 0, Str+i, 1);
				
				int OldX = X;
				X += CharX;
				if (abs(x-OldX) < abs(x-X))
				{
					return i;
				}
			}
		}

		return Len;
		*/
	}

	return 0;
}

int GFont::XW(char16 *Str, int Len)
{
	int x = 0;
	SizeW(&x, 0, Str, Len, 0);
	return x;
}

int GFont::YW(char16 *Str, int Len)
{
	int y = 0;
	SizeW(0, &y, Str, Len, 0);
	return y;
}

