#include <stdio.h>
#include "Lgi.h"
#include "IDns.h"
#include "GToken.h"
#include "INet.h"

#define swap2(s) ( (((s)&0xff00)>>8) | (((s)&0xff)<<8) )
#define swap4(s) ( (((s)&0xff000000)>>24) | (((s)&0xff0000)>>8) | (((s)&0xff00)<<8) | (((s)&0xff)<<24) )

bool GetNameServers(GArray<char*> &Ns)
{
	bool Status = false;

	#ifdef WIN32
	GRegKey k("HKEY_LOCAL_MACHINE\\SYSTEM\\ControlSet001\\Services\\Tcpip\\Parameters\\Interfaces");
	List<char> Sub;
	k.GetKeyNames(Sub);
	if (Sub.First())
	{
		for (char *s = Sub.First(); s; s = Sub.Next())
		{
			GRegKey Interface("HKEY_LOCAL_MACHINE\\SYSTEM\\ControlSet001\\Services\\Tcpip\\Parameters\\Interfaces\\%s", s);
			
			char *NameServers = Interface.GetStr("DhcpNameServer");
			if (NameServers)
			{
				GToken Servers(NameServers, " ");
				for (int i=0; i<Servers.Length(); i++)
				{
					bool Has = false;
					for (int n=0; n<Ns.Length(); n++)
					{
						if (strcmp(Ns[n], Servers[i]) == 0)
						{
							Has = true;
							break;
						}
					}
					if (NOT Has)
					{
						Ns[Ns.Length()] = NewStr(Servers[i]);
						Status = true;
					}
				}
			}
		}
		Sub.DeleteArrays();
	}
	#endif

	return Status;
}

uchar *DnsMessage::ReadLabel(uchar *s, char *&Label)
{
	GStringPipe p;
	while (*s)
	{
		if (p.GetSize()) p.Push(".");
		if ((*s & 0xc0) == 0xc0)
		{
			int Offset = ((s[0] & 0x3f) << 8) | s[1];
			uchar *Ptr = (uchar*)&Header + Offset;
			
			char *l = 0;
			ReadLabel(Ptr, l);
			if (l)
			{
				p.Push(l);
			}
			
			s += 2;
			Label = p.NewStr();
			return s;
		}
		else
		{
			p.Push((char*)s + 1, *s);
			s += *s + 1;
		}
	}
	s++;
	Label = p.NewStr();
	return s;
}

uchar *DnsMessage::WriteLabel(uchar *s, char *Label)
{
	GToken n(Label, ".");
	for (int i=0; i<n.Length(); i++)
	{
		*s = min(strlen(n[i]), 255);
		memcpy(s + 1, n[i], *s);
		s += *s + 1;
	}
	*s++ = 0;
	return s;
}

// Functions
char *IDnsResolve(char *Name, int Type, int Class)
{
	char *Status = 0;
	LgiAssert(sizeof(DnsHeader) == 12);

	if (Name)
	{
		int NameLen = strlen(Name);

		GArray<char*> Ns;
		if (GetNameServers(Ns))
		{
			GSocket *s = NEW(GSocketImpl);
			if (s)
			{
				if (s->Open(Ns[0], DNS_PORT))
				{
					DnsMessage m;
					ZeroObj(m);
					m.Header.Id = swap2(DNS_MY_ID);
					m.Header.QueryType = DNS_Q_QUERY;
					m.Header.QueryCount = swap2(1);
					m.Header.RecursionDesired = 1;

					uint16 *p = (uint16*) m.WriteLabel(m.Data, Name);
					*p++ = swap2(Type);
					*p++ = swap2(Class);

					int Length = (int)p - (int)&m;
					int Bytes;
					m.Length = swap2(Length);
					if ((Bytes = s->Write(&m, Length + 2)) > 0)
					{
						if ((Bytes = s->Read(&m, sizeof(m))) > 0)
						{
							m.Length = swap2(m.Length);
							m.Header.Id = swap2(m.Header.Id);
							m.Header.QueryCount = swap2(m.Header.QueryCount);
							m.Header.ResourceCount = swap2(m.Header.ResourceCount);
							m.Header.NameServerCount = swap2(m.Header.NameServerCount);
							m.Header.AdditionalCount = swap2(m.Header.AdditionalCount);
							
							uchar *d = m.Data;
							
							// Parse queries
							for (int q=0; q<m.Header.QueryCount; q++)
							{
								char *Label = 0;
								p = (uint16*) m.ReadLabel(d, Label);
								DeleteArray(Label);
								p += 2;
								d = (uchar*) p;
							}

							// Parse resource records
							for (int r=0; r<m.Header.ResourceCount; r++)
							{
								char *Name = 0;
								p = (uint16*) m.ReadLabel(d, Name);
								uint16 Type = swap2(*p); p++;
								uint16 Class = swap2(*p); p++;
								uint32 *l = (uint32*)p;
								uint32 Ttl = swap4(*l); l++;
								p = (uint16*) l;
								uint16 DataLen = swap2(*p); p++;
								
								if (Type == DNS_TYPE_A AND
									Class == DNS_CLASS_IN AND
									DataLen == 4)
								{
									char Ip[32];
									uchar *j = (uchar*)p;
									sprintf(Ip, "%u.%u.%u.%u", j[0], j[1], j[2], j[3]);
									Status = NewStr(Ip);
								}
								else if (Type == DNS_TYPE_MX AND
									Class == DNS_CLASS_IN)
								{
									char *Mx = 0;
									m.ReadLabel((uchar*)(p+1), Mx);
									Status = Mx;
								}

								d = (uchar*)p + DataLen;
							}
						}
					}
				}

				DeleteObj(s);
			}

			Ns.DeleteArrays();
		}
	}

	return Status;
}

