/*hdr
**      FILE:           INet.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           28/5/98
**      DESCRIPTION:    Internet sockets
**
**      Copyright (C) 1998, Matthew Allen
**              fret@memecode.com
*/

#include <ctype.h>

#include "GFile.h"
#include "INet.h"
#include "GToken.h"
#include "GString.h"
#include "LgiCommon.h"

#define USE_BSD_SOCKETS			1
#define DEBUG_CONNECT			0
#define ETIMEOUT				400

#if defined WIN32

 	#include <stdio.h>
	#include <stdlib.h>
	#include <ras.h>

	typedef HOSTENT HostEnt;
	typedef int socklen_t;
	typedef unsigned long in_addr_t;

	#define MSG_NOSIGNAL		0

#elif defined LINUX

	#include <stdio.h>
	#include <sys/types.h>
	#include <sys/socket.h>
	#include <netinet/in.h>
	#include <arpa/inet.h>
	#include <ctype.h>
	#include <netdb.h>
	#include <errno.h>
	#include <linux/tcp.h>
    #include "LgiCommon.h"
	
	#define SOCKET_ERROR -1
	typedef hostent HostEnt;
	
#elif defined BEOS

	#include <stdio.h>
	#include <NetEndpoint.h>
	#include <socket.h>

	#define SOCKET_ERROR -1
	#define MSG_NOSIGNAL 0

	typedef struct hostent HostEnt;
	typedef int socklen_t;

#elif defined ATHEOS

	#include <stdio.h>
	#include <atheos/socket.h>
	typedef hostent HostEnt;

	int htons(int i) { return i; }
	long inet_addr(char *s) { return 0; }

#endif

#include "LgiNetInc.h"

///////////////////////////////////////////////////////////////////////////////
#ifdef BEOS
uint32 inet_addr(char *HostAddr)
{
	uint32 s = 0;
	GToken Ip(HostAddr, ".");
	if (Ip.Length() == 4)
	{
		uchar *IpAddress = (uchar*)&s;
		for (int i=0; i<4; i++)
		{
			IpAddress[i] = atoi(Ip[i]);
		}				
	}
	return s;
}
#endif


///////////////////////////////////////////////////////////////////////////////
GNetwork::GNetwork()
{
	#ifdef WIN32
	SocketsOpen = false;

	uchar Major = 1;
	uchar Minor = 1;
	ushort Version = MAKEWORD(Major, Minor);
	WSADATA WsaData;

	// Start sockets
	int Result = WSAStartup(Version, &WsaData);
	if (NOT Result)
	{
		if (LOBYTE(WsaData.wVersion) == Major AND HIBYTE(WsaData.wVersion) == Minor)
		{
			SocketsOpen = Result == 0;
		}
		else
		{
			WSACleanup();
		}
	}

	#else
	SocketsOpen = true;
	#endif
}

GNetwork::~GNetwork()
{
	#ifdef WIN32
	if (SocketsOpen)
	{
		WSACleanup();
		SocketsOpen = false;
	}
	#endif
}

#ifdef WIN32
#include "..\..\Win32\INet\MibAccess.h"
#endif

bool GNetwork::EnumInterfaces(List<char> &Lst)
{
	bool Status = false;

	#ifdef WIN32
	UINT IPArray[50];
	UINT Size = 50;
	MibII m;

	if (m.Init() AND
		m.GetIPAddress(IPArray, Size))
	{
		/*
		UINT MaskArray[50];
		UINT MaskArraySize =50;
		
		if (m.GetIPMask(MaskArray,MaskArraySize))
		{
			m_Edit="";
			UINT i;
			for(i=0; i< MaskArraySize; i++)
			{
				m_Edit += "ip: " + TypeIP(IPArray[i])+" mask: " + TypeIP(MaskArray[i])+ "\r\n";
			}
		}
		*/

		for (int i=0; i<Size; i++)
		{
			if (IPArray[i])
			{
				char Str[256];
				uchar *Ip = (uchar*) &IPArray[i];
				sprintf(Str, "%i.%i.%i.%i", Ip[0], Ip[1], Ip[2], Ip[3]);
				Lst.Insert(NewStr(Str));

				Status = true;
			}
		}
	}	
	#endif

	return Status;
}

////////////////////////////////////////////////////////////////////////////////////////////////
class GSocketImplPrivate
{
public:
	// Data
	GNetwork	*Parent;
	bool		Async;
	bool		Canceled;
	int			LogType;
	char		*LogFile;
	int			TimeOut;
	OsSocket	Socket;
	int			Flags;

	GSocketImplPrivate()
	{	
		Async = false;
		Canceled = false;
		LogType = NET_LOG_NONE;
		LogFile = 0;
		TimeOut = 15;
		Socket = INVALID_SOCKET;
		Flags = 0;
	}

	~GSocketImplPrivate()
	{
		DeleteArray(LogFile);
	}
};

GSocketImpl::GSocketImpl()
{
	BytesWritten = 0;
	BytesRead = 0;
	d = NEW(GSocketImplPrivate);
}

GSocketImpl::~GSocketImpl()
{
	Close();
	DeleteObj(d);
}

bool GSocketImpl::IsOK()
{
	return	this != 0 AND
			d != 0 AND
			(d->Parent ? d->Parent->IsValid() : false);
}

bool GSocketImpl::Cancel()
{
	return d->Canceled;
}

void GSocketImpl::Cancel(bool i)
{
	d->Canceled = i;
}

void GSocketImpl::OnDisconnect()
{
}

OsSocket GSocketImpl::Handle(OsSocket Set)
{
	if (Set != INVALID_SOCKET)
	{
		d->Socket = Set;
	}

	return d->Socket;
}

int GSocketImpl::Is(GSocketProp p)
{
	switch (p)
	{
		case GReadable:
		{
			return IsReadable();
			break;
		}
		case GWriteable:
		{
			fd_set s;
			FD_ZERO(&s);
			FD_SET(d->Socket, &s);

			timeval t = {0, 0};
			int r = select(d->Socket+1, 0, &s, 0, &t);
			return r > 0;
			break;
		}
		case GAcceptable:
		{
			fd_set s;
			FD_ZERO(&s);
			FD_SET(d->Socket, &s);

			timeval t = {0, 0};
			int r = select(d->Socket+1, &s, 0, 0, &t);
			return r > 0;
			break;
		}
		case GException:
		{
			fd_set s;
			FD_ZERO(&s);
			FD_SET(d->Socket, &s);

			timeval t = {0, 0};
			int r = select(d->Socket+1, 0, 0, &s, &t);
			return r > 0;
			break;
		}
		case GAsync:
		{
			return d->Async;
			break;
		}
		case GNoDelay:
		{
			return TestFlag(d->Flags, TCP_NODELAY);
		}
	}

	return 0;
}

void GSocketImpl::Is(GSocketProp p, int i)
{
	switch (p)
	{
		case GReadable:
		{
			break;
		}
		case GWriteable:
		{
			break;
		}
		case GAcceptable:
		{
			break;
		}
		case GAsync:
		{
			d->Async = i;
			#if defined WIN32
			ulong NonBlocking = d->Async;
			ioctlsocket(d->Socket, FIONBIO, &NonBlocking);
			#elif defined LINUX
			fcntl(d->Socket, F_SETFL, (d->Async = i) ? O_NONBLOCK : 0);
			#elif defined BEOS
			uint32 b = d->Async ? -1 : 0;
			setsockopt(d->Socket, SOL_SOCKET, SO_NONBLOCK, &b, sizeof(b));
			#else
			#error Impl me.
			#endif
			break;
		}
		case GNoDelay:
		{
			d->Flags = TCP_NODELAY;
			break;
		}
	}
}

bool GSocketImpl::IsOpen()
{
	if (ValidSocket(d->Socket))
	{
		return true;
	}

	return false;
}

bool GSocketImpl::IsReadable()
{
	#if defined(WIN32) || defined(LINUX) || defined(BEOS)
	ulong Val = 0;
	
	// Assign to local var to avoid a thread changing it
	// on us between the validity check and the select.
	// Which is important because a socket value of -1
	// (ie invalid) will crash the FD_SET macro.
	OsSocket s = d->Socket; 
	if (ValidSocket(s))
	{
		struct timeval t = {0, 0};

		fd_set r;
		FD_ZERO(&r);
		FD_SET(s, &r);		
		if (select(s+1, &r, 0, 0, &t) > 0)
		{
			return true;
		}
	}
	return false;
	#else
	#error Not implemented.
	#endif
}

int GSocketImpl::WhatsMyPort()
{
	struct sockaddr_in addr;
	socklen_t size;
	
	ZeroObj(addr);
	size = sizeof(addr);
	if ((getsockname(Handle(), (sockaddr*)&addr, &size)) >= 0)
	{
		return ntohs(addr.sin_port);
	}
	return 0;
}

bool GSocketImpl::WhatsMyIp(char *IpAddr)
{
	if (IpAddr)
	{
		struct sockaddr_in addr;
		socklen_t size;
		
		size = sizeof(addr);
		if ((getsockname(Handle(), (sockaddr*)&addr, &size)) < 0)
			return false;
		
		if (addr.sin_addr.s_addr == INADDR_ANY)
			return false;
		
		uchar *a = (uchar*)&addr.sin_addr.s_addr;
		sprintf(	IpAddr,
					"%i.%i.%i.%i",
					a[0],
					a[1],
					a[2],
					a[3]);
	
		return true;
	}

	return false;
}

bool GSocketImpl::GetRemoteIp(char *IpAddr)
{
	if (IpAddr)
	{
		struct sockaddr_in a;
		socklen_t addrlen = sizeof(a);
		if (NOT getpeername(Handle(), (sockaddr*)&a, &addrlen))
		{
			uchar *addr = (uchar*)&a.sin_addr.s_addr;
			sprintf(IpAddr,
					"%u.%u.%u.%u",
					addr[0],
					addr[1],
					addr[2],
					addr[3]);
			return true;
		}
	}

	return false;
}

int GSocketImpl::Open(char *HostAddr, int Port)
{
	bool Status = false;
	
	Close();

	if (HostAddr)
	{
		char Buf[256];

		BytesWritten = 0;
		BytesRead = 0;
		
		sockaddr_in RemoteAddr;
		HostEnt *Host = 0;
		in_addr_t IpAddress = 0;

		d->Socket = socket(AF_INET, SOCK_STREAM, 0);
		if (d->Socket >= 0)
		{
			if (d->Flags & TCP_NODELAY)
			{
				int flag = 1;
				setsockopt
				(
					d->Socket,
					IPPROTO_TCP,
					TCP_NODELAY,
					(char *) &flag,
					sizeof(int)
				);
			}

			if (isdigit(*HostAddr) AND strchr(HostAddr, '.'))
			{
				// Ip address
				IpAddress = inet_addr(HostAddr);
				
				#if defined(WIN32) || defined(BEOS)

				Host = gethostbyaddr((const char*) &IpAddress, 4, AF_INET);
				#ifdef WIN32
				if (NOT Host)
				{
					int e = WSAGetLastError();
					int asd=0;
				}
				#endif

				#else

				Host = gethostbyaddr(&IpAddress, 4, AF_INET);

				#endif
			}
			else
			{
				// Name address
				#ifdef LINUX

				Host = NEW(HostEnt);
				if (Host)
				{
					memset(Host, 0, sizeof(*Host));
					
					HostEnt *Result = 0;
					int Err = 0;
					if (gethostbyname_r(HostAddr, Host, Buf, sizeof(Buf), &Result, &Err))
					{
						char *ErrStr = GetErrorName(Err);
						printf("%s:%i - gethostbyname_r('%s') failed (%i - %s)\n",
							__FILE__, __LINE__,
							HostAddr, Err, ErrStr);
						DeleteObj(Host);
					}
				}
				
				#else
				
				Host = gethostbyname(HostAddr);
				
				#endif
				
				if (NOT Host)
				{
					Error();
					return false;
				}
			}
			
			if (1)
			{
				RemoteAddr.sin_family = AF_INET;
				RemoteAddr.sin_port = htons(Port);
				if (Host)
				{
					memcpy(&RemoteAddr.sin_addr, Host->h_addr_list[0], sizeof(in_addr));
				}
				else
				{
					memcpy(&RemoteAddr.sin_addr, &IpAddress, sizeof(in_addr));
				}

				#ifdef LINUX

				int Err = -1;
				bool As = Is(GAsync);
				int Start = LgiCurrentTime();

				// Non-blocking
				Is(GAsync, true);
				
				// Do connect..
				errno = 0;
				Err = connect(d->Socket, (sockaddr*) &RemoteAddr, sizeof(sockaddr));
				
				#if DEBUG_CONNECT
				printf("connect=%i(%s)\n", Err, GetErrorName(errno));
				#endif
				
				if (Err < 0)
				{				
					// Wait for results
					while (ValidSocket(d->Socket))
					{
						Err = connect(d->Socket, (sockaddr*) &RemoteAddr, sizeof(sockaddr));
						
						#if DEBUG_CONNECT
						printf("connect=%i(%s)\n", Err, GetErrorName(errno));
						#endif
						if (Err == 0 OR (Err == -1 AND errno == EISCONN))
						{
							// Connection state changed, is it ok or failed?
							Err = false;
							
							#if DEBUG_CONNECT
							printf("Connect = EISCONN\n");
							#endif
							break;
						}
						else if (errno == ECONNREFUSED)
						{
							#if DEBUG_CONNECT
							printf("%s:%i - connect(%s, %i)=ECONNREFUSED\n",
								__FILE__, __LINE__,
								HostAddr, Port);
							#endif

							Close();
							break;
						}
						else if ((LgiCurrentTime() - Start) > (d->TimeOut * 1000))
						{
							// Timeout
							#if DEBUG_CONNECT
							printf("%s:%i - connect(%s, %i) timed out (%is)\n", __FILE__, __LINE__, HostAddr, Port, d->TimeOut);
							#endif
							
							Close();
							errno = ETIMEOUT;
							break;
						}
						else
						{
							// Wait
							LgiSleep(50);
						}
					}
				}

				if (Err)
				#else
				if (connect(d->Socket, (sockaddr*) &RemoteAddr, sizeof(sockaddr)))
				#endif
				{
					Error();
				}
				else
				{
					char Info[256];
					sprintf(Info,
							"[INET] Socket Connect: %s [%i.%i.%i.%i], port: %i",
							HostAddr,
							(RemoteAddr.sin_addr.s_addr) & 0xFF,
							(RemoteAddr.sin_addr.s_addr >> 8) & 0xFF,
							(RemoteAddr.sin_addr.s_addr >> 16) & 0xFF,
							(RemoteAddr.sin_addr.s_addr >> 24) & 0xFF,
							Port);
					OnInformation(Info);

					Status = true;
				}
			}

			if (NOT Status)
			{
				#ifdef WIN32
				closesocket(d->Socket);
				#else
				close(d->Socket);
				#endif
				d->Socket = INVALID_SOCKET;
			}
		}
		else
		{
			Error();
		}
	}
	
	return Status;
}

bool GSocketImpl::Listen(int Port)
{
	Close();

	d->Socket = socket(AF_INET, SOCK_STREAM, 0);
	if (d->Socket >= 0)
	{
		BytesWritten = 0;
		BytesRead = 0;

		sockaddr Addr;
		sockaddr_in *a = (sockaddr_in*) &Addr;
		ZeroObj(Addr);
		a->sin_family = AF_INET;
		a->sin_port = htons(Port);
		#if defined WIN32
		a->sin_addr.S_un.S_addr = INADDR_ANY;
		#elif defined LINUX
		a->sin_addr.s_addr = INADDR_ANY;
		#endif

		if (bind(d->Socket, &Addr, sizeof(Addr)) >= 0)
		{
			if (listen(d->Socket, SOMAXCONN) != SOCKET_ERROR)
			{
				return true;
			}
			else
			{
				Error();
			}
		}
		else
		{
			Error();
		}
	}
	else
	{
		Error();
	}
	return false;
}

bool GSocketImpl::Accept(GSocket *c)
{
	if (c)
	{
		OsSocket NewSocket = INVALID_SOCKET;
		sockaddr Address;
		
		#ifdef WIN32
		int Length = sizeof(Address);
		NewSocket = accept(d->Socket, &Address, &Length);
		#else
		socklen_t Length = sizeof(Address);
		while (ValidSocket(d->Socket))
		{
			if (Is(GAcceptable))
			{
				NewSocket = accept(d->Socket, &Address, &Length);
				break;
			}
			else
			{
				LgiSleep(500);
			}
		}
		#endif
		
		if (ValidSocket(NewSocket))
		{
			c->Handle(NewSocket);
			return true;
		}
		
		return false;
	}

	return false;
}

int GSocketImpl::Close()
{
	#if defined WIN32
	
	if (ValidSocket(d->Socket))
	{
		closesocket(d->Socket);
		OnDisconnect();
	}
	d->Socket = INVALID_SOCKET;
	
	#else
	
	if (ValidSocket(d->Socket))
	{
		close(d->Socket);
		OnDisconnect();
		d->Socket = INVALID_SOCKET;
	}
	
	#endif

	return true;
}

void GSocketImpl::Log(char *Msg, int Ret, char *Buf, int Len)
{
	if (d->LogFile)
	{
		GFile f;
		if (f.Open(d->LogFile, O_WRITE))
		{
			f.Seek(f.GetSize(), SEEK_SET);

			switch (d->LogType)
			{
				case NET_LOG_HEX_DUMP:
				{
					char s[256];

					sprintf(s, "%s = %i\r\n", Msg, Ret);
					f.Write(s, strlen(s));

					for (int i=0; i<Len;)
					{
						char Ascii[32], *a = Ascii;
						sprintf(s, "%08.8X\t", i);
						for (int n = i + 16; i<Len AND i<n; i++)
						{
							*a = (uchar)Buf[i];
							if (*a < ' ') *a = '.';
							a++;

							sprintf(s+strlen(s), "%02.2X ", (uchar)Buf[i]);
						}
						*a++ = 0;
						while (strlen(s) < 58)
						{
							strcat(s, " ");
						}
						strcat(s, Ascii);
						strcat(s, (char*)"\r\n");
						f.Write(s, strlen(s));
					}
					break;
				}
				case NET_LOG_ALL_BYTES:
				{
					f.Write(Buf, Len);
					break;
				}
			}
		}
	}
}

int GSocketImpl::Write(void *Data, int Len, int Flags)
{
	int Total = 0;

	LgiAssert(Data != 0 AND Len >= 0);

	if (Data)
	{
		d->Canceled = false;
		while (	ValidSocket(d->Socket) AND
				Total < Len AND
				NOT d->Canceled)
		{
			int Block = Len-Total;
			Block = min(Block, 4096);

			#ifndef WIN32
			errno = 0;
			#endif

			int i = send(d->Socket, (char*)Data+Total, Block, Flags | MSG_NOSIGNAL);
			// printf("send=%i\n", i);

			#ifdef LINUX
			if (i < 0 AND errno == EAGAIN)
			{
				continue;
			}
			#endif

			if (i > 0)
			{
				Log("Write", i, (char*)Data+Total, i>0 ? i : 0);

				OnWrite((char*)Data+Total, i);
				BytesWritten += i;
				Total += i;
			}
			else
			{
				Error();
				return -1;
			}
		}
	}
	
	return Total;
}

int GSocketImpl::Read(void *Data, int Len, int Flags)
{
	if (ValidSocket(d->Socket) AND Data)
	{
		#ifndef WIN32
		errno = 0;
		#endif

		#if defined(LINUX)
		int Code;
		int Start = LgiCurrentTime();
		do
		{
			errno = 0;
			Code = recv(d->Socket, (char*)Data, Len, Flags | MSG_NOSIGNAL);
			if (Code == 0)
			{
				break;
			}
			else if (Code < 0)
			{
				if (errno == ECONNREFUSED OR
					errno == EBADF)
				{
					Close();
					break;
				}
				else if (errno == EAGAIN)
				{
					LgiSleep(50);
				}
				else
				{
					printf("recv=%i (errno=%s)\n", Code, GetErrorName(errno));
					LgiSleep(50);
				}

				if ((LgiCurrentTime() - Start) > (d->TimeOut * 1000))
				{
					// Timeout
					printf("Read timeout (%is).\n", d->TimeOut);
					break;
				}
			}
			else
			{
				//printf("recv=%i\n", Code);
			}
		}
		while (Code < 0);
		#else
		int Code = recv(d->Socket, (char*)Data, Len, Flags);
		#endif

		Log("Read", Code, (char*)Data, Code>0 ? Code : 0);

		if (Code <= 0)
		{
			Error();
		}
		else
		{
			if (Code < Len)
			{
				// Just in case it's a string lets be safe
				((char*)Data)[Code] = 0;
			}

			BytesRead += Code;
			OnRead((char*)Data, Code);
		}

		return Code;
	}
	return 0;
}

int GSocketImpl::Error()
{
	class ErrorMsg {
	public:
		int Code;
		char *Msg;
	}
	ErrorCodes[] =
	{
		{0,					"Socket disconnected."},
		#if defined WIN32

		{WSAEACCES,			"Permission denied."},
		{WSAEADDRINUSE,		"Address already in use."},
		{WSAEADDRNOTAVAIL,	"Cannot assign requested address."},
		{WSAEAFNOSUPPORT,	"Address family not supported by protocol family."},
		{WSAEALREADY,		"Operation already in progress."},
		{WSAECONNABORTED,	"Software caused connection abort."},
		{WSAECONNREFUSED,	"Connection refused."},
		{WSAECONNRESET,		"Connection reset by peer."},
		{WSAEDESTADDRREQ,	"Destination address required."},
		{WSAEFAULT,			"Bad address."},
		{WSAEHOSTDOWN,		"Host is down."},
		{WSAEHOSTUNREACH,	"No route to host."},
		{WSAEINPROGRESS,	"Operation now in progress."},
		{WSAEINTR,			"Interrupted function call."},
		{WSAEINVAL,			"Invalid argument."},
		{WSAEISCONN,		"Socket is already connected."},
		{WSAEMFILE,			"Too many open files."},
		{WSAEMSGSIZE,		"Message too long."},
		{WSAENETDOWN,		"Network is down."},
		{WSAENETRESET,		"Network dropped connection on reset."},
		{WSAENETUNREACH,	"Network is unreachable."},
		{WSAENOBUFS,		"No buffer space available."},
		{WSAENOPROTOOPT,	"Bad protocol option."},
		{WSAENOTCONN,		"Socket is not connected."},
		{WSAENOTSOCK,		"Socket operation on non-socket."},
		{WSAEOPNOTSUPP,		"Operation not supported."},
		{WSAEPFNOSUPPORT,	"Protocol family not supported."},
		{WSAEPROCLIM,		"Too many processes."},
		{WSAEPROTONOSUPPORT,"Protocol not supported."},
		{WSAEPROTOTYPE,		"Protocol wrong type for socket."},
		{WSAESHUTDOWN,		"Cannot send after socket shutdown."},
		{WSAESOCKTNOSUPPORT,"Socket type not supported."},
		{WSAETIMEDOUT,		"Connection timed out."},
		{WSAEWOULDBLOCK,	"Resource temporarily unavailable."},
		{WSAHOST_NOT_FOUND,	"Host not found."},
		{WSANOTINITIALISED,	"Successful WSAStartup not yet performed."},
		{WSANO_DATA,		"Valid name, no data record of requested type."},
		{WSANO_RECOVERY,	"This is a non-recoverable error."},
		{WSASYSNOTREADY,	"Network subsystem is unavailable."},
		{WSATRY_AGAIN,		"Non-authoritative host not found."},
		{WSAVERNOTSUPPORTED,"WINSOCK.DLL version out of range."},
		{WSAEDISCON,		"Graceful shutdown in progress."},

		#else
		
		{EACCES,			"Permission denied."},
		{EADDRINUSE,		"Address already in use."},
		{EADDRNOTAVAIL,		"Cannot assign requested address."},
		{EAFNOSUPPORT,		"Address family not supported by protocol family."},
		{EALREADY,			"Operation already in progress."},
		{ECONNABORTED,		"Software caused connection abort."},
		{ECONNREFUSED,		"Connection refused."},
		{ECONNRESET,		"Connection reset by peer."},
		{EFAULT,			"Bad address."},
		{EHOSTUNREACH,		"No route to host."},
		{EINPROGRESS,		"Operation now in progress."},
		{EINTR,				"Interrupted function call."},
		{EINVAL,			"Invalid argument."},
		{EISCONN,			"Socket is already connected."},
		{EMFILE,			"Too many open files."},
		{EMSGSIZE,			"Message too long."},
		{ENETDOWN,			"Network is down."},
		{ENETRESET,			"Network dropped connection on reset."},
		{ENETUNREACH,		"Network is unreachable."},
		{ENOBUFS,			"No buffer space available."},
		{ENOPROTOOPT,		"Bad protocol option."},
		{ENOTCONN,			"Socket is not connected."},
		{ENOTSOCK,			"Socket operation on non-socket."},
		{EOPNOTSUPP,		"Operation not supported."},
		{EPFNOSUPPORT,		"Protocol family not supported."},
		{EPROTONOSUPPORT,	"Protocol not supported."},
		{EPROTOTYPE,		"Protocol wrong type for socket."},
		{ESHUTDOWN,			"Cannot send after socket shutdown."},
		{ETIMEDOUT,			"Connection timed out."},
		{EWOULDBLOCK,		"Resource temporarily unavailable."},
		{HOST_NOT_FOUND,	"Host not found."},
		{NO_DATA,			"Valid name, no data record of requested type."},
		{NO_RECOVERY,		"This is a non-recoverable error."},
		{TRY_AGAIN,			"Non-authoritative host not found."},
		{ETIMEOUT,			"Operation timed out."},

		#ifndef BEOS
		{EDESTADDRREQ,		"Destination address required."},
		{EHOSTDOWN,			"Host is down."},
		{ESOCKTNOSUPPORT,	"Socket type not supported."},
		#endif
		
		#endif
		{-1, 0}
	};

	int LastError =
		#ifdef WIN32
		WSAGetLastError();
		#else
		errno;
		#endif
	
	ErrorMsg *Error = ErrorCodes;
	while (Error->Code >= 0 AND Error->Code != LastError)
	{
		Error++;
	}
	OnError(LastError, (Error->Code >= 0) ? Error->Msg : (char*)"<unknown error>");

	switch (LastError)
	{
		case 0:
		#ifdef WIN32
		case 183: // I think this is a XP 'disconnect' ???
		case WSAECONNABORTED:
		case WSAECONNRESET:
		case WSAENETRESET:
		#else
		case ECONNABORTED:
		case ECONNRESET:
		case ENETRESET:
		#endif
		{
			Close();
			break;
		}
	}

	return LastError;
}

void GSocketImpl::SetLogFile(char *FileName, int Type)
{
	DeleteArray(d->LogFile);

	if (FileName)
	{
		switch (Type)
		{
			case NET_LOG_HEX_DUMP:
			case NET_LOG_ALL_BYTES:
			{
				d->LogFile = NewStr(FileName);
				d->LogType = Type;
				break;
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////////
bool HaveNetConnection()
{
	bool Status = false;

	// Check for dial up connection
	#if defined WIN32
	typedef DWORD (__stdcall *RasEnumConnections_Proc)(LPRASCONN lprasconn, LPDWORD lpcb, LPDWORD lpcConnections); 
	typedef DWORD (__stdcall *RasGetConnectStatus_Proc)(HRASCONN hrasconn, LPRASCONNSTATUS lprasconnstatus);

	HMODULE hRas = (HMODULE) LoadLibrary("rasapi32.dll");
	if (hRas)
	{
		RASCONN Con[10];
		DWORD Connections = 0;
		DWORD Bytes = sizeof(Con);

		ZeroObj(Con);
		Con[0].dwSize = sizeof(Con[0]);

		RasEnumConnections_Proc pRasEnumConnections = (RasEnumConnections_Proc) GetProcAddress(hRas, "RasEnumConnectionsA"); 
		RasGetConnectStatus_Proc pRasGetConnectStatus = (RasGetConnectStatus_Proc) GetProcAddress(hRas, "RasGetConnectStatusA"); 

		if (pRasEnumConnections AND
			pRasGetConnectStatus)
		{
			pRasEnumConnections(Con, &Bytes, &Connections);

			for (int i=0; i<Connections; i++)
			{
				RASCONNSTATUS Stat;
				char *Err = "";

				ZeroObj(Stat);
				Stat.dwSize = sizeof(Stat);

				DWORD Error = 0;				
				if (NOT (Error=pRasGetConnectStatus(Con[i].hrasconn, &Stat)))
				{
					if (Stat.rasconnstate == RASCS_Connected)
					{
						Status = true;
					}
				}

				/*
				else
				{
					switch (Error)
					{
						case ERROR_NOT_ENOUGH_MEMORY:
						{
							Err = "Not enuf mem.";
							break;
						}
						case ERROR_INVALID_HANDLE:
						{
							Err = "Invalid Handle.";
							break;
						}
					}
				}

				int n=0;
				*/
			}
		}

		FreeLibrary(hRas);
	}
	#endif

	// Check for ethernet connection
	if (NOT Status)
	{
	}

	return Status;
}


///////////////////////////////////////////////////////////////////////////////////////////
bool WhatsMyIp(char *IpAddr)
{
	bool Status = false;

	strcpy(IpAddr, "(none)");
	GNetwork Sockets;
	if (Sockets.IsValid())
	{
		#if defined WIN32
		char hostname[256];
		HostEnt *e = NULL;
		if (gethostname(hostname, sizeof(hostname)) != SOCKET_ERROR)
		{
			e = gethostbyname(hostname);
		}

		if (e)
		{
			int Which = 0;
			for (; e->h_addr_list[Which]; Which++);
			Which--;

			sprintf(	IpAddr,
						"%i.%i.%i.%i",
						(uchar)e->h_addr_list[Which][0],
						(uchar)e->h_addr_list[Which][1],
						(uchar)e->h_addr_list[Which][2],
						(uchar)e->h_addr_list[Which][3]);
			Status = true;
		}
		#elif defined BEOS

		struct sockaddr_in addr;
		int sock, size;
		
		memset((char *)&addr, 0, sizeof(addr));
		addr.sin_family = AF_INET;
		addr.sin_addr.s_addr = INADDR_ANY;
		addr.sin_port = 0;
		
		if ((sock = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
			return false;
		
		if (bind(sock, (struct sockaddr *)&addr, sizeof(addr)) < 0)
			return false;
		
		size = sizeof(addr);
		if ((getsockname(sock, (struct sockaddr *)&addr, &size)) < 0)
			return false;
		
		close(sock);
		
		if (addr.sin_addr.s_addr == INADDR_ANY)
			return false;
		
		uchar *a = (uchar*)&addr.sin_addr.s_addr;
		sprintf(	IpAddr,
					"%i.%i.%i.%i",
					a[0],
					a[1],
					a[2],
					a[3]);
	
		Status = true;
		
		#endif
	}

	return Status;
}

//////////////////////////////////////////////////////////////////////
#define SOCKS5_VER					5

#define SOCKS5_CMD_CONNECT			1
#define SOCKS5_CMD_BIND				2
#define SOCKS5_CMD_ASSOCIATE		3

#define SOCKS5_ADDR_IPV4			1
#define SOCKS5_ADDR_DOMAIN			3
#define SOCKS5_ADDR_IPV6			4

#define SOCKS5_AUTH_NONE			0
#define SOCKS5_AUTH_GSSAPI			1
#define SOCKS5_AUTH_USER_PASS		2

GSocks5Socket::GSocks5Socket(	char *proxy,
										int port,
										char *username,
										char *password)
{
	Socks5Connected = false;
	Proxy = NewStr(proxy);
	Port = port;
	UserName = NewStr(username);
	Password = NewStr(password);
}

GSocks5Socket::~GSocks5Socket()
{
	DeleteObj(Proxy);
	DeleteObj(UserName);
	DeleteObj(Password);
}

int GSocks5Socket::Open(char *HostAddr, int port)
{
	bool Status = false;
	if (HostAddr)
	{
		char Msg[256];
		sprintf(Msg, "[SOCKS5] Connecting to proxy server '%s'", HostAddr);
		OnInformation(Msg);

		Status = GSocketImpl::Open(Proxy, Port);
		if (Status)
		{
			char Buf[1024];
			Buf[0] = SOCKS5_VER;
			Buf[1] = 2; // methods
			Buf[2] = SOCKS5_AUTH_NONE;
			Buf[3] = SOCKS5_AUTH_USER_PASS;
			
			// No idea how to implement this.
			// AuthReq[3] = SOCKS5_AUTH_GSSAPI;

			OnInformation("[SOCKS5] Connected, Requesting authentication type.");
			GSocketImpl::Write(Buf, 4, 0);
			if (GSocketImpl::Read(Buf, 2, 0) == 2)
			{
				if (Buf[0] == SOCKS5_VER)
				{
					bool Authenticated = false;
					switch (Buf[1])
					{
						case SOCKS5_AUTH_NONE:
						{
							Authenticated = true;
							OnInformation("[SOCKS5] No authentication needed.");
							break;
						}
						case SOCKS5_AUTH_USER_PASS:
						{
							OnInformation("[SOCKS5] User/Pass authentication needed.");
							if (UserName AND Password)
							{
								char *b = Buf;
								*b++ = 1; // ver of sub-negotiation ??
								
								int NameLen = strlen(UserName);
								*b++ = NameLen;
								strcpy(b, UserName);
								b += NameLen;

								int PassLen = strlen(Password);
								*b++ = PassLen;
								strcpy(b, Password);
								b += PassLen;

								GSocketImpl::Write(Buf, 3 + NameLen + PassLen, 0);
								if (GSocketImpl::Read(Buf, 2, 0) == 2)
								{
									Authenticated = (Buf[0] == 1 AND Buf[1] == 0);
								}

								if (NOT Authenticated)
								{
									OnInformation("[SOCKS5] User/Pass authentication failed.");
								}
							}
							break;
						}
					}

					if (Authenticated)
					{
						OnInformation("[SOCKS5] Authentication successful.");

						int HostPort = htons(port);
						
						// Header
						char *b = Buf;
						*b++ = SOCKS5_VER;
						*b++ = SOCKS5_CMD_CONNECT;
						*b++ = 0; // reserved

						long IpAddr = inet_addr(HostAddr);
						if (IpAddr != -1)
						{
							// Ip
							*b++ = SOCKS5_ADDR_IPV4;
							memcpy(b, &IpAddr, 4);
							b += 4;
						}
						else
						{
							// Domain Name
							*b++ = SOCKS5_ADDR_DOMAIN;
							int Len = strlen(HostAddr);
							*b++ = Len;
							strcpy(b, HostAddr);
							b += Len;
						}

						// Port
						memcpy(b, &HostPort, 2);
						b += 2;

						GSocketImpl::Write(Buf, (int)b - (int)Buf, 0);
						if (GSocketImpl::Read(Buf, 10, 0) == 10)
						{
							if (Buf[0] == SOCKS5_VER)
							{
								switch (Buf[1])
								{
									case 0:
										Socks5Connected = true;
										OnInformation("[SOCKS5] Connected!");
										break;
									case 1:
										OnInformation("[SOCKS5] General SOCKS server failure");
										break;
									case 2:
										OnInformation("[SOCKS5] Connection not allowed by ruleset");
										break;
									case 3:
										OnInformation("[SOCKS5] Network unreachable");
										break;
									case 4:
										OnInformation("[SOCKS5] Host unreachable");
										break;
									case 5:
										OnInformation("[SOCKS5] Connection refused");
										break;
									case 6:
										OnInformation("[SOCKS5] TTL expired");
										break;
									case 7:
										OnInformation("[SOCKS5] Command not supported");
										break;
									case 8:
										OnInformation("[SOCKS5] Address type not supported");
										break;
									default:
										OnInformation("[SOCKS5] Unknown SOCKS server failure");
										break;
								}
							}
							else
							{
								OnInformation("[SOCKS5] Wrong socks version.");
							}
						}
						else
						{
							OnInformation("[SOCKS5] Connection request read failed.");
						}
					}
					else
					{
						OnInformation("[SOCKS5] Not authenticated.");
					}
				}
				else
				{
					OnInformation("[SOCKS5] Wrong socks version.");
				}
			}
			else
			{
				OnInformation("[SOCKS5] Authentication type read failed.");
			}

			Status = Socks5Connected;
			if (NOT Status)
			{
				GSocketImpl::Close();
				OnInformation("[SOCKS5] Failure: Disconnecting.");
			}
		}
	}

	return Status;
}

//////////////////////////////////////////////////////////
GHost::GHost(int d, char *s)
{
	DefaultPort = d;
	Server = 0;
	Port = 0;

	if (s) *this = s;
}

GHost::~GHost()
{
	Empty();
}

void GHost::Empty()
{
	DeleteArray(Server);
	Port = 0;
}

GHost &GHost::operator =(char *s)
{
	Empty();
	if (s)
	{
		char *c = strchr(s, ':');
		if (c)
		{
			*c = 0;
			Server = NewStr(s);
			Port = atoi(c+1);
			*c = ':';
		}
		else
		{
			Server = NewStr(s);
		}

		if (Port == 0)
		{
			Port = DefaultPort;
		}
	}

	return *this;
}


