/*
**	FILE:			GuiDataDlg.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			30/10/98
**	DESCRIPTION:	Data serialisation classes
**
**	Copyright (C) 1998, Matthew Allen
**		fret@memecode.com
*/


#include <stdio.h>

#include "Lgi.h"
#include "GPassword.h"
#include "GDataDlg.h"
#include "GTextLabel.h"
#include "GScrollBar.h"

////////////////////////////////////////////////////////////////////////////////////////////////
DataDlgField::DataDlgField(int type, int ctrlid, char *opt, char *desc)
{
	Type = type;
	CtrlId = ctrlid;
	Option = NewStr(opt);
	Desc = NewStr(desc);
}

DataDlgField::~DataDlgField()
{
	DeleteArray(Option);
	DeleteArray(Desc);
}

/////////////////////////////////////////////////////////////////////////////////
DataDlgTools::DataDlgTools()
{
	DeleteEmptyStrings = true;
	Dlg = 0;
	Options = 0;
}

void DataDlgTools::Set(GView *dlg, ObjProperties *options)
{
	Dlg = dlg;
	Options = options;
}

ObjProperties *DataDlgTools::GetOptions()
{
	return Options;
}

bool DataDlgTools::ProcessField(DataDlgField *f, bool Write, char *OptionOverride)
{
	char *Opt = (OptionOverride) ? OptionOverride : f->Option;
	GView *v = Dlg->FindControl(f->CtrlId);
	if (Options AND Dlg AND Opt AND v)
	{
		char OsOpt[256] = "";
		switch (f->Type)
		{
			case DATA_PASSWORD:
			{
				GPassword g;
				if (Write) // Ctrl -> Opts
				{
					if (Dlg->GetCtrlEnabled(f->GetCtrl()))
					{
						char *Pass = v->Name();
						if (Pass)
						{
							g.Set(Pass);
							g.Serialize(Options, f->GetOption(), Write);
						}
					}
					else
					{
						Options->DeleteKey(f->GetOption());
					}
				}
				else // Opts -> Ctrl
				{
					bool HasPass = g.Serialize(Options, f->GetOption(), Write);
					Dlg->SetCtrlEnabled(f->GetCtrl(), HasPass);
				}
				break;
			}
			case DATA_STR_SYSTEM:
			{
				sprintf(OsOpt, "%s-%s", Opt, LgiGetOsName());
				Opt = OsOpt;
				// fall through
			}
			case DATA_STR:
			{
				if (Write) // Ctrl -> Opts
				{
					char *s = v->Name();
					if (DeleteEmptyStrings AND NOT ValidStr(s))
					{
						Options->DeleteKey(Opt);
					}
					else
					{
						Options->Set(Opt, ValidStr(s) ? s : (char*)"");
					}
				}
				else // Opts -> Ctrl
				{
					char *n = 0;
					Options->Get(Opt, n);
					v->Name(n);
				}
				break;
			}
			case DATA_BOOL:
			case DATA_INT:
			{
				if (Write) // Ctrl -> Opts
				{
					Options->Set(Opt, v->Value());
				}
				else // Opts -> Ctrl
				{
					int n = 0;
					Options->Get(Opt, n);
					v->Value(n);
				}
				break;
			}
			default:
			{
				LgiAssert(0);
				return false;
				break;
			}
		}

		return true;
	}

	return false;
}

bool DataDlgTools::ProcessFields(DataDlgFieldList &Field, bool Write)
{
	int Processed = 0;
	
	for (DataDlgField *f = Field.First(); f; f = Field.Next())
	{
		if (ProcessField(f, Write))
		{
			Processed++;
		}
	}

	return Processed == Field.GetItems();
}

bool DataDlgTools::LoadFields(DataDlgFieldList &Field)
{
	return ProcessFields(Field, false);
}

bool DataDlgTools::SaveFields(DataDlgFieldList &Field)
{
	return ProcessFields(Field, true);
}

////////////////////////////////////////////////////////////////////////////////
DRecordSetCtrls::DRecordSetCtrls(	GView *window,
									DataDlgFieldList *fields,
									List<DRecordSetObj> *records,
									int DescId,
									int ScrollId,
									int NewId,
									int DelId,
									char *Template)

{
	Current = 0;
	RecordStringTemplate = NewStr(Template);
	Dlg = window;
	Fields = fields;
	Records = records;
	NewRecordId = NewId;
	DeleteRecordId = DelId;

	if (Dlg)
	{
		Description = dynamic_cast<GText*>(Dlg->FindControl(DescId));
		Scroll = dynamic_cast<GScrollBar*>(Dlg->FindControl(ScrollId));
		OnMoveRecord(Records->First());
	}
}

DRecordSetCtrls::~DRecordSetCtrls()
{
	DeleteArray(RecordStringTemplate);
}

int DRecordSetCtrls::GetCurrentIndex()
{
	return (Records AND Current) ? Records->IndexOf(Current) : -1;
}

void DRecordSetCtrls::Serialize(bool Write)
{
	if (Fields AND Dlg AND Current)
	{
		ObjProperties Temp;
		Options = &Temp;

		if (Write)
		{
			// Move data from controls to propertiy list
			SaveFields(*Fields);

			// Move data from property list to record
			Current->Serialize(Temp, false);
		}
		else
		{
			// Move data from record to property list
			Current->Serialize(Temp, true);

			// Load the new data
			LoadFields(*Fields);
		}

		Options = 0;
	}
}

void DRecordSetCtrls::OnMoveRecord(DRecordSetObj *r)
{
	if (Dlg AND Fields)
	{
		Serialize(true);
		Current = r;
		Serialize(false);

		// Set controls enabled flag correctly
		for (DataDlgField *f = Fields->First(); f; f = Fields->Next())
		{
			GView *Ctrl = dynamic_cast<GView*>(Dlg->FindControl(f->GetCtrl()));
			if (Ctrl)
			{
				if (NOT Current)
				{
					// clear controls
				}

				Ctrl->Enabled(Current != 0);
			}
		}

		// Set the scrollbar
		if (Scroll)
		{
			Scroll->SetLimits(0, Records->GetItems() - 1);
			Scroll->Value((Current) ? Records->IndexOf(Current) : 0);
			Scroll->SetPage(1);
		}

		// Set the description
		if (Description)
		{
			char Str[256];
			sprintf(Str, RecordStringTemplate ? RecordStringTemplate : (char*)"Record: %i of %i", GetCurrentIndex()+1, (Records) ? Records->GetItems() : 0);
			Description->Name(Str);
		}
	}
}

int DRecordSetCtrls::OnNotify(GView *Col, int Flags)
{
	if (Fields AND Dlg AND Records)
	{
		if (Col == Scroll)
		{
			int CurrentIndex = (Current) ? Records->IndexOf(Current) : -1;
			int NewIndex = Col->Value();
			if (CurrentIndex != NewIndex)
			{
				OnMoveRecord(Records->ItemAt(NewIndex));
			}
		}

		if (Col->GetId() == NewRecordId)
		{
			DRecordSetObj *r = NewObj();
			if (r)
			{
				Records->Insert(r);
				OnMoveRecord(r);
			}
		}

		if (Col->GetId() == DeleteRecordId)
		{
			int Index = GetCurrentIndex();
			DRecordSetObj *r = Records->ItemAt(Index);
			if (r)
			{
				Records->Delete(r);
				OnMoveRecord(Records->ItemAt(limit(Index, 0, Records->GetItems()-1)));
				DelObj(r);
			}
		}
	}

	return 0;
}
