/*hdr
**      FILE:           GDocApp.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           20/4/01
**      DESCRIPTION:    Generic Document Application Impl
**
*/

#include <stdlib.h>
#include "GDocApp.h"
#include "GToken.h"

//////////////////////////////////////////////////////////////////////////////////////////
class GDocAppPrivate
{
public:
	// Data
	char			*OptionsFile;
	char			*AppName;
	char			*CurFile;
	char			*Icon;
	ObjProperties	*Options;
	bool			Dirty;

	GDocAppPrivate()
	{
		OptionsFile = 0;
		Options = 0;
		AppName = 0;
		CurFile = 0;
		Icon = 0;
		Dirty = 0;
	}
	
	~GDocAppPrivate()
	{
		// Release memory
		DeleteArray(OptionsFile);
		DeleteArray(AppName);
		DeleteArray(Icon);
		DeleteObj(Options);
	}
	
	char *GetOptionsFile()
	{
		// Get options file
		char *Status = 0;
		char Exe[256];
		LgiGetExeFile(Exe, sizeof(Exe));
		char *File = strrchr(Exe, DIR_CHAR);
		if (File)
		{
			File++;
			#ifdef WIN32
			char *End = strchr(File, '.');
			if (End)
			{
				strcpy(End+1, "r");
			}
			else
			#endif
			{
				// unix apps have no '.' in their name
				strcat(File, ".r");
			}

			Status = LgiFindFile(File);
			if (NOT Status)
			{
				// No options file yet... so create one in the
				// app's directory
				Status = NewStr(Exe);
			}
		}

		return Status;
	}
};

//////////////////////////////////////////////////////////////////////////////////////////
GDocApp::GDocApp(char *appname, char *icon)
{
	d = NEW(GDocAppPrivate);

	GRect r(0, 0, 500, 400);
	SetPos(r);
	MoveToCenter();
	_FileMenu = 0;
	d->AppName = NewStr(appname?appname:(char*)"LGI::GDocApp");

	SetQuitOnClose(true);

	// Setup class
	if (icon)
	{
		#if defined WIN32
		GWin32Class *c = GWin32Class::Create(d->AppName);
		if (c)
		{
			c->Class.a.hIcon = LoadIcon(LgiProcessInst(), ((int)icon&0xffff0000)?icon:MAKEINTRESOURCE(icon));
		}
		#else
		d->Icon = NewStr(icon);
		#endif
	}
}

GDocApp::~GDocApp()
{
	DeleteObj(d);
}

void GDocApp::_Close()
{
	Empty();
	SetCurFile(0);
}

bool GDocApp::_DoSerialize(bool Write)
{
	bool Status = false;

	if (NOT d->OptionsFile)
	{
		d->OptionsFile = d->GetOptionsFile();
	}

	if (NOT d->Options)
	{
		d->Options = NEW(ObjProperties);
	}

	GFile f;
	if (f.Open(d->OptionsFile, Write?O_WRITE:O_READ))
	{
		if (Write)
		{
			// clear existing contents
			f.SetSize(0);

			// Save window position
			#if defined WIN32
			if (NOT IsZoomed(Handle()) AND NOT IsIconic(Handle()))
			#endif
			{
				GRect r = GetPos();
				d->Options->Set("Pos", r.Describe());
			}

			// save misc options
			SerializeOptions(d->Options, true);
			GMru::Serialize(d->Options, "Mru", true);
		}

		// do the work
		Status = _SerializeFile(f, Write);

		if (NOT Write)
		{
			// read misc options
			GMru::Serialize(d->Options, "Mru", false);
			SerializeOptions(d->Options, false);

			// window pos
			GRect r(100, 100, 500, 400);
			char *Pos;
			if (d->Options->Get("Pos", Pos))
			{
				GToken p(Pos, ",");
				if (p.Length() == 4)
				{
					r.x1 = atoi(p[0]);
					r.y1 = atoi(p[1]);
					r.x2 = atoi(p[2]);
					r.y2 = atoi(p[3]);
				}
			}

			SetPos(r);
		}
	}

	return Status;
}

bool GDocApp::_SerializeFile(GFile &f, bool Write)
{
	return d->Options->Serialize(f, Write);
}

bool GDocApp::_Create()
{
	// Load options
	_DoSerialize(false);

	// Create and setup the main application window
	#ifdef WIN32
	CreateClass(d->AppName);
	#endif

	if (Attach(0))
	{
		Name(d->AppName);
	
		#if defined LINUX
		if (d->Icon)
		{
			char *Path = LgiFindFile(d->Icon);
			if (Path)
			{
				Handle()->setIcon(Path);
				DeleteArray(Path);
			}
		}
		#endif

		if (X() < 1)
		{	
			GRect r(100, 100, 600, 500);
			SetPos(r);
		}
	}
	else return false;

	return true;
}

bool GDocApp::_Destroy()
{
	// Save options
	_DoSerialize(true);
	LgiApp->AppWnd = 0;
	return true;
}

bool GDocApp::_LoadMenu(char *Resource)
{
	Menu = NEW(GMenu);
	if (Menu)
	{
		Menu->Attach(this);

		if (Resource)
		{
			Menu->Load(this, Resource);
			_FileMenu = Menu->AppendSub("&File", 0);
		}
		else
		{
			_FileMenu = Menu->AppendSub("&File");
		}

		if (_FileMenu)
		{
			_FileMenu->AppendItem("&Open", IDM_OPEN, true);
			_FileMenu->AppendItem("&Save", IDM_SAVE, true);
			_FileMenu->AppendItem("Save &As", IDM_SAVEAS, true);
			_FileMenu->AppendItem("&Close", IDM_CLOSE, true);
			_FileMenu->AppendSeparator();

			GSubMenu *Recent = _FileMenu->AppendSub("Recent...");
			if (Recent)
			{
				Set(Recent);
				_FileMenu->AppendSeparator();
				GMru::Serialize(d->Options, "Mru", false);
			}

			_FileMenu->AppendItem("E&xit", IDM_EXIT, true);
		}
	}

	return Menu != 0;
}

bool GDocApp::_OpenFile(char *File)
{
	bool Status = false;
	if (SetDirty(false))
	{
		char RealPath[256];
		if (ResolveShortcut(File, RealPath, sizeof(RealPath)))
		{
			File = RealPath;
		}

		Status = GMru::_OpenFile(File);
		if (Status)
		{
			d->Dirty = false;
			SetCurFile(File);
		}
	}

	return Status;
}

bool GDocApp::_SaveFile(char *File)
{
	char RealPath[256];
	if (ResolveShortcut(File, RealPath, sizeof(RealPath)))
	{
		File = RealPath;
	}
	else
	{
		strsafecpy(RealPath, File, sizeof(RealPath));
	}

	bool Status = GMru::_SaveFile(RealPath);
	if (Status)
	{
		d->Dirty = false;
		SetCurFile(RealPath);
		OnDirty(d->Dirty);
	}

	return Status;
}

char *GDocApp::GetAppName()
{
	return d->AppName;
}

void GDocApp::SetCurFile(char *f)
{
	if (f != d->CurFile)
	{
		DeleteArray(d->CurFile);
		d->CurFile = NewStr(f);
	}

	char s[256];
	if (d->CurFile)
	{
		sprintf(s, "%s [%s%s]", d->AppName, d->CurFile, d->Dirty ? " changed" : "");
	}
	else if (d->Dirty)
	{
		sprintf(s, "%s [changed]", d->AppName);
	}
	else
	{
		strcpy(s, d->AppName);
	}

	Name(s);
}

char *GDocApp::GetCurFile()
{
	return d->CurFile;
}

bool GDocApp::SetDirty(bool Dirty)
{
	if (d->Dirty ^ Dirty)
	{
		// Changing...
		if (Dirty)
		{
			// Setting dirty
			d->Dirty = true;
			SetCurFile(d->CurFile);
		}
		else
		{
			// Clearing dirty
			int Result = LgiMsg(this, "Do you want to save your changes?", d->AppName, MB_YESNOCANCEL);
			if (Result == IDYES)
			{
				if (NOT ValidStr(d->CurFile))
				{
					DoFileDlg(false);
				}
				else
				{
					_SaveFile(d->CurFile);
				}
			}
			else if (Result == IDCANCEL)
			{
				return false;
			}

			d->Dirty = false;
			SetCurFile(d->CurFile);
		}

		OnDirty(d->Dirty);
	}

	return true;
}

bool GDocApp::GetDirty()
{
	return d->Dirty;
}

ObjProperties *GDocApp::GetOptions()
{
	return d->Options;
}

void GDocApp::OnReceiveFiles(List<char> &Files)
{
	char *f = Files.First();
	if (f AND _OpenFile(f))
	{
		GMru::AddFile(f);
	}
}

bool GDocApp::OnRequestClose(bool OsShuttingDown)
{
	if (SetDirty(false))
	{
		return GWindow::OnRequestClose(OsShuttingDown);
	}

	return false;
}

int GDocApp::OnCommand(int Cmd, int Event, OsView Window)
{
	switch (Cmd)
	{
		case IDM_SAVE:
		{
			if (NOT GetCurFile())
			{
				GMessage m = CreateMsg(M_COMMAND, IDM_SAVEAS, 0);
				GMru::OnEvent(&m);
			}
			else
			{
				_SaveFile(GetCurFile());
			}
			break;
		}
		case IDM_CLOSE:
		{
			_Close();
			break;
		}
		case IDM_EXIT:
		{
			LgiCloseApp();
			break;
		}
	}

	return 0;
}

int GDocApp::OnEvent(GMessage *m)
{
	GMru::OnEvent(m);

	switch (MsgCode(m))
	{
		#ifdef WIN32
		case WM_CLOSE:
		{
			if (NOT SetDirty(false))
			{
				return 0;
			}
			break;
		}
		#endif
	}

	return GWindow::OnEvent(m);
}



