#include <stdlib.h>
#include <stdio.h>

#include "Lgi.h"
#include "GVariant.h"
#include "GToken.h"

GVariant::GVariant()
{
	Type = GV_NULL;
	Id = 0;
	Dirty = 0;
	ZeroObj(Value);
}

GVariant::GVariant(int i)
{
	Type = GV_INT32;
	Id = 0;
	Dirty = 0;
	Value.Int = i;
}

GVariant::GVariant(int64 i)
{
	Type = GV_INT64;
	Id = 0;
	Dirty = 0;
	Value.Int64 = i;
}

GVariant::GVariant(double i)
{
	Type = GV_DOUBLE;
	Id = 0;
	Dirty = 0;
	Value.Dbl = i;
}

GVariant::GVariant(char *s)
{
	Type = GV_STRING;
	Id = 0;
	Dirty = 0;
	Value.String = NewStr(s);
}

GVariant::GVariant(void *p)
{
	Type = GV_NULL;
	Id = 0;
	Dirty = 0;
	*this = p;
}

GVariant::GVariant(GDom *p)
{
	Type = GV_NULL;
	Id = 0;
	Dirty = 0;
	*this = p;
}

GVariant::GVariant(GDateTime *d)
{
	Type = GV_NULL;
	Id = 0;
	Dirty = 0;
	*this = d;
}

GVariant::~GVariant()
{
	Empty();
}

GVariant &GVariant::operator =(GDateTime *d)
{
	Empty();

	if (d)
	{
		Type = GV_DATETIME;
		Value.Date = NEW(GDateTime);
		if (Value.Date)
		{
			*Value.Date = *d;
			if (Dirty) *Dirty = true;
		}
	}

	return *this;
}

GVariant &GVariant::operator =(int i)
{
	Empty();
	Type = GV_INT32;
	Value.Int = i;
	if (Dirty) *Dirty = true;

	return *this;
}

GVariant &GVariant::operator =(bool i)
{
	Empty();
	Type = GV_BOOL;
	Value.Bool = i;
	if (Dirty) *Dirty = true;

	return *this;
}

GVariant &GVariant::operator =(int64 i)
{
	Empty();
	Type = GV_INT64;
	Value.Int64 = i;
	if (Dirty) *Dirty = true;

	return *this;
}

GVariant &GVariant::operator =(double i)
{
	Empty();
	Type = GV_DOUBLE;
	Value.Dbl = i;
	if (Dirty) *Dirty = true;

	return *this;
}

GVariant &GVariant::operator =(char *s)
{
	Empty();
	if (s)
	{
		Type = GV_STRING;
		Value.String = NewStr(s);
	}
	if (Dirty) *Dirty = true;

	return *this;
}

GVariant &GVariant::operator =(void *p)
{
	Empty();
	if (p)
	{
		Type = GV_VOID_PTR;
		Value.Ptr = p;
		if (Dirty) *Dirty = true;
	}

	return *this;
}

GVariant &GVariant::operator =(GDom *p)
{
	Empty();
	if (p)
	{
		Type = GV_DOM;
		Value.Dom = p;
		if (Dirty) *Dirty = true;
	}

	return *this;
}

GVariant &GVariant::operator =(class GView *p)
{
	Empty();
	if (p)
	{
		Type = GV_GVIEW;
		Value.Ptr = p;
		if (Dirty) *Dirty = true;
	}

	return *this;
}

GVariant &GVariant::operator =(class GMouse *p)
{
	Empty();
	if (p)
	{
		Type = GV_GMOUSE;
		Value.Ptr = p;
		if (Dirty) *Dirty = true;
	}

	return *this;
}

GVariant &GVariant::operator =(class GKey *p)
{
	Empty();
	if (p)
	{
		Type = GV_GKEY;
		Value.Ptr = p;
		if (Dirty) *Dirty = true;
	}

	return *this;
}

GVariant &GVariant::operator =(GVariant &i)
{
	Empty();
	Type = i.Type;
	Id = i.Id;

	switch (Type)
	{
		case GV_INT32:
		{
			Value.Int = i.Value.Int;
			break;
		}
		case GV_BOOL:
		{
			Value.Bool = i.Value.Bool;
			break;
		}
		case GV_INT64:
		{
			Value.Int64 = i.Value.Int64;
			break;
		}
		case GV_DOUBLE:
		{
			Value.Dbl = i.Value.Dbl;
			break;
		}
		case GV_STRING:
		{
			Value.String = NewStr(i.Str());
			break;
		}
		case GV_BINARY:
		{
			SetBinary(i.Value.Binary.Length, i.Value.Binary.Data);
			break;
		}
		case GV_LIST:
		{
			SetList(i.Value.Lst);
			break;
		}
		case GV_DOM:
		{
			Value.Dom = i.Value.Dom;
			break;
		}
		case GV_VOID_PTR:
		{
			Value.Ptr = i.Value.Ptr;
			break;
		}
		case GV_DATETIME:
		{
			if (i.Value.Date)
			{
				Value.Date = NEW(GDateTime);
				if (Value.Date)
				{
					*Value.Date = *i.Value.Date;
				}
			}
			break;
		}
		case GV_NULL:
		{
			break;
		}
		default:
		{
			LgiAssert(0);
			break;
		}
	}

	if (Dirty) *Dirty = true;

	return *this;
}

bool GVariant::SetBinary(int Len, void *Data)
{
	bool Status = false;

	Empty();

	Type = GV_BINARY;
	Value.Binary.Length = Len;
	Value.Binary.Data = NEW(uchar[Value.Binary.Length]);
	if (Value.Binary.Data)
	{
		if (Data)
		{
			memcpy(Value.Binary.Data, Data, Value.Binary.Length);
		}
		else
		{
			memset(Value.Binary.Data, 0, Value.Binary.Length);
		}

		Status = true;
	}

	if (Dirty) *Dirty = true;

	return Status;
}

bool GVariant::SetList(List<GVariant> *Lst)
{
	Empty();
	Type = GV_LIST;
	Value.Lst = NEW(List<GVariant>);
	if (Value.Lst AND Lst)
	{
		for (GVariant *s=Lst->First(); s; s=Lst->Next())
		{
			GVariant *New = NEW(GVariant);
			if (New)
			{
				*New = *s;
				Value.Lst->Insert(New);
			}
		}
	}

	return Value.Lst != 0;
}

char *GVariant::Str()
{
	return (Type == GV_STRING) ? Value.String : 0;
}

void GVariant::Empty()
{
	switch (Type)
	{
		case GV_STRING:
		{
			DeleteArray(Value.String);
			break;
		}
		case GV_BINARY:
		{
			char *d = (char*) Value.Binary.Data;
			DeleteArray(d);
			Value.Binary.Data = 0;
			break;
		}
		case GV_DATETIME:
		{
			DeleteObj(Value.Date);
			break;
		}
		case GV_LIST:
		{
			if (Value.Lst)
			{
				Value.Lst->DeleteObjects();
				DeleteObj(Value.Lst);
			}
			break;
		}
	}

	Type = GV_NULL;
	ZeroObj(Value);
}

bool GVariant::IsInt()
{
	return Type == GV_INT32;
}

bool GVariant::IsBool()
{
	return Type == GV_BOOL;
}

bool GVariant::IsDouble()
{
	return Type == GV_DOUBLE;
}

bool GVariant::IsString()
{
	return Type == GV_STRING;
}

bool GVariant::IsBinary()
{
	return Type == GV_BINARY;
}

bool GVariant::IsNull()
{
	return Type == GV_NULL;
}

#define IsList() (Type == GV_LIST AND Value.Lst)

int GVariant::GetItems()
{
	return IsList() ? Value.Lst->GetItems() : 0;
}

bool GVariant::HasItem(GVariant *s)
{
	return IsList() ? Value.Lst->HasItem(s) : false;
}

GVariant *GVariant::ItemAt(int i)
{
	return IsList() ? Value.Lst->ItemAt(i) : 0;
}

GVariant *GVariant::First()
{
	return IsList() ? Value.Lst->First() : 0;
}

GVariant *GVariant::Next()
{
	return IsList() ? Value.Lst->Next() : 0;
}

GVariant *GVariant::Prev()
{
	return IsList() ? Value.Lst->Prev() : 0;
}

GVariant *GVariant::Last()
{
	return IsList() ? Value.Lst->Last() : 0;
}

bool GVariant::Insert(GVariant *s, int Where)
{
	return IsList() ? Value.Lst->Insert(s, Where) : false;
}

bool GVariant::Delete(GVariant *s)
{
	return IsList() ? Value.Lst->Delete(s) : false;
}

int32 GVariant::CastInt32()
{
	switch (Type)
	{
		case GV_INT32:
			return Value.Int;
		case GV_INT64:
			return Value.Int64;
		case GV_STRING:
			return Value.String ? atoi(Value.String) : 0;
	}
	
	return 0;
}

char *GVariant::CastString()
{
	static char i[40];
	switch (Type)
	{
		case GV_INT32:
		{
			sprintf(i, "%i", Value.Int);
			return i;
		}
		case GV_INT64:
		{
			#ifdef WIN32
			sprintf(i, "%I64i", Value.Int64);
			#else
			sprintf(i, "%Li", Value.Int64);
			#endif			
			return i;
		}
		case GV_STRING:
			return Value.String;
	}
	
	return 0;
}

/////////////////////////////////////////////////////////////////////////////////
GDom *GDom::ResolveObject(char *Var, char *Name, char *Array)
{
	GDom *Object = this;

	// Tokenise the string
	GArray<char*> t;
	for (char *s = Var; s AND *s; )
	{
		char *e = s;
		while (*e AND *e != '.')
		{
			if (*e == '[')
			{
				e++;
				while (*e AND *e != ']')
				{
					if (*e == '\"' OR *e == '\'')
					{
						char d = *e++;
						while (*e AND *e != d)
						{
							e++;
						}
						if (*e == d) e++;
					}
					else e++;
				}
				if (*e == ']') e++;
			}
			else e++;
		}
		
		int Len = e - s;
		if (Len < 1 OR Len > 255) break;
		
		char Part[256];
		memcpy(Part, s, Len);
		Part[Len] = 0;
		while (strchr(" \t", Part[Len-1]))
		{
			Len--;
		}

		t[t.Length()] = NewStr(Part, Len);
		s = *e ? e + 1 : e;
	}

	// Process elements
	for (int i=0; i<t.Length(); i++)
	{
		char Base[256];
		char Index[256] = "";

		// Parse out object name parts
		char *Obj = t[i];
		char *ArrStart = strchr(Obj, '[');
		if (ArrStart)
		{
			*ArrStart++ = 0;
			char *ArrEnd = 0;
			if (strchr("\'\"", *ArrStart))
			{
				ArrEnd = strchr(ArrStart + 1, ArrStart[0]);
				ArrStart++;
			}
			else
			{
				ArrEnd = strchr(ArrStart, ']');
			}
			if (ArrEnd)
			{
				*ArrEnd = 0;
				strcpy(Index, ArrStart);
			}
		}
		strcpy(Base, Obj);

		// Process parts
		if (i == t.Length()-1)
		{
			strcpy(Name, Base);
			strcpy(Array, Index);
			
			char *e = Name + strlen(Name) - 1;
			while (e > Name AND strchr(" \t\r\n", *e))
			{
				*e-- = 0;
			}
		}
		else
		{
			GVariant v;
			if (Index[0])
			{
				if (Object->GetVariant(Base, v, Index))
				{
					if (v.Type == GV_LIST)
					{
						int N = atoi(Index);
						GVariant *Element = v.ItemAt(N);
						if (Element AND Element->Type == GV_DOM)
						{
							Object = Element->Value.Dom;
						}
						else
						{
							Object = 0;
							goto ResolveDone;
						}
					}
					else if (v.Type == GV_DOM)
					{
						Object = v.Value.Dom;
					}
					else
					{
						Object = 0;
						goto ResolveDone;
					}
				}
				else
				{
					Object = 0;
					goto ResolveDone;
				}
			}
			else
			{
				if (Object->GetVariant(Obj, v) AND
					v.Type == GV_DOM)
				{
					Object = v.Value.Dom;
				}
				else
				{
					Object = 0;
					goto ResolveDone;
				}
			}
		}
	}

ResolveDone:
	t.DeleteArrays();
	return Object;
}

bool GDom::GetValue(char *Var, GVariant &Value)
{
	if (Var)
	{
		char Name[256], Arr[256] = "";
		GDom *Object = ResolveObject(Var, Name, Arr);
		if (Object)
		{
			return Object->GetVariant(Name, Value, ValidStr(Arr) ? Arr : 0);
		}
	}

	return false;
}

bool GDom::SetValue(char *Var, GVariant &Value)
{
	if (Var)
	{
		char Name[256], Arr[256] = "";
		GDom *Object = ResolveObject(Var, Name, Arr);
		if (Object)
		{
			return Object->SetVariant(Name, Value, ValidStr(Arr) ? Arr : 0);
		}
	}

	return false;
}

