/** \file
	\author Matthew Allen
 */

#include "Lgi.h"

/** \brief The main entry point of a Lgi program

	To hide the differences between the different OS's standard entry points for programs
	the file LgiMain.cpp implements a simple platform specific entry point and then calls
	LgiMain() to pass execution onto your program.

	A simple LgiMain() looks like this:
	\code
	int LgiMain(OsAppArguments &Args)
	{
		GApp *App = new GApp("application/x-MyProgram", Args);
		if (App && App->IsOk())
		{
			App->AppWnd = new MyWindow;
			App->Run();
			delete App;
		}

		return 0;
	}
	\endcode
*/
extern int LgiMain
(
	/// The arguments passed in from the OS.
	OsAppArguments &AppArgs
);

bool _BuildCheck()
{
	#ifdef _DEBUG
	char *AppBuild = "Debug";
	#else
	char *AppBuild = "Release";
	#endif
	char *LgiBuild = (char*) (LgiIsReleaseBuild() ? "Release" : "Debug");

	if (stricmp(AppBuild, LgiBuild))
	{
		return LgiMsg(	NULL,
						"This application and it's Dll's are mismatched:\n"
						"\n"
						"   Application:\t%s\n"
						"   Lgi:\t\t%s\n"
						"\n"
						"If you continue you may encounter crashes and other undesirable problems.\n"
						"Do you want to continue anyway?",
						"Warning",
						MB_YESNO,
						AppBuild,
						LgiBuild) == IDYES;
	}

	return true;
}

#if defined WIN32

int
WINAPI
WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	int Status = 0;
	if (_BuildCheck())
	{
		OsAppArguments AppArgs;
		AppArgs.hInstance = hInstance;

		char16 *CL = GetCommandLineW();
		if (*CL == '\"')
		{
			AppArgs.lpCmdLine = StrchrW(CL+1, '\"');
		}
		else
		{
			AppArgs.lpCmdLine = StrchrW(CL+1, ' ');
		}
		if (AppArgs.lpCmdLine) AppArgs.lpCmdLine++;

		AppArgs.nCmdShow = nCmdShow;
		Status = LgiMain(AppArgs);
	}
	return Status;
}

#else

int main(int Args, char **Arg)
{
	int Status = 0;
	if (_BuildCheck())
	{
		OsAppArguments AppArgs;
		AppArgs.Args = Args;
		AppArgs.Arg = Arg;
		Status = LgiMain(AppArgs);
	}

	return Status;
}

#endif

