#include "Mfs.h"
#include "MfsPriv.h"

//////////////////////////////////////////////////////////////////////////////////////
MFileSystem::MFileSystem(char *File)
{
	d = NEW(MFileSystemPrivate);
	if (File)
	{
		OpenSystem(File);
	}

	LgiAssert(sizeof(MSuperNode) == d->BlockSize);
}

MFileSystem::~MFileSystem()
{
	CloseSystem();
	DeleteObj(d);
}

bool MFileSystem::CreateNewSystem(char *FileName)
{
	bool Status = false;
	
	if (FileName AND
		d->File.Open(FileName, O_READWRITE))
	{
		d->File.SetSize(0);
		
		// Write super
		MSuperNode s;
		ZeroObj(s);
		s.Magic = MFS_SUPER_BLOCK;
		strcpy(s.Id, MfsSystemId);
		s.BlockSize = 1024;

		int64 Pos = 0;
		d->WriteAt(Pos, &s, d->BlockSize);
		
		Status = true;
	}
	
	return Status;
}

bool MFileSystem::OpenSystem(char *FileName)
{
	bool Status = false;
	
	if (FileName)
	{
		if (d->File.Open(FileName, O_READWRITE))
		{
			Status = true;
		}
		else if (d->File.Open(FileName, O_READ))
		{
			d->ReadOnly = true;
			Status = true;
		}
	}
	
	return Status;
}

bool MFileSystem::CreateFolder(char *Path)
{
	return false;
}

bool MFileSystem::DeleteFolder(char *Path)
{
	return false;
}

bool MFileSystem::CloseSystem()
{
	bool Status = false;

	if (d->File.IsOpen())
	{
		Commit();
		d->File.Close();
		Status = true;
	}

	return Status;
}

char *MFileSystem::GetCurrentFile()
{
	return d->File.GetName();
}

typedef GArray<MTreeNode*> NodeArr;

bool MFileSystem::Commit()
{
	bool Status = false;
	
	// Make all the dirty node's parents dirty as well
	int i, Len = d->Dirty.Length();
	for (i=0; i<Len; i++)
	{
		for (MTreeNode *n=d->Dirty[i]; n; n=n->Parent)
		{
			if (NOT n->Dirty)
			{
				n->SetDirty();
			}
			else break;
		}
	}
	
	// Build arrays by depth so that we can write out the deepest
	// nodes first, moving up the tree until the super block is
	// written.
	GArray<NodeArr*> a;
	NodeArr *na;
	for (i=0; i<d->Dirty.Length(); i++)
	{
		MTreeNode *n = d->Dirty[i];
		if (n->Depth >= a.Length())
		{
			a[n->Depth] = na = NEW(NodeArr);
		}
		else
		{
			na = a[n->Depth];
		}
		if (na)
		{
			(*na)[na->Length()] = n;
		}
	}
	
	// Write the nodes out lowest first, all the way up to the super
	// block. Thus completing the transaction.
	for (int Depth=a.Length()-1; Depth>=0; Depth--)
	{
		if (na = a[Depth])
		{
			for (i=0; i<na->Length(); i++)
			{
				(*na)[i]->SetClean();
			}
		}
	}
	
	// Clean up the memory
	a.DeleteObjects();
	d->Dirty.Length(0);
	
	return Status;
}

bool MFileSystem::DeleteFile(char *File)
{
	bool Status = false;
	
	// Loop through all the meta data, deleting the indexed data
	
	// Delete the UID entry
	
	// Commit it all...
	
	return Status;
}

bool MFileSystem::MoveFile(char *To, char *From, int64 *Progress)
{
	bool Status = false;
	return Status;
}

bool MFileSystem::CopyFile(char *To, char *From, int64 *Progress)
{
	bool Status = false;
	return Status;
}

//////////////////////////////////////////////////////////////////////////////////////
MFileSystemPrivate::MFileSystemPrivate()
{
	Uid = 0;
	BlockBmp = 0;
	BlockSize = 1024;
	BlockShift = 10;
}

MFileSystemPrivate::~MFileSystemPrivate()
{
	DeleteObj(Uid);
	DeleteObj(BlockBmp);
	Indexes.DeleteObjects();
}

int MFileSystemPrivate::ReadAt(int64 &Pos, void *Buf, int Len, int Flags)
{
	int r = 0;

	if (FileLock.Lock())
	{
		File.SetPos(Pos);
		r = File.Read(Buf, Len, Flags);
		Pos = File.GetPos();
		FileLock.Unlock();
	}

	return r;
}

int MFileSystemPrivate::WriteAt(int64 &Pos, void *Buf, int Len, int Flags)
{
	int w = 0;

	if (FileLock.Lock())
	{
		File.SetPos(Pos);
		w = File.Write(Buf, Len, Flags);
		Pos = File.GetPos();
		FileLock.Unlock();
	}

	return w;
}

bool MFileSystemPrivate::AllocateBlocks(GArray<uint32> &Blocks, int Num)
{
	return false;
}
