#include "Mfs.h"
#include "MfsPriv.h"

MTreeNode::MTreeNode(MFileSystem *sys, uint32 block)
{
	Sys = sys;
	Left = Right = Parent = 0;
	Dirty = false;
	Block = block;
	Depth = 0;
	Data = NEW(char[Sys->d->BlockSize]);
	if (Data)
	{
		Load();
	}
}

MTreeNode::~MTreeNode()
{
	DeleteObj(Left);
	DeleteObj(Right);
	DeleteArray(Data);
}

void MTreeNode::SetDirty()
{
	if (NOT Dirty)
	{
		Dirty = true;
		Sys->d->Dirty[Sys->d->Dirty.Length()] = this;
	}
}

bool MTreeNode::SetClean()
{
	if (Dirty)
	{
		if (Parent)
		{
			// Pick new location...
			GArray<uint32> b;
			if (Sys->d->AllocateBlocks(b))
			{
				// Update the parent block with our new location
				if (Parent->OnChildMoved(this, b[0]))
				{
					// Save to the new location
					if (Save())
					{
						Dirty = false;
						return true;
					}
					else
					{
						LgiAssert(0);
					}
				}
				else
				{
					LgiAssert(0);
				}
			}
			else
			{
				LgiAssert(0);
			}
		}
		else
		{
			// I'm the super block, and thus never move location.
			return true;
		}
	}
	
	return false;
}

bool MTreeNode::Load(uint32 block)
{
	if (Data)
	{
		if (block)
		{
			Block = block;
		}

		int64 Pos = (int64)Block << Sys->d->BlockShift;
		return Sys->d->ReadAt(Pos, Data, Sys->d->BlockSize)
			==
			Sys->d->BlockSize;
	}

	return false;
}

bool MTreeNode::Save()
{
	if (Data)
	{
		int64 Pos = (int64)Block << Sys->d->BlockShift;
		return Sys->d->WriteAt(Pos, Data, Sys->d->BlockSize)
			==
			Sys->d->BlockSize;
	}

	return false;
}

bool MTreeNode::Walk(NodeWalker Callback, int Data)
{
	if (Left)
	{
		if (NOT Left->Walk(Callback, Data))
		{
			return false;
		}
	}

	if (NOT Callback(this, Data))
	{
		return false;
	}

	if (Right)
	{
		if (NOT Right->Walk(Callback, Data))
		{
			return false;
		}
	}

	return true;
}

bool MTreeNode::OnChildMoved(MTreeNode *Child, uint32 NewPos)
{
	bool Status = false;
	
	return Status;	
}
