// Basic tokeniser

#include <string.h>
#include "GMem.h"
#include "GContainers.h"
#include "GToken.h"
#include "GString.h"

/////////////////////////////////////////////////////////////////////////
bool IsNumber(char *p)
{
	if (p)
	{
		for (uint i=0; i<strlen(p); i++)
		{
			if ((p[i]<'0' OR p[i]>'9') AND p[i]!='-')
			{
				return false;
			}
		}
		return true;
	}
	return false;
}

char *Skip(char *p, char *Delimiter)
{
	while (p AND *p AND strchr(Delimiter, *p)) p++;
	return p;
}

////////////////////////////////////////////////////////////////////////
GToken::GToken()
{
	Item = 0;
	Items = 0;
	Raw = 0;
}

GToken::GToken(char *Str, char *Delimiters, bool GroupDelim, int Length)
{
	Item = 0;
	Items = 0;
	Raw = 0;
	Parse(Str, Delimiters, GroupDelim, Length);
}

GToken::~GToken()
{
	Empty();
}

void GToken::Parse(char *Str, char *Delimiters, bool GroupDelim, int Length)
{
	Empty();
	if (Str)
	{
		List<char> l;

		// Prepare lookup table
		bool Lut[256];
		ZeroObj(Lut);
		if (Delimiters)
		{
			for (char *d=Delimiters; *d; d++)
			{
				Lut[(uchar)*d] = true;
			}
		}

		// Do tokenisation
		Raw = NewStr(Str, Length);
		if (Raw)
		{
			if (Delimiters)
			{
				if (GroupDelim)
				{
					for (char *s=Raw; *s;)
					{
						// Skip Delimiters
						while (*s && Lut[(uchar)*s]) *s++ = 0;

						// Emit pointer
						if (*s)
						{
							l.Insert(s);
						}

						// Skip Non-delimiters
						while (*s && !Lut[(unsigned char)*s]) s++;

					}
				}
				else
				{
					for (char *s=Raw; *s;)
					{
						// Emit pointer
						if (*s)
						{
							l.Insert(Lut[(uchar)*s] ? (char*)-1 : s);
						}

						// Skip Non-delimiters
						while (*s && !Lut[(unsigned char)*s]) s++;

						// Skip Delimiter
						if (*s && Lut[(uchar)*s]) *s++ = 0;
					}
				}
			}
			else
			{
				l.Insert(Raw);
			}
		}

		Items = l.GetItems();
		if (Items > 0)
		{
			Item = NEW(char*[Items]);
			if (Item)
			{
				int i=0;
				for (char *t=l.First(); t; t=l.Next())
				{
					if (t == (char*)-1)
					{
						Item[i++] = 0;
					}
					else
					{
						Item[i++] = t;
					}
				}
			}
		}
	}
}

void GToken::Empty()
{
	DeleteArray(Raw);
	DeleteArray(Item);
	Items = 0;
}

void GToken::AppendTokens(GToken *T)
{
	if (T)
	{
		int Total = T->Items + Items;
		char **All = NEW(char*[Total]);
		if (All)
		{
			int i = 0;
			for (; i<Items; i++)
			{
				All[i] = Item[i];
			}
			for (; i<Total; i++)
			{
				All[i] = NewStr(T->Item[i-Items]);
			}
			DeleteArray(Item);
			Item = All;
			Items = Total;
		}
	}
}



