#include <stdlib.h>
#include <stdio.h>

#include "Lgi.h"
#include "GWordStore.h"
// #define DBG_MSGS

GWordStore::GWordStore(char *file)
{
	Dirty = false;
	Table.IsCase(false);
	Items = 0;
	File = NewStr(file);
	Table.SetStringPool(true);
	if (File)
	{
		Serialize(File, true);
	}
}

GWordStore::~GWordStore()
{
	if (Dirty)
	{
		Serialize(File, false);
	}
	DeleteArray(File);
}

bool GWordStore::Serialize(char *FileName, bool Load)
{
	static const char *ObjName = "GWordStore";
	bool Status = false;
	GFile f;
	
	char *Name = FileName ? FileName : File;
	if (Load)
	{
		if (f.Open(Name, O_READ))
		{
			if (NOT File)
			{
				File = NewStr(Name);
			}

			bool GotVer = false;
			double FileFormat = 0.0;
			int64 Start = LgiCurrentTime();
			
			GStringPipe p;
			char Buf[4<<10];
			int Words = 0;
			while (NOT f.Eof())
			{
				int r = f.Read(Buf, sizeof(Buf));
				if (r > 0)
				{
					p.Push(Buf, r);

					while (p.Pop(Buf, sizeof(Buf)))
					{
						if (GotVer)
						{
							char *Comma = strchr(Buf, ',');
							if (Comma)
							{
								*Comma++ = 0;
								Table.Add(Buf, (void*) atoi(Comma));
							}
							else
							{
								LgiAssert(0);
							}
						}
						else
						{
							char *Obj;
							if (Obj = stristr(Buf, (char*)ObjName))
							{
								char *v = strchr(Buf + 10, 'v');
								if (v)
								{
									FileFormat = atof(v + 1);
									GotVer = true;
								}
								
								char *i = stristr(Buf, "Items=");
								if (i)
								{
									Items = atoi(i+6);
								}

								i = stristr(Buf, "Words=");
								if (i)
								{
									Words = atoi(i+6);
									Table.SetSize(Words * 5 / 2);
								}
							}
						}
					}
				}
				else break;
			}

			#if DBG_MSGS
			LgiTrace("Load '%s' time: %ims\n", Name, LgiCurrentTime() - Start);
			#endif
			
			Status = GotVer;
		}
	}	
	else
	{
		if (f.Open(Name, O_WRITE))
		{
			f.SetSize(0);
			
			f.Print("%s v1.00 Items=%i, Words=%i\n", ObjName, Items, Table.Length());
			
			int i = 0;
			char *Word;
			for (void *Count = Table.First(&Word); Count; Count = Table.Next(&Word))
			{
				i++;
				f.Print("%s,%i\n", Word, (int)Count);
			}

			#if DBG_MSGS
			LgiTrace("Write to '%s': %i words in %i lines\n", Name, Table.Length(), i);
			#endif
			
			Status = true;
		}
	}
	
	if (Status)
	{
		Dirty = false;
	}
	
	return Status;
}

int GWordStore::GetItems()
{
	return Items;
}

void GWordStore::SetItems(int i)
{
	if (i != Items)
	{
		Items = i;
		Dirty = true;
	}
}

bool GWordStore::Insert(char *Word)
{
	bool Status = false;
	
	if (Word)
	{
		int c = (int)Table.Find(Word);
		Status = Table.Add(Word, (void*) (c + 1));
		Dirty = true;
	}
	
	return Status;

}

int GWordStore::GetWordCount(char *Word)
{
	return Word ? (int)Table.Find(Word) : 0;
}

void GWordStore::Empty()
{
	Table.Empty();
	Dirty = true;
}

#ifdef _DEBUG
int64 GWordStore::Sizeof()
{
	return sizeof(*this) + Table.Sizeof();
}
#endif