#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>

#include "Lgi.h"
#include "vCard-vCal.h"

/////////////////////////////////////////////////////////////
// General IO class
VIo::~VIo()
{
	Object.DeleteObjects();
}

bool VIo::IsBeginEnd()
{
	if (Var.Length() == 2)
	{
		if (stricmp(Var[0], "begin") == 0)
		{
			Object.Insert(NewStr(Var[1]));
			return true;
		}
		else if (stricmp(Var[0], "end") == 0)
		{
			char *o=Object.Last();
			if (o AND stricmp(o, Var[1]) == 0)
			{
				Object.Delete(o);
				DeleteArray(o);
			}
			return true;
		}
	}

	return false;
}

bool VIo::IsObject(char *Obj)
{
	if (Obj)
	{
		for (char *o=Object.Last(); o; o=Object.Prev())
		{
			if (stricmp(Obj, o) == 0)
			{
				return true;
			}
		}
	}

	return false;
}

bool VIo::IsTyped()
{
	Type.Empty();
	Value.Empty();
	Variables.Empty();

	char *Semi = Var[0] ? strchr(Var[0], ';') : 0;
	if (Semi)
	{
		GToken V(Var[0], ";", false);
		*Semi++ = 0;

		if (strchr(Semi, '='))
		{
			for (int i=0; i<V.Length(); i++)
			{
				if (strnicmp(V[i], "type=", 5) == 0)
				{
					GToken T(V[i]+5, ",");
					Type.AppendTokens(&T);
				}
				else if (strnicmp(V[i], "value=", 5) == 0)
				{
					GToken T(V[i]+6, ",");
					Value.AppendTokens(&T);
				}
				else
				{
					char *v = V[i];
					char *e = strchr(v, '=');
					if (e)
					{
						*e++ = 0;
						if (strchr("\'\"", *e))
						{
							char d = *e++;
							char *End = strchr(e, d);
							if (End) *End = 0;
						}
						Variables[v] = e;
					}
				}
			}
		}
		else
		{
			Type.Parse(Semi, ";", false);
		}

		return true;
	}

	return false;
}

bool VIo::IsEncoded()
{
	if (IsType("quoted-printable"))
	{
		if (Var[1])
		{
			char *In = Var[1];
			char *Out = In;
			while (*In)
			{
				if (*In == '=' AND In[1])
				{
					char Hex[3] = {In[1], In[2], 0};
					*Out++ = htoi(Hex);
					In += 3;
				}
				else
				{
					*Out++ = *In++;
				}
			}
			*Out++ = 0;
		}

		return true;
	}

	return false;
}

bool VIo::IsVar(char *a, char *b)
{
	if (Var[0] AND a)
	{
		if (stricmp(a, Var[0]) == 0)
		{
			if (Var[1] AND b)
			{
				return stricmp(Var[1], b) == 0;
			}
			else
			{
				return true;
			}
		}
	}

	return false;
}

bool VIo::IsType(char *type)
{
	for (int t=0; t<Type.Length(); t++)
	{
		if (type AND stricmp(Type[t], type) == 0)
		{
			return true;
		}
	}

	return false;
}

bool VIo::IsValue(char *value)
{
	for (int i=0; i<Value.Length(); i++)
	{
		if (value AND stricmp(Value[i], value) == 0)
		{
			return true;
		}
	}

	return false;
}

bool VIo::ParseDate(GDateTime &Out, char *In)
{
	bool Status = false;

	if (In)
	{
		GToken v(In, "T");
		if (v.Length() > 0)
		{
			char *d = v[0];
			if (d AND strlen(d) == 8)
			{
				char Year[5] = {d[0], d[1], d[2], d[3], 0};
				char Month[3] = {d[4], d[5], 0};
				char Day[3] = {d[6], d[7], 0};
				Out.Year(atoi(Year));
				Out.Month(atoi(Month));
				Out.Day(atoi(Day));
				Status = true;
			}

			char *t = v[1];
			if (t AND strlen(t) >= 6)
			{
				char Hour[3] = {t[0], t[1], 0};
				char Minute[3] = {t[2], t[3], 0};
				char Second[3] = {t[4], t[5], 0};
				Out.Hours(atoi(Hour));
				Out.Minutes(atoi(Minute));
				Out.Seconds(atoi(Second));
				Status = true;
			}
		}
	}

	return Status;
}

bool VIo::ParseDuration(GDateTime &Out, int &Sign, char *In)
{
	bool Status = false;

	if (In)
	{
		Sign = 1;
		if (*In == '-')
		{
			Sign = -1;
			In++;
		}

		if (toupper(*In++) == 'P' AND
			toupper(*In++) == 'T')
		{
			while (*In)
			{
				int i = atoi(In);
				while (isdigit(*In)) In++;

				switch (toupper(*In++))
				{
					case 'W':
					{
						Out.Day(Out.Day() + (i * 7));
						break;
					}
					case 'D':
					{
						Out.Day(Out.Day() + i);
						break;
					}
					case 'H':
					{
						Out.Hours(Out.Hours() + i);
						break;
					}
					case 'M':
					{
						Out.Minutes(Out.Minutes() + i);
						break;
					}
					case 'S':
					{
						Out.Seconds(Out.Seconds() + i);
						break;
					}
				}
			}

			Status = true;
		}
	}

	return Status;
}

void VIo::Fold(GStringPipe &o, char *i)
{
	int x = 0;
	for (char *s=i; s AND *s;)
	{
		if (x >= 74)
		{
			o.Push(i, (int)s-(int)i);
			o.Push("\n\t");
			x = 0;
			i = s;
		}
		else if (*s == '\n')
		{
			o.Push(i, (int)s-(int)i);
			o.Push("\\n");
			x += 2;
			i = ++s;
		}
		else
		{
			s++;
			x++;
		}
	}
	o.Push(i);
}

char *VIo::Unfold(char *In)
{
	if (In)
	{
		GStringPipe p;

		char *n;
		for (char *i=In; i AND *i; i=n)
		{
			n = strchr(i, '\n');
			if (n)
			{
				bool Un = n[1] AND strchr(" \t", n[1]);
				if (Un)
				{
					p.Push(i, (int)n-(int)i-1);
					n += 2;
				}
				else
				{
					n++;
					p.Push(i, (int)n-(int)i);
				}
			}
			else
			{
				p.Push(i);
			}
		}

		return p.NewStr();
	}

	return 0;
}

char *VIo::UnMultiLine(char *In)
{
	if (In)
	{
		GStringPipe p;
		char *n;
		for (char *i=In; i AND *i; i=n)
		{
			n = stristr(i, "\\n");
			if (n)
			{
				p.Push(i, (int)n-(int)i);
				p.Push("\n");
				n += 2;
			}
			else
			{
				p.Push(i);
			}
		}

		return p.NewStr();
	}

	return 0;
}

/////////////////////////////////////////////////////////////
// VCard class
bool VCard::Import(char *Text)
{
	bool Status = false;
	if (Text)
	{
		char *Unfolded = Unfold(Text);
		GToken L(Unfolded, "\r\n");
		if (L.Length() > 0)
		{
			for (int i=0; i<L.Length(); i++)
			{
				char *Line = L[i];
				if (NOT Line) break;
				char *Eol = Line+strlen(Line)-1;
				while (Eol > Line AND strchr(" \t", *Eol)) *Eol-- = 0;

				Var.Empty();

				char *Colon = strchr(L[i], ':');
				if (Colon)
				{
					*Colon++ = 0;
					GToken Temp(L[i], 0);
					Var.AppendTokens(&Temp);
					Temp.Parse(Colon, 0);
					Var.AppendTokens(&Temp);

					if (Var.Length() > 0)
					{
						if (NOT IsBeginEnd() AND
							IsObject("vcard"))
						{
							IsTyped();
							IsEncoded();

							if (IsVar("n"))
							{
								GToken Name(Var[1], ";", false);
								char *First = Name[1];
								char *Last = Name[0];
								char *Title = Name[3];

								if (First)
								{
									c->SetField(FIELD_FIRST_NAME, First);
									Status = true;
								}

								if (Last)
								{
									c->SetField(FIELD_LAST_NAME, Last);
									Status = true;
								}

								if (Title)
								{
									c->SetField(FIELD_TITLE, Title);
									Status = true;
								}
							}
							else if (IsVar("nick"))
							{
								GToken Nick(Var[1], ";", false);
								c->SetField(FIELD_NICK, Nick[0]);
							}
							else if (IsVar("tel"))
							{
								GToken Phone(Var[1], ";", false);
								for (int p=0; p<Phone.Length(); p++)
								{
									if (IsType("cell"))
									{
										if (IsType("Work"))
										{
											c->SetField(FIELD_WORK_MOBILE, Phone[p]);
										}
										else
										{
											c->SetField(FIELD_HOME_MOBILE, Phone[p]);
										}
									}
									else if (IsType("fax"))
									{
										if (IsType("Work"))
										{
											c->SetField(FIELD_WORK_FAX, Phone[p]);
										}
										else
										{
											c->SetField(FIELD_HOME_FAX, Phone[p]);
										}
									}
									else
									{
										if (IsType("Work"))
										{
											c->SetField(FIELD_WORK_PHONE, Phone[p]);
										}
										else
										{
											c->SetField(FIELD_HOME_PHONE, Phone[p]);
										}
									}
								}
							}
							else if (IsVar("email"))
							{
								GToken Email(Var[1], ";", false);
								for (int e=0; e<Email.Length(); e++)
								{
									if (IsType("internet"))
									{
										c->SetField(FIELD_EMAIL_ADDR, Email[e]);
									}
								}
							}
							else if (IsVar("org"))
							{
								GToken Org(Var[1], ";", false);
								if (Org[0]) c->SetField(FIELD_COMPANY, Org[0]);
							}
							else if (IsVar("adr"))
							{
								bool IsWork = IsType("work");
								bool IsHome = IsType("home");
								GToken Addr(Var[1], ";", false);
								if (Addr[2])
								{
									GToken A(Addr[2], "\r\n");
									if (A.Length() > 1)
									{
										c->SetField(IsWork ? FIELD_WORK_STREET : FIELD_HOME_STREET, A[0]);
										if (A[1]) c->SetField(IsWork ? FIELD_WORK_SUBURB : FIELD_HOME_SUBURB, A[1]);
										if (A[2]) c->SetField(IsWork ? FIELD_WORK_COUNTRY : FIELD_HOME_COUNTRY, A[2]);
									}
									else
									{
										c->SetField(IsWork ? FIELD_WORK_STREET : FIELD_HOME_STREET, Addr[2]);
									}
								}

								if (Addr[3]) c->SetField(IsWork ? FIELD_WORK_SUBURB : FIELD_HOME_SUBURB, Addr[3]);
								if (Addr[4]) c->SetField(IsWork ? FIELD_WORK_STATE : FIELD_HOME_STATE, Addr[4]);
								if (Addr[5]) c->SetField(IsWork ? FIELD_WORK_POSTCODE : FIELD_HOME_POSTCODE, Addr[5]);
								if (Addr[6]) c->SetField(IsWork ? FIELD_WORK_COUNTRY : FIELD_HOME_COUNTRY, Addr[6]);
							}
							else if (IsVar("note"))
							{
								GToken n(Var[1], ";", false);
								c->SetField(FIELD_NOTE, n[0]);
							}
							else if (IsVar("uid"))
							{
								GToken n(Var[1], ";", false);
								char *Uid = n[0];
								if (isdigit(*Uid))
								{
									c->SetField(FIELD_UID, atoi(n[0]));
								}
								else
								{
									c->SetField(FIELD_UID, n[0]);
								}
							}
							else if (IsVar("x-perm"))
							{
								int Perms = atoi(Var[1]);
								c->SetField(FIELD_PERMISSIONS, Perms);
							}
							else if (IsVar("url"))
							{
								bool IsWork = IsType("work");
								bool IsHome = IsType("home");
								if (IsWork)
								{
									c->SetField(FIELD_HOME_WEBPAGE, Var[i]);
								}
								else if (IsHome)
								{
									c->SetField(FIELD_WORK_WEBPAGE, Var[i]);
								}
								else
								{
									// ???
								}								
							}
						}
					}
				}
			}
		}

		DeleteArray(Unfolded);
	}

	return Status;
}

char *VCard::Export()
{
	GStringPipe o;
	char s[512];

	#define OutputTypedField(Field, Name, Type)					\
		{ char *d = 0;											\
		if (c->GetField(Name, d))								\
		{														\
			sprintf(s, Field ";type=" Type ":%s\r\n", d);		\
			o.Push(s);											\
		} }

	#define OutputField(Field, Name)							\
		{ char *d = 0;											\
		if (c->GetField(Name, d))								\
		{														\
			sprintf(s, Field ":%s\r\n", d);						\
			o.Push(s);											\
		} }

	o.Push("begin:vcard\r\n");
	o.Push("version:3.0\r\n");

	char *First = 0, *Last = 0, *Title = 0;
	c->GetField(FIELD_FIRST_NAME, First);
	c->GetField(FIELD_LAST_NAME, Last);
	c->GetField(FIELD_TITLE, Title);
	if (First OR Last)
	{
		strcpy(s, "n:");
		if (Last)
		{
			strcat(s, Last);
		}
		if (First)
		{
			if (Last) strcat(s, ";");
			strcat(s, First);
		}
		if (Title)
		{
			strcat(s, ";;");
			strcat(s, Title);
		}
		strcat(s, "\r\n");
		o.Push(s);
	}

	OutputTypedField("tel", FIELD_WORK_PHONE, "Work");
	OutputTypedField("tel", FIELD_WORK_MOBILE, "Work,Cell");
	OutputTypedField("tel", FIELD_WORK_FAX, "Work,Fax");
	OutputTypedField("tel", FIELD_HOME_PHONE, "Home");
	OutputTypedField("tel", FIELD_HOME_MOBILE, "Home,Cell");
	OutputTypedField("tel", FIELD_HOME_FAX, "Home,Fax");
	OutputField("org", FIELD_COMPANY);
	OutputTypedField("email", FIELD_EMAIL_ADDR, "internet");

	int Uid;
	if (c->GetField(FIELD_UID, Uid))
	{
		sprintf(s, "UID:%i\r\n", Uid);
		o.Push(s);
	}

	char *Street, *Suburb, *PostCode, *State, *Country;
	Street = Suburb = PostCode = State = Country = 0;
	c->GetField(FIELD_HOME_STREET, Street);
	c->GetField(FIELD_HOME_SUBURB, Suburb);
	c->GetField(FIELD_HOME_POSTCODE, PostCode);
	c->GetField(FIELD_HOME_STATE, State);
	c->GetField(FIELD_HOME_COUNTRY, Country);
	if (Street OR Suburb OR PostCode OR State OR Country)
	{
		char *Empty = "";
		sprintf(s,
				"adr;type=home;encoding=quoted-printable:;;%s;%s;%s;%s;%s\r\n",
				Street?Street:Empty,
				Suburb?Suburb:Empty,
				State?State:Empty,
				PostCode?PostCode:Empty,
				Country?Country:Empty);
		o.Push(s);
	}

	Street = Suburb = PostCode = State = Country = 0;
	c->GetField(FIELD_WORK_STREET, Street);
	c->GetField(FIELD_WORK_SUBURB, Suburb);
	c->GetField(FIELD_WORK_POSTCODE, PostCode);
	c->GetField(FIELD_WORK_STATE, State);
	c->GetField(FIELD_WORK_COUNTRY, Country);
	if (Street OR Suburb OR PostCode OR State OR Country)
	{
		char *Empty = "";
		sprintf(s,
				"adr;type=work;encoding=quoted-printable:;;%s;%s;%s;%s;%s\r\n",
				Street?Street:Empty,
				Suburb?Suburb:Empty,
				State?State:Empty,
				PostCode?PostCode:Empty,
				Country?Country:Empty);
		o.Push(s);
	}

	OutputField("X-Perm", FIELD_PERMISSIONS);

	char *Url;
	if (c->GetField(FIELD_HOME_WEBPAGE, Url))
	{
		sprintf(s, "url;type=home:%s\r\n", Url);
		o.Push(s);
	}
	if (c->GetField(FIELD_WORK_WEBPAGE, Url))
	{
		sprintf(s, "url;type=work:%s\r\n", Url);
		o.Push(s);
	}

	o.Push("end:vcard\r\n");

	return o.NewStr();
}

/////////////////////////////////////////////////////////////
// VCal class
bool VCal::Import(char *Text)
{
	bool Status = false;
	if (Text)
	{
		char *Unfolded = Unfold(Text);
		GToken L(Unfolded, "\r\n");
		if (L.Length() > 0)
		{
			for (int i=0; i<L.Length(); i++)
			{
				char *Line = L[i];
				if (NOT Line) break;
				char *Eol = Line+strlen(Line)-1;
				while (Eol > Line AND strchr(" \t", *Eol)) *Eol-- = 0;

				Var.Empty();

				char *Colon = strchr(L[i], ':');
				if (Colon)
				{
					*Colon++ = 0;
					GToken Temp(L[i], 0);
					Var.AppendTokens(&Temp);
					Temp.Parse(Colon, 0);
					Var.AppendTokens(&Temp);

					if (NOT IsBeginEnd() AND
						IsObject("vcalendar"))
					{
						Status = true;
						IsTyped();
						IsEncoded();

						if (IsObject("vevent"))
						{
							if (IsObject("valarm"))
							{
								if (IsVar("trigger"))
								{
									if (IsValue("date-time"))
									{
										// don't really support this...
									}
									else
									{
										c->SetField(FIELD_CAL_REMINDER_ACTION, true);
										c->SetField(FIELD_CAL_REMINDER_TIME, 0); // reasonable default

										GDateTime Dt;
										int Sign = 0;
										if (ParseDuration(Dt, Sign, Var[1]))
										{
											int d = Dt.Day() * 24 * 60;
											int h = Dt.Hours() * 60;
											int m = Dt.Minutes();
											c->SetField(FIELD_CAL_REMINDER_TIME, -(d + h + m)); 
										}
									}
								}
							}
							else
							{
								if (IsVar("dtstart"))
								{
									GDateTime d;
									if (ParseDate(d, Var[1]))
									{
										c->SetField(FIELD_CAL_START, d);
									}
								}
								else if (IsVar("dtend"))
								{
									GDateTime d;
									if (ParseDate(d, Var[1]))
									{
										c->SetField(FIELD_CAL_END, d);
									}
								}
								else if (IsVar("summary"))
								{
									c->SetField(FIELD_CAL_SUBJECT, Var[1]);
								}
								else if (IsVar("description"))
								{
									char *Sum = UnMultiLine(Var[1]);
									if (Sum)
									{
										c->SetField(FIELD_CAL_NOTES, Sum);
										DeleteArray(Sum);
									}
								}
								else if (IsVar("location"))
								{
									c->SetField(FIELD_CAL_LOCATION, Var[1]);
								}
								else if (IsVar("uid"))
								{
									char *Uid = Var[1];
									if (isdigit(*Uid))
									{
										c->SetField(FIELD_UID, atoi(Uid));
									}
									else
									{
										c->SetField(FIELD_UID, Uid);
									}
								}
								else if (IsVar("x-showas"))
								{
									if (stricmp(Var[1], "TENTATIVE") == 0)
									{
										c->SetField(FIELD_CAL_SHOW_TIME_AS, 1);
									}
									else if (stricmp(Var[1], "BUSY") == 0)
									{
										c->SetField(FIELD_CAL_SHOW_TIME_AS, 2);
									}
									else if (stricmp(Var[1], "OUT") == 0)
									{
										c->SetField(FIELD_CAL_SHOW_TIME_AS, 3);
									}
									else
									{
										c->SetField(FIELD_CAL_SHOW_TIME_AS, 0);
									}
								}
								else if (IsVar("attendee"))
								{
									char *e = stristr(Var[1], "mailto=");
									if (e)
									{
										e += 7;

										char *Name = Variables["CN"];
										char *Role = Variables["Role"];

										VObject *a = c->CreateAttendee();
										if (a)
										{
											a->SetField(FIELD_ATTENDEE_NAME, Name);
											a->SetField(FIELD_ATTENDEE_EMAIL, e);
											if (stricmp(Role, "CHAIR") == 0)
											{
												a->SetField(FIELD_ATTENDEE_ATTENDENCE, 1);
											}
											else if (stricmp(Role, "REQ-PARTICIPANT") == 0)
											{
												a->SetField(FIELD_ATTENDEE_ATTENDENCE, 2);
											}
											else if (stricmp(Role, "OPT-PARTICIPANT") == 0)
											{
												a->SetField(FIELD_ATTENDEE_ATTENDENCE, 3);
											}
										}
									}
								}
							}
						}
					}
				}
			}
		}

		DeleteArray(Unfolded);
	}

	return Status;
}

char *VCal::Export()
{
	GStringPipe o;
	char s[512];
	int Type = 0;
	c->GetField(FIELD_CAL_TYPE, Type);
	char *TypeStr = Type == 1 ? (char*)"VTODO" : (char*)"VEVENT";

	o.Push("BEGIN:VCALENDAR\r\n");
	o.Push("VERSION:1.0\r\n");

	if (c)
	{
		o.Print("BEGIN:%s\r\n", TypeStr);

		#define OutputStr(Field, Name)						\
		{													\
			char *_s = 0;									\
			if (c->GetField(Field, _s))						\
			{												\
				o.Push(Name ":");							\
				Fold(o, _s);								\
				o.Push("\r\n");								\
			}												\
		}

		OutputStr(FIELD_CAL_SUBJECT, "SUMMARY");
		OutputStr(FIELD_CAL_LOCATION, "LOCATION");
		OutputStr(FIELD_CAL_NOTES, "DESCRIPTION");

		int ShowAs;
		if (c->GetField(FIELD_CAL_SHOW_TIME_AS, ShowAs))
		{
			switch (ShowAs)
			{
				default:
				case 0:
				{
					o.Push("X-SHOWAS:FREE\r\n");
					break;
				}
				case 1:
				{
					o.Push("X-SHOWAS:TENTATIVE\r\n");
					break;
				}
				case 2:
				{
					o.Push("X-SHOWAS:BUSY\r\n");
					break;
				}
				case 3:
				{
					o.Push("X-SHOWAS:OUT\r\n");
					break;
				}
			}
		}

		int Uid;
		if (c->GetField(FIELD_UID, Uid))
		{
			sprintf(s, "UID:%i\r\n", Uid);
			o.Push(s);
		}

		GDateTime Dt;
		if (c->GetField(FIELD_CAL_START, Dt))
		{
			sprintf(s,
					"DTSTART:%04.4i%02.2i%02.2iT%02.2i%02.2i%02.2i\r\n",
					Dt.Year(), Dt.Month(), Dt.Day(),
					Dt.Hours(), Dt.Minutes(), Dt.Seconds());
			o.Push(s);
		}
		if (c->GetField(FIELD_CAL_END, Dt))
		{
			sprintf(s,
					"DTEND:%04.4i%02.2i%02.2iT%02.2i%02.2i%02.2i\r\n",
					Dt.Year(), Dt.Month(), Dt.Day(),
					Dt.Hours(), Dt.Minutes(), Dt.Seconds());
			o.Push(s);
		}

		int AlarmAction;
		int AlarmTime;
		if (c->GetField(FIELD_CAL_REMINDER_ACTION, AlarmAction) AND
			AlarmAction AND
			c->GetField(FIELD_CAL_REMINDER_TIME, AlarmTime))
		{
			o.Push("BEGIN:VALARM\r\n");
			sprintf(s, "TRIGGER:%sPT%iM\r\n", (AlarmTime<0) ? "-" : "", abs(AlarmTime));
			o.Push(s);
			o.Push("END:VALARM\r\n");
		}

		List<VObject> Attendees;
		if (c->GetAttendees(Attendees))
		{
			for (VObject *a=Attendees.First(); a; a=Attendees.Next())
			{
				// ATTENDEE;CN="Matthew Allen";ROLE=REQ-PARTICIPANT;RSVP=TRUE:MAILTO:matthew@cisra.canon.com.au

				int Attend = 2;
				char *Name = 0;
				char *Email = 0;
				int Response = 0;

				a->GetField(FIELD_ATTENDEE_ATTENDENCE, Attend);
				a->GetField(FIELD_ATTENDEE_NAME, Name);
				a->GetField(FIELD_ATTENDEE_EMAIL, Email);
				a->GetField(FIELD_ATTENDEE_RESPONSE, Response);

				if (Email)
				{
					strcpy(s, "ATTENDEE");
					if (Name)
					{
						sprintf(s+strlen(s), ";CN=\"%s\"", Name);
					}
					switch (Attend)
					{
						case 1:
						{
							strcat(s, ";ROLE=CHAIR");
							break;
						}
						default:
						case 2:
						{
							strcat(s, ";ROLE=REQ-PARTICIPANT");
							break;
						}
						case 3:
						{
							strcat(s, ";ROLE=OPT-PARTICIPANT");
							break;
						}
					}
					if (Email)
					{
						sprintf(s+strlen(s), ":MAILTO=%s", Email);
					}
					strcat(s, "\r\n");
					o.Push(s);
				}				
			}
		}

		o.Print("END:%s\r\n", TypeStr);
	}

	o.Push("END:VCALENDAR\r\n");

	return o.NewStr();
}
