#include "Lgi.h"
#include "GBitmap.h"

//////////////////////////////////////////////////////////////////////////////////
#ifdef _MT
class GBitmapThread : public GThread
{
	GBitmap *Bmp;
	char *File;
	GThread **Owner;

public:
	GBitmapThread(GBitmap *bmp, char *file, GThread **owner)
	{
		Bmp = bmp;
		File = NewStr(file);
		Owner = owner;
		if (Owner)
		{
			*Owner = this;
		}
		Run();
	}

	~GBitmapThread()
	{
		if (Owner)
		{
			*Owner = 0;
		}
		DeleteArray(File);
	}

	int Main()
	{
		if (Bmp)
		{
			GSurface *pDC = LoadDC(File);
			if (pDC)
			{
				Bmp->SetDC(pDC);
				GRect r = Bmp->GetPos();
				r.Dimension(pDC->X()+4, pDC->Y()+4);

				while (NOT Bmp->Handle())
				{
					LgiSleep(10);
				}

				Bmp->SetPos(r, true);
				Bmp->PostEvent(M_CHANGE, (int)Bmp, 0);
			}
		}

		return 0;
	}
};
#endif

GBitmap::GBitmap(int id, int x, int y, char *FileName, bool Async)
	: ResObject(Res_Bitmap)
{
	pDC = 0;
	pThread = 0;

	SetClass(LGI_BITMAP);
	SetStyle(WS_CHILD | WS_VISIBLE);

	SetId(id);
	GRect r(0, 0, 4, 4);
	r.Offset(x, y);

	if (ValidStr(FileName))
	{
		#ifdef _MT
		if (NOT Async)
		{
		#endif
			pDC = LoadDC(FileName);
			if (pDC)
			{
				r.Dimension(pDC->X()+4, pDC->Y()+4);
			}
		#ifdef _MT
		}
		else
		{
			NEW(GBitmapThread(this, FileName, &pThread));
		}
		#endif

	}

	SetPos(r);
	CreateClass(LGI_BITMAP);
}

GBitmap::~GBitmap()
{
	#ifdef _MT
	if (pThread)
	{
		pThread->Terminate();
		DeleteObj(pThread);
	}
	#endif
	DeleteObj(pDC);
}

void GBitmap::SetDC(GSurface *pNewDC)
{
	DeleteObj(pDC);
	pDC = pNewDC;
	if (pDC)
	{
		GRect r = GetPos();
		r.Dimension(pDC->X() + 4, pDC->Y() + 4);
		SetPos(r);
		Invalidate();
	}
}

GSurface *GBitmap::GetSurface()
{
	return pDC;
}

int GBitmap::OnEvent(GMessage *Msg)
{
	return GView::OnEvent(Msg);
}

void GBitmap::OnPaint(GSurface *pScreen)
{
	GRect a = GetClient();
	if (pDC)
	{
		pScreen->Blt(0, 0, pDC);

		pScreen->Colour(LC_MED, 24);
		if (pDC->X() < a.X())
		{
			pScreen->Rectangle(pDC->X(), 0, a.x2, pDC->Y());
		}
		if (pDC->Y() < a.Y())
		{
			pScreen->Rectangle(0, pDC->Y(), a.x2, a.y2);
		}
	}
	else
	{
		pScreen->Colour(LC_MED, 24);
		pScreen->Rectangle(&a);
	}
}

void GBitmap::OnMouseClick(GMouse &m)
{
	if (NOT m.Down() AND GetParent())
	{
		((GDialog*) GetParent())->OnNotify(this, 0);
	}
}
