#include <stdlib.h>
#include <stdio.h>

#include "Lgi.h"
#include "GSkinEngine.h"
#include "GButton.h"

class GButtonPrivate
{
public:
	int Down;
	bool Over;
	bool WantsDefault;
	GDisplayString *Txt;

	GButtonPrivate()
	{
		Down = 0;
		Over = 0;
		Txt = 0;
		WantsDefault = false;
	}

	~GButtonPrivate()
	{
		DeleteObj(Txt);
	}

	void Layout(GFont *f, char *s)
	{
		DeleteObj(Txt);
		Txt = NEW(GDisplayString(f, s));
	}
};

GButton::GButton(int id, int x, int y, int cx, int cy, char *name) :
	ResObject(Res_Button)
{
	d = NEW(GButtonPrivate);
	Name(name);
	GRect r(x, y, x+cx, y+cy);
	SetPos(r);
	SetId(id);
	SetTabStop(true);
	
	if (LgiApp->SkinEngine)
	{
		SetFont(LgiApp->SkinEngine->GetDefaultFont(_ObjName));
	}
	
	#ifdef WIN32
	CreateClass("Lgi.Button");
	#endif
}

GButton::~GButton()
{
	if (GetWindow() AND
		GetWindow()->_Default == this)
	{
		GetWindow()->_Default = 0;
	}

	DeleteObj(d);
}

bool GButton::Default()
{
	if (GetWindow())
	{
		return GetWindow()->_Default == this;
	}
	else
	{
		printf("%s:%i - No window.\n", __FILE__, __LINE__);
	}
	
	return false;
}

void GButton::Default(bool b)
{
	if (GetWindow())
	{
		GetWindow()->_Default = b ? this : 0;
		if (IsAttached())
		{
			Invalidate();
		}
	}
	else
	{
		d->WantsDefault = b;
	}
}

bool GButton::Name(char *n)
{
	bool Status = GView::Name(n);
	d->Layout(GetFont(), GObject::Name());
	return Status;
}

bool GButton::NameW(char16 *n)
{
	bool Status = GView::NameW(n);
	d->Layout(GetFont(), GObject::Name());
	return Status;
}

void GButton::SetFont(GFont *Fnt, bool OwnIt)
{
	GView::SetFont(Fnt, OwnIt);
	d->Layout(GetFont(), GObject::Name());
	Invalidate();
}

int GButton::OnEvent(GMessage *Msg)
{
	return GView::OnEvent(Msg);
}

void GButton::OnMouseClick(GMouse &m)
{
	if (Enabled())
	{
		bool Click = IsCapturing();
		
		Capture(m.Down());
		d->Down = m.Down();

		if (m.Down())
		{
			Focus(true);
		}

		Invalidate();
		
		GRect r(0, 0, X()-1, Y()-1);

		// printf("click overlap=%i m.Down=%i click=%i\n", r.Overlap(m.x, m.y), m.Down(), Click);

		if (r.Overlap(m.x, m.y) AND
			NOT m.Down() AND
			Click)
		{
			GView *n = GetNotify() ? GetNotify() : GetParent();
			if (n)
			{
				// This may delete ourself, so do it last.
				n->OnNotify(this, 0);
			}
		}
	}
}

void GButton::OnMouseEnter(GMouse &m)
{
	d->Over = Enabled();
	if (IsCapturing())
	{
		Value(true);
	}
	else if (Enabled())
	{
		Invalidate();
	}
}

void GButton::OnMouseExit(GMouse &m)
{
	d->Over = false;
	if (IsCapturing())
	{
		Value(false);
	}
	else if (Enabled())
	{
		Invalidate();
	}
}

bool GButton::OnKey(GKey &k)
{
	if (NOT Enabled())
	{
		return false;
	}
	
	switch (k.c16)
	{
		case VK_ESCAPE:
		{
			if (GetId() != IDCANCEL)
			{
				break;
			}
			// else fall thru
		}
		case ' ':
		case VK_RETURN:
		{
			int WasDown = d->Down;
			
			d->Down = k.Down();
			Invalidate();

			if (WasDown)
			{
				GView *n = GetNotify() ? GetNotify() : GetParent();
				if (n)
				{
					// This may delete ourself, so do it last.
					n->OnNotify(this, 0);
				}
			}
			return true;
			break;
		}
	}
	
	return false;
}

void GButton::OnFocus(bool f)
{
	Invalidate();
}

void GButton::OnPaint(GSurface *pDC)
{
	if (GApp::SkinEngine AND
		TestFlag(GApp::SkinEngine->GetFeatures(), GSKIN_BUTTON))
	{
		GSkinState State;
		State.pScreen = pDC;
		State.MouseOver = d->Over;
		State.Text = &d->Txt;
		GApp::SkinEngine->OnPaint_GButton(this, &State);
	}
	else
	{
		COLOUR Back = d->Over ? LC_HIGH : LC_MED;
		GRect r(0, 0, X()-1, Y()-1);
		if (Default())
		{
			pDC->Colour(LC_BLACK, 24);
			pDC->Box(&r);
			r.Size(1, 1);
		}
		LgiWideBorder(pDC, r, d->Down ? SUNKEN : RAISED);

		if (d->Txt)
		{
			int x = d->Txt->X(), y = d->Txt->Y();
			int Tx = r.x1 + ((r.X()-x)/2) + d->Down;
			int Ty = r.y1 + ((r.Y()-y)/2) + d->Down;

			GFont *f = GetFont();
			f->Transparent(false);
			if (Enabled())
			{
				f->Colour(LC_TEXT, Back);
				d->Txt->Draw(pDC, Tx, Ty, &r);
			}
			else
			{
				f->Colour(LC_LIGHT, Back);
				d->Txt->Draw(pDC, Tx+1, Ty+1, &r);

				f->Transparent(true);
				f->Colour(LC_LOW, Back);
				d->Txt->Draw(pDC, Tx, Ty, &r);
			}

			if (Focus())
			{
				pDC->Colour(LC_LOW, 24);
				pDC->Box(Tx-2, Ty, Tx+x+2, Ty+y);
			}
		}
		else
		{
			pDC->Colour(Back, 24);
			pDC->Rectangle(&r);
		}
	}
}

int GButton::Value()
{
	return d->Down;
}

void GButton::Value(int i)
{
	d->Down = i;
	Invalidate();
}

void GButton::OnAttach()
{
	if (d->WantsDefault)
	{
		d->WantsDefault = false;
		if (GetWindow())
		{
			GetWindow()->_Default = this;
		}
	}
}
