#include <stdio.h>

#include "Lgi.h"
#include "GColour.h"

class GColourPopup : public GPopup
{
	GColour *Colour;

public:
	GColourPopup(GColour *Parent) : GPopup(Parent)
	{
		SetParent(Colour = Parent);
		GRect r(0, 0, 100, 16 + 4);
		SetPos(r);
	}

	void OnPaint(GSurface *pDC)
	{
		GRect r = GetClient();
		LgiWideBorder(pDC, r, RAISED);
		pDC->Colour(LC_MED, 24);
		pDC->Rectangle(&r);

		SysFont->Transparent(true);
		if (Colour->List)
		{
			char s[64];
			
			SysFont->Colour(LC_BLACK, LC_MED);
			SysFont->Text(pDC, r.x1 + 2, r.y1 + 2, LgiLoadString(L_COLOUR_NONE, "No Colour"));

			for (int i=0; i<Colour->Items; i++)
			{
				int y = r.y1 + ((i+1) * 16);
				pDC->Colour(Colour->List[i], 24);
				pDC->Rectangle(r.x1+1, y+1, r.x1 + 15, y+15);
				
				sprintf(s, "%02.2X,%02.2X,%02.2X", R24(Colour->List[i]), G24(Colour->List[i]), B24(Colour->List[i]));
				SysFont->Text(pDC, r.x1 + 20, y + 2, s);
			}
		}
	}

	void OnMouseClick(GMouse &m)
	{
		GRect r = GetClient();
		r.Size(2, 2);
		if (r.Overlap(m.x, m.y))
		{
			int i = m.y / 16;
			if (i == 0)
			{
				Colour->Value(-1);
			}
			else
			{
				Colour->Value(Colour->List[i-1]);
			}

			Visible(false);
		}
	}

	int Value() { return Colour->Value(); }
	void Value(int i) { Colour->Value(i); }
};

GColour::GColour(COLOUR *list, int items) :
	ResObject(Res_Custom), GDropDown(-1, 0, 0, 10, 10, 0)
{
	List = 0;
	Items = 0;
	c = Rgb24(0, 0, 255);

	SetPopup(NEW(GColourPopup(this)));
	SetColourList(list, items);
}

GColour::~GColour()
{
	DeleteArray(List);
}

void GColour::SetColourList(COLOUR *list, int items)
{
	DeleteArray(List);
	if (list)
	{
		List = NEW(COLOUR[Items = items]);
		if (List)
		{
			memcpy(List, list, sizeof(COLOUR) * Items);
		}
	}

	if (GetPopup())
	{
		GRect r(0, 0, 100, (max(Items+1, 1) * 16) + 4);
		GetPopup()->SetPos(r);
	}
}

int GColour::Value()
{
	return c;
}

void GColour::Value(int i)
{
	if (c != (COLOUR)i)
	{
		c = (COLOUR)i;
		Invalidate();

		GView *n = GetNotify() ? GetNotify() : GetParent();
		if (n)
		{
			n->OnNotify(this, 0);
		}
	}
}

void GColour::OnPaint(GSurface *pDC)
{
	GDropDown::OnPaint(pDC);

	GRect r = GetClient();
	r.x2 -= 9;
	r.Size(5, 5);
	if (IsOpen()) r.Offset(1, 1);

	bool HasColour = Enabled() AND c != 0xffffffff;
	pDC->Colour(HasColour ? LC_BLACK : LC_LOW, 24);
	pDC->Box(&r);

	r.Size(1, 1);
	pDC->Colour(HasColour ? Value() : LC_LOW, 24);
	if (HasColour)
	{
		pDC->Rectangle(&r);
	}
	else
	{
		pDC->Line(r.x1, r.y1, r.x2, r.y2);
		pDC->Line(r.x1, r.y2, r.x2, r.y1);
	}
}

class GColourFactory : public GViewFactory
{
	GView *NewView(char *Class, GRect *Pos, char *Text)
	{
		if (Class AND
			stricmp(Class, "GColour") == 0)
		{
			return NEW(GColour);
		}

		return 0;
	}

} ColourFactory;
