#include <stdio.h>
#include "Lgi.h"
#include "GPopup.h"

/////////////////////////////////////////////////////////////////////////////////////
#if 0

#ifndef WH_MOUSE_LL
#define WH_MOUSE_LL        14
#endif

class GMouseHookPrivate
{
public:
	HWND hMouseOver;
	HHOOK hMouse;
	List<GPopup> Popups;

	GMouseHookPrivate()
	{
		hMouseOver = 0;
		hMouse = 0;
	}
};

GMouseHook::GMouseHook()
{
	GApp::MouseHook = this;
	d = NEW(GMouseHookPrivate);
	d->hMouse = SetWindowsHookEx(WH_MOUSE, MouseProc, LgiProcessInst(), LgiGetCurrentThread());
	if (NOT d->hMouse)
	{
		DWORD e = GetLastError();
		LgiAssert(d->hMouse);
	}
}

GMouseHook::~GMouseHook()
{
	if (d->hMouse)
	{
		UnhookWindowsHookEx(d->hMouse);
	}

	GApp::MouseHook = 0;
	DeleteObj(d);
}

void GMouseHook::RegisterPopup(GPopup *p)
{
	d->Popups.Insert(p);
}

void GMouseHook::UnregisterPopup(GPopup *p)
{
	d->Popups.Delete(p);
}

LRESULT CALLBACK GMouseHook::MouseProc(int Code, WPARAM a, LPARAM b)
{
	GMouseHook *This = GApp::MouseHook;
	MOUSEHOOKSTRUCT *Mhs = (MOUSEHOOKSTRUCT*)b;

	if (Mhs)
	{
		switch (a)
		{
			case WM_LBUTTONDOWN:
			case WM_RBUTTONDOWN:
			case WM_MBUTTONDOWN:
			{
				POINT Screen = Mhs->pt;

				for (GView *v=This->d->Popups.First(); v; v=This->d->Popups.Next())
				{
					GRect p = v->GetPos();
					if (NOT v->GetPos().Overlap(Screen.x, Screen.y))
					{
						v->Visible(false);
						break;
					}
				}
				break;
			}
			case WM_MOUSEMOVE:
			{
				DWORD hWndProcess = 0;
				DWORD hProcess = GetCurrentProcessId();
				GetWindowThreadProcessId(Mhs->hwnd, &hWndProcess);
				if (hWndProcess == hProcess)
				{
					RECT rc;
					GetClientRect(Mhs->hwnd, &rc);
					ScreenToClient(Mhs->hwnd, &Mhs->pt);

					// is the mouse inside the client area?
					bool Inside = NOT  (Mhs->pt.x < 0 OR
										Mhs->pt.y < 0 OR
										Mhs->pt.x >= rc.right-rc.left OR
										Mhs->pt.y >= rc.bottom-rc.top);

					HWND hWnd = (Inside) ? Mhs->hwnd : 0;
					if (hWnd != This->d->hMouseOver)
					{
						// window has changed
						if (This->d->hMouseOver AND IsWindow(This->d->hMouseOver))
						{
							// window is LOSING mouse
							SendMessage(This->d->hMouseOver, M_MOUSEEXIT, 0, MAKELONG((short) Mhs->pt.x, (short) Mhs->pt.y));
						}

						// set current window
						This->d->hMouseOver = hWnd;

						if (This->d->hMouseOver AND IsWindow(This->d->hMouseOver))
						{
							// window is GETTING mouse
							SendMessage(This->d->hMouseOver, M_MOUSEENTER, Inside, MAKELONG((short) Mhs->pt.x, (short) Mhs->pt.y));
						}
					}
				}
				break;
			}
		}
	}

	return CallNextHookEx(This->d->hMouse, Code, a, b);
}

#endif

#ifdef WIN32

typedef HWND HookWnd;

#else

#define MAKELONG(low, high) \
	( ((low) & 0xffff) | ((high) << 16) )

#ifdef XWIN

typedef Window HookWnd;

HookWnd WindowFromPoint(int x, int y)
{
	Display *Dsp = QObject::XDisplay();
	Window Root = RootWindow(Dsp, 0);

	Window w1 = Root;
	Window w2 = Root;
	Window Child;
	int x1 = x, y1 = y, x2, y2;
	
	while (XTranslateCoordinates(Dsp, w1, w2, x1, y1, &x2, &y2, &Child))
	{
		if (Child == XNone)
		{
			return w2;
		}

		w1 = w2;
		w2 = Child;
		x1 = x2;
		y1 = y2;
	}
	
	return 0;
}

bool GetWindowRect(HookWnd Wnd, GRect &r)
{
	Window Root;
	int x, y;
	uint w, h, border, depth;
	int Err = XGetGeometry(QObject::XDisplay(),
							Wnd,
							&Root,
							&x,
							&y,
							&w,
							&h,
							&border,
							&depth);
	if (Err == 1)
	{
		r.Set(x, y, x + w - 1, y + h - 1);
		return true;
	}
	return false;
}

bool ScreenToClient(HookWnd Wnd, GdcPt2 &p)
{
	Display *Dsp = QObject::XDisplay();
	Window Root = RootWindow(Dsp, 0);

	Window w1 = Root;
	Window w2 = Wnd;
	Window Child;
	int x1 = p.x, y1 = p.y, x2, y2;
	
	if (XTranslateCoordinates(Dsp, w1, w2, x1, y1, &x2, &y2, &Child))
	{
		p.x = x2;
		p.y = y2;
		return true;
	}
	
	return 0;
}

#endif

bool IsWindow(HookWnd Wnd)
{
	// Do any available validation of the Wnd we can here
	return true;
}

#endif

uint32 LgiGetViewPid(OsView View)
{
	#if defined(WIN32)
	DWORD hWndProcess = 0;
	GetWindowThreadProcessId(View, &hWndProcess);
	return hWndProcess;
	#endif
	
	// FIXME: Linux and BeOS
	
	return LgiProcessId();
}

class GMouseHookPrivate : public GSemaphore, public GThread
{
	bool Loop;
	HookWnd hMouseOver;

public:
	List<GPopup> p;

	GMouseHookPrivate() : GSemaphore("MouseHookLock")
	{
		Loop = true;
		hMouseOver = 0;
		Run();
	}

	~GMouseHookPrivate()
	{
		Loop = false;
		while (NOT IsExited())
		{
			LgiSleep(10);
		}
	}
	
	void PostEvent(HookWnd h, int c, int a, int b)
	{
		#ifdef XWIN
		QWidget q(h);
		LgiPostEvent(&q, c, a, b);
	    q.detachHandle();
		#else
		LgiPostEvent(h, c, a, b);
		#endif
	}

	int Main()
	{
		GMouse Old;
		GView v;

		while (Loop)
		{
			if (Lock())
			{
				GMouse m;
				v.GetMouse(m, true);

				if (m.Down() AND NOT Old.Down()) 
				{
					// Down click....
					int64 Now = LgiCurrentTime();
					bool Over = false;
					for (GPopup *w = p.First(); w; w = p.Next())
					{
						if (w->GetPos().Overlap(m.x, m.y))
						{
							Over = true;
							break;
						}
					}
					
					if (NOT Over)
					{

						for (GPopup *w = p.First(); w; w = p.Next())
						{
							if (w->Visible() AND
								w->Start < Now - 100)
							{
								bool Close = true;

								#ifdef WIN32
								// This is a bit of a hack to prevent GPopup's with open context menus from
								// closing when the user clicks on the context menu.
								//
								// FIXME: Linux/BeOS/Mac

								// Scan the window under the mouse up the parent tree
								POINT p = { m.x, m.y };
								for (HWND hnd = WindowFromPoint(p); hnd; hnd = ::GetParent(hnd))
								{
									// Is this a popup?
									ULONG Style = GetWindowLong(hnd, GWL_STYLE);
									if (NOT TestFlag(Style, WS_POPUP))
									{
										// No it's a normal window, so close the popup
										break;
									}

									// Are we over a popup?
									RECT rc;
									GetWindowRect(hnd, &rc);
									GRect gr = rc;
									if (gr.Overlap(m.x, m.y))
									{
										// Yes, so don't close it
										Close = false;
										break;
									}
								}
								#endif

								if (Close)
								{
									w->Visible(false);
								}
							}
						}
					}
				}

				if (m.x != Old.x OR
					m.y != Old.y)
				{
					// Mouse moved...
					HookWnd hOver = 0;

					#if defined WIN32

					POINT WinPt = { m.x, m.y };
					hOver = WindowFromPoint(WinPt);

					RECT WinRect;
					GetClientRect(hOver, &WinRect);
					ScreenToClient(hOver, &WinPt);
					GRect rc = WinRect;
					GdcPt2 p(WinPt.x, WinPt.y);

					#elif defined LINUX
					
					hOver = WindowFromPoint(m.x, m.y);
					GRect rc;
					GdcPt2 p(m.x, m.y);
					if (hOver)
					{
						if (NOT GetWindowRect(hOver, rc))
						{
							printf("No Rect for over\n");
						}
						if (NOT ScreenToClient(hOver, p))
						{
							printf("No conversion for point.\n");
						}
					}
					else
					{
						printf("No hOver\n");
					}
					
					#else
					#error "Not Impl."
					#endif

					// is the mouse inside the client area?
					bool Inside = NOT  (p.x < 0 OR
										p.y < 0 OR
										p.x >= rc.X() OR
										p.y >= rc.Y());
					
					HookWnd hWnd = (Inside) ? hOver : 0;

					if (hWnd != hMouseOver)
					{
						uint32 hProcess = LgiProcessId();

						// Window has changed
						if (hMouseOver AND IsWindow(hMouseOver))
						{
							// Window is LOSING mouse
							// Using 'post' because send can cause a deadlock.
							uint32 hWndProcess = LgiGetViewPid(hMouseOver);
							if (hWndProcess == hProcess)
							{
								PostEvent(hMouseOver, M_MOUSEEXIT, 0, MAKELONG((short) p.x, (short) p.y));
							}
						}

						// Set current window
						hMouseOver = hWnd;

						if (hMouseOver AND IsWindow(hMouseOver))
						{
							// Window is GETTING mouse
							// Using 'post' because send can cause a deadlock.
							uint32 hWndProcess = LgiGetViewPid(hMouseOver);
							if (hWndProcess == hProcess)
							{
								PostEvent(hMouseOver, M_MOUSEENTER, Inside, MAKELONG((short) p.x, (short) p.y));
							}
						}
					}
				}

				Old = m;

				Unlock();
			}

			LgiSleep(40);
		}

		return 0;
	}
};

GMouseHook::GMouseHook()
{
	d = NEW(GMouseHookPrivate);
}

GMouseHook::~GMouseHook()
{
	d->Lock();
	DeleteObj(d);
}

bool GMouseHook::OnViewKey(GView *v, GKey &k)
{
	bool Status = false;

	if (d->Lock())
	{
		GView *l = d->p.Last();
		if (l)
		{
			if (l->OnKey(k))
			{
				Status = true;
			}
		}
		
		d->Unlock();
	}
	
	return Status;
}

void GMouseHook::RegisterPopup(GPopup *p)
{
	if (d->Lock())
	{
		if (NOT d->p.HasItem(p))
		{
			d->p.Insert(p);
		}
		d->Unlock();
	}
}

void GMouseHook::UnregisterPopup(GPopup *p)
{
	if (d->Lock())
	{
		d->p.Delete(p);
		d->Unlock();
	}
}

#ifdef WIN32
LRESULT CALLBACK GMouseHook::MouseProc(int Code, WPARAM a, LPARAM b)
{
	return 0;
}
#endif

#if 0

class _QPopup : public QView<QPopup>
{
	GPopup *Popup;

public:
	_QPopup(GPopup *popup) : QView<QPopup>(popup)
	{
		Popup = popup;
	}

	void _QPopup::show()
	{
		QView<QPopup>::show();
		Popup->GView::Visible(true);
	}

	void _QPopup::hide()
	{
		bool Was = Popup->GView::Visible();
		if (GetCancelled()) Popup->Cancelled = true;
				
		QView<QPopup>::hide();
		Popup->GView::Visible(false);

		if (Popup->GetParent())
		{
			Popup->GetParent()->OnNotify(Popup, Was);
		}
	}

	QWidget *_QPopup::parentWidget()
	{
		if (Popup->GetParent())
		{
			return Popup->GetParent()->Handle();
		}

		return 0;
	}
};

#endif

/////////////////////////////////////////////////////////////////////////////////////
GPopup::GPopup(GView *owner)
#if defined XWIN
	: GView(NEW(DefaultOsView(this)))
#endif
{
	Start = 0;
	Cancelled = false;
	if (Owner = owner)
	{
		#ifndef WIN32
		Owner->PopupChild() = this;
		#endif
		
		_Window = Owner->GetWindow();
	}
	
	GView::Visible(false);
}

GPopup::~GPopup()
{
	if (Owner)
	{
		#ifndef WIN32
		Owner->PopupChild() = 0;
		#endif
	}

	GMouseHook *Hook = LgiApp->GetMouseHook();
	if (Hook) Hook->UnregisterPopup(this);

	Children.DeleteObjects();
}

bool GPopup::Attach(GView *p)
{
	if (p) SetParent(p);
	if (GetParent())
	{
		#if defined WIN32

		SetStyle(WS_POPUP);
		CreateClass("GPopup");
		GView::Attach(GetParent());
		
		#endif

		AttachChildren();

		if (NOT _Window)
		{
			if (Owner)
			{
				_Window = Owner->GetWindow();
			}
			else
			{
				_Window = p->GetWindow();
			}
		}
	}

	return Handle() != 0;
}

void GPopup::Visible(bool i)
{
	#ifdef XWIN
	if (i)
	{
		Handle()->SetPopup(true);
	}
	#endif

	bool Was = GFlags::Visible();
	GView::Visible(i);

	if (i)
	{
		Start = LgiCurrentTime();

		GMouseHook *Hook = LgiApp->GetMouseHook();
		if (Hook) Hook->RegisterPopup(this);

		if (NOT _Window)
		{
			if (Owner)
			{
				_Window = Owner->GetWindow();
			}
			else if (GetParent())
			{
				_Window = GetParent()->GetWindow();
			}
		}
	}
	else
	{
		GMouseHook *Hook = LgiApp->GetMouseHook();
		if (Hook) Hook->UnregisterPopup(this);

		SendNotify(Was);
	}
}

bool GPopup::Visible()
{
	return GView::Visible();
}

/////////////////////////////////////////////////////////////////////////////////////
GDropDown::GDropDown(int Id, int x, int y, int cx, int cy, GPopup *popup)
{
	SetId(Id);
	GRect r(x, y, x+cx, y+cy);
	SetPos(r);
	Popup = popup;
	SetTabStop(true);
}

GDropDown::~GDropDown()
{
	DeleteObj(Popup);
}

void GDropDown::OnFocus(bool f)
{
	Invalidate();
}

GPopup *GDropDown::GetPopup()
{
	return Popup;
}

void GDropDown::SetPopup(GPopup *popup)
{
	DeleteObj(Popup);
	Popup = popup;
	Invalidate();
}

bool GDropDown::IsOpen()
{
	return Popup AND Popup->Visible();
}

void GDropDown::OnPaint(GSurface *pDC)
{
	GRect r = GetClient();
	r.Offset(-r.x1, -r.y1);

	LgiWideBorder(pDC, r, IsOpen() ? SUNKEN : RAISED);
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(&r);
	if (Focus())
	{
		#ifdef WIN32
		DrawFocusRect(pDC->Handle(), &((RECT)r));
		#else
		pDC->Colour(LC_LOW, 24);
		pDC->Box(&r);
		#endif
	}

	int Cx = r.x2 - 7;
	int Cy = r.y1 + ((r.Y() - 3) >> 1);
	pDC->Colour(Enabled() AND Popup ? LC_TEXT : LC_LOW, 24);
	if (IsOpen())
	{
		Cx++;
		Cy++;
	}
	for (int i=0; i<3; i++)
	{
		pDC->Line(Cx+i, Cy+i, Cx+5-i, Cy+i);
	}
}

void GDropDown::Activate()
{
	if (IsOpen())
	{
		// Hide
		Popup->Visible(false);
	}
	else // Show
	{
		// Locate under myself
		GdcPt2 p(X()-1, Y());
		PointToScreen(p);
		GRect r(p.x-Popup->X()+1, p.y, p.x, p.y+Popup->Y()-1);
		
		// Show the popup
		if (NOT Popup->IsAttached())
		{
			Popup->Attach(this);
		}

		Popup->Cancelled = false;
		Popup->SetPos(r);
		Popup->Visible(true);
	}

	Invalidate();
}

bool GDropDown::OnKey(GKey &k)
{
	if (k.c == ' ' OR k.c == VK_RETURN)
	{
		if (k.Down())
		{
			Activate();
		}
		
		return true;
	}
	
	return false;
}

void GDropDown::OnMouseClick(GMouse &m)
{
	if (Popup AND m.Down())
	{
		Focus(true);
		Activate();
	}
}

int GDropDown::OnNotify(GView *c, int f)
{
	if (c == Popup AND f)
	{
		Invalidate();
		OnPopupClose();
	}

	return false;
}
