/*hdr
**      FILE:           GTabView.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           20/10/2000
**      DESCRIPTION:    Lgi self-drawn tab control
**
**      Copyright (C) 2000 Matthew Allen
**              fret@memecode.com
*/

#include <stdio.h>
#include "Lgi.h"
#include "GTabView.h"
#include "GSkinEngine.h"

#ifdef WIN32

#ifdef TOOL_VLOW
#undef TOOL_VLOW
#endif

#ifdef TOOL_LOW
#undef TOOL_LOW
#endif

#ifdef TOOL_HIGH
#undef TOOL_HIGH
#endif

#ifdef TOOL_VHIGH
#undef TOOL_VHIGH
#endif


#define TOOL_VLOW	GetSysColor(COLOR_3DDKSHADOW)
#define TOOL_LOW	GetSysColor(COLOR_3DSHADOW)
#define TOOL_HIGH	GetSysColor(COLOR_3DLIGHT)
#define TOOL_VHIGH	GetSysColor(COLOR_3DHILIGHT)

#endif

////////////////////////////////////////////////////////////////////////////////////////////
class GTabViewPrivate
{
public:
	// General
	int Current;
	GRect TabClient;
	
	// Scrolling
	int Scroll;			// number of buttons scrolled off the left of the control
	GRect LeftBtn;	// scroll left button
	GRect RightBtn;	// scroll right button

	GTabViewPrivate()
	{
		Current = 0;
		TabClient.ZOff(-1, -1);

		Scroll = 0;
		LeftBtn.ZOff(-1, -1);
		RightBtn.ZOff(-1, -1);
	}
};

////////////////////////////////////////////////////////////////////////////////////////////
// Tab Control
GTabView::GTabView(int id, int x, int y, int cx, int cy, char *name, int Init) :
	ResObject(Res_TabView)
{
	d = NEW(GTabViewPrivate);
	d->Current = Init;

	SetId(id);
	GRect r(x, y, x+cx, y+cy);
	SetPos(r);
	Name(name);
	_BorderSize = 0;
	SetTabStop(true);

	#if defined WIN32
	SetDlgCode(DLGC_WANTARROWS);
	CreateClass("Lgi.TabView");
	#elif defined BEOS
	Handle()->SetFlags(	Handle()->Flags() |
						B_FULL_UPDATE_ON_RESIZE);
	#endif
}

GTabView::~GTabView()
{
	List<GTabPage> Temp;
	for (GTabPage *p=Tabs.First(); p; p=Tabs.Next())
	{
		Temp.Insert(p);
	}
	Tabs.Empty();
	Temp.DeleteObjects();
	DeleteObj(d);
}

GTabPage *GTabView::GetCurrent()
{
	return Tabs[d->Current];
}

int GTabView::TabY()
{
	return max(16, GetFont()->Y("A")) + 4;
}

void GTabView::OnChildrenChanged(GView *Wnd, bool Attaching)
{
}

#ifdef WIN32
GView *GTabView::FindControl(HWND hCtrl)
{
	GView *Ctrl = 0;

	if (hCtrl == Handle())
	{
		return this;
	}

	for (	GTabPage *Page = Tabs.First();
			Page AND NOT Ctrl;
			Page = Tabs.Next())
	{
		Ctrl = Page->FindControl(hCtrl);
	}
	return Ctrl;
}
#endif

GView *GTabView::FindControl(int Id)
{
	if (GetId() == Id)
	{
		return this;
	}

	for (	GTabPage *Page = Tabs.First();
			Page;
			Page = Tabs.Next())
	{
		GView *Ctrl = Page->FindControl(Id);
		if (Ctrl) return Ctrl;
	}

	return 0;
}

bool GTabView::Attach(GView *parent)
{
	bool Status = GView::Attach(parent);
	if (Status)
	{
		GTabPage *p = Tabs.ItemAt(d->Current);
		if (p)
		{
			p->Attach(this);
		}
		
		for (p=Tabs.First(); p; p=Tabs.Next())
		{
			p->_Window = _Window;
		}

		OnPosChange();
	}

	return Status;
}

int GTabView::Value()
{
	return d->Current;
}

void GTabView::Value(int i)
{
	if (i != d->Current)
	{
		// change tab
		GTabPage *Old = Tabs.ItemAt(d->Current);
		if (Old)
		{
			Old->Visible(false);
		}

		d->Current = i;

		GTabPage *p = Tabs.ItemAt(d->Current);
		if (p)
		{
			p->Attach(this);
			p->Visible(true);
		}

		OnPosChange();
		Invalidate();

		GView *n = (GetNotify()) ? GetNotify() : GetParent();
		if (n)
		{
			n->OnNotify(this, d->Current);
		}
	}
}

int GTabView::OnEvent(GMessage *Msg)
{
	return GView::OnEvent(Msg);
}

int GTabView::OnNotify(GView *Ctrl, int Flags)
{
	if (GetParent())
	{
		return GetParent()->OnNotify(Ctrl, Flags);
	}

	return 0;
}

bool GTabView::Append(GTabPage *Page, int Where)
{
	if (Page)
	{
		Page->TabCtrl = this;
		Page->_Window = _Window;
		Page->SetParent(this);
		Tabs.Insert(Page, Where);

		if (IsAttached() AND Tabs.GetItems() == 1)
		{
			Page->Attach(this);
			OnPosChange();
		}

		Invalidate();

		return true;
	}
	return false;
}

GTabPage *GTabView::Append(char *name, int Where)
{
	GTabPage *Page = NEW(GTabPage(name));
	if (Page)
	{
		Page->TabCtrl = this;
		Page->_Window = _Window;
		Page->SetParent(this);
		Tabs.Insert(Page, Where);

		if (IsAttached() AND Tabs.GetItems() == 1)
		{
			Page->Attach(this);
			OnPosChange();
		}

		Invalidate();
	}
	return Page;
}

bool GTabView::Delete(GTabPage *Page)
{
	bool Status = false;
	int Index = Tabs.IndexOf(Page);
	if (Index >= 0)
	{
		if (Index == d->Current)
		{
			Page->Detach();
		}
		
		Tabs.Delete(Page);
		Status = false;
	}
	
	return Status;
}

GRect &GTabView::GetTabClient()
{
	d->TabClient = GView::GetClient();
	d->TabClient.Offset(-d->TabClient.x1, -d->TabClient.y1);
	d->TabClient.Size(2, 2);
	d->TabClient.y1 += TabY();

	return d->TabClient;
}

void GTabView::OnMouseClick(GMouse &m)
{
	if (m.Down())
	{
		if (m.Left())
		{
			if (d->LeftBtn.Overlap(m.x, m.y))
			{
				d->Scroll++;
				Invalidate();
			}
			else if (d->RightBtn.Overlap(m.x, m.y))
			{
				d->Scroll = max(0, d->Scroll-1);
				Invalidate();
			}
			else
			{
				int Hit = -1;
				GTabPage *p = 0;
				for (p = Tabs.First(); p; p = Tabs.Next())
				{
					if (p->TabPos.Overlap(m.x, m.y))
					{
						Hit = Tabs.IndexOf(p);
						break;
					}
				}
				
				if (Hit >= 0)
				{
					Value(Hit);
				}
			}

			Focus(true);
		}
	}
}

bool GTabView::OnKey(GKey &k)
{
	if (k.Down())
	{
		switch (k.c)
		{
			case VK_LEFT:
			{
				if (d->Current > 0)
				{
					GTabPage *p = Tabs[d->Current - 1];
					if (p AND NOT p->TabPos.Valid())
					{
						if (d->Scroll) d->Scroll--;
					}

					Value(d->Current - 1);
				}
				return true;
				break;
			}
			case VK_RIGHT:
			{
				if (d->Current < Tabs.GetItems() - 1)
				{
					GTabPage *p = Tabs[d->Current + 1];
					if (p AND NOT p->TabPos.Valid())
					{
						d->Scroll++;
					}
					
					Value(d->Current + 1);
				}
				return true;
				break;
			}
		}
	}
	
	return false;
}

void GTabView::OnFocus(bool f)
{
	GTabPage *p = Tabs.ItemAt(d->Current);
	if (p)
	{
		GRect r = p->TabPos;
		r.y2 += 2;
		Invalidate(&r);
	}
}

void GTabView::OnPaint(GSurface *pDC)
{
	if (GApp::SkinEngine AND
		TestFlag(GApp::SkinEngine->GetFeatures(), GSKIN_TABVIEW))
	{
		GSkinState State;
		State.pScreen = pDC;
		State.MouseOver = false;
		State.Text = 0;
		GApp::SkinEngine->OnPaint_GTabView(this, &State);
	}
	else
	{
		GRect r = GetTabClient();

		r.Size(-2, -2);
		LgiWideBorder(pDC, r, RAISED);

		pDC->Colour(LC_MED, 24);
		pDC->Rectangle(0, 0, X()-1, d->TabClient.y1-3);

		GTabPage *Sel = 0;
		int x = r.x1;
		int n=0;
		
		if (d->Scroll)
		{
			d->RightBtn.ZOff(12, TabY() - 2);
			x = d->RightBtn.x2 + 4;
		}
		else
		{
			d->RightBtn.ZOff(-1, -1);
		}
		d->LeftBtn.ZOff(-1, -1);
		
		
		for (GTabPage *p = Tabs.First(); p; p = Tabs.Next(), n++)
		{
			if (n < d->Scroll)
			{
				p->TabPos.ZOff(-1, -1);
			}
			else
			{
				char *Text = p->Name();
				int Wid = ((Text) ? p->GetFont()->X(Text) : 0) + 13;
				p->TabPos.ZOff(Wid, TabY()-3);
				p->TabPos.Offset(x, 2);
				
				if (p->TabPos.x2 > r.x2 - 16)
				{
					d->LeftBtn.x2 = X()-1;
					d->LeftBtn.x1 = d->LeftBtn.x2 - 12;
					d->LeftBtn.y1 = 0;
					d->LeftBtn.y2 = TabY() - 2;
					
					p->TabPos.ZOff(-1, -1);
					break;
				}
				
				if (d->Current != n)
				{
					p->PaintTab(pDC, false);
				}
				else
				{
					Sel = p;
				}
				x += Wid+1;
			}
		}
		
		if (Sel)
		{
			Sel->PaintTab(pDC, true);
		}
		
		if (d->LeftBtn.Valid())
		{
			r = d->LeftBtn;
			LgiWideBorder(pDC, r, RAISED);

			int x = r.x1 + (r.X() >> 1) + 1;
			int y = r.y1 + (r.Y() >> 1) - 1;
			pDC->Colour(LC_TEXT, 24);
			for (int i=0; i<4; i++)
			{
				pDC->Line(x-i, y-i, x-i, y+i);
			}
		}
		if (d->RightBtn.Valid())
		{
			r = d->RightBtn;
			LgiWideBorder(pDC, r, RAISED);

			int x = r.x1 + (r.X() >> 1) - 2;
			int y = r.y1 + (r.Y() >> 1) - 1;
			pDC->Colour(LC_TEXT, 24);
			for (int i=0; i<4; i++)
			{
				pDC->Line(x+i, y-i, x+i, y+i);
			}
		}
	}
}

void GTabView::OnPosChange()
{
	GetTabClient();
	GTabPage *p = Tabs.ItemAt(d->Current);
	if (p)
	{
		p->SetPos(d->TabClient, true);
	}
}

//////////////////////////////////////////////////////////////////////////////////
char *_lgi_gview_cmp(GView *a, GView *b)
{
	static char Str[256];
	if (a AND b)
	{
		sprintf(Str,
				"GView: %p,%p Hnd: %x,%x",
				dynamic_cast<GView*>(a),
				dynamic_cast<GView*>(b),
				a->Handle(),
				b->Handle());
	}
	else
	{
		Str[0] = 0;
	}
	return Str;
}


GTabPage::GTabPage(char *name) :
#if defined WIN32
	GView(0)
#else
	GView(NEW(DefaultOsView(this)))
#endif
	, ResObject(Res_Tab)
{
	GRect r(0, 0, 100, 100);
	SetPos(r);
	Name(name);
	TabCtrl = 0;
	TabPos.ZOff(-1, -1);

	#if defined BEOS
	Handle()->SetViewColor(216, 216, 216);
	Handle()->SetResizingMode(B_FOLLOW_ALL_SIDES);
	Handle()->SetFlags(Handle()->Flags() & ~B_NAVIGABLE);
	#elif defined WIN32
	SetStyle(GetStyle() | WS_CLIPCHILDREN);
	CreateClass("LGI_TabPage", 0, CS_HREDRAW | CS_VREDRAW);
	#endif
}

GTabPage::~GTabPage()
{
}

void GTabPage::PaintTab(GSurface *pDC, bool Selected)
{
	GRect r = TabPos;
	if (Selected)
	{
		r.Size(-2, -2);
	}
	
	pDC->Colour(LC_LIGHT, 24);
	
	bool First = false;
	if (TabCtrl)
	{
		Iterator<GTabPage> Iter(&TabCtrl->Tabs);
		First = Iter.First() == this;
	}

	if (First)
	{
		pDC->Line(r.x1, r.y1+1, r.x1, r.y2);
	}
	else
	{
		pDC->Line(r.x1, r.y1+1, r.x1, r.y2-1);
	}
	pDC->Line(r.x1+1, r.y1, r.x2-1, r.y1);

	pDC->Colour(LC_HIGH, 24);
	pDC->Line(r.x1+1, r.y1+1, r.x1+1, r.y2);
	pDC->Line(r.x1+1, r.y1+1, r.x2-1, r.y1+1);

	pDC->Colour(LC_LOW, 24);
	pDC->Line(r.x2-1, r.y1+1, r.x2-1, r.y2);

	pDC->Colour(LC_SHADOW, 24);
	pDC->Line(r.x2, r.y1+1, r.x2, r.y2-1);

	r.x2 -= 2;
	r.x1 += 2;
	r.y1 += 2;
	
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(&r);
	pDC->Set(r.x1, r.y1);
	pDC->Set(r.x2, r.y1);
	
	char *t = Name();
	if (t)
	{
		int Tx, Ty;
		GFont *f = GetFont();
		f->Size(&Tx, &Ty, t);
		f->Colour(LC_TEXT, LC_MED);
		f->Transparent(true);
		
		int x = r.x1 + ((r.X()-Tx)/2), y = r.y1 + ((r.Y()-Ty)/2);
		f->Text(pDC, x, y, t);
		
		if (TabCtrl->Focus() AND Selected)
		{
			r.Set(x, y, x+Tx, y+Ty);
			r.Size(-2, -1);
			r.y1++;
			pDC->Colour(LC_LOW, 24);
			pDC->Box(&r);
		}
	}
}

bool GTabPage::Attach(GView *parent)
{
	bool Status = false;

	if (TabCtrl)
	{
		if (NOT IsAttached())
		{
			Status = GView::Attach(parent);
		}
		else
		{
			Status = true;
		}
		
		GView *View = dynamic_cast<GView*>(this);
		for (GView *w = Children.First(); w; w = Children.Next())
		{
			if (NOT w->IsAttached())
			{
				w->Attach(View);
				w->SetNotify(TabCtrl->GetParent());
			}
		}
	}

	return Status;
}

char *GTabPage::Name()
{
	return GObject::Name();
}

bool GTabPage::Name(char *name)
{
	bool Status = GView::Name(name);
	if (GetParent()) GetParent()->Invalidate();
	return Status;
}

int GTabPage::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		#ifdef WIN32
		#endif
	}

	return GView::OnEvent(Msg);
}

void GTabPage::Append(GView *Wnd)
{
	if (Wnd)
	{
		Wnd->SetNotify(TabCtrl);

		if (IsAttached() AND
			TabCtrl)
		{
			// AND TabCtrl->Value() == TabCtrl->Tabs.IndexOf(this)
			Wnd->Attach(this);
		}
		else if (NOT Children.HasItem(Wnd))
		{
			Wnd->SetParent(this);
			Children.Insert(Wnd);
		}

	}
}

bool GTabPage::Remove(GView *Wnd)
{
	if (Wnd)
	{
		Children.Delete(Wnd);
		if (Wnd->GetParent())
		{
			Wnd->Detach();
		}
		return true;
	}
	return false;
}

void GTabPage::OnPaint(GSurface *pDC)
{
	GRect r(0, 0, X()-1, Y()-1);
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle(&r);
}

void GTabPage::OnFocus(bool b)
{
}

bool GTabPage::OnKey(GKey &k)
{
	return false;
}

bool GTabPage::AppendDialogFromResource(int Res)
{
	char n[256] = "";
	List<GView> l;
	bool Status = GLgiRes::LoadFromResource(Res, l, 0, n);

	for (GView *v=l.First(); v; v=l.Next())
	{
		Append(v);
	}

	if (NOT ValidStr(Name()))
	{
		Name(n);
	}

	return Status;
}




