#include "Lgi.h"
#include "GTableLayout.h"
#include "GToken.h"
#include "GVariant.h"
#include "GTextLabel.h"
#include "GButton.h"
#include "GEdit.h"
#include "GCombo.h"
#include "GList.h"
#include "GTree.h"
#include "GCheckBox.h"

#define Izza(c)				dynamic_cast<c*>(v)
#define CELL_SPACING		4
#define DEBUG_TRACE			0
#define DEBUG_PROFILE		0

enum CellFlag
{
	SizeUnknown,
	SizeFixed,
	SizeGrow,
};

template <class T>
T CountRange(GArray<T> &a, int Start, int End)
{
	T c = 0;
	for (int i=Start; i<=End; i++)
	{
		c += a[i];
	}
	return c;
}

struct UnderInfo
{
	int Col;		// index of column
	int Grow;		// in px
};

int Cmp(UnderInfo *a, UnderInfo *b)
{
	return b->Grow - a->Grow;
}

void DistributeUnusedSpace(GArray<int> &Min, GArray<int> &Max, GArray<CellFlag> &Flags, int Total, int CellSpacing)
{
	// Now allocate unused space
	int Sum = CountRange<int>(Min, 0, Min.Length()-1) + ((Min.Length() - 1) * CellSpacing);
	if (Sum < Total)
	{
		int i, Avail = Total - Sum;

		// Do growable ones first
		GArray<UnderInfo> Small, Large;
		for (i=0; i<Min.Length(); i++)
		{
			CellFlag f = Flags[i];
			if (Max[i] > Min[i])
			{
				UnderInfo u;
				u.Col = i;
				u.Grow = Max[i] - Min[i];
				if (u.Grow < Avail >> 1)
				{
					Small.Add(u);
				}
				else
				{
					Large.Add(u);
				}
			}
		}
		Small.Sort(Cmp);
		Large.Sort(Cmp);

		for (i=0; Avail>0 AND i<Small.Length(); i++)
		{
			int Col = Small[i].Col;
			Min[Col] += Small[i].Grow;
			Avail -= Small[i].Grow;
		}

		for (i=0; Avail>0 AND i<Large.Length(); i++)
		{
			int Col = Large[i].Col;
			int Part = Avail / Large.Length();
			Min[Col] += Part;
			Avail -= Part;
		}
	}
}

void PreLayoutTextCtrl(GView *v, int &Min, int &Max)
{
	int MyMin = 0;
	int MyMax = 0;

	char *t = v->Name();
	GFont *f = v->GetFont();
	if (t AND f)
	{
		int Sp = f->X(" ");
		int X = 0;

		char White[] = " \t\r\n";
		for (char *s = t; s AND *s; )
		{
			while (*s AND strchr(White, *s))
			{
				if (*s == '\n')
				{
					X = 0;
				}
				s++;
			}
			char *e = s;
			while (*e AND NOT strchr(White, *e)) e++;

			GDisplayString d(f, s, e - s);
			MyMin = max(d.X(), MyMin);
			X += d.X() + Sp;
			MyMax = max(X, MyMax);

			s = *e ? e + 1 : e;
		}
	}

	Min = max(Min, MyMin);
	Max = max(Max, MyMax);
}

int LayoutTextCtrl(GView *v, int Offset, int Width)
{
	int Ht = 0;

	char *t = v->Name();
	GFont *f = v->GetFont();
	if (t AND f)
	{
		int y = f->Y("A");

		char *e = t + strlen(t);
		char White[] = " \t\r\n";
		for (char *s = t; s AND *s; )
		{
			GDisplayString d(f, s, min(1000, e - s));
			int Ch = d.CharAt(Width);			

			e = LgiSeekUtf8(s, Ch);
			while (	*e AND
					*e != '\n' AND
					e > s AND
					NOT strchr(White, e[-1]))
			{
				e--;
			}

			s = *e ? e + 1 : e;
			Ht += y;
		}
	}

	return Ht;
}

enum CellAlign
{
	AlignMin,
	AlignCenter,
	AlignMax,
};

CellAlign ConvertAlign(char *s)
{
	if (s)
	{
		if (stricmp(s, "Center") == 0)
			return AlignCenter;
		if (stricmp(s, "Max") == 0)
			return AlignMax;
	}

	return AlignMin;
}

class TableCell : public GLayoutCell
{
public:
	GTableLayout *Table;
	GRect Cell;		// Cell position
	GRect Pos;		// Pixel position
	GArray<GView*> Children;
	CellAlign AlignX, AlignY;

	TableCell(GTableLayout *t, int Cx, int Cy)
	{
		AlignX = AlignY = AlignMin;
		Table = t;
		Cell.ZOff(0, 0);
		Cell.Offset(Cx, Cy);
	}

	bool GetVariant(char *Name, GVariant &Value, char *Array)
	{
		return false;
	}

	bool SetVariant(char *Name, GVariant &Value, char *Array)
	{
		if (stricmp(Name, "span") == 0)
		{
			GRect r;
			if (r.SetStr(Value.Str()))
			{
				Cell = r;
			}
			else return false;
		}
		else if (stricmp(Name, "children") == 0)
		{
			for (GVariant *v = Value.First(); v; v = Value.Next())
			{
				if (v->Type == GV_VOID_PTR)
				{
					ResObject *o = (ResObject*)v->Value.Ptr;
					if (o)
					{
						GView *gv = dynamic_cast<GView*>(o);
						if (gv)
						{
							Children.Add(gv);
							Table->Children.Insert(gv);
							gv->SetParent(Table);

							GText *t = dynamic_cast<GText*>(gv);
							if (t)
							{
								t->SetWrap(true);
							}
						}
					}
				}
			}
		}
		else if (stricmp(Name, "align") == 0)
		{
			AlignX = ConvertAlign(Value.Str());
		}
		else if (stricmp(Name, "valign") == 0)
		{
			AlignY = ConvertAlign(Value.Str());
		}
		else return false;

		return true;
	}

	/// Calculates the minimum and maximum widths this cell can occupy.
	void PreLayout(int &Min, int &Max, CellFlag &Flag)
	{
		for (int i=0; i<Children.Length(); i++)
		{
			GView *v = Children[i];
			if (Izza(GText))
			{
				PreLayoutTextCtrl(v, Min, Max);
				if (Max > Min)
				{
					Flag = SizeGrow;
				}
			}
			else if (Izza(GCheckBox))
			{
				int cmin = 0, cmax = 0;
				PreLayoutTextCtrl(v, cmin, cmax);
				Min = max(Min, cmin + 16);
				Max = max(Max, cmax + 16);
			}
			else if (Izza(GButton))
			{
				GFont *f = v->GetFont();
				char *t = v->Name();
				int x = (f AND t) ? max(v->X(), f->X(t) + 16) : v->X();
				if (x > v->X())
				{
					GRect r = v->GetPos();
					r.x2 = r.x1 + x - 1;
					v->SetPos(r);
				}

				Min = max(Min, x);
				Max = max(Max, x);
				Flag = SizeFixed;
			}
			else if (Izza(GEdit))
			{
				Min = max(Min, 40);
				Max = max(Max, 1000);
			}
			else if (Izza(GCombo))
			{
				GCombo *Cbo = Izza(GCombo);
				GFont *f = Cbo->GetFont();
				char *t = Cbo->Name();
				int x = (f AND t) ? f->X(t) + 32 : 200;
				Min = max(Min, 40);
				Max = max(Max, x);
			}
			else if (Izza(GList) OR
					 Izza(GTree))
			{
				Min = max(Min, 40);
				Max = max(Max, 1000);
				Flag = SizeGrow;
			}
			else
			{
				Min = max(Min, v->X());
				Max = max(Max, v->X());
			}
		}
	}

	/// Calculate the height of the cell based on the given width
	void Layout(int Width, int &Min, int &Max)
	{
		Pos.ZOff(Width-1, 0);
		for (int i=0; i<Children.Length(); i++)
		{
			GView *v = Children[i];
			if (v)
			{
				if (i)
				{
					Pos.y2 += CELL_SPACING;
				}

				if (Izza(GText))
				{
					int y = LayoutTextCtrl(v, 0, Pos.X());
					Pos.y2 += y;
					
					GRect r = v->GetPos();
					r.y2 = r.y1 + y - 1;
					v->SetPos(r);
				}
				else if (Izza(GEdit) OR
						 Izza(GButton) OR
						 Izza(GCombo))
				{
					int y = v->GetFont()->GetHeight() + 8;
					
					GRect r = v->GetPos();
					r.y2 = r.y1 + y - 1;
					v->SetPos(r);

					Pos.y2 += y;

					if (Izza(GEdit) AND
						Izza(GEdit)->MultiLine())
					{
						Max = max(Max, 1000);
					}
				}
				else if (Izza(GCheckBox))
				{
					int y = v->GetFont()->GetHeight() + 2;
					
					GRect r = v->GetPos();
					r.y2 = r.y1 + y - 1;
					v->SetPos(r);

					Pos.y2 += y;
				}
				else if (Izza(GList) OR
						 Izza(GTree))
				{
					Pos.y2 += v->GetFont()->GetHeight() + 8;
					Max = max(Max, 1000);
				}
				else
				{
					Pos.y2 += v->Y();
				}
			}
		}

		Min = max(Min, Pos.Y());
		Max = max(Max, Pos.Y());
	}

	/// Called after the layout has been done to move the controls into place
	void PostLayout()
	{
		int Cy = 0;
		for (int i=0; i<Children.Length(); i++)
		{
			GView *v = Children[i];
			if (v)
			{
				GRect r = v->GetPos();
				r.Offset(Pos.x1 - r.x1, Pos.y1 - r.y1 + Cy);
				
				if (Izza(GList) OR
					Izza(GTree) OR
					(Izza(GEdit) AND Izza(GEdit)->MultiLine()))
				{
					int Avail = Pos.Y() - Cy - v->Y();
					if (Avail > 0)
					{
						r.y2 += Avail;
					}
				}
				if (NOT Izza(GButton))
				{
					r.x2 = r.x1 + Pos.X() - 1;
				}

				if (AlignX == AlignCenter)
				{
					r.Offset( Pos.x1 + ((Pos.X() - r.X()) / 2) - r.x1, 0);
				}
				else if (AlignX == AlignMax)
				{
					r.Offset( Pos.x1 + (Pos.X() - r.X()) - r.x1, 0);
				}

				v->SetPos(r);
				Cy += r.Y() + CELL_SPACING;
			}
		}
	}
};

void DistributeSize(GArray<int> &a, GArray<CellFlag> &Flags, int Start, int Span, int Size, int Border)
{
	// Calculate the current size of the cells
	int Cur = -Border;
	for (int i=0; i<Span; i++)
	{
		Cur += a[Start+i] + Border;
	}

	// Is there more to allocate?
	if (Cur < Size)
	{
		// Get list of growable cells
		GArray<int> Pref;
		for (int i=Start; i<Start+Span; i++)
		{
			if (Flags[i] != SizeFixed)
			{
				Pref.Add(i);
			}
		}

		if (Pref.Length())
		{
			// Distribute size amongst the cells
			int Needs = Size - Cur;
			int Part = Needs / Pref.Length();
			for (int i=0; i<Pref.Length(); i++)
			{
				int Cell = Pref[i];
				int Add = i < Pref.Length() - 1 ? Part : Needs;
				a[i] = a[i] + Add;
				Needs -= Add;
			}
		}
		else
		{
			// Distribute size amongst the cells
			int Needs = Size - Cur;
			int Part = Needs / Span;
			for (int k=0; k<Span; k++)
			{
				int Add = k < Span - 1 ? Part : Needs;
				a[Start+k] = a[Start+k] + Add;
				Needs -= Add;
			}
		}
	}
}

class GTableLayoutPrivate
{
public:
	GArray<double> Rows, Cols;
	GArray<TableCell*> Cells;
	int CellSpacing;
	bool FirstLayout;

	GTableLayoutPrivate()
	{
		FirstLayout = false;
		CellSpacing = CELL_SPACING;
	}

	TableCell *GetCellAt(int cx, int cy)
	{
		for (int i=0; i<Cells.Length(); i++)
		{
			if (Cells[i]->Cell.Overlap(cx, cy))
			{
				return Cells[i];
			}
		}

		return 0;
	}

	void Layout(GRect &Client)
	{
		#if DEBUG_PROFILE
		int64 Start = LgiCurrentTime();
		#endif

		int Px = 0, Py = 0, Cx, Cy, i;
		GArray<int> MinCol, MaxCol;
		GArray<CellFlag> ColFlags, RowFlags;

		ColFlags.Length(Cols.Length());
		RowFlags.Length(Rows.Length());

		// Do pre-layout to determin minimum and maximum column sizes
		for (Cy=0; Cy<Rows.Length(); Cy++)
		{
			for (Cx=0; Cx<Cols.Length(); )
			{
				TableCell *c = GetCellAt(Cx, Cy);
				if (c)
				{
					// Non-spanned cells
					if (c->Cell.x1 == Cx AND
						c->Cell.y1 == Cy AND
						c->Cell.X() == 1 AND
						c->Cell.Y() == 1)
					{
						CellFlag Flag = SizeUnknown;
						c->PreLayout(MinCol[Cx], MaxCol[Cx], Flag);
						if (NOT ColFlags[Cx])
						{
							ColFlags[Cx] = Flag;
						}
					}

					Cx += c->Cell.X();
				}
				else break;
			}
		}

		// Pre-layout for spanned cells
		for (Cy=0; Cy<Rows.Length(); Cy++)
		{
			for (Cx=0; Cx<Cols.Length(); )
			{
				TableCell *c = GetCellAt(Cx, Cy);
				if (c)
				{
					if (c->Cell.x1 == Cx AND
						c->Cell.y1 == Cy AND
						(c->Cell.X() > 1 OR c->Cell.Y() > 1))
					{
						int Min = 0, Max = 0;
						CellFlag Flag = SizeUnknown;
						c->PreLayout(Min, Max, Flag);
						if (Flag)
						{
							for (i=c->Cell.x1; i<=c->Cell.x2; i++)
							{
								if (NOT ColFlags[i])
								{
									ColFlags[i] = Flag;
								}
							}
						}

						DistributeSize(MinCol, ColFlags, c->Cell.x1, c->Cell.X(), Min, CellSpacing);
						DistributeSize(MaxCol, ColFlags, c->Cell.x1, c->Cell.X(), Max, CellSpacing);
					}

					Cx += c->Cell.X();
				}
				else break;
			}
		}

		#if DEBUG_TRACE
		for (i=0; i<Cols.Length(); i++)
		{
			LgiTrace("ColBefore[%i]: min=%i max=%i\n", i, MinCol[i], MaxCol[i]);
		}
		#endif

		// Now allocate unused space
		DistributeUnusedSpace(MinCol, MaxCol, ColFlags, Client.X(), CellSpacing);

		#if DEBUG_TRACE
		for (i=0; i<Cols.Length(); i++)
		{
			LgiTrace("ColAfter[%i]: min=%i max=%i\n", i, MinCol[i], MaxCol[i]);
		}
		#endif

		// Do row height layout
		GArray<int> MinRow, MaxRow;
		for (Cy=0; Cy<Rows.Length(); Cy++)
		{
			for (Cx=0; Cx<Cols.Length(); )
			{
				TableCell *c = GetCellAt(Cx, Cy);
				if (c)
				{
					// Single cell
					if (c->Cell.x1 == Cx AND
						c->Cell.y1 == Cy)
					{
						int x = CountRange<int>(MinCol, c->Cell.x1, c->Cell.x2) +
								((c->Cell.X() - 1) * CellSpacing);

						c->Layout(x, MinRow[Cy], MaxRow[Cy]);
					}

					Cx += c->Cell.X();
				}
				else break;
			}
		}

		#if DEBUG_TRACE
		for (i=0; i<Rows.Length(); i++)
		{
			LgiTrace("RowBefore[%i]: min=%i max=%i\n", i, MinRow[i], MaxRow[i]);
		}
		#endif

		// Allocate remaining vertical space
		DistributeUnusedSpace(MinRow, MaxRow, RowFlags, Client.Y(), CellSpacing);
		
		#if DEBUG_TRACE
		for (i=0; i<Rows.Length(); i++)
		{
			LgiTrace("RowAfter[%i]: min=%i max=%i\n", i, MinRow[i], MaxRow[i]);
		}
		#endif

		#if DEBUG_TRACE
		for (i=0; i<Cols.Length(); i++)
		{
			LgiTrace("Final[%i]=%i\n", i, MinCol[i]);
		}
		#endif

		// Move cells into their final positions
		for (Cy=0; Cy<Rows.Length(); Cy++)
		{
			int MaxY = 0;
			for (Cx=0; Cx<Cols.Length(); )
			{
				TableCell *c = GetCellAt(Cx, Cy);
				if (c)
				{
					if (c->Cell.x1 == Cx AND
						c->Cell.y1 == Cy)
					{
						int y = CountRange<int>(MinRow, c->Cell.y1, c->Cell.y2) +
								((c->Cell.Y() - 1) * CellSpacing);

						c->Pos.y2 = c->Pos.y1 + y - 1;
						c->Pos.Offset(Client.x1 + Px, Client.y1 + Py);
						c->PostLayout();

						MaxY = max(MaxY, c->Pos.y2);
					}

					Px += c->Pos.X() + CellSpacing;
					Cx += c->Cell.X();
				}
				else break;
			}

			Py += MinRow[Cy] + CellSpacing;
			Px = 0;
		}

		#if DEBUG_PROFILE
		LgiTrace("GTableLayout::Layout = %i ms\n", (int)(LgiCurrentTime()-Start));
		#endif
	}
};

GTableLayout::GTableLayout() : ResObject(Res_Table)
{
	d = NEW(GTableLayoutPrivate);
	SetPourLargest(true);
}

GTableLayout::~GTableLayout()
{
	DeleteObj(d);
}

void GTableLayout::OnCreate()
{
	AttachChildren();
}

int GTableLayout::CellX()
{
	return d->Cols.Length();
}

int GTableLayout::CellY()
{
	return d->Rows.Length();
}

GLayoutCell *GTableLayout::CellAt(int x, int y)
{
	return d->GetCellAt(x, y);
}

void GTableLayout::OnPosChange()
{
	GRect r = GetClient();
	r.Size(CELL_SPACING, CELL_SPACING);
	d->Layout(r);
}

GRect GTableLayout::GetUsedArea()
{
	TableCell *c = d->GetCellAt(d->Cols.Length()-1, d->Rows.Length()-1);
	GRect r(0, 0, c ? c->Pos.x2 : 0, c ? c->Pos.y2 : 0);
	return r;
}

void GTableLayout::InvalidateLayout()
{
	d->FirstLayout = false;
	Invalidate();
}

void GTableLayout::OnPaint(GSurface *pDC)
{
	if (NOT d->FirstLayout)
	{
		d->FirstLayout = true;
		OnPosChange();
	}

	pDC->Colour(LC_MED, 24);
	pDC->Rectangle();

	#if 0
	for (int i=0; i<d->Cells.Length(); i++)
	{
		GRect r = d->Cells[i]->Pos;
		r.Size(-2, -2);
		pDC->Colour(Rgb24(192, 192, 222), 24);
		pDC->Box(&r);
		pDC->Line(r.x1, r.y1, r.x2, r.y2);
		pDC->Line(r.x2, r.y1, r.x1, r.y2);
	}
	#endif
}

bool GTableLayout::GetVariant(char *Name, GVariant &Value, char *Array)
{
	return false;
}

bool ConvertNumbers(GArray<double> &a, char *s)
{
	GToken t(s, ",");
	for (int i=0; i<t.Length(); i++)
	{
		a.Add(atof(t[i]));
	}
	return a.Length() > 0;
}

bool GTableLayout::SetVariant(char *Name, GVariant &Value, char *Array)
{
	if (stricmp(Name, "cols") == 0)
	{
		return ConvertNumbers(d->Cols, Value.Str());
	}
	else if (stricmp(Name, "rows") == 0)
	{
		return ConvertNumbers(d->Rows, Value.Str());
	}
	else if (stricmp(Name, "cell") == 0)
	{
		if (Array)
		{
			GToken t(Array, ",");
			if (t.Length() == 2)
			{
				int cx = atoi(t[0]);
				int cy = atoi(t[1]);
				TableCell *c = NEW(TableCell(this, cx, cy));
				if (c)
				{
					d->Cells.Add(c);
					if (Value.Type == GV_VOID_PTR)
					{
						GDom **d = (GDom**)Value.Value.Ptr;
						if (d)
						{
							*d = c;
						}
					}
				}
				else return false;
			}
			else return false;
		}
		else return false;
	}
	else return false;

	return true;
}

void GTableLayout::OnChildrenChanged(GView *Wnd, bool Attaching)
{
	if (NOT Attaching)
	{
		for (int i=0; i<d->Cells.Length(); i++)
		{
			TableCell *c = d->Cells[i];
			for (int n=0; n<c->Children.Length(); n++)
			{
				if (c->Children[n] == Wnd)
				{
					c->Children.DeleteAt(n);
					return;
				}
			}
		}

		LgiAssert(0);
	}
}
