#include <stdlib.h>
#include <stdio.h>

#include "Lgi.h"
#include "GTextLabel.h"

class GTextPrivate : public GSemaphore
{
public:
	int Mx, My;
	List<GDisplayString> Strs;
	COLOUR Background;
	bool Wrap;

	GTextPrivate()
	{
		Wrap = false;
		Mx = My = 0;
		Background = -1;
	}

	~GTextPrivate()
	{
		Strs.DeleteObjects();
	}

	void Layout(GFont *f, char *s, int Width)
	{
		if (Lock())
		{
			Mx = My = 0;
			Strs.DeleteObjects();
			while (s AND *s)
			{
				char *e = strchr(s, '\n');
				if (NOT e) e = s + strlen(s);

				GDisplayString *n = NEW(GDisplayString(f, s, (int)e - (int)s));
				if (Wrap)
				{
					int All = LgiCharLen((OsChar*)*n, LGI_WideCharset);
					int Ch = n->CharAt(Width);
					if (Ch == 0)
					{
						break;
					}
					else if (Ch < All)
					{
						// Break string into chunks
						e = LgiSeekUtf8(s, Ch);
						while (e > s AND e[-1] != ' ')
							e--;
						if (e == s)
						{
							e = LgiSeekUtf8(s, Ch);
							while (*e AND *e != ' ')
								e++;
						}

						DeleteObj(n);
						n = NEW(GDisplayString(f, s, (int)e - (int)s));
					}
				}

				Strs.Insert(n);
				s = *e == '\n' ? e + 1 : e;

				if (n)
				{
					Mx = max(Mx, n->X());
					My += n->Y();
				}
			}
			Unlock();
		}
	}
};

GText::GText(int id, int x, int y, int cx, int cy, char *name) :
	ResObject(Res_StaticText)
{
	d = NEW(GTextPrivate);
	Name(name);

	if (cx < 0) cx = d->Mx;
	if (cy < 0) cy = d->My;

	GRect r(x, y, x+cx, y+cy);
	SetPos(r);
	SetId(id);

	#ifdef WIN32
	CreateClass("Lgi.TextLabel");
	#endif
}

GText::~GText()
{
	DeleteObj(d);
}

bool GText::GetWrap()
{
	return d->Wrap;
}

void GText::SetWrap(bool b)
{
	d->Wrap = b;
	d->Layout(GetFont(), GObject::Name(), X());
	Invalidate();
}

bool GText::Name(char *n)
{
	bool Status = GView::Name(n);
	d->Layout(GetFont(), GObject::Name(), X());
	Invalidate();
	return Status;
}

bool GText::NameW(char16 *n)
{
	bool Status = GView::NameW(n);
	d->Layout(GetFont(), GObject::Name(), X());
	Invalidate();
	return Status;
}

void GText::SetFont(GFont *Fnt, bool OwnIt)
{
	GView::SetFont(Fnt, OwnIt);
	d->Layout(GetFont(), GObject::Name(), X());
	Invalidate();
}

int GText::Value()
{
	int Status = 0;
	char *n = Name();
	if (n)
	{
		Status = atoi(n);
	}
	return Status;
}

void GText::Value(int i)
{
	char n[32];
	sprintf(n, "%i", i);
	Name(n);
}

void GText::OnPosChange()
{
	if (d->Wrap)
	{
		d->Layout(GetFont(), GObject::Name(), X());
	}
}

void GText::OnPaint(GSurface *pDC)
{
	bool Status = false;
	COLOUR Back = d->Background == -1 ? LC_MED : d->Background;
	
	GFont *f = GetFont();
	if (d->Lock())
	{
		if (d->Strs.First())
		{
			f->Transparent(false);
			f->Colour(LC_TEXT, Back);

			int y = 0;
			for (GDisplayString *s = d->Strs.First(); s; s = d->Strs.Next())
			{
				if (Enabled())
				{
					GRect r(0, y, X()-1, y + s->Y() - 1);
					s->Draw(pDC, 0, y, &r);
				}
				else
				{
					GRect r(0, y, X()-1, y + s->Y() - 1);
					f->Transparent(false);
					f->Colour(LC_LIGHT, LC_MED);
					s->Draw(pDC, 1, y+1, &r);
					
					f->Transparent(true);
					f->Colour(LC_LOW, LC_MED);
					s->Draw(pDC, 0, y, &r);
				}
				
				y += s->Y();
			}

			if (y < Y())
			{
				pDC->Colour(Back, 24);
				pDC->Rectangle(0, y, X()-1, Y()-1);
			}
			Status = true;
		}
		d->Unlock();
	}
	
	if (NOT Status)
	{
		pDC->Colour(Back, 24);
		pDC->Rectangle();
	}
}

COLOUR GText::Background()
{
	return d->Background;
}

void GText::Background(COLOUR c)
{
	if (d->Background != c)
	{
		d->Background = c;
		Invalidate();
	}
}

