/*
**	FILE:			GToolbar.cpp
**	AUTHOR:			Matthew Allen
**	DATE:				18/10/2001
**	DESCRIPTION:	Toolbar classes
**
**	Copyright (C) 2001, Matthew Allen
**		fret@memecode.com
*/

#include <stdlib.h>
#include <stdio.h>
#include "Lgi.h"
#include "GToken.h"

#ifdef WIN32

HPALETTE GetSystemPalette();
bool BltBmpToBmp(HBITMAP hDest, int xDst, int yDst, int cx, int cy, HBITMAP hSrc, int xSrc, int ySrc, DWORD dwRop);
bool BltBmpToDc(HDC DestDC, int xDst, int yDst, int cx, int cy, HBITMAP hSrc, int xSrc, int ySrc,  DWORD dwRop);
bool BltDcToBmp(HBITMAP hDest, int xDst, int yDst, int cx, int cy, HDC SrcDC, int xSrc, int ySrc, DWORD dwRop);

#define AttachButton(b)	Children.Insert(b);

#else

#define AttachButton(b) b->Attach(this);

#endif

#ifdef BEOS
#include <Bitmap.h>
#endif

COLOUR Map(GSurface *pDC, COLOUR c);

////////////////////////////////////////////////////////////////////////
GImageList *LgiLoadImageList(char *File, int x, int y)
{
	GImageList *ImgList = 0;
	char *Path = FileExists(File) ? NewStr(File) : LgiFindFile(File);
	if (Path)
	{
		GSurface *pDC = LoadDC(Path);
		if (pDC)
		{
			ImgList = NEW(GImageList(x, y, pDC));
			
			#ifdef WIN32
			if (ImgList)
			{
				ImgList->Create(pDC->X(), pDC->Y(), GdcD->GetBits());
			}
			#endif

			DeleteObj(pDC);
		}

		DeleteArray(Path);
	}
	else
	{
		printf("LgiLoadImageList: Couldn't find '%s'\n", File);
	}

	return ImgList;
}

GToolBar *LgiLoadToolbar(GView *Parent, char *File, int x, int y)
{
	GToolBar *Toolbar = NEW(GToolBar);
	if (Toolbar)
	{
		char *FileName = LgiFindFile(File);
		if (FileName)
		{
			bool Success = FileName AND Toolbar->SetBitmap(FileName, x, y);
			if (NOT Success)
			{
				LgiMsg(Parent,
						"Can't load '%s' for the toolbar.\n"
						"You can find it in this program's archive.",
						"Lgi::LgiLoadToolbar",
						MB_OK,
						File);
				DeleteObj(Toolbar);
			}

			DeleteArray(FileName);
		}
		else
		{
			LgiMsg(Parent,
					"Can't find the graphic '%s' for the toolbar.\n"
					"You can find it in this program's archive.",
					"Lgi::LgiLoadToolbar",
					MB_OK,
					File);
		}
	}

	return Toolbar;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
#define ImgLst_Empty	0x40000000
#define IgmLst_Add		0x80000000

GImageList::GImageList(int x, int y, GSurface *pDC)
{
	Sx = x;
	Sy = y;
	pCreate = pDC;
	Bounds = 0;

	#if defined BEOS

	if (pDC)
	{
		int Bits = (GdcD->GetBits() == 32) ? 32 : 15;
		if (Create(pDC->X(), pDC->Y(), Bits))
		{
			GPalette *Pal = pDC->Palette();
			if (Pal)
			{
				Palette(NEW(GPalette(Pal)));
			}

			Blt(0, 0, pDC);
			
			COLOUR InitKey = Get(0, 0);

			for (int cy=0; cy<Y(); cy++)
			{
				switch (GetBits())
				{
					case 8:
					{
						uchar Key = InitKey;
						uchar *p = (*this)[cy];
						for (int cx=0; cx<X(); cx++)
						{
							if (p[cx] == Key)
							{
								p[cx] = B_TRANSPARENT_MAGIC_CMAP8;
							}
						}
						break;
					}
					case 15:
					{
						ushort Key = InitKey;
						ushort *p = (ushort*) (*this)[cy];
						ushort Trans = B_TRANSPARENT_MAGIC_RGBA15;
						
						for (int cx=0; cx<X(); cx++)
						{
							if (p[cx] == Trans)
							{
								p[cx] = Trans+1;
							}
							else if (p[cx] == Key)
							{
								p[cx] = Trans;
							}
						}
						break;
					}
					case 32:
					{
						ulong Key = InitKey;
						ulong *p = (ulong*) ((*this)[cy]);
						for (int cx=0; cx<X(); cx++)
						{
							if (	R32(p[cx]) == R32(Key) AND
									G32(p[cx]) == G32(Key) AND
									B32(p[cx]) == B32(Key))
							{
								p[cx] = B_TRANSPARENT_MAGIC_RGBA32;
							}
						}
						break;
					}
				}

			}
		}
	}

	#else

	if (pDC AND
		Create(pDC->X(), pDC->Y(), GdcD->GetBits()))
	{
		Blt(0, 0, pDC);
		
		#ifdef LINUX

		GetBounds();

		IsAlpha(true);
		GSurface *Alpha = AlphaDC();
		if (Alpha)
		{
			COLOUR Key = Get(0, 0);
			for (int y=0; y<Y(); y++)
			{
				uchar *a = (*Alpha)[y];
				for (int x=0; x<X(); x++)
				{
					COLOUR c = Get(x, y);
					*a++ = c == Key ? 0 : 0xff;
				}
			}
			
			Unlock();
		}
		
		
		#endif
	}

	#endif
}

GImageList::~GImageList()
{
	DeleteArray(Bounds);
}

bool GImageList::Create(int x, int y, int Bits, int LineLen, bool KeepData)
{
	#ifdef WIN32

	bool Status = GMemDC::Create(x, y, (Bits < 8) ? 24 : Bits, LineLen, KeepData);
	if (Status)
	{
		if (pCreate)
		{
			RGBQUAD *Rgb = GetInfo()->bmiColors;

			if (GdcD->GetBits() <= 8)
			{
				// create palette system palette
				PALETTEENTRY Entries[256];
				HDC hDC = CreateCompatibleDC(0);
				GetSystemPaletteEntries(hDC, 0, 256, Entries);
				DeleteDC(hDC);

				GPalette *Pal = NEW(GPalette(0, 256));
				if (Pal)
				{
					for (int i=0; i<256; i++)
					{
						GdcRGB *p = (*Pal)[i];
						if (p)
						{
							Rgb[i].rgbRed = p->R = Entries[i].peRed;
							Rgb[i].rgbGreen = p->G = Entries[i].peGreen;
							Rgb[i].rgbBlue = p->B = Entries[i].peBlue;
							Rgb[i].rgbReserved = 0;
						}
					}
	
					Palette(Pal);
				}
			}

			if (pCreate->GetBits() == 8 AND GetBits() == 8)
			{
				// Set the palette
				HDC hDC = StartDC();
				if (hDC)
				{
					SetDIBColorTable(hDC, 0, 256, Rgb);
					EndDC();
				}

				// gotta remap here
				char Remap[256];
				GPalette *SrcPal = pCreate->Palette();
				GPalette *DstPal = Palette();
				if (SrcPal AND DstPal)
				{
					// build remap table
					ZeroObj(Remap);
					for (int i=0; i<256; i++)
					{
						GdcRGB *From = (*SrcPal)[i];
						if (From)
						{
							Remap[i] = DstPal->MatchRgb(Rgb24(From->R, From->G, From->B));
						}
					}

					// do blt
					for (int y=0; y<Y(); y++)
					{
						uchar *s = (*pCreate)[y];
						uchar *d = (*this)[y];
						for (int x=0; x<X(); x++)
						{
							*d++ = Remap[*s++];
						}
					}
				}
			}
			else
			{
				// just standard copy blt
				Blt(0, 0, pCreate);
			}
		}
	}
	return Status;

	#else

	return GMemDC::Create(x, y, Bits, LineLen, KeepData);

	#endif
}

void GImageList::Update(int Flags)
{
}

void GImageList::Draw(GSurface *pDest, int Dx, int Dy, int Image, int Flags)
{
	if (pDest)
	{
		GRect r;
		r.ZOff(Sx-1, Sy-1);
		r.Offset(Image * Sx, 0);

		if ((*this)[0])
		{
			COLOUR Key = Get(0, 0);

			if (Flags & IMGLST_DISABLED)
			{
				COLOUR Low = CBit(pDest->GetBits(), LC_LOW);
				COLOUR High = CBit(pDest->GetBits(), LC_LIGHT);

				for (int y=0; y<r.Y(); y++)
				{
					for (int x=0; x<r.X(); x++)
					{
						if (Get(r.x1 + x, r.y1 + y) != Key)
						{
							pDest->Colour(Low);
							pDest->Set(Dx + x, Dy + y);
							pDest->Colour(High);
							pDest->Set(Dx + x + 1, Dy + y + 1);
						}
					}
				}
			}
			else
			{
				#if defined BEOS

				GScreenDC *pViewDC = dynamic_cast<GScreenDC*>(pDest);
				if (pDest->IsScreen() AND pViewDC)
				{
					BView *owner = pViewDC->Handle();
					BBitmap *Bmp = GetBitmap();

					if (owner AND Bmp)
					{
						int x = Image * Sx;
						BRect S(x, 0, x+Sx-1, Sy-1);
						BRect D(Dx, Dy, Dx+(Sx-1), Dy+(Sx-1));

						drawing_mode Mode = owner->DrawingMode();
						owner->SetDrawingMode(B_OP_OVER);
						owner->DrawBitmap(Bmp, S, D);
						owner->SetDrawingMode(Mode);
					}
				}
				else

				#elif defined LINUX

				GMemDC Buf(r.X(), r.Y(), GdcD->GetBits());
				GSurface *Screen = pDest;
				GRect Src;
				Src.ZOff(r.X()-1, r.Y()-1);
				Src.Offset(Dx, Dy);
				Buf.Blt(0, 0, Screen, &Src);
				int ScrX = Dx;
				int ScrY = Dy;
				pDest = &Buf;
				Dx = Dy = 0;

				#endif
				{
					COLOUR Map[256];
					int b = GetBits();
					if (b == 8)
					{
						GPalette *Pal = Palette();
						if (Pal)
						{
							for (int i=0; i<Pal->GetSize(); i++)
							{
								Map[i] = CBit(24, i, 8, Pal);
							}
						}
					}

					for (int y=0; y<r.Y(); y++)
					{
						if (b == 8)
						{
							// Paletted
							for (int x=0; x<r.X(); x++)
							{
								COLOUR c = Get(r.x1 + x, r.y1 + y);
								if (c != Key)
								{
									pDest->Colour(Map[c], 24);
									pDest->Set(Dx + x, Dy + y);
								}
							}
						}
						else
						{
							// True colour
							for (int x=0; x<r.X(); x++)
							{
								COLOUR c = Get(r.x1 + x, r.y1 + y);
								if (c != Key)
								{
									pDest->Colour(c, b);
									pDest->Set(Dx + x, Dy + y);
								}
							}
						}
					}
				}
				
				#if defined LINUX
				Screen->Blt(ScrX, ScrY, pDest);
				#endif
			}
		}
		else
		{
			// Pixmap!
			pDest->Blt(Dx, Dy, this, &r);
		}
	}
}

GRect *GImageList::GetBounds()
{
	if (NOT Bounds AND (*this)[0])
	{
		Bounds = NEW(GRect[GetItems()]);
		if (Bounds)
		{
			COLOUR Key = Get(0, 0);

			for (int i=0; i<GetItems(); i++)
			{
				Bounds[i].x1 = Sx-1;
				Bounds[i].y1 = Sy-1;
				Bounds[i].x2 = 0;
				Bounds[i].y2 = 0;

				int Bx = i * Sx;
				for (int y=0; y<Sy; y++)
				{
					for (int x=0; x<Sx; x++)
					{
						if (Get(Bx+x, y) != Key)
						{
							Bounds[i].x1 = min(Bounds[i].x1, x);
							Bounds[i].y1 = min(Bounds[i].y1, y);
							Bounds[i].x2 = max(Bounds[i].x2, x);
							Bounds[i].y2 = max(Bounds[i].y2, y);
						}
					}
				}
			}
		}
	}

	return Bounds;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
class GToolBarPrivate
{
public:
	int Bx, By;
	int Sx, Sy;
	bool Vertical;
	bool Text;
	int LastIndex;
	bool OwnImgList;
	GImageList *ImgList;
	GFont *Font;
	GToolTip *Tip;
	
	// Customization menu
	ObjProperties *CustomStore;
	char *CustomProp;

	// bitmap cache
	GMemDC *pColour;
	GMemDC *pDisabled;

	GToolBarPrivate()
	{
		Bx = By = 16;
		Sx = Sy = 10;
		Vertical = false;
		Text = false;
		Font = 0;
		Tip = 0;
		CustomStore = 0;
		CustomProp = 0;
	}
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////
GToolButton::GToolButton(int Bx, int By)
#ifdef WIN32
	: GView(0)
#else
	: GView(NEW(DefaultOsView(this)))
#endif
{
	Type = TBT_PUSH;
	SetId(IDM_NONE);
	SetDown(false);
	Clicked = false;
	Over = false;
	ImgIndex = -1;

	GRect r(0, 0, Bx+1, By+1);
	SetPos(r);
	SetParent(0);
	TipId = -1;

	_BorderSize = 0;
	#ifdef BEOS
	Handle()->SetViewColor(B_TRANSPARENT_COLOR);
	#endif
}

GToolButton::~GToolButton()
{
}

void GToolButton::OnPaint(GSurface *pDC)
{
	GToolBar *Par = (GToolBar*) GetParent();
	bool e = Enabled();

	if (Par)
	{
		GRect p = GetClient();

		// Draw Background
		COLOUR Grey = (e AND Over) ? LC_HIGH : LC_MED;

		int BaseX = (X() - (2*BORDER_BUTTON) - Par->d->Bx) / 2;
		int Off = BORDER_BUTTON + ((Down) ? 1 : 0);

		if (GetId() >= 0)
		{
			// Button
			int Flags = IMGLST_GDC;

			// Draw border
			if (Down)
			{
				LgiThinBorder(pDC, p, SUNKEN);
			}
			else if (e AND Over)
			{
				LgiThinBorder(pDC, p, RAISED);
			}

			pDC->Colour(Grey, 24);
			pDC->Rectangle(&p);
			pDC->ClipRgn(&p);

			if (Par->d->ImgList)
			{
				if (ImgIndex >= 0)
				{
					if (e AND Over)
					{
						// Draw transparent
						Par->d->ImgList->Draw(pDC, BaseX+Off, Off, ImgIndex, Flags);
					}
					else
					{
						// Draw cached
						Par->_DrawFromCache(pDC, BaseX+Off, Off, ImgIndex, NOT e);
					}
				}
			}
			else
			{
				// Draw a red cross indicating no icons.
				pDC->Colour(Rgb24(255, 0, 0), 24);
				pDC->Line(p.x1, p.y1, p.x2, p.y2);
				pDC->Line(p.x2, p.y1, p.x1, p.y2);
			}

			pDC->ClipRgn(0);

			// Text
			char *s = Name();
			if (Par->d->Text AND
				Par->d->Font AND
				s)
			{
				// Write each word centered on a different line
				int Ty = Off + Par->d->By + 2;
				COLOUR a = e ? LC_TEXT : LC_LOW;
				COLOUR b = LC_MED;
				char Buf[256] = "", *e;

				Par->d->Font->Colour(a, b);

				for (e = s; s AND *s; s = *e ? e + 1 : e)
				{
					char *Last = Buf + strlen(Buf);
					char *Add = Last;
					char *Src = s;
					int PrevX = 0;
					int Sy = 0;
					
					#define iswhite(c) ( (c)==' ' OR (c)=='\t' )

					while (*s)
					{
						// Add a word
						while (iswhite(*s))
						{
							*Add++ = *s++;
						}
						while (*s AND NOT iswhite(*s))
						{
							if (*s == '&')
							{
								s++;
							}

							*Add++ = *s++;
						}
						*Add = 0;

						// Measure it
						int Sx = 0;
						Par->d->Font->Size(&Sx, &Sy, Buf);
						if (Sx > X() - 2)
						{
							// Too long... revert to last known size
							*Last = 0;
							break;
						}

						Last = Add;
						Src = s;
						PrevX = Sx;
					}

					Par->d->Font->Text(	pDC,
										Off + ((X()-PrevX)/2),
										Ty,
										Buf);
					e = Src;
					Buf[0] = 0;
					Ty += Sy;
				}
			}
		}
		else
		{
			// Separator
			int Px = X()-1;
			int Py = Y()-1;
			pDC->Colour(Grey, 24);
			pDC->Rectangle();

			if (X() > Y())
			{
				int c = Y()/2-1;
				pDC->Colour(Map(pDC, LC_LOW), 24);
				pDC->Line(2, c, Px-2, c);
				pDC->Colour(Map(pDC, LC_LIGHT), 24);
				pDC->Line(2, c+1, Px-2, c+1);
			}
			else
			{
				int c = X()/2-1;
				pDC->Colour(Map(pDC, LC_LOW), 24);
				pDC->Line(c, 2, c, Py-2);
				pDC->Colour(Map(pDC, LC_LIGHT), 24);
				pDC->Line(c+1, 2, c+1, Py-2);
			}
		}
	}
}

void GToolButton::Image(int i)
{
	if (ImgIndex != i)
	{
		ImgIndex = i;
		Invalidate();
	}
}

void GToolButton::Value(int b)
{
	switch (Type)
	{
		case TBT_PUSH:
		{
			// do nothing... can't set value
			break;
		}

		case TBT_TOGGLE:
		{
			if (Value() != b)
			{
				SetDown(b);
				Invalidate();
			}
			break;
		}

		case TBT_RADIO:
		{
			if (GetParent() AND b)
			{
				// Clear any other radio buttons that are down
				GToolButton *But;
				for (	But = (GToolButton*) (GetParent()->Children.HasItem(this) ? this : 0);
						But AND But->GetId() >= 0;
						But = (GToolButton*) GetParent()->Children.Next())
				{
					if (But->Type == TBT_RADIO AND
						But != this AND
						But->Down)
					{
						But->SetDown(false);
						But->Invalidate();
					}
				}

				for (	But = (GToolButton*) (GetParent()->Children.HasItem(this) ? this : 0);
						But AND But->GetId() >= 0;
						But = (GToolButton*) GetParent()->Children.Prev())
				{
					if (But->Type == TBT_RADIO AND
						But != this AND
						But->Down)
					{
						But->SetDown(false);
						But->Invalidate();
					}
				}
			}

			SetDown(b);
			Invalidate();
			break;
		}
	}
}

void GToolButton::OnCommand()
{
	if (GetParent())
	{
		((GToolBar*) GetParent())->OnButtonClick(this);
	}
}

void GToolButton::SendNotify(int Flags)
{
	GView *w = (GetNotify()) ? GetNotify() : GetParent();
	if (w)
	{
		w->PostEvent(M_CHANGE, (int)((GView*)this), Flags);
	}
}

void GToolButton::OnMouseClick(GMouse &m)
{
	GToolBar *ToolBar = (GToolBar*) GetParent();

	if (ToolBar AND
		ToolBar->d->CustomStore AND
		m.Right())
	{
		m.ToScreen();
		ToolBar->ContextMenu(m);
	}
	else
	{
		// left click action...
		if (GetId() >= 0 AND Enabled())
		{
			switch (Type)
			{
				case TBT_PUSH:
				{
					bool Old = Down;
	
					Clicked = m.Down();
					Capture(m.Down());
					
					if (Old AND IsOver(m))
					{
						if (m.Left())
						{
							OnCommand();
						}
						SendNotify(m.Flags);
					}

					SetDown(m.Down());
					if (Old != Down)
					{
						Invalidate();
					}
					break;
				}

				case TBT_TOGGLE:
				{
					if (m.Down())
					{
						if (m.Left())
						{
							Value(NOT Down);
							OnCommand();
						}
						SendNotify(m.Flags);
					}
					break;
				}

				case TBT_RADIO:
				{
					if (m.Down())
					{
						if (NOT Down AND m.Left())
						{
							Value(true);
							OnCommand();
						}
						SendNotify(m.Flags);
					}
					break;
				}
			}
		}
		else
		{
			printf("%s:%i\n", __FILE__, __LINE__);
		}
	}
}

void GToolButton::OnMouseEnter(GMouse &m)
{
	if (NOT Separator() AND Enabled())
	{
		Over = true;
		Invalidate();
	}

	if (Clicked)
	{
		SetDown(true);
		Invalidate();
	}
	else
	{
		GToolBar *Bar = dynamic_cast<GToolBar*>(GetParent());
		if (Bar)
		{
			Bar->OnMouseEnter(m);
			if (Bar->d->Tip AND TipId < 0)
			{
				TipId = Bar->d->Tip->NewTip(Name(), GetPos());
			}
		}

		if (GetParent())
		{
			((GToolBar*) GetParent())->PostDescription(this, Name());
		}
	}
}

void GToolButton::SetDown(bool d)
{
	Down = d;
}

void GToolButton::OnMouseMove(GMouse &m)
{
	#ifdef BEOS
	if (GetParent())
	{
		((GToolBar*) GetParent())->PostDescription(this, Name());
	}
	#endif
}

void GToolButton::OnMouseExit(GMouse &m)
{
	if (Over)
	{
		Over = false;
		Invalidate();
	}

	if (Clicked)
	{
		SetDown(false);
		Invalidate();
	}
	else if (GetParent())
	{
		((GToolBar*) GetParent())->PostDescription(this, "");
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////
GToolBar::GToolBar()
{
	d = NEW(GToolBarPrivate);
	Name("LGI_Toolbar");
	_BorderSize = 1;
	Raised(true);

	// Setup tool button font
	GFontType SysFontType;
	if (SysFontType.GetSystemFont("Small"))
	{
		d->Font = SysFontType.Create();
		if (d->Font)
		{
			d->Font->PointSize(min(d->Font->PointSize(), SysFont->PointSize()));
			d->Font->Colour(0);
			d->Font->Bold(false);
			d->Font->Transparent(true);
		}
	}

	d->LastIndex = 0;
	d->OwnImgList = false;
	d->ImgList = 0;
	d->pColour = 0;
	d->pDisabled = 0;

	#if defined BEOS
	Handle()->SetViewColor(B_TRANSPARENT_COLOR);
	#endif
}

GToolBar::~GToolBar()
{
	DeleteObj(d->Tip);
	_BuildCache(0);

	if (d->OwnImgList)
	{
		DeleteObj(d->ImgList);
	}

	DeleteObj(d->Font);
	DeleteObj(d);
}

int GToolBar::GetBx()
{
	return d->Bx;
}

int GToolBar::GetBy()
{
	return d->By;
}

void GToolBar::ContextMenu(GMouse &m)
{
	if (d->CustomStore AND d->CustomProp)
	{
		GSubMenu *Sub = NEW(GSubMenu);
		if (Sub)
		{
			int n = 1;
			GView *v;
			for (v = Children.First(); v; v = Children.Next(), n++)
			{
				GToolButton *Btn = dynamic_cast<GToolButton*>(v);
				if (Btn AND Btn->Separator())
				{
					Sub->AppendSeparator();
				}
				else
				{
					GMenuItem *Item = Sub->AppendItem(v->Name(), n, true);
					if (Item)
					{
						Item->Checked(v->Visible());
					}
				}
			}
			Sub->AppendSeparator();
			GMenuItem *Txt = Sub->AppendItem(LgiLoadString(L_TOOLBAR_SHOW_TEXT, "Show Text Labels"), 1000, true);
			Txt->Checked(d->Text);

			bool Save = false;
			int Pick = Sub->Float(this, m.x, m.y);
			switch (Pick)
			{
				case 1000:
				{
					d->Text = NOT d->Text;
					Save = true;
					break;
				}
				default:
				{
					GView *Ctrl = Children[Pick - 1];
					if (Ctrl)
					{
						Ctrl->Visible(NOT Ctrl->Visible());
						Save = true;
					}
					break;
				}
			}
			
			DeleteObj(Sub);
			
			if (Save)
			{
				GStringPipe p(256);
				p.Push((char*) (d->Text ? "text" : "no"));
				for (v = Children.First(); v; v = Children.Next())
				{
					if (v->Visible())
					{
						p.Printf(",%i", v->GetId());
					}
				}
				char *o = p.NewStr();
				if (o)
				{
					d->CustomStore->Set(d->CustomProp, o);
					DeleteArray(o);
				}
				
				FixSeparators();

				for (GView *v = Children.First(); v; v = Children.Next())
				{
					GToolButton *b = dynamic_cast<GToolButton*>(v);
					if (b AND b->TipId >= 0)
					{
						d->Tip->DeleteTip(b->TipId);
						b->TipId = -1;
					}
				}

				GetWindow()->Pour();
			}
		}
	}
}

bool GToolBar::Customizable()
{
	return d->CustomStore != 0;
}

void GToolBar::FixSeparators()
{
	// Fix up separators so that no 2 separators are next to each other. I.e.
	// all the buttons between them are switched off.
	GToolButton *Last = 0;
	bool HasVis = false;
	for (GView *v = Children.First(); v; v = Children.Next())
	{
		GToolButton *Btn = dynamic_cast<GToolButton*>(v);
		if (Btn)
		{
			if (Btn->Separator())
			{
				Btn->Visible(HasVis);
				if (HasVis)
				{
					Last = Btn;
				}
				HasVis = false;
			}
			else
			{
				HasVis |= Btn->Visible();
			}
		}
	}
	
	if (Last)
	{
		Last->Visible(HasVis);
	}
}

void GToolBar::Customizable(ObjProperties *Store, char *Option)
{
	d->CustomStore = Store;
	d->CustomProp = Option;
	
	if (d->CustomStore)
	{
		char *o;
		if (d->CustomStore->Get(d->CustomProp, o))
		{
			GToken t(o, ",");
			if (t.Length() >= 1)
			{
				d->Text = stricmp(t[0], "text") == 0;
				
				// Make all controls not visible.
				GView *v;
				for (v = Children.First(); v; v = Children.Next())
				{
					GToolButton *Btn = dynamic_cast<GToolButton*>(v);
					if (Btn) v->Visible(false);
				}
				
				// Set sub-set of ctrls visible according to saved ID list
				for (int i=1; i<t.Length(); i++)
				{
					SetCtrlVisible(atoi(t[i]), true);
				}
				
				FixSeparators();				
			}
		}
	}
}

bool GToolBar::IsVertical()
{
	return d->Vertical;
}

void GToolBar::IsVertical(bool v)
{
	d->Vertical = v;
}

bool GToolBar::TextLabels()
{
	return d->Text;
}

void GToolBar::TextLabels(bool i)
{
	d->Text = i;
}

GFont *GToolBar::GetFont()
{
	return d->Font;
}

void GToolBar::_BuildCache(GImageList *From)
{
	DeleteObj(d->pColour);
	DeleteObj(d->pDisabled);

	if (From)
	{
		#ifdef BEOS
		int Bits = From->GetBits();
		#else
		int Bits = GdcD->GetBits();
		#endif

		if (Bits == 8)
		{
			// Generate global palette for the full colour and
			// alpha blended versions of the icons by colour reducing
			// them both on the same DC.
			GSurface *r = NEW(GMemDC(From->X(), From->Y() * 2, 24));
			if (r)
			{
				// Background
				r->Colour(LC_MED, 24);
				r->Rectangle();

				// Draw the full colour icons
				int i;
				for (i=0; i<From->GetItems(); i++)
				{
					From->Draw(r, i * From->TileX(), 0, i);
				}

				// Draw the alpha blended icons
				r->Op(GDC_ALPHA);
				GApplicator *pApp = r->Applicator();
				if (pApp)
				{
					pApp->SetVar(GAPP_ALPHA_A, 40);
					pApp->SetVar(GAPP_ALPHA_PAL, (int)r->Palette());
				}
				for (i=0; i<From->GetItems(); i++)
				{
					From->Draw(r, i * From->TileX(), From->TileY(), i);
				}

				// Put the full colour icons in their DC
				d->pColour = NEW(GMemDC);
				if (d->pColour AND
					d->pColour->Create(From->X(), From->Y(), Bits))
				{
					d->pColour->Palette(NEW(GPalette(GdcD->GetGlobalColour()->GetPalette())));

					int OldType = GdcD->GetOption(GDC_REDUCE_TYPE);
					GdcD->SetOption(GDC_REDUCE_TYPE, REDUCE_ERROR_DIFFUSION);
					d->pColour->Blt(0, 0, r);
					GdcD->SetOption(GDC_REDUCE_TYPE, OldType);
				}

				// Put the alpha blended icons in their DC
				//
				// Because the colour and alpha icons have the same palette
				// they should appear fine together on the screen, as the 2
				// palettes are the same.
				d->pDisabled = NEW(GMemDC);
				if (d->pDisabled AND
					d->pDisabled->Create(From->X(), From->Y(), Bits))
				{
					d->pDisabled->Palette(NEW(GPalette(GdcD->GetGlobalColour()->GetPalette())));

					GRect t(0, From->Y(), From->X()-1, (From->Y() * 2)-1);

					int OldType = GdcD->GetOption(GDC_REDUCE_TYPE);
					GdcD->SetOption(GDC_REDUCE_TYPE, REDUCE_ERROR_DIFFUSION);
					d->pDisabled->Blt(0, 0, r, &t);
					GdcD->SetOption(GDC_REDUCE_TYPE, OldType);
				}

				DeleteObj(r);
			}
		}
		else
		{
			From->Lock();
			
			// True colour screen
			d->pColour = NEW(GMemDC);
			if (d->pColour AND
				d->pColour->Create(From->X(), From->Y(), Bits))
			{
				d->pColour->Colour(LC_MED, 24);
				d->pColour->Rectangle();

				for (int i=0; i<From->GetItems(); i++)
				{
					From->Draw(d->pColour, i * From->TileX(), 0, i);
				}
			}

			d->pDisabled = NEW(GMemDC);
			if (d->pDisabled AND
				d->pDisabled->Create(From->X(), From->Y(), Bits == 8 ? 24 : Bits))
			{
				d->pDisabled->Colour(LC_MED, 24);
				d->pDisabled->Rectangle();

				d->pDisabled->Op(GDC_ALPHA);
				GApplicator *pApp = d->pDisabled->Applicator();
				if (pApp)
				{
					pApp->SetVar(GAPP_ALPHA_A, 40);
					pApp->SetVar(GAPP_ALPHA_PAL, (int)d->pDisabled->Palette());
				}

				for (int i=0; i<From->GetItems(); i++)
				{
					From->Draw(d->pDisabled, i * From->TileX(), 0, i, IMGLST_GDC);
				}
			}

			From->Unlock();
			d->pColour->Unlock();
			d->pDisabled->Unlock();
		}
	}
}

void GToolBar::_DrawFromCache(GSurface *pDC, int x, int y, int Index, bool Disabled)
{
	if (pDC AND
		d->pDisabled AND
		d->pColour AND
		d->ImgList)
	{
		GRect s;
		
		s.ZOff(d->ImgList->TileX()-1, d->ImgList->TileY()-1);
		s.Offset(d->ImgList->TileX()*Index, 0);

		pDC->Blt(x, y, Disabled ? d->pDisabled : d->pColour, &s);
	}
}

bool GToolBar::Pour(GRegion &r)
{
	int PosX = BORDER_SPACER;
	int PosY = BORDER_SPACER;
	int SrcX = 0;
	int SrcY = 0;
	int EndX = 0;
	int EndY = 0;
	int MaxDim = 0;

	GRect ButPos;
	GView *But = Children.First();
	while (But)
	{
		if (But->Visible())
		{
			int Tx = 0, Ty = 2;

			if (d->Text)
			{
				char *Name = But->Name();
				if (Name)
				{
					// Process '&'s
					Name = NewStr(Name);
					if (Name)
					{
						char *Src = Name;
						char *Dest = Name;
						while (*Src)
						{
							if (Src[0] == '&')
							{
								if (Src[1] == '&') *Dest++ = '&';
								Src++;
							}
							else
							{
								*Dest++ = *Src++;
							}
						}
						*Dest++ = 0;
					}

					// Work out the space required to write each word 
					// centered on a different line
					if (d->Font AND
						d->Text)
					{
						char *n = Name;
						while (*n)
						{
							char *e;
							for (e=n; *e; e++)
							{
								if (*e == ' ')
								{
									if (e[1] AND e[2] == ' ')
									{
										continue;
									}
									break;
								}
							}
							int l = (int)e-(int)n;
							int _x = 0, _y = 0;

							d->Font->Size(&_x, &_y, n, l);
							Tx = max(Tx, _x);
							Ty += _y;

							n = e;
							if (*n == ' ') n++;
						}
					}
					DeleteArray(Name);
				}
			}

			ButPos = But->GetPos();
			GToolButton *Button = dynamic_cast<GToolButton*>(But);
			if (Button)
			{
				if (Button->Separator())
				{
					// This will be stretched out later by the code that makes
					// everything the same height.
					ButPos.ZOff(BORDER_SEPARATOR+1, BORDER_SEPARATOR+1);
				}
				else
				{
					if (Button->Image() >= 0)
					{
						// Set initial size to the icon size
						ButPos.ZOff(d->Bx + 2, d->By + 2);
					}
					else
					{
						// Otherwise default to text size
						ButPos.ZOff(7, 0);
					}
					
					Tx += 4;
					if (ButPos.X() < Tx)
					{
						// Make button wider for text label
						ButPos.x2 = Tx - 1;
					}
					
					ButPos.y2 += Ty;
				}
			}
			
			if (d->Vertical)
			{
				MaxDim = max(MaxDim, ButPos.X());
			}
			else
			{
				MaxDim = max(MaxDim, ButPos.Y());
			}

			ButPos.Offset(PosX - ButPos.x1, PosY - ButPos.y1);

			if (But->GetId() == IDM_BREAK)
			{
				ButPos.ZOff(0, 0);
				if (d->Vertical)
				{
					PosX = MaxDim;
					PosY = BORDER_SHADE + BORDER_SPACER;
				}
				else
				{
					PosX = BORDER_SHADE + BORDER_SPACER;
					PosY = MaxDim;
				}
			}
			else
			{
				if (d->Vertical)
				{
					PosY = ButPos.y2 + 1;
				}
				else
				{
					PosX = ButPos.x2 + 1;
				}
			}

			But->SetPos(ButPos);
		}
		else
		{
			GRect p(0, 0, 0, 0);
			But->SetPos(p);
		}
		
		But = Children.Next();
	}

	for (GView *w = Children.First(); w; w = Children.Next())
	{
		GRect p = w->GetPos();

		if (d->Vertical)
		{
			if (w->X() < MaxDim)
			{
				p.x2 = p.x1 + MaxDim - 1;
				w->SetPos(p);
			}
		}
		else
		{
			if (w->Y() < MaxDim)
			{
				p.y2 = p.y1 + MaxDim - 1;
				w->SetPos(p);
			}
		}

		EndX = max(EndX, p.x2 + 1);
		EndY = max(EndY, p.y2 + 1);
	}

	d->Sx = EndX + BORDER_SPACER;
	d->Sy = EndY + BORDER_SPACER;

	int BorderPx = Raised() OR Sunken() ? _BorderSize : 0;

	GRect n;
	n.ZOff(max(7, d->Sx)+BorderPx-1, max(7, d->Sy)+BorderPx-1);

	// GRect *Best = FindLargest(r);
	GRect *Best = FindLargestEdge(r, GV_EDGE_TOP);
	if (Best)
	{
		n.Offset(Best->x1, Best->y1);
		n.Bound(Best);
		SetPos(n, true);

		return true;
	}

	return false;
}

void GToolBar::OnButtonClick(GToolButton *Btn)
{
	GView *w = (GetNotify()) ? GetNotify() : GetParent();
	if (w AND Btn)
	{
		w->PostEvent(M_COMMAND, Btn->GetId(), (int) Handle());
	}
}

int GToolBar::PostDescription(GView *Ctrl, char *Text)
{
	if (GetParent())
	{
		return GetParent()->PostEvent(M_DESCRIBE, (int) Ctrl, (int) Text);
	}
	return 0;
}

int GToolBar::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case M_CHANGE:
		{
			if (GetParent())
			{
				return GetParent()->OnEvent(Msg);
			}
			break;
		}
	}

	return GView::OnEvent(Msg);
}

void GToolBar::OnPaint(GSurface *pDC)
{
	GRect r = GetClient();
	pDC->Colour(LC_MED, 24);
	pDC->Box(&r);
}

void GToolBar::OnMouseClick(GMouse &m)
{
}

void GToolBar::OnMouseEnter(GMouse &m)
{
	if (NOT d->Tip)
	{
		d->Tip = NEW(GToolTip);
		if (d->Tip)
		{

			d->Tip->Attach(this);
		}
	}
}

void GToolBar::OnMouseExit(GMouse &m)
{
}

void GToolBar::OnMouseMove(GMouse &m)
{
}

bool GToolBar::SetBitmap(char *File, int bx, int by)
{
	bool Status = false;

	GSurface *pDC = LoadDC(File);
	if (pDC)
	{
		Status = SetDC(pDC, bx, by);
		DeleteObj(pDC);
	}

	return Status;
}

bool GToolBar::SetDC(GSurface *pNewDC, int bx, int by)
{
	if (d->OwnImgList)
	{
		DeleteObj(d->ImgList);
	}

	d->Bx = bx;
	d->By = by;

	if (pNewDC)
	{
		d->ImgList = NEW(GImageList(bx, by, pNewDC));
		if (d->ImgList)
		{
			d->OwnImgList = true;
			#ifdef WIN32
			if (d->ImgList->Create(pNewDC->X(), pNewDC->Y(), GdcD->GetBits()))
			#endif
			{
				_BuildCache(d->ImgList);
				return true;
			}
		}
	}
	return false;
}

GImageList *GToolBar::GetImageList()
{
	return d->ImgList;
}

bool GToolBar::SetImageList(GImageList *l, int bx, int by, bool Own)
{
	if (d->OwnImgList)
	{
		DeleteObj(d->ImgList);
	}
	d->OwnImgList = Own;
	d->Bx = bx;
	d->By = by;

	d->ImgList = l;
	_BuildCache(d->ImgList);

	return d->ImgList != 0;
}

GToolButton *GToolBar::AppendButton(char *Tip, int Id, int Type, int Enabled, int IconId)
{
	bool HasIcon = IconId != TOOL_ICO_NONE;

	GToolButton *But = NEW(GToolButton(d->Bx, d->By));
	if (But)
	{
		But->Name(Tip);
		But->SetId(Id);
		But->Type = Type;
		But->SetParent(this);
		But->Enabled(Enabled);

		if (IconId >= 0)
		{
			But->ImgIndex = IconId;
		}
		else if (IconId == TOOL_ICO_NEXT)
		{
			But->ImgIndex = d->LastIndex++;
		}
		else if (IconId == TOOL_ICO_NONE)
		{
			But->ImgIndex = -1;
		}

		AttachButton(But);
	}

	return But;
}

bool GToolBar::AppendSeparator()
{
	GToolButton *But = NEW(GToolButton(d->Bx, d->By));
	if (But)
	{
		But->SetId(IDM_SEPARATOR);
		But->SetParent(this);
		AttachButton(But);
		return true;
	}
	return false;
}

bool GToolBar::AppendBreak()
{
	GToolButton *But = NEW(GToolButton(d->Bx, d->By));
	if (But)
	{
		But->SetId(IDM_BREAK);
		But->SetParent(this);
		AttachButton(But);
		return true;
	}
	return false;
}

bool GToolBar::AppendControl(GView *Ctrl)
{
	bool Status = false;
	if (Ctrl)
	{
		Ctrl->SetParent(this);
		AttachButton(Ctrl);
		Status = true;
	}
	return Status;
}

void GToolBar::Empty()
{
	for (GView *But = Children.First(); But; But = Children.Next())
	{
		DeleteObj(But);
	}
}

///////////////////////////////////////////////////////////////////////
COLOUR Map(GSurface *pDC, COLOUR c)
{
	if (pDC AND pDC->GetBits() <= 8)
	{
		if (pDC->IsScreen())
		{
			c = CBit(24, c);
		}
		#ifdef WIN32
		else
		{
			HPALETTE hPal = GetSystemPalette();
			if (hPal)
			{
				c = GetNearestPaletteIndex(hPal, c);
				DeleteObject(hPal);
			}
		}
		#endif
	}
	return c;
}

#ifdef WIN32
HPALETTE GetSystemPalette()
{
	HPALETTE hPal = 0;
	LOGPALETTE *Log = (LOGPALETTE*) NEW(uchar[sizeof(LOGPALETTE) + (sizeof(PALETTEENTRY) * 255)]);
	if (Log)
	{
		Log->palVersion = 0x300;
		Log->palNumEntries = 256;

		HDC hDC = CreateCompatibleDC(0);
		GetSystemPaletteEntries(hDC, 0, 256, Log->palPalEntry);
		DeleteDC(hDC);

		hPal = CreatePalette(Log);
	}

	return hPal;
}

bool BltBmpToBmp(HBITMAP hDest,
		int xDst,
		int yDst,
		int cx,
		int cy,
		HBITMAP hSrc,
		int xSrc,
		int ySrc,
		DWORD dwRop)
{
	bool Status = false;
	HDC DestDC = CreateCompatibleDC(0);
	HDC SrcDC = CreateCompatibleDC(0);

	if (DestDC AND SrcDC)
	{
		hDest = (HBITMAP) SelectObject(DestDC, hDest);
		hSrc = (HBITMAP) SelectObject(SrcDC, hSrc);

		Status = BitBlt(DestDC,
				xDst,
				yDst,
				cx,
				cy,
				SrcDC,
				xSrc,
				ySrc,
				dwRop);

		hDest = (HBITMAP) SelectObject(DestDC, hDest);
		hSrc = (HBITMAP) SelectObject(SrcDC, hSrc);
	}

	if (DestDC)
	{
		DeleteDC(DestDC);
	}
	if (SrcDC)
	{
		DeleteDC(SrcDC);
	}

	return Status;
}

bool BltBmpToDc(HDC DestDC,
		int xDst,
		int yDst,
		int cx,
		int cy,
		HBITMAP hSrc,
		int xSrc,
		int ySrc,
		DWORD dwRop)
{
	bool Status = false;
	HDC SrcDC = CreateCompatibleDC(0);

	if (DestDC AND SrcDC)
	{
		hSrc = (HBITMAP) SelectObject(SrcDC, hSrc);

		Status = BitBlt(DestDC,
				xDst,
				yDst,
				cx,
				cy,
				SrcDC,
				xSrc,
				ySrc,
				dwRop);

		hSrc = (HBITMAP) SelectObject(SrcDC, hSrc);
	}

	if (SrcDC)
	{
		DeleteDC(SrcDC);
	}

	return Status;
}

bool BltDcToBmp(HBITMAP hDest,
		int xDst,
		int yDst,
		int cx,
		int cy,
		HDC SrcDC,
		int xSrc,
		int ySrc,
		DWORD dwRop)
{
	bool Status = false;
	HDC DestDC = CreateCompatibleDC(0);

	if (DestDC AND SrcDC)
	{
		hDest = (HBITMAP) SelectObject(DestDC, hDest);

		Status = BitBlt(DestDC,
				xDst,
				yDst,
				cx,
				cy,
				SrcDC,
				xSrc,
				ySrc,
				dwRop);

		hDest = (HBITMAP) SelectObject(DestDC, hDest);
	}

	if (DestDC)
	{
		DeleteDC(DestDC);
	}

	return Status;
}
#endif






