#include <stdio.h>

#include "Lgi.h"
#include "GDragAndDrop.h"
#include "GToken.h"
#include "GPopup.h"

///////////////////////////////////////////////////////////////////////
#if defined XWIN

QWindow::QWindow(GWindow *wnd)
{
	Wnd = wnd;
	setBackgroundMode(NoBackground);

	// Setup our protocols
	XSetWMProtocols(XDisplay(), handle(), &QApp()->XA_WmDeleteWindow, 1);
}

QWindow::~QWindow()
{
	if (Wnd)
	{
		Wnd->_View = 0;
		Wnd->_OnViewDelete();
	}
}

void QWindow::_SetWndPtr(void *p)
{
	Wnd = (GWindow*)p;
}

void *QWindow::_GetWndPtr()
{
	return Wnd;
}

void QWindow::SetModal()
{
}

void QWindow::closeEvent(QEvent *ce)
{
	if (Wnd->OnRequestClose(false))
	{
		_SetDeleteMe();
	}
}

void QWindow::resizeEvent(QEvent *re)
{
	if (Wnd)
	{
		Wnd->GetPos();

		// printf("QWindow('%s')::resizeEvent pos=%s\n", Wnd->Name(), Wnd->GetPos().Describe());
		
		Wnd->OnPosChange();
	}
}

void QWindow::paintEvent(QEvent *e)
{
	if (Wnd)
	{
		GScreenDC Dc(this);
		Wnd->_Paint(&Dc);
	}
}

void QWindow::customEvent(QEvent *e)
{
	GMessage *m = dynamic_cast<GMessage*>(e);
	if (Wnd AND m)
	{
		Wnd->OnEvent(m);
	}
}

void QWindow::propertyEvent(QEvent *e)
{
	char *Prop = XGetAtomName(XDisplay(), e->GetEvent()->xproperty.atom);
	if (Prop)
	{
		if (stricmp(Prop, "_NET_WM_STATE") == 0)
		{
			Atom Type = 0;
			int Format = 0;
			ulong Items = 0, Bytes = 0;
			uchar *Data = 0;
			
			XGetWindowProperty(	XDisplay(),
								handle(),
								e->GetEvent()->xproperty.atom,
								0,
								1024,
								false,
								AnyPropertyType,
								&Type,
								&Format,
								&Items,
								&Bytes,
								&Data);
								
			if (Data AND Type)
			{
				Atom *a = (Atom*)Data;
				if (*a)
				{
					char *Event = XGetAtomName(XDisplay(), *a);
					if (Event)
					{
						if (strcmp(Event, "_NET_WM_STATE_HIDDEN") == 0)
						{
							Wnd->OnZoom(GZoomMin);
						}
						else if (strstr(Event, "_NET_WM_STATE_MAXIMIZED"))
						{
							Wnd->OnZoom(GZoomMax);
						}
					}
					else if (Wnd)
					{
						Wnd->OnZoom(GZoomNormal);
					}
				}
			}
			
			if (Data)
			{
				XFree(Data);
			}
		}
		
		XFree(Prop);
	}
}

#elif defined YWIN

#include <stdint.h>
#include <sigc++/signal_system.h>
#include <sigc++/func_slot.h>
#include <sigc++/bind.h>

using std::cout;
using std::endl;
using SigC::slot;
using SigC::bind;

class YWindowPrivate
{
public:
	GWindow *This;
	Y::Window *w;
	Y::Canvas *c;
	GRect Pos;
	
	YWindowPrivate()
	{
		Pos.ZOff(-1, -1);
	}
};

YWindow::YWindow(GWindow *t) : YView(t)
{
	d = NEW(YWindowPrivate);
	d->This = t;
	SetHandle(d->w = NEW(Y::Window));
	d->c = NEW(Y::Canvas);
	if (d->c)
	{
		d->w->requestClose.connect (slot (this, &YWindow::OnRequestExit));
		d->c->requestSize(300, 200);
		d->w->setChild(d->c);
		d->c->resize.connect (slot (this, &YWindow::OnDraw));
	}
}

YWindow::~YWindow()
{
	DeleteObj(d);
}

bool YWindow::reparent(void *parent, OsPoint &p, bool vis)
{
	printf("YWindow::reparent\n");
	return true;
}

void YWindow::show()
{
	if (d->w)
	{
		d->w->show();
	}
}

void YWindow::hide()
{
}

void YWindow::OnRequestExit()
{
	if (d->This->OnRequestClose(false))
	{
		printf("delete this\n");
		delete this;
	}
}

void YWindow::OnDraw()
{
	GScreenDC Dc(this, d->c);
	d->This->_Paint(&Dc);
}

GRect &YWindow::geometry()
{
	return d->Pos;
}

void YWindow::setGeometry(int x, int y, int w, int h)
{
	if (d->c AND
		w > 0 AND
		h > 0 AND
		w < 2000 AND
		h < 2000)
	{
		printf("requestSize %i,%i\n", w, h);
		d->c->requestSize(w, h);
		d->Pos.Set(x, y, x + w - 1, y + h - 1);
	}
}

#endif

///////////////////////////////////////////////////////////////////////
class HookInfo
{
public:
	int Flags;
	GView *Target;
};

class GWindowPrivate
{
public:
	int Sx, Sy;
	bool Dynamic;
	GKey LastKey;
	GArray<HookInfo> Hooks;
	bool SnapToEdge;

	GWindowPrivate()
	{
		Sx = Sy = 0;
		Dynamic = true;
		SnapToEdge = false;
		ZeroObj(LastKey);		
	}
	
	int GetHookIndex(GView *Target, bool Create = false)
	{
		for (int i=0; i<Hooks.Length(); i++)
		{
			if (Hooks[i].Target == Target) return i;
		}
		
		if (Create)
		{
			HookInfo *n = &Hooks[Hooks.Length()];
			if (n)
			{
				n->Target = Target;
				n->Flags = 0;
				return Hooks.Length() - 1;
			}
		}
		
		return -1;
	}
};

///////////////////////////////////////////////////////////////////////
#define GWND_CREATE		0x0010000

GWindow::GWindow() :
	GView(0)
{
	d = NEW(GWindowPrivate);
	_QuitOnClose = false;
	Menu = 0;
	_Default = 0;
	_Window = this;
	WndFlags |= GWND_CREATE;
	GFlags::Visible(false);

    _Lock = NEW(GSemaphore);
    #if defined XWIN
	_View = Wnd = NEW(QWindow(this));
	#elif defined YWIN
	_View = Wnd = NEW(YWindow(this));
	#endif
}

GWindow::~GWindow()
{
	if (LgiApp->AppWnd == this)
	{
		LgiApp->AppWnd = 0;
	}

	DeleteObj(Menu);
	DeleteObj(d);
	DeleteObj(_Lock);
}

bool GWindow::GetSnapToEdge()
{
	return d->SnapToEdge;
}

void GWindow::SetSnapToEdge(bool s)
{
	d->SnapToEdge = s;
}

bool GWindow::Visible()
{
	return GView::Visible();
}

void GWindow::Visible(bool i)
{
	GView::Visible(i);
	OnPosChange();
}

void GWindow::_SetDynamic(bool i)
{
	d->Dynamic = i;
}

void GWindow::_OnViewDelete()
{
	if (d->Dynamic)
	{
		delete this;
	}
}

bool GWindow::Attach(GView *p)
{
	bool Status = GView::Attach(p);
	if (Status)
	{
		// Setup default button...
		if (NOT _Default)
		{
			_Default = FindControl(IDOK);
		}
		if (_Default)
		{
			_Default->Invalidate();
		}
	}
	return Status;
}

bool GWindow::OnRequestClose(bool OsShuttingDown)
{
	if (GetQuitOnClose())
	{
		LgiCloseApp();
	}

	return GView::OnRequestClose(OsShuttingDown);
}

bool GWindow::OnViewMouse(GView *v, GMouse &m)
{
	for (int i=0; i<d->Hooks.Length(); i++)
	{
		if (d->Hooks[i].Flags & GMouseEvents)
		{
			if (NOT d->Hooks[i].Target->OnViewMouse(v, m))
			{
				return false;
			}
		}
	}
	
	return true;
}

#define DEBUG_KEYS			0

/*
	// Any window in a popup always gets the key...
	for (GView *p = v; p; p = p->GetParent())
	{
		GPopup *Popup;
		if (Popup = dynamic_cast<GPopup*>(p))
		{
			Status = v->OnKey(k);
			if (NOT Status)
			{
				if (k.c16 == VK_ESCAPE)
				{
					// Popup window (or child) didn't handle the key, and the user
					// pressed ESC, so do the default thing and close the popup.
					Popup->Cancelled = true;
					Popup->Visible(false);
				}
				else
				{
					#if DEBUG_KEYS
					printf("Popup ignores '%c' down=%i alt=%i ctrl=%i sh=%i\n", k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
					#endif
					break;
				}
			}
			
			#if DEBUG_KEYS
			printf("Popup ate '%c' down=%i alt=%i ctrl=%i sh=%i\n", k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
			#endif
			
			goto AllDone;
		}
	}
*/

bool GWindow::OnViewKey(GView *v, GKey &k)
{
	bool Status = false;
	GView *Ctrl = 0;

	// Any window in a popup always gets the key...
	GView *p;
	for (p = v->GetParent(); p; p = p->GetParent())
	{
		if (dynamic_cast<GPopup*>(p))
		{
			return v->OnKey(k);
		}
	}

	// Give key to popups
	if (LgiApp AND
		LgiApp->GetMouseHook() AND
		LgiApp->GetMouseHook()->OnViewKey(v, k))
	{
		goto AllDone;
	}

	// Allow any hooks to see the key...
	for (int i=0; i<d->Hooks.Length(); i++)
	{
		if (d->Hooks[i].Flags & GKeyEvents)
		{
			if (d->Hooks[i].Target->OnViewKey(v, k))
			{
				Status = true;
				
				#if DEBUG_KEYS
				printf("Hook ate '%c'(%i) down=%i alt=%i ctrl=%i sh=%i\n", k.c16, k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
				#endif
				
				goto AllDone;
			}
		}
	}

	// Give the key to the window...
	if (v->OnKey(k))
	{
		#if DEBUG_KEYS
		printf("View ate '%c'(%i) down=%i alt=%i ctrl=%i sh=%i\n",
			k.c16, k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
		#endif
		
		Status = true;
		goto AllDone;
	}
	
	// Window didn't want the key...
	switch (k.c16)
	{
		case VK_RETURN:
		{
			Ctrl = _Default;
			break;
		}
		case VK_ESCAPE:
		{
			Ctrl = FindControl(IDCANCEL);
			break;
		}
	}

	if (Ctrl AND Ctrl->Enabled())
	{
		if (Ctrl->OnKey(k))
		{
			Status = true;

			#if DEBUG_KEYS
			printf("Default Button ate '%c'(%i) down=%i alt=%i ctrl=%i sh=%i\n",
				k.c16, k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
			#endif
			
			goto AllDone;
		}
	}

	if (Menu)
	{
		Status = Menu->OnKey(v, k);
		if (Status)
		{
			#if DEBUG_KEYS
			printf("Menu ate '%c' down=%i alt=%i ctrl=%i sh=%i\n", k.c16, k.Down(), k.Alt(), k.Ctrl(), k.Shift());
			#endif
		}
	}

AllDone:
	d->LastKey = k;

	return Status;
}


void GWindow::Raise()
{
	if (_View->handle())
	{
		#if defined XWIN
		XWindowChanges c;
		c.stack_mode = XAbove;
		XConfigureWindow(	_View->XDisplay(),
							_View->handle(),
							CWStackMode,
							&c);
		#endif							
	}
}


GWindowZoom GWindow::GetZoom()
{
	#if defined XWIN
	if (Handle())
	{
		Display *Dsp = Handle()->XDisplay();
		Window Win = Handle()->handle();
		Atom Type;
		int Format;
		ulong Items, Bytes;
		uchar *Prop = 0;

		XGetWindowProperty(	Dsp,
							Win,
							XInternAtom(Dsp, "_NET_WM_STATE", false),
							0, 1024,
							false,
							AnyPropertyType,
							&Type,
							&Format,
							&Items,
							&Bytes,
							&Prop);
		if (Prop)
		{
			Atom s[] =
			{
				XInternAtom(Dsp, "_NET_WM_STATE_MAX_V", false),
				XInternAtom(Dsp, "_NET_WM_STATE_MAX_H", false)
			};

			for (int i=0; i<Items; i++)
			{
				Atom a = ((Atom*)Prop)[i];
				if (a == s[0] OR a == s[1])
				{
					XFree(Prop);
					return GZoomMax;
				}
			}

			XFree(Prop);
		}

		XWindowAttributes a;
		ZeroObj(a);
		XGetWindowAttributes(Dsp, Win, &a);
		
		if (a.map_state == IsUnmapped)
		{
			return GZoomMin;
		}
	}
	#endif

	return GZoomNormal;
}

void GWindow::SetZoom(GWindowZoom i)
{
	#if defined XWIN
	Display *Dsp = Handle()->XDisplay();
	Window Win = Handle()->handle();

	switch (i)
	{
		case GZoomMax:
		{
			/*
			_NET_WM_STATE_MAX_V
			_NET_WM_STATE_MAX_H
		    _NET_WM_STATE_STAYS_ON_TOP
            _NET_WM_STATE_MODAL
			*/

			Atom s[] =
			{
				XInternAtom(Dsp, "_NET_WM_STATE_MAX_V", false),
				XInternAtom(Dsp, "_NET_WM_STATE_MAX_H", false)
			};

		    XChangeProperty(Dsp,
							Win,
							XInternAtom(Dsp, "_NET_WM_STATE", false),
							XA_ATOM,
							32,
							PropModeReplace,
							(unsigned char *) s,
							CountOf(s));
			break;
		}
		case GZoomMin:
		{
			XIconifyWindow(Dsp, Win, DefaultScreen(Dsp) );
			break;
		}
		case GZoomNormal:
		{
		    XChangeProperty(Dsp,
							Win,
							XInternAtom(Dsp, "_NET_WM_STATE", false),
							XA_ATOM,
							32,
							PropModeReplace,
							0,
							0);

			XMapRaised(Dsp, Win);
			break;
		}
	}
	#endif
}

GView *GWindow::GetDefault()
{
	return _Default;
}

void GWindow::SetDefault(GView *v)
{
	if (v AND
		v->_Window == this)
	{
		if (_Default != v)
		{
			GView *Old = _Default;
			_Default = v;

			if (Old) Old->Invalidate();
			if (_Default) _Default->Invalidate();
		}
	}
	else
	{
		_Default = 0;
	}
}

bool GWindow::Name(char *n)
{
	bool Status = GObject::Name(n);

	if (Wnd)
	{
		_View->setText(n);
	}

	return Status;
}

char *GWindow::Name()
{
	return GObject::Name();
}

GRect &GWindow::GetClient()
{
	static GRect r;
	
	r = GView::GetClient();
	if (Menu AND
		Menu->Handle() AND
		Menu->Handle()->View())
	{
		/*
		#if defined XWIN
		QMenuBar *Bar = dynamic_cast<QMenuBar*>(Menu->Handle());
		if (Bar)
		{
			GRect p = Bar->geometry();
			r.y1 = p.y2 + 1;
		}
		#endif
		*/
		
		GRect p = Menu->Handle()->View()->GetPos();
		r.y1 = p.y2 + 1;
	}
	
	return r;
}

GRect &GWindow::GetPos()
{
	if (Wnd AND _View)
	{
		OsRect r = _View->geometry();

		Pos.x1 = r.left;
		Pos.y1 = r.top;
		Pos.x2 = r.right;
		Pos.y2 = r.bottom;

		#if defined XWIN
		XWindowAttributes a;
		ZeroObj(a);
		XGetWindowAttributes(Handle()->XDisplay(), Handle()->handle(), &a);
		if (NOT a.override_redirect)
		{
			int Dx, Dy;
			_View->GetDecorationSize(Dx, Dy);
			Pos.Offset(Dx, Dy);
		}
		#endif
	}

	return Pos;
}

bool GWindow::SetPos(GRect &p, bool Repaint)
{
	Pos = p;

	if (Wnd)
	{
		_View->setGeometry(p.x1, p.y1, p.X(), p.Y());
	}

	return true;
}

void GWindow::OnChildrenChanged(GView *Wnd, bool Attaching)
{
	// Force repour
	d->Sx = d->Sy = -1;
}

void GWindow::OnCreate()
{
}

void GWindow::_Paint(GSurface *pDC)
{
	GRect r = GetClient();
	pDC->Handle()->SetClient(&r);
	OnPaint(pDC);
	pDC->Handle()->SetClient(0);
}

void GWindow::OnPaint(GSurface *pDC)
{
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle();
}

void GWindow::OnPosChange()
{
	GView::OnPosChange();

	if (d->Sx != X() OR	d->Sy != Y())
	{
		// printf("Window '%s' Layout: x,y=%i,%i\n", Name(), X(), Y());
		
		Pour();
		d->Sx = X();
		d->Sy = Y();
	}
	else
	{
		// printf("Window '%s' Not Layout: x,y=%i,%i == %i,%i\n", Name(), d->Sx, d->Sy, X(), Y());
	}
}

void GWindow::Pour()
{
	GRect r = GetClient();

	if (Menu AND
		Menu->Handle() AND
		Menu->Handle()->View())
	{
		GRegion MenuRgn(0, 0, X()-1, Y()-1);
		Menu->Handle()->View()->Pour(MenuRgn);
	}
	else if (Menu)
	{
		printf("%s:%i - Menu has no view!\n", __FILE__, __LINE__);
	}
	
	GRegion Client(r);
	GRegion Update;

	Iterator<GView> Lst(&Children);
	for (GView *w = Lst.First(); w; w = Lst.Next())
	{
		GRect OldPos = w->Pos;
		Update.Union(&OldPos);

		if (w->Pour(Client))
		{
			GRect p = w->GetPos();

			if (NOT w->Visible())
			{
				w->Visible(true);
			}

			w->Invalidate();

			Client.Subtract(&w->Pos);
			Update.Subtract(&w->Pos);
		}
		else
		{
			// non-pourable
		}
	}
}

int GWindow::WillAccept(List<char> &Formats, GdcPt2 Pt, int KeyState)
{
	int Status = DROPEFFECT_NONE;
	
	// printf("First=%p\n", &Formats);
	
	for (char *f=Formats.First(); f; )
	{
		if (stricmp(f, LGI_FileDropFormat) == 0)
		{
			f = Formats.Next();
			Status = DROPEFFECT_COPY;
		}
		else
		{
			Formats.Delete(f);
			DeleteArray(f);
			f = Formats.Current();
		}
	}
	
	return Status;
}

int GWindow::OnDrop(char *Format, GVariant *Data, GdcPt2 Pt, int KeyState)
{
	int Status = DROPEFFECT_NONE;

	if (Format AND Data)
	{
		if (stricmp(Format, LGI_FileDropFormat) == 0)
		{
			List<char> Files;
			if (Data->IsBinary())
			{
				GToken Uri(	(char*)Data->Value.Binary.Data,
							"\r\n,",
							true,
							Data->Value.Binary.Length);
				for (int i=0; i<Uri.Length(); i++)
				{
					char *File = Uri[i];
					if (strnicmp(File, "file:", 5) == 0)
						File += 5;
					
					char *i = File, *o = File;
					while (*i)
					{
						if (i[0] == '%' AND
							i[1] AND
							i[2])
						{
							char h[3] = { i[1], i[2], 0 };
							*o++ = htoi(h);
							i += 3;
						}
						else
						{
							*o++ = *i++;
						}
					}
					*o++ = 0;
					
					if (FileExists(File))
					{
						Files.Insert(NewStr(File));
					}
				}
			}
			
			if (Files.First())
			{
				Status = DROPEFFECT_COPY;
				OnReceiveFiles(Files);
				Files.DeleteArrays();
			}
		}
	}
	
	return Status;
}

int GWindow::OnEvent(GMessage *m)
{
	switch (MsgCode(m))
	{
		case M_CLOSE:
		{
			if (OnRequestClose(false))
			{
				Quit();
				return 0;
			}
			break;
		}
	}

	return GView::OnEvent(m);
}

bool GWindow::RegisterHook(GView *Target, int EventType, int Priority)
{
	bool Status = false;
	
	if (Target AND EventType)
	{
		int i = d->GetHookIndex(Target, true);
		if (i >= 0)
		{
			d->Hooks[i].Flags = EventType;
			Status = true;
		}
	}
	
	return Status;
}

bool GWindow::UnregisterHook(GView *Target)
{
	int i = d->GetHookIndex(Target);
	if (i >= 0)
	{
		d->Hooks.DeleteAt(i);
		return true;
	}
	return false;
}

