#include <ctype.h>

#include "qevent.h"
#include "qwidget.h"
#include "qapplication.h"

/////////////////////////////////////////////////////////////////////////////////////////////
QInput::QInput()
{
	Context = 0;
	Method = XOpenIM(XDisplay(), 0, 0, 0);
	if (Method)
	{
		Context = XCreateIC(Method);
	}
	else
	{
		printf("QInput, no method!\n");
	}
}

QInput::~QInput()
{
	if (Context)
	{
		XDestroyIC(Context);
	}
	if (Method)
	{
		XCloseIM(Method);
	}
}

/////////////////////////////////////////////////////////////////////////////////////////////
QEvent::QEvent(XEvent *e, QInput *i)
{
	Event = e;

	Char = 0;
	switch (Event->type)
	{
		case KeyPress:
		case KeyRelease:
		{
			// Do this here, so that it runs in the GUI thread and
			// can't get stuck in a deadlock getting the mutex
			//
			// Also it's nice that it's only called once too.

			// k = XLookupKeysym(&Event->xkey, 0); // hangs after locale change

			KeySym k;
			char c[256] = "";
			int Chars = 0;
			bool Shifted = (Event->xkey.state & ShiftMask) != 0; // ShiftMask == 0x01

			/*
			if (Event->type == KeyPress)
			{
				// Xutf8LookupString(
			}
			else
			*/
			{
				int State = Event->xkey.state;
				Event->xkey.state &= ~0x4;
				Chars = XLookupString(&Event->xkey, c, sizeof(c), &k, 0);
				Event->xkey.state = State;
			}

			if (Chars > 0)
			{
				c[Chars] = 0;
				Char = c[0];
			}

			if (Char >= 0x0e AND Char <= 0x1a)
			{
				// This range is assigned to virtual keys, X11 shouldn't generate these
				// directly. We assign this range of values in the if (k == 0x####) code.
				Char = 0;
			}


			#if 0
			// if (Chars > 0) printf("Chars=%i c='%s'\n", Chars, c);
			printf("::ascii Char=%i,0x%x,'%c' keycode=%i keysym=%i (0x%x) state=%x (%s %s)\n",
				Char, Char, Char<' '?'.':Char,
				Event->xkey.keycode, k, k, Event->xkey.state,
				(char*)(Shifted ? "Shifted" : ""),
				(char*)(down()?"Down":"")
				);
			#endif

			if (Event->type == KeyRelease)
			{
				// printf("Keyup k=%i (%x)\n", k, k);
			}			

			if (k == 65505) Char = VK_SHIFT;
			else if (k == 65307) Char = VK_ESCAPE;
			else if (k == 65361) Char = VK_LEFT;
			else if (k == 65362) Char = VK_UP;
			else if (k == 65363) Char = VK_RIGHT;
			else if (k == 65364) Char = VK_DOWN;
			else if (k == 65365) Char = VK_PRIOR;
			else if (k == 65366) Char = VK_NEXT;
			else if (k == 65360) Char = VK_HOME;
			else if (k == 65367) Char = VK_END;
			else if (k == 65379) Char = VK_INSERT;
			else if (k == 65535) Char = VK_DELETE;
			else if (k == 65289 OR k == 65056) Char = '\t';
			else if (k == 65293 OR k == 65421) Char = VK_RETURN;
			else if (k == 65288) Char = VK_BACK;
			else if (k == 65470) Char = VK_F1;
			else if (k == 65471) Char = VK_F2;
			else if (k == 65472) Char = VK_F3;
			else if (k == 65473) Char = VK_F4;
			else if (k == 65474) Char = VK_F5;
			else if (k == 65475) Char = VK_F6;
			else if (k == 65476) Char = VK_F7;
			else if (k == 65477) Char = VK_F8;
			else if (k == 65478) Char = VK_F9;
			else if (k == 65479) Char = VK_F10;
			else if (k == 65480) Char = VK_F11;
			else if (k == 65481) Char = VK_F12;
			
			/*
			else
			{
				if (Shifted)
				{
					int n = toupper(k);

					if (n >= 'A' AND n <= 'Z')
					{
						return n;
					}
					else if (k >= '0' AND k <= '9')
					{
						char *Map = ")!@#$%^&*(";
						return Map[k - '0'];
					}
					else if (k == '-') return '_';
					else if (k == '=') return '+';
					else if (k == '[') return '{';
					else if (k == ']') return '}';
					else if (k == '\\') return '|';
					else if (k == ';') return ':';
					else if (k == '\'') return '\"';
					else if (k == ',') return '<';
					else if (k == '.') return '>';
					else if (k == '/') return '?';
					else if (k == '`') return '~';
					else if (k == ' ') return ' ';
				}
				else if ((k&0xff00) == 0)
				{
					int n = k & 0xff;
					return tolower(n);
				}
			}
			*/
			break;
		}
	}
}

int QEvent::type()
{
	return Event->type;
}

// Mouse
int QEvent::x()
{
	switch (Event->type)
	{
		case ButtonPress:
		case ButtonRelease:
		{
			return Event->xbutton.x - 1;
			break;
		}
		case MotionNotify:
		{
			return Event->xmotion.x - 1;
			break;
		}
	}

	return 0;
}

int QEvent::ScreenX()
{
	switch (Event->type)
	{
		case ButtonPress:
		case ButtonRelease:
		{
			return Event->xbutton.x_root - 1;
			break;
		}
		case MotionNotify:
		{
			return Event->xmotion.x_root - 1;
			break;
		}
	}

	return 0;
}

int QEvent::y()
{
	switch (Event->type)
	{
		case ButtonPress:
		case ButtonRelease:
		{
			return Event->xbutton.y - 1;
			break;
		}
		case MotionNotify:
		{
			return Event->xmotion.y - 1;
			break;
		}
	}

	return 0;
}

int QEvent::ScreenY()
{
	switch (Event->type)
	{
		case ButtonPress:
		case ButtonRelease:
		{
			return Event->xbutton.y_root - 1;
			break;
		}
		case MotionNotify:
		{
			return Event->xmotion.y_root - 1;
			break;
		}
	}

	return 0;
}

int QEvent::button()
{
	int Status = 0;
	int State = state();

	switch (Event->type)
	{
		case ButtonPress:
		case ButtonRelease:
		{
			switch (Event->xbutton.button)
			{
				case 1:
					Status |= QWidget::LeftButton;
					break;
				case 2:
					Status |= QWidget::MidButton;
					break;
				case 3:
					Status |= QWidget::RightButton;
					break;
			}
			break;
		}
		case MotionNotify:
		{
			if (State & Button1MotionMask)
			{
				Status |= QWidget::LeftButton;
			}
			else if (State & Button2MotionMask)
			{
				Status |= QWidget::MidButton;
			}
			else if (State & Button3MotionMask)
			{
				Status |= QWidget::RightButton;
			}
			break;
		}
	}

	if (State & 0x8)
	{
		Status |= QWidget::AltButton;
	}

	if (State & ControlMask)
	{
		Status |= QWidget::ControlButton;
	}

	if (State & ShiftMask)
	{
		Status |= QWidget::ShiftButton;
	}

	return Status;
}

bool QEvent::down()
{
	switch (Event->type)
	{
		case ButtonPress:
		case KeyPress:
		{
			return true;
			break;
		}
		case MotionNotify:
		{
			return (button() & (QWidget::LeftButton | QWidget::MidButton | QWidget::RightButton)) != 0;
		}
	}

	return 0;
}

bool QEvent::doubleclick()
{
	if (Event->type == ButtonPress AND
		Event->xbutton.button >= 0 AND
		Event->xbutton.button < 4)
	{
		uint64 Now = LgiCurrentTime();
		uint64 Last = QApp()->GetLastButtonPressTime
			(
				Event->xbutton.button-1,
				Event->xbutton.x_root,
				Event->xbutton.y_root
			);
			
		return (Now - Last) < DOUBLE_CLICK_TIME;
	}

	return false;
}

// Keyboard
#include "LgiOsDefs.h"
int QEvent::ascii()
{
	return Char;
}

int QEvent::state()
{
	switch (Event->type)
	{
		case KeyPress:
		case KeyRelease:
		{
			return Event->xkey.state;
			break;
		}
		case MotionNotify:
		{
			return Event->xmotion.state;
			break;
		}
		case ButtonPress:
		case ButtonRelease:
		{
			return Event->xbutton.state;
			break;
		}
	}
	return 0;
}

// Wheel
int QEvent::delta()
{
	switch (Event->type)
	{
		case ButtonPress:
		{
			if (Event->xbutton.button == 4) return 120;
			if (Event->xbutton.button == 5) return -120;
			break;
		}
		case ButtonRelease:
		{
			break;
		}
	}

	return 0;
}

// Focus
bool QEvent::focus()
{
	return 0;
}

// Expose event
GRect &QEvent::exposed()
{
	static GRect r;

	switch (Event->type)
	{
		case Expose:
		{
			r.Set(	Event->xexpose.x,
					Event->xexpose.y,
					Event->xexpose.x + Event->xexpose.width,
					Event->xexpose.y + Event->xexpose.height);
			break;
		}
        default:
        {
        	r.ZOff(-1, -1);
            break;
        }
	}

	return r;
}

