#include "Gdc2.h"
#include "qapplication.h"
#include "GString.h"
#include "qpopup.h"
#include "LgiClasses.h"

enum MapState
{
	Unmapped,
	Mapping,
	Unmapping,
	Mapped,	
};

class QWidgetPrivate : public QObject
{
public:
	enum WaitState
	{
		WaitNone,
		WaitMapped,
		WaitExpose,
		WaitExiting
	} State;
	
	char *StateName(WaitState s)
	{
		switch (s)
		{
			case WaitNone: return "WaitNone";
			case WaitMapped: return "WaitMapped";
			case WaitExpose: return "WaitExpose";
			case WaitExiting: return "WaitExiting";
		}
		
		return 0;
	}
	
	static OsPoint *DecorationSize;

	QWidget *Widget;
	QWidget *Parent;
	Window Win;
	char *Text;
	int x, y, w, h;
	bool PosDirty;

	bool Exposed;				// true if the widget has received an expose event (ie visible)
	bool Enabled;				// our own enabled start
	bool TabStop;				// is a tabstop
	int WantKeys;
	bool Debug;					// Print debugging info for this window
	bool AddToList;
	bool Popup;

	// Mapping State
	MapState Map;
	char *MapName()
	{
		switch (Map)
		{
			case Unmapped: return "Unmapped";
			case Mapping: return "Mapping";
			case Unmapping: return "Unmapping";
			case Mapped: return "Mapped";
		}
		return "(error)";
	}

	/*
	int Rects;					// Clip region
	XRectangle *Rect;
	*/
	GRegion *Clip;

	QList<QWidget> Children;

	QWidget *Top()
	{
	    QWidget *p = Widget;
		while (p->d->Parent) p = p->d->Parent;
		return p;
	}

	QWidgetPrivate(QWidget *This, bool top, Window Existing)
	{
		PosDirty = false;
		TabStop = false;
		WantKeys = 0;
		Widget = This;
		Parent = 0;
		Text = 0;
		x = y = 3000;
		w = h = 1;
		Map = Unmapped;
		Enabled = true;
		Exposed = false;
		Debug = false;
		State = WaitNone;
		Clip = 0;
		AddToList = true;
		Popup = false;

		if (Existing)
		{
			Win = Existing;
		}
		else
		{
			// Create the window
			XSetWindowAttributes v;
			v.cursor = XNone;
			v.background_pixmap = XNone;
			v.event_mask =			ExposureMask |
									KeyPressMask | KeyReleaseMask |
									ButtonPressMask | ButtonReleaseMask |
									PointerMotionMask |
									EnterWindowMask | LeaveWindowMask |
									StructureNotifyMask |
									PropertyChangeMask |
									(top ? FocusChangeMask : 0);

			Win = XCreateWindow(	XDisplay(),
									RootWindow(XDisplay(), 0),
									x, y, w, h,
									0,								// border width
									CopyFromParent,					// depth
									InputOutput,					// class
									XDefaultVisual(XDisplay(), 0),	// visual
									CWBackPixmap | CWEventMask | CWCursor,
									&v);
		}
	}

	~QWidgetPrivate()
	{
		Destroy();
	}

	void Destroy()
	{
		if (Win)
		{
			if (Debug)
			{
				// printf("%p: destroy\n", Widget);
			}

			if (WaitForNone(3000))
			{
				XDestroyWindow(XDisplay(), Win);
				Win = 0;
			}
			else
			{
				printf("QWidgetPrivate::Destroy() couldn't get exit loops.\n");
				int *i = 0;
				*i = 0;
			}
		}

		DeleteArray(Text);
		DeleteObj(Clip);
	}

	bool ParentChainMap()
	{
		for (QWidgetPrivate *p=this; p; p=p->Parent ? p->Parent->d : 0)
		{
			if (NOT p->Map == Mapped AND NOT p->Map == Mapping) return false;
		}
		
		return true;
	}
	
	bool WaitForNone(int Time)
	{
		if (State != WaitNone)
		{
			WaitState Old = State; 
			State = WaitExiting;
			
			int Start = LgiCurrentTime();
			while (State != WaitNone)
			{
				LgiSleep(5);

				int Now = LgiCurrentTime();
				if (Now > Start + Time)
				{
					printf("%s,%i - WaitNone timeout at +%ims (%i) State=%s Txt=%s\n", __FILE__, __LINE__, Now-Start, Time, StateName(Old), Text);
					break;
				}
			}
		}

		return State == WaitNone;
	}

	bool WaitForMapped(int Time)
	{
		int Start = LgiCurrentTime();
		if (Map == Mapping AND State != WaitMapped)
		{
			if (ParentChainMap())
			{
				if (WaitForNone(Time))
				{
					State = WaitMapped;
					while (	State == WaitMapped AND
							(Map == Mapping OR Map == Unmapping) AND
							LgiCurrentTime() < Start + Time)
					{
						QApp()->processEvents();
					}
					
					if (Map == Mapping)
					{
						printf("%s,%i - WaitForMapped timed out (%i,%i).\n", __FILE__, __LINE__, Mapping, Mapped);
					}
					
					State = WaitNone;
				}
				else
				{
					printf("%s,%i - WaitForNone timed out.\n", __FILE__, __LINE__);
				}
			}
			else
			{
				// printf("%s,%i - WaitForMapped, parent isn't visible yet.\n", __FILE__, __LINE__);
			}
		}

		return Map == Mapped;
	}

	void WaitForExpose(int Time)
	{
		if (Mapped AND WaitForNone(Time))
		{
			int Start = LgiCurrentTime();
			
			State = WaitExpose;
			while (State == WaitExpose AND Mapped)
			{
				XEvent e;
				if (XCheckTypedWindowEvent(XDisplay(), Win, Expose, &e))
				{
					Start = LgiCurrentTime();
					QApp()->onEvent(e);

					// We got our expose, now leave
					break;
				}
				else
				{
					LgiSleep(5);

					int Now = LgiCurrentTime();
					if (LgiCurrentTime() > Start + Time)
					{
						break;
					}
				}
			}
			
			State = WaitNone;
		}
	}
	
	void PollExpose()
	{
		XEvent e;
		while (XCheckTypedEvent(XDisplay(), Expose, &e))
		{
			QApp()->onEvent(e);
		}
	}

	OsPoint *GetDecorationSize()
	{
		if (NOT Parent AND
			NOT DecorationSize)
		{
			Window Root = 0, Wind = 0, Parent = 0, *Child = 0;
			int tx = 0, ty = 0;
			uint Children;

			XQueryTree(XDisplay(), Win, &Root, &Parent, &Child, &Children);
			if (Child) XFree(Child);
			if (Parent)
			{
				XQueryTree(XDisplay(), Parent, &Root, &Parent, &Child, &Children);
				if (Child) XFree(Child);
				if (Parent)
				{
					XTranslateCoordinates(	XDisplay(),
											Parent,
											Win,
											0,
											0,
											&tx, &ty,
											&Wind);

					DecorationSize = NEW(OsPoint(tx, ty));
				}
				else
				{
					// printf("%s,%i - No parent.\n", __FILE__, __LINE__);
				}
			}
			else
			{
				// printf("%s,%i - No parent.\n", __FILE__, __LINE__);
			}
		}
		
		return DecorationSize;
	}
};

OsPoint *QWidgetPrivate::DecorationSize = 0;
QWidget *QWidget::OwnsClipboard = 0;
QList<QWidget> QWidget::AllWidgets;

QWidget::QWidget(Window Existing, bool AddToList)
{
	_event = 0;
	d = NEW(QWidgetPrivate(this, false, Existing));
	
	d->AddToList = AddToList;
	if (AddToList AND
		QApp()->Lock())
	{
		AllWidgets.Insert(this);
		QApp()->Unlock();
	}
}

QWidget::QWidget(char *name, bool top)
{
	_event = 0;
	d = NEW(QWidgetPrivate(this, top, 0));
	
	if (QApp()->Lock())
	{
		AllWidgets.Insert(this);
		QApp()->Unlock();
	}
}

QWidget::~QWidget()
{
	if (d->AddToList AND
		QApp()->Lock())
	{
		AllWidgets.Delete(this);
		QApp()->Unlock();
	}
		
	if (d->Debug)
	{
		printf("%p: destroy\n", this);
	}

	if (QApplication::WantsFocus == this)
	{
		QApplication::WantsFocus = 0;
	}
	
	if (MouseGrabber == this)
	{
		MouseGrabber = 0;
	}

	if (d->Parent)
	{
		d->Parent->d->Children.Delete(this);
	}

	for (QWidget *c=d->Children.First(); c; c=d->Children.Next())
	{
		// printf("~QWidget clearing childs ptr to me=%p, child=%p childdata=%p\n", this, c, c->d);
		if (c->d)
		{
			c->d->Parent = 0;
		}
	}

	QApp()->OnDeleteWidget(this);
	DeleteObj(d);
}

QMainWindow *QWidget::GetWindow()
{
	QWidget *Top = this;

	while (Top AND Top->d->Parent)
	{
		Top = Top->d->Parent;
	}	
	
	return dynamic_cast<QMainWindow*>(Top);
}

QWidget *QWidget::Find(Window f)
{
	QWidget *Status = 0;
	
	if (QApp()->Lock())
	{
		for (QWidget *w = AllWidgets.First(); w; w = AllWidgets.Next())
		{
			if (w->d->Win == f)
			{
				Status = w;
				break;
			}
		}
		QApp()->Unlock();
	}
	
	return Status;
}

void QWidget::detachHandle()
{
	d->Win = 0;
}

Window QWidget::handle()
{
	return d ? d->Win : 0;
}

bool QWidget::IsPopup()
{
	for (QWidget *w = this; w; w = w->d->Parent)
	{
		if (d->Popup)
		{
			return true;
		}
	}
	
	return false;
}

bool QWidget::IsMouseOver(QEvent *e)
{
	if (e)
	{
		OsPoint p;
		p = mapToGlobal(p);
		GRect r;
		r.ZOff(width()-1, height()-1);
		r.Offset(p.x, p.y);
		return r.Overlap(e->ScreenX(), e->ScreenY());
	}

	return false;
}

void QWidget::_SetDeleteMe()
{
	if (NOT QApp()->Delete.HasItem(this))
	{
		QApp()->Delete.Insert(this);
	}
}

void QWidget::_dump(int i)
{
	printf("%p: QWidget name=%s exposed=%i map=%s\n", this, d->Text, d->Exposed, d->MapName());
	d->Debug = true;
}

GRegion *QWidget::_GetClipRgn()
{
	return d->Clip;
}

void QWidget::_SetClipRgn(GRegion *c)
{
	DeleteObj(d->Clip);
	d->Clip = c;
}

bool QWidget::_isdebug()
{
	return d->Debug;
}

void QWidget::_paint(XExposeEvent *e, QEvent *q)
{
	LgiAssert(this != 0 AND this->d != 0);
	
	if (d->Debug)
	{
		printf("%p: paint pos=%i,%i-%i,%i map=%s\n", this, d->x, d->y, d->w, d->h, d->MapName());
	}
	
	if (d->Map == Mapping)
	{
		d->Map = Mapped;
	}
	
	if (d->Map == Mapped)
	{
		d->Exposed = true;

		if (this == QApplication::WantsFocus)
		{
			// at least we're garenteed to be visible
			QApplication::WantsFocus->setFocus();
			QApplication::WantsFocus = 0;
		}

		paintEvent(q);
	}
}

bool QWidget::_wait_mapped(int timeout)
{
	return d->WaitForMapped(timeout);
}

void QWidget::_map(bool m)
{
	if (d->Debug)
	{
		printf("%p: map(%i) PosDirty=%i\n", this, m, d->PosDirty);
	}

	d->Map = m ? Mapped : Unmapped;

	if (m)
	{
		if (d->PosDirty)
		{
			setGeometry(d->x, d->y, d->w, d->h);
			
			LgiAssert(d->PosDirty == false);
		}
	}
	else
	{
		d->Exposed = false;
	}
}

void QWidget::_resize(XConfigureEvent *e, QEvent *q)
{
	if (NOT d->Parent)
	{
		// OsPoint *DecSize = d->GetDecorationSize();

		if (d->x != e->x OR
			d->y != e->y OR
			d->w != e->width OR
			d->h != e->height)
		{
			d->x = e->x;
			d->y = e->y;
			d->w = e->width;
			d->h = e->height;
			
			resizeEvent(q);
		}
	}
}

int QWidget::wantKeys()
{
	return d->WantKeys;
}

void QWidget::wantKeys(int keys)
{
	d->WantKeys = keys;
}

bool QWidget::GetPopup()
{
	return d->Popup;
}

void QWidget::SetPopup(bool b)
{
	d->Popup = b;
	
	XSetWindowAttributes a;
	a.override_redirect = true;
	XChangeWindowAttributes(QWidget::XDisplay(), d->Win, CWOverrideRedirect, &a);
}

bool QWidget::isTabStop()
{
	return d->TabStop;
}

void QWidget::isTabStop(bool i)
{
	d->TabStop = i;
}

bool QWidget::hasFocus()
{
    QWidget *p = d->Top();

	// printf("p=%p FocusWindow=%p Top=%p\n", p, QApp()->GetFocusWindow(), d->Top());
	if (p AND
		(QApp()->GetFocusWindow() == p OR QApp()->GetFocusWindow() == 0))
	{
		if (d->Top() == p)
		{
			return QApp()->GetFocus() == this;
		}
	}
	
	return false;
}

void QWidget::setFocus()
{
    QWidget *p = d->Top();
    
	if (p)
	{
		if (p->d->Win AND
			p->d->Exposed AND
			p->d->Map == Mapped)
		{
			int e = XSetInputFocus(	XDisplay(),
									p->d->Win,
									XNone,
									CurrentTime);
		}
		else
		{
			QApplication::WantsFocus = this;
		}
		
		if (GetWindow())
		{
			GetWindow()->SetLastFocus(this);
		}
	}
	
	QApp()->SetFocus(this);
}

void QWidget::clearFocus()
{
	QApp()->SetFocus(0);
}

int QWidget::height()
{
	return d->h;
}

int QWidget::width()
{
	return d->w;
}

void QWidget::setIcon(char *Path)
{
	if (Path)
	{
		int Old = GdcD->SetOption(GDC_PROMOTE_ON_LOAD, 0);
		GSurface *Icon = LoadDC(Path);
		GdcD->SetOption(GDC_PROMOTE_ON_LOAD, Old);
		if (Icon)
		{
			int Len = Icon->X() * Icon->Y() + 2;
			uint32 *Buf = NEW(uint32[Len]);
			if (Buf)
			{
				// Slap the X,Y in the first 2 32bit values
				Buf[0] = Icon->X();
				Buf[1] = Icon->Y();

				// Convert the bits to the right ARGB format.
				for (int y=0; y<Icon->Y(); y++)
				{
					uint32 *p = Buf + (y * Icon->X()) + 2;
					for (int x=0; x<Icon->X(); x++)
					{
						uint32 c = Icon->GetBits() == 32 ? Icon->Get(x, y) : CBit(32, Icon->Get(x, y), Icon->GetBits(), Icon->Palette());
						p[x] = (A32(c) << 24) | (R32(c) << 16) | (G32(c) << 8) | B32(c);
					}
				}

				// Set the property
				XChangeProperty(XDisplay(),
								d->Win,
								QApp()->XA_NetWmIcon,
								QApp()->XA_Cardinal,
								32,
								PropModeReplace,
								(uchar*)Buf,
								Len);

				DeleteArray(Buf);
			}

			DeleteObj(Icon);
		}
	}
}

void QWidget::setGeometry(int X, int Y, int W, int H)
{
	if (X)
	{
		int asd=0;
	}

	bool Changed =	d->x != X OR
					d->y != Y OR
					d->w != W OR
					d->h != H;
	
	if (Changed AND d->Debug)
	{
		printf("%p: set geos (%i,%i-%i,%i) -> (%i,%i-%i,%i) Mapping=%s\n", this,
			d->x, d->y, d->w, d->h,
			X, Y, W, H,
			d->MapName());
	}

	d->x = X;
	d->y = Y;
	d->w = max(W, 1);
	d->h = max(H, 1);

	if (d->Win)
	{
		int Err = XMoveResizeWindow(XDisplay(), d->Win, d->x, d->y, d->w, d->h);
		if (Err != 1)
		{
			printf("%s,%i - XMoveResizeWindow failed: %s\n", __FILE__, __LINE__, XErr(Err));
		}

		XConfigureEvent e;
		e.type = ConfigureNotify;
		QEvent q((XEvent*) &e);
		resizeEvent(&q);
		
		if (d->Map == Mapping)
		{
			d->PosDirty = Changed;
		}
		else
		{
			d->PosDirty = false;
		}
	}
	else
	{
		printf("%s,%i - Not setting pos, map=%s (%i,%i - %i,%i)\n",
				__FILE__, __LINE__,
				d->MapName(),
				d->x, d->y, d->w, d->h);
	}
}

void QWidget::GetDecorationSize(int &x, int &y)
{
	OsPoint *Ds = d->GetDecorationSize();
	x = Ds ? Ds->x : 0;
	y = Ds ? Ds->y : 0;
}

GRect &QWidget::geometry()
{
	static GRect r;
	r.Set(d->x, d->y, d->x + d->w - 1, d->y + d->h - 1);
	return r;
}

void QWidget::setBackgroundMode(Background b)
{
}

void QWidget::repaint()
{
	if (d->Win AND d->Map == Mapped)
	{
		if (d->Debug)
		{
			printf("%p: repaint\n", this);
		}

		int e = XClearArea(	XDisplay(),
							d->Win,
							0, 0,
							d->w + 1, d->h + 1,
							true);
		// printf("XClearArea(%i)=%i\n", d->Win, e);
		d->PollExpose();
	}
}

void QWidget::repaint(int x, int y, int wid, int ht)
{
	if (d->Win AND d->Map == Mapped)
	{
		if (d->Debug)
		{
			printf("%p: repaint(x, y)\n", this);
		}

		GRect r(x, y, x+wid, y+ht);
		GRect k(0, 0, d->w, d->h);
		if (r.Overlap(&k))
		{
			int e = XClearArea(	XDisplay(),
								d->Win,
								x, y,
								wid + 1, ht + 1,
								true);
			// printf("XClearArea(%i, %i, %i, %i, %i)=%i\n", d->Win, x, y, wid + 1, ht + 1, e);
			d->PollExpose();
		}
		else
		{
			/*
			printf("QWidget::repaint(%i,%i,%i,%i) No overlap (%i,%i)\n",
				x, y, wid, ht,
				d->w, d->h);
			*/
		}
	}
}

void QWidget::update()
{
	if (d->Win AND
		d->Map == Mapped)
	{
		if (d->Debug)
		{
			printf("%p: update\n", this);
		}

		XClearArea(	XDisplay(),
					d->Win,
					0, 0,
					d->w + 1, d->h + 1,
					true);
	}
}

void QWidget::update(int x, int y, int wid, int ht)
{
	if (d->Win AND
		d->Map == Mapped)
	{
		if (d->Debug)
		{
			printf("%p: update(%i,%i-%i,%i)\n", this, x, y, wid, ht);
		}

		GRect r(x, y, x+wid, y+ht);
		GRect k(0, 0, d->w, d->h);
		if (r.Overlap(&k))
		{
			XClearArea(	XDisplay(),
						d->Win,
						x, y,
						wid + 1, ht + 1,
						true);
		}
	}
}

bool QWidget::isEnabled()
{
	return d->Enabled;
}

void QWidget::setEnabled(bool i)
{
	d->Enabled = i;
}

bool QWidget::isVisible()
{
	return d->Map == Mapped;
}

void QWidget::show(bool Raise)
{
	if (d->Win AND NOT d->WaitForMapped(200))
	{
		int e;
		
		if (d->Debug)
		{
			printf("%p: show\n", this);
		}
		
		if (NOT parentWidget())
		{
			XSizeHints *s = XAllocSizeHints();
			if (s)
			{
				s->flags = USPosition | USSize;
				s->x = d->x;
				s->y = d->y;
				s->width = d->w;
				s->height = d->h;
				XSetWMNormalHints(XDisplay(), d->Win, s);
				XFree(s);
			}
		}
		else
		{
			d->PosDirty = true;
		}

		if (Raise)
		{
			e = XMapRaised(XDisplay(), d->Win);
		}
		else
		{
			e = XMapWindow(XDisplay(), d->Win);
		}
		
		if (e != 1)
		{
			printf("%s,%i - XMapWindow=%s\n", __FILE__, __LINE__, XErr(e));
		}
		
		d->Map = Mapping;
	}
	else
	{
		printf("QWidget::show() failed, Win=%i Map=%s Text=%s\n", d->Win, d->MapName(), d->Text);
	}
}

void QWidget::hide()
{
	if (d->Win AND
		(d->Map == Mapped OR d->Map == Mapping))
	{
		d->Exposed = false;
		d->Map = Unmapping;

		if (d->Debug)
		{
			printf("%p: hide\n", this);
		}

		XUnmapWindow(XDisplay(), d->Win);
	}
}

OsPoint &QWidget::mapFromGlobal(OsPoint &p)
{
	if (d->Win)
	{
		Window Win;
		int x, y;

		if (XTranslateCoordinates(	XDisplay(),
									RootWindow(XDisplay(), 0),
									d->Win,
									p.x,
									p.y,
									&x, &y,
									&Win))
		{
			p.set(x, y);
		}
	}

	return p;
}

OsPoint &QWidget::mapToGlobal(OsPoint &p)
{
	if (d->Win)
	{
		Window Win;
		int x, y;

		if (XTranslateCoordinates(	XDisplay(),
									d->Win,
									RootWindow(XDisplay(), 0),
									p.x,
									p.y,
									&x, &y,
									&Win))
		{
			p.set(x, y);
		}
	}

	return p;
}

void QWidget::setParentWidget(QWidget *p)
{
	d->Parent = p;
}

QWidget *QWidget::parentWidget()
{
	return d->Parent;
}

QList<QWidget> *QWidget::childrenWidgets()
{
	return &d->Children;
}

bool QWidget::reparent(QWidget *parent, OsPoint &p, bool showNow)
{
	Window ParWin = parent ? parent->d->Win : RootWindow(XDisplay(), 0);

	if (NOT ParWin)
	{
		printf("Error: QWidget::reparent(%p, [%i,%i], %i)\n", parent, p.x, p.y, showNow);
		printf("\tParent->Win=%i\n", parent ? parent->d->Win : 0);
	}
	else
	{
		// Remove from existing location
		if (d->Parent)
		{
			// printf("delete parent ptr this=%p d->Parent=%p d->Parent->d=%p\n", this, d->Parent, d->Parent->d);
			if (d->Parent->d) d->Parent->d->Children.Delete(this);
			else printf("delete parent ptr has bad data part this=%p parent=%p\n", this, d->Parent);
		}

		if (d->Win)
		{
			hide();
		}

		// Add to new location
		d->x = p.x;
		d->y = p.y;
		d->Parent = parent;

		if (d->Parent)
		{
			QList<QWidget> *l = &d->Parent->d->Children;
			
			bool Has = false;
			for (QWidget *	w = l->First();
							w;
							w = l->Next())
			{
				if (w == this)
				{
					Has = true;
					break;
				}
			}
			
			if (NOT Has)
			{
				d->Parent->d->Children.Insert(this);
			}
		}

		if (d->Debug)
		{
			printf("%p: reparent(%i,[%i,%i],%i)\n", this, ParWin, p.x, p.y, showNow);
		}
		
		XReparentWindow(XDisplay(), d->Win, ParWin, d->x, d->y);

		if (showNow)
		{
			show();
		}

		return true;
	}

	return false;
}

char *QWidget::getText()
{
	return d->Text;
}

void QWidget::setText(char *text)
{
	char *t = NewStr(text);
	DeleteArray(d->Text);
	d->Text = t;

	if (d->Debug)
	{
		printf("%p: settext\n", this);
	}
	XStoreName(XDisplay(), d->Win, d->Text ? d->Text : "");
	
	update();
}

int QWidget::value()
{
	return 0;
}

void QWidget::setValue(int i)
{
}

QWidget *QWidget::MouseGrabber = 0;

void QWidget::grabMouse()
{
	if (d->Win)
	{
		if (d->Debug)
		{
			printf("%p: grab mouse win=%i\n", this, d->Win);
		}

		if (MouseGrabber)
		{
			// printf("Ungrabbing previous grab to grab on new window.\n");
			MouseGrabber->ungrabMouse();
		}
		
		int e;
		if ((e = XGrabPointer(	XDisplay(),
								d->Win,
								false,				// owner events
								ButtonPressMask |
									ButtonReleaseMask |
									PointerMotionMask |
									EnterWindowMask |
									LeaveWindowMask |
									FocusChangeMask,// event mask
								GrabModeAsync,		// pointer mode
								GrabModeAsync,		// keyboard mode
								XNone,				// confine to
								XNone,				// cursor
								CurrentTime)) == GrabSuccess)
		{
			MouseGrabber = this;
		}
		else
		{
			printf("Error: QWidget::grabMouse failed (%s). Win=%i Map=%s Exposed=%i\n", XErr(e), d->Win, d->MapName(), d->Exposed);
		}
	}
}

void QWidget::ungrabMouse()
{
	if (d->Win)
	{
		if (d->Debug)
		{
			printf("%p: ungrab mouse\n", this);
		}
		
		XUngrabPointer(XDisplay(), CurrentTime);
		MouseGrabber = 0;
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////
template <class q>
void QView<q>::OnClick(QEvent *e, bool down, bool dbl, bool move)
{
	if (v)
	{
		GMouse m;

		m.Target = v;
		m.Flags = 0;
		m.x = e->x();
		m.y = e->y();
		m.Down(e->down());
		m.Double(e->doubleclick());

		if (e->button() & QWidget::LeftButton) m.Left(true);
		if (e->button() & QWidget::RightButton) m.Right(true);
		if (e->button() & QWidget::MidButton) m.Middle(true);

		if (e->button() & QWidget::ShiftButton) m.Shift(true);
		if (e->button() & QWidget::ControlButton) m.Ctrl(true);
		if (e->button() & QWidget::AltButton) m.Alt(true);

		v->_Mouse(m, move);
	}
}

template <class q>
bool QView<q>::OnKey(QEvent *e, bool down)
{
	if (v)
	{
		GKey k;

		k.c16 = k.c = e->ascii();
		k.Flags = 0;
		k.Down(down);
		k.Data = 0;

		if (e->button() & QWidget::ShiftButton) k.Shift(true);
		if (e->button() & QWidget::ControlButton) k.Ctrl(true);
		if (e->button() & QWidget::AltButton) k.Alt(true);

		k.IsChar =	(k.c >= ' ' OR
					k.c == '\t' OR
					k.c == VK_RETURN OR
					k.c == 8) AND
					NOT k.Alt() AND
					NOT k.Ctrl();
			
		GWindow *w = v->GetWindow();
		if (w)
		{
			return w->OnViewKey(v, k);
		}
		else
		{
			printf("%s:%i - No window.\n", __FILE__, __LINE__);
		}
	}
		
	return false;
}

template <class q>
	QView<q>::QView(GView *view, bool p = true) : q()
	{
		v = view;
		_Paint = p;
		setBackgroundMode(QWidget::NoBackground);
	}

template <class q>
	QView<q>::QView(GView *view, Window Existing) : q(Existing)
	{
		v = view;
		_Paint = true;
		setBackgroundMode(QWidget::NoBackground);
	}

template <class q>
	QView<q>::~QView()
	{
		if (v)
		{
			v->_View = 0;
			DeleteObj(v);
		}
	}

template <class q>
	void QView<q>::_SetWndPtr(void *p)
	{
		v = (GView*)p;
	}
	
template <class q>
	void *QView<q>::_GetWndPtr()
	{
		return v;
	}

	// Events
template <class q>
	void QView<q>::resizeEvent(QEvent *e)
	{
		q::resizeEvent(e);
		if (v) v->OnPosChange();
	}

template <class q>
	void QView<q>::paintEvent(QEvent *e)
	{
		if (_Paint) q::paintEvent(e);
		if (v) v->_Paint();
	}

template <class q>
	void QView<q>::customEvent(QEvent *e)
	{
		GMessage *m = dynamic_cast<GMessage*>(e);
		if (m AND v)
		{
			v->OnEvent(m);
		}
	}

template <class q>
	void QView<q>::notifyEvent(int i = 0)
	{
		if (v)
		{
			GView *n = v->GetNotify() ? v->GetNotify() : v->GetParent();
			if (n)
			{
				n->OnNotify(v, i);
			}
		}
	}

	// Mouse events
template <class q>
	void QView<q>::mousePressEvent(QEvent *e)
	{
		q::mousePressEvent(e);
		OnClick(e, true, false, false);
	}

template <class q>
	void QView<q>::mouseDoubleClickEvent(QEvent *e)
	{
		q::mouseDoubleClickEvent(e);
		OnClick(e, true, true, false);
	}

template <class q>
	void QView<q>::mouseReleaseEvent(QEvent *e)
	{
		q::mouseReleaseEvent(e);
		OnClick(e, false, false, false);
	}

template <class q>
	void QView<q>::mouseMoveEvent(QEvent *e)
	{
		q::mouseMoveEvent(e);
		OnClick(e, false, false, true);
	}

template <class q>
	void QView<q>::leaveEvent(QEvent *e)
	{
		q::leaveEvent(e);

		GMouse m;
		m.x = m.y = m.Flags = 0;
		if (v) v->OnMouseExit(m);
	}

template <class q>
	void QView<q>::enterEvent(QEvent *e)
	{
		q::enterEvent(e);

		GMouse m;
		m.x = m.y = m.Flags = 0;
		if (v) v->OnMouseEnter(m);
	}

template <class q>
	void QView<q>::wheelEvent(QEvent *e)
	{
		q::wheelEvent(e);

		if (v) v->OnMouseWheel(- ((double) e->delta() * QApplication::wheelScrollLines() / 120) );
	}

	// Focus events
template <class q>
	void QView<q>::focusInEvent(QEvent *e)
	{
		q::focusInEvent(e);
		if (v) v->OnFocus(true);
	}

template <class q>
	void QView<q>::focusOutEvent(QEvent *e)
	{
		q::focusOutEvent(e);
		if (v) v->OnFocus(false);
	}

	// Keyboard events
template <class q>
	bool QView<q>::keyPressEvent(QEvent *e)
	{
		q::keyPressEvent(e);
		return OnKey(e, true);
	}

template <class q>
	bool QView<q>::keyReleaseEvent(QEvent *e)
	{
		q::keyReleaseEvent(e);
		return OnKey(e, false);
	}

// Instances
template class QView<QWidget>;
