/*hdr
**	FILE:			GMemDC.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			14/10/2000
**	DESCRIPTION:	GDC v2.xx header
**
**	Copyright (C) 2000, Matthew Allen
**		fret@memecode.com
*/

#include <stdio.h>
#include <math.h>

#include "Gdc2.h"
#include "GString.h"

#define UnTranslate()		// if (p) p->translate(OriginX, OriginY);
#define Translate()			// if (p) p->translate(-OriginX, -OriginY);

/////////////////////////////////////////////////////////////////////////////////////////////////////
GMemDC::GMemDC(int x, int y, int bits)
{
	Bmp = 0;
	p = 0;
	
	if (bits > 0)
	{
		Create(x, y, bits);
	}
}

GMemDC::GMemDC(GSurface *pDC)
{
	Bmp = 0;
	p = 0;
	
	if (pDC AND
		Create(pDC->X(), pDC->Y(), pDC->GetBits()))
	{
		Blt(0, 0, pDC);
	}
}

GMemDC::~GMemDC()
{
	if (pMem)
	{
		DeleteArray(pMem->Base);
	}
}

OsPainter GMemDC::Handle()
{
	return 0;
}

void GMemDC::SetOrigin(int x, int y)
{
	Handle();
	UnTranslate();
	GSurface::SetOrigin(x, y);
	Translate();
}

bool GMemDC::Create(int x, int y, int Bits, int LineLen, bool KeepData)
{
	if (pMem)
	{
		DeleteArray(pMem->Base);
	}

	if (x < 1 OR y < 1 OR Bits < 1) return false;
	
	if (NOT pMem) pMem = NEW(GBmpMem);
	if (pMem)
	{
		pMem->x = x;
		pMem->y = y;
		pMem->Bits = Bits;
		pMem->Line = ((Bits * x + 31) / 32) * 4;
		pMem->Base = NEW(uchar[pMem->Line * pMem->y]);
		pMem->Flags = 0;

		int NewOp = (pApp) ? Op() : GDC_SET;

		if (	(Flags & GDC_OWN_APPLICATOR) AND
			NOT (Flags & GDC_CACHED_APPLICATOR))
		{
			DeleteObj(pApp);
		}

		for (int i=0; i<GDC_CACHE_SIZE; i++)
		{
			DeleteObj(pAppCache[i]);
		}

		if (NewOp < GDC_CACHE_SIZE AND NOT DrawOnAlpha())
		{
			pApp = (pAppCache[NewOp]) ? pAppCache[NewOp] : pAppCache[NewOp] = CreateApplicator(NewOp);
			Flags &= ~GDC_OWN_APPLICATOR;
			Flags |= GDC_CACHED_APPLICATOR;
		}
		else
		{
			pApp = CreateApplicator(NewOp, pMem->Bits);
			Flags &= ~GDC_CACHED_APPLICATOR;
			Flags |= GDC_OWN_APPLICATOR;
		}

		if (NOT pApp)
		{
			printf("GMemDC::Create(%i,%i,%i,%i,%i) No Applicator.\n", x, y, Bits, LineLen, KeepData);
			LgiAssert(0);
		}

		Clip.ZOff(X()-1, Y()-1);

		return true;
	}

	return false;
}


void GMemDC::Blt(int x, int y, GSurface *Src, GRect *a)
{
	if (Src)
	{
		if (Src->IsScreen())
		{
			bool Status = false;
			GScreenDC *Screen = dynamic_cast<GScreenDC*>(Src);
			if (Screen)
			{
				/*
				OsView Wnd = Screen->Handle()->Handle();
				GRect *Client = Screen->Handle()->GetClient();

				GRect SrcRect(0, 0, Src->X()-1, Src->Y()-1);
				if (Client)
				{
					SrcRect = *Client;
				}
				
				GRect s;
				if (a)
				{
					s = *a;
					s.Offset(SrcRect.x1, SrcRect.y1);
				}
				else
				{
					s = SrcRect;
				}
				
				GRect b;
				if (a)
				{
					b = *a;
					b.Offset(SrcRect.x1, SrcRect.y1);
					b.Bound(&SrcRect);
				}
				else				
				{
					b = SrcRect;
				}
				
				if (b.Valid())
				{
					GRect Dst, e(0, 0, X()-1, Y()-1);
					Dst.ZOff(b.X()-1, b.Y()-1);
					Dst.Offset(x, y);
					Dst.Bound(&e);
					if (Dst.Valid())
					{
						// Clip the input rectangle against the size of the dest
						if (b.X() > Dst.X())
						{
							b.x2 = b.x1 + Dst.X() - 1;
						}
						if (b.Y() > Dst.Y())
						{
							b.y2 = b.y1 + Dst.Y() - 1;
						}

						Display *Dsp = Handle()->XDisplay();
						if (GetBitmap()->GetImage())
						{
							XGetSubImage(	Dsp,
											Wnd->handle(),
											b.x1, b.y1,
											b.X(), b.Y(),
											-1,
											ZPixmap,
											GetBitmap()->GetImage(),
											b.x1-s.x1, b.y1-s.y1);
							
							Status = true;
						}
					}
				}
				*/
			}
			
			if (NOT Status)
			{
				Colour(Rgb24(255, 0, 255), 24);
				Rectangle();
			}
		}
		else
		{
			GSurface::Blt(x, y, Src, a);
		}
	}
}

void GMemDC::StretchBlt(GRect *d, GSurface *Src, GRect *s)
{

}

void GMemDC::HLine(int x1, int x2, int y, COLOUR a, COLOUR b)
{
	if (x1 > x2) LgiSwap(x1, x2);

	if (x1 < Clip.x1) x1 = Clip.x1;
	if (x2 > Clip.x2) x2 = Clip.x2;
	if (	x1 <= x2 AND
		y >= Clip.y1 AND
		y <= Clip.y2)
	{
		COLOUR Prev = pApp->c;

		pApp->SetPtr(x1, y);
		for (; x1 <= x2; x1++)
		{
			if (x1 & 1)
			{
				pApp->c = a;
			}
			else
			{
				pApp->c = b;
			}

			pApp->Set();
			pApp->IncX();
		}

		pApp->c = Prev;
	}
}

void GMemDC::VLine(int x, int y1, int y2, COLOUR a, COLOUR b)
{
	if (y1 > y2) LgiSwap(y1, y2);
	
	if (y1 < Clip.y1) y1 = Clip.y1;
	if (y2 > Clip.y2) y2 = Clip.y2;
	if (	y1 <= y2 AND
		x >= Clip.x1 AND
		x <= Clip.x2)
	{
		COLOUR Prev = pApp->c;

		pApp->SetPtr(x, y1);
		for (; y1 <= y2; y1++)
		{
			if (y1 & 1)
			{
				pApp->c = a;
			}
			else
			{
				pApp->c = b;
			}

			pApp->Set();
			pApp->IncY();
		}

		pApp->c = Prev;
	}
}
