#include "Lgi.h"
#include <cups/ipp.h>
#include <cups/cups.h>

#define PS_SCALE			10

///////////////////////////////////////////////////////////////////////////////////////
#include "GFilterUtils.h"

class GCups : public GLibrary
{
public:
	GCups() : GLibrary("libcups")
	{
		if (NOT IsLoaded())
		{
			printf("%s,%i - CUPS (Common Unix Printing System) not found.\n", __FILE__, __LINE__);
		}
	}
	
	~GCups() {}

	DynFunc5(int, cupsPrintFile,
			const char *, printer,
			const char *, filename,
			const char *, title,
			int, num_options,
			cups_option_t *, options);
	DynFunc0(int, cupsLastError);
	DynFunc1(int, cupsGetDests, cups_dest_t **, dests);
	DynFunc4(cups_dest_t *, cupsGetDest, const char *, name,
			const char *, instance, int, num_dests, cups_dest_t *, dests);
	DynFunc4(int, cupsAddOption, const char *, name, const char *, value, int, num_options, cups_option_t **, options);
};

///////////////////////////////////////////////////////////////////////////////////////
class GPrintDCPrivate : public GCups
{
public:
	class PrintPainter *p;
	void *Handle;
	char *PrintJobName;
	bool DocOpen;
	bool PageOpen;
	int Pages;
	COLOUR c;
	
	char *FileName;
	GFile Ps;
	
	GPrintDCPrivate()
	{
		p = 0;
		c = -1;
		Pages = 0;
		FileName = 0;
		Handle = 0;
		PrintJobName = 0;
		PageOpen = false;
		DocOpen = false;
	}
	
	~GPrintDCPrivate()
	{
		if (DocOpen)
		{
		}
		
		DeleteArray(FileName);
		DeleteArray(PrintJobName);
	}
	
	bool IsOk()
	{
		return	this != 0 AND
				IsLoaded();
	}
	
	bool StartPs()
	{
		char Temp[256];
		char Part[64];
		do
		{
			LgiGetSystemPath(LSP_TEMP, Temp, sizeof(Temp));
			sprintf(Part, "_lgi_printjob_%x.ps", LgiRand());
			LgiMakePath(Temp, sizeof(Temp), Temp, Part);
		}
		while (FileExists(Temp));
		
		FileName = NewStr(Temp);
		if (FileName)
		{
			if (Ps.Open(FileName, O_WRITE))
			{
				Ps.Print(	"%!PS-Adobe-3.0\n"
							"%%BoundingBox: left bottom right top\n"
							"%%Pages: (atend)\n"
							"%%EndComments\n"
							"newpath clippath pathbbox\n"
							"/papery2 exch def\n"
							"/paperx2 exch def\n"
							"/papery1 exch def\n"
							"/paperx1 exch def\n"
							"paperx1 papery2 translate\n"
							"1 setlinewidth\n");
				
				return true;
			}
		}			
		
		return false;
	}
	
	bool StartPage()
	{
		Ps.Print(	"%%Page: Content %i\n"
					"gsave\n",
					Pages + 1);
	}

	bool EndPage()
	{
		Ps.Print(	"grestore\n"
					"showpage\n");
		Pages++;
	}
	
	bool EndPs()
	{
		Ps.Print(	"%%Trailer\n"
					"%%Pages: %i\n"
					"%%EOF\n",
					Pages);
		Ps.Close();
		if (FileName) FileDev->DeleteFile(FileName, false);
	}
};

///////////////////////////////////////////////////////////////////////////////////////
class PrintPainter
{
	GPrintDC *pDC;
	int Fore, Back;
	List<GRect> Clip;
	int Ox, Oy;

	char Face[256];
	int PtSize;
	int Ascent;

public:
	PrintPainter(GPrintDC *dc)
	{
		Face[0] = 0;
		pDC = dc;
		Fore = Back = 0;
		Ox = Oy = 0;
		PtSize = 0;
		Ascent = 0;
	}
	
	~PrintPainter()
	{
		EmptyClip();
	}

	// Don't need these
	int X() { return pDC->X(); }
	int Y() { return pDC->Y(); }
	void drawPoint(int x, int y) {}
	void drawLine(int x1, int y1, int x2, int y2) {}
	void drawRect(int x, int y, int wid, int height) {}
	void drawImage(int x, int y, OsBitmap image, int sx, int sy, int sw, int sh, int op) {}
	
	GFile &File()
	{
		return pDC->d->Ps;
	}

	void GetScale(double &x, double &y)
	{
		x = PS_SCALE;
		y = PS_SCALE;
	}
	
	void setFore(int c)
	{
		Fore = c;
	}
	
	void setBack(int c)
	{
		Back = c;
	}
	
	void setFont(OsFont f)
	{
		/*
		if (File().IsOpen() AND
			(stricmp(f.GetFamily(), Face) != 0 OR PtSize != f.GetPointSize()))
		{
			strcpy(Face, f.GetFamily());
			PtSize = f.GetPointSize();
			
			QFontMetrics m(&f);
			Ascent = m.ascent();
			
			char c[256], *o = c;
			for (char *i=Face; *i; i++)
			{
				if (*i != ' ') *o++ = *i;
			}
			*o = 0;
			
			File().Print(	"/%s findfont\n"
							"%i scalefont\n"
							"setfont\n",
							c,
							f.GetPointSize());
		}
		*/
	}
	
	void translate(int x, int y)
	{
		Ox += x;
		Oy += y;
	}
	
	void PushClip(int x1, int y1, int x2, int y2)
	{
		GRect *Next = NEW(GRect(Ox + x1, Oy + y1, Ox + x2, Ox + y2));
		if (Next)
		{
			GRect *Last = Clip.Last();
			if (Last)
			{
				Next->Bound(Last);
			}
			else
			{
				GRect r(0, 0, X()-1, Y()-1);
				Next->Bound(&r);
			}
			
			Clip.Insert(Last);
		}
	}
	
	void PopClip()
	{
		GRect *Last = Clip.Last();
		if (Last)
		{
			Clip.Delete(Last);
			DeleteObj(Last);
		}
	}
	
	void EmptyClip()
	{
		Clip.DeleteObjects();
	}
	
	void SetClient(GRect *Client)
	{
	}
	
	void drawText(int x, int y, char16 *text, int len, int *backColour, GRect *clip)
	{
		if (ValidStrW(text) AND File().IsOpen())
		{
			if (len < 0)
			{
				len = StrlenW(text);
			}
			
			File().Print(	"newpath\n"
							"%f %f moveto\n"
							"(%.*S) show\n",
							pDC->Xc(x), pDC->Yc(y),
							len, text);
		}
	}
};

/////////////////////////////////////////////////////////////////////////////////////
GPrintDC::GPrintDC(void *Handle, char *PrintJobName)
{
	d = NEW(GPrintDCPrivate);
	d->Handle = Handle;
	d->PrintJobName = NewStr(PrintJobName);
}

GPrintDC::~GPrintDC()
{
	EndPage();
	
	if (d->Pages)
	{
		char Copies[32] = "1";
		
		d->Ps.Close();
		printf("FileName=%s Exists=%i\n", d->FileName, FileExists(d->FileName));

		int Job = 0;
		cups_dest_t *Dests = 0;
		int DestCount = d->cupsGetDests(&Dests);
		if (Dests)
		{
			cups_dest_t *Dest = d->cupsGetDest(NULL, NULL, DestCount, Dests);
			if (Dest)
			{
				int Options = 0;
				cups_option_t *Option = 0;
				Options = d->cupsAddOption("copies", Copies, Options, &Option);				
				Job = d->cupsPrintFile(Dest->name, d->FileName, d->PrintJobName, Options, Option);
			}
		}

		d->EndPs();
		
		if (NOT Job)
		{
			int Err = d->cupsLastError();
  			printf("Cups Error: %x\n", (int)Err);
  		}
  		else
  		{
			printf("Successful Job=%i\n", Job);
		}
	}	
	
	DeleteObj(d);
}

double GPrintDC::Xc(int x)
{
	return (double) x / PS_SCALE;
}

double GPrintDC::Yc(int y)
{
	return (double) -y / PS_SCALE;
}

OsPainter GPrintDC::Handle()
{
	return 0;
}

int GPrintDC::X()
{
	return (int) (8.26 * (double)DpiX());
}

int GPrintDC::Y()
{
	return (int) (10.27 * (double)DpiY()); // 11.69
}

// This is arbitary, chosen because it's easy to use.
int GPrintDC::GetBits()
{
	return 24;
}

// This resolution is chosen to match postscripts internal
// user space in pt's.
int GPrintDC::DpiX()
{
	return 72 * PS_SCALE;
}

int GPrintDC::DpiY()
{
	return 72 * PS_SCALE;
}

bool GPrintDC::StartPage()
{
	bool Status = false;
	
	if (d->IsOk())
	{
		if (NOT d->Ps.IsOpen())
		{
			d->StartPs();
		}
		
		Status = d->PageOpen = d->StartPage();
	}

	return Status;
}

void GPrintDC::EndPage()
{
	if (d->IsOk() AND
		d->PageOpen)
	{
		d->EndPage();
		d->PageOpen = false;
	}
}

///////////////////////////////////////////////////////////////////////////////
// Print primitives
COLOUR GPrintDC::Colour()
{
	return d->c;
}

COLOUR GPrintDC::Colour(COLOUR c, int Bits)
{
	COLOUR c24 = CBit(24, c, Bits?Bits:24);
	if (c24 != d->c)
	{
		d->c = c24;
		if (d->Ps.IsOpen())
		{
			d->Ps.Print("%f %f %f setrgbcolor\n",
				(double) R24(d->c) / 255.0,
				(double) G24(d->c) / 255.0,
				(double) B24(d->c) / 255.0);
		}
	}
}

void GPrintDC::Set(int x, int y)
{
	if (d->Ps.IsOpen())
	{
		d->Ps.Print("newpath\n"
					"%f %f moveto\n"
					"%f %f lineto\n"
					"stroke\n",
					Xc(x), Yc(y),
					Xc(x), Yc(y));
	}
}

COLOUR GPrintDC::Get(int x, int y)
{
}

void GPrintDC::HLine(int x1, int x2, int y)
{
	if (d->Ps.IsOpen())
	{
		d->Ps.Print("newpath\n"
					"%f %f moveto\n"
					"%f %f lineto\n"
					"stroke\n",
					Xc(x1), Yc(y),
					Xc(x2), Yc(y));
	}
}

void GPrintDC::VLine(int x, int y1, int y2)
{
	if (d->Ps.IsOpen())
	{
		d->Ps.Print("newpath\n"
					"%f %f moveto\n"
					"%f %f lineto\n"
					"stroke\n",
					Xc(x), Yc(y1),
					Xc(x), Yc(y2));
	}
}

void GPrintDC::Line(int x1, int y1, int x2, int y2)
{
	if (d->Ps.IsOpen())
	{
		d->Ps.Print("newpath\n"
					"%f %f moveto\n"
					"%f %f lineto\n"
					"stroke\n",
					Xc(x1), Yc(y1),
					Xc(x2), Yc(y2));
	}
}

void GPrintDC::Circle(double cx, double cy, double radius)
{
}

void GPrintDC::FilledCircle(double cx, double cy, double radius)
{
}

void GPrintDC::Arc(double cx, double cy, double radius, double start, double end)
{
}

void GPrintDC::FilledArc(double cx, double cy, double radius, double start, double end)
{
}

void GPrintDC::Ellipse(double cx, double cy, double x, double y)
{
}

void GPrintDC::FilledEllipse(double cx, double cy, double x, double y)
{
}

void GPrintDC::Box(int x1, int y1, int x2, int y2)
{
	GRect a(x1, y1, x2, y2);
	Box(&a);
}

void GPrintDC::Box(GRect *a)
{
	GRect r;
	if (a) r = *a;
	else r.ZOff(X()-1, Y()-1);

	if (d->Ps.IsOpen())
	{
		d->Ps.Print("newpath\n"
					"%f %f moveto\n"
					"%f %f lineto\n"
					"%f %f lineto\n"
					"%f %f lineto\n"
					"closepath\n"
					"stroke\n",
					Xc(r.x1), Yc(r.y1),
					Xc(r.x2), Yc(r.y1),
					Xc(r.x2), Yc(r.y2),
					Xc(r.x1), Yc(r.y2));
	}
}

void GPrintDC::Rectangle(int x1, int y1, int x2, int y2)
{
	GRect a(x1, y1, x2, y2);
	Rectangle(&a);
}

void GPrintDC::Rectangle(GRect *a)
{
	GRect r;
	if (a) r = *a;
	else r.ZOff(X()-1, Y()-1);
	
	if (d->Ps.IsOpen())
	{
		d->Ps.Print("newpath\n"
					"%f %f moveto\n"
					"%f %f lineto\n"
					"%f %f lineto\n"
					"%f %f lineto\n"
					"closepath\n"
					"fill\n",
					Xc(r.x1), Yc(r.y1),
					Xc(r.x2), Yc(r.y1),
					Xc(r.x2), Yc(r.y2),
					Xc(r.x1), Yc(r.y2));
	}
}

void GPrintDC::Blt(int x, int y, GSurface *Src, GRect *a)
{
}

void GPrintDC::StretchBlt(GRect *d, GSurface *Src, GRect *s)
{
}

void GPrintDC::Polygon(int Points, GdcPt2 *Data)
{
}

void GPrintDC::Bezier(int Threshold, GdcPt2 *Pt)
{
}

void GPrintDC::FloodFill(int x, int y, int Mode, COLOUR Border, GRect *Bounds)
{
}

