/*hdr
**	FILE:			GScreenDC.cpp
**	AUTHOR:			Matthew Allen
**	DATE:			14/10/2000
**	DESCRIPTION:	GDC v2.xx header
**
**	Copyright (C) 2000, Matthew Allen
**		fret@memecode.com
*/

#include <stdio.h>
#include <math.h>

#include "Gdc2.h"

class GScreenPrivate
{
public:
	COLOUR Col;
	OsView View;
	Y::Canvas *c;

	GScreenPrivate()
	{
		View = 0;
		Col = 0;
		c = 0;
	}
};

// Translates are cumulative... so we undo the last one before resetting it.
#define UnTranslate()		// d->p.translate(OriginX, OriginY);
#define Translate()			// d->p.translate(-OriginX, -OriginY);

/////////////////////////////////////////////////////////////////////////////////////////////////////
GScreenDC::GScreenDC()
{
	d = NEW(GScreenPrivate);
}

GScreenDC::GScreenDC(OsView view, void *Param)
{
	d = NEW(GScreenPrivate);
	d->View = view;
	d->c = (Y::Canvas*)Param;
	if (d->c)
	{
		d->c->savePainterState();
		int x = X(), y = Y();
		d->c->reset(x, y);
	}
}

GScreenDC::~GScreenDC()
{
	if (d->c)
	{
		d->c->swapBuffers();
		d->c->restorePainterState();
	}

	UnTranslate();
	DeleteObj(d);
}

OsPainter GScreenDC::Handle()
{
	return 0;
}

bool GScreenDC::IsScreen()
{
	return true;
}

uint GScreenDC::LineStyle()
{
	return GSurface::LineStyle();
}

uint GScreenDC::LineStyle(uint Bits)
{
	return GSurface::LineStyle(Bits);
}

int GScreenDC::GetFlags()
{
	return 0;
}

void GScreenDC::GetOrigin(int &x, int &y)
{
	return GSurface::GetOrigin(x, y);
}

void GScreenDC::SetOrigin(int x, int y)
{
	UnTranslate();
	GSurface::SetOrigin(x, y);
	Translate();
}

GRect GScreenDC::ClipRgn()
{
	return Clip;
}

GRect GScreenDC::ClipRgn(GRect *c)
{
	GRect Prev = Clip;
	UnTranslate();

	if (c)
	{
		Clip = *c;
		// d->p.EmptyClip();
		// d->p.PushClip(Clip.x1, Clip.y1, Clip.x2, Clip.y2);
	}
	else
	{
		// d->p.EmptyClip();
	}

	Translate();
	return Prev;
}

COLOUR GScreenDC::Colour()
{
	return d->Col;
}

COLOUR GScreenDC::Colour(COLOUR c, int Bits)
{
	COLOUR Prev = Colour();

	d->Col = CBit(32, c, Bits>0 ? Bits : GdcD->GetBits());
	if (d->c)
	{
		d->c->setPenColour(d->Col);
		d->c->setFillColour(d->Col);
	}

	return Prev;
}

int GScreenDC::Op(int ROP)
{
	int Prev = Op();

	switch (ROP)
	{
		case GDC_SET:
		{
			// d->p.setRasterOp(QPainter::CopyROP);
			break;
		}
		case GDC_OR:
		{
			// d->p.setRasterOp(QPainter::OrROP);
			break;
		}
		case GDC_AND:
		{
			// d->p.setRasterOp(QPainter::AndROP);
			break;
		}
		case GDC_XOR:
		{
			// d->p.setRasterOp(QPainter::XorROP);
			break;
		}
	}

	return Prev;
}

int GScreenDC::Op()
{
	/*
	switch (d->p.rasterOp())
	{
		case QPainter::CopyROP:
		{
			return GDC_SET;
			break;
		}
		case QPainter::OrROP:
		{
			return GDC_OR;
			break;
		}
		case QPainter::AndROP:
		{
			return GDC_AND;
			break;
		}
		case QPainter::XorROP:
		{
			return GDC_XOR;
			break;
		}
	}
	*/
	
	return GDC_SET;
}

int GScreenDC::X()
{
	if (d->View)
	{
		GRect r = d->View->geometry();
		return r.X();
	}
	return 0;
}

int GScreenDC::Y()
{
	if (d->View)
	{
		GRect r = d->View->geometry();
		return r.Y();
	}
	return 0;
}

int GScreenDC::GetBits()
{
	return 32; // GdcD->GetBits();
}

GPalette *GScreenDC::Palette()
{
	return 0;
}

void GScreenDC::Palette(GPalette *pPal, bool bOwnIt)
{
}

void GScreenDC::Set(int x, int y)
{
	if (d->c)
	{
		d->c->drawRectangle(x, y, 1, 1);
	}
}

COLOUR GScreenDC::Get(int x, int y)
{
	return 0;
}

void GScreenDC::HLine(int x1, int x2, int y)
{
	if (d->c)
	{
		d->c->drawHLine(x1, y, x2-x1);
	}
}

void GScreenDC::VLine(int x, int y1, int y2)
{
	if (d->c)
	{
		d->c->drawVLine(x, y1, y2-y1);
	}
}

void GScreenDC::Line(int x1, int y1, int x2, int y2)
{
	if (d->c)
	{
		d->c->drawLine(x1, y1, x2-x1, y2-y1);
	}
}

void GScreenDC::Circle(double cx, double cy, double radius)
{
	// d->p.drawArc(cx, cy, radius);
}

void GScreenDC::FilledCircle(double cx, double cy, double radius)
{
	// d->p.fillArc(cx, cy, radius);
}

void GScreenDC::Arc(double cx, double cy, double radius, double start, double end)
{
	// d->p.drawArc(cx, cy, radius, start, end);
}

void GScreenDC::FilledArc(double cx, double cy, double radius, double start, double end)
{
	// d->p.fillArc(cx, cy, radius, start, end);
}

void GScreenDC::Ellipse(double cx, double cy, double x, double y)
{
}

void GScreenDC::FilledEllipse(double cx, double cy, double x, double y)
{
}

void GScreenDC::Box(int x1, int y1, int x2, int y2)
{
	Line(x1, y1, x2-1, y1);
	Line(x2, y1, x2, y2-1);
	Line(x2, y2, x1+1, y2);
	Line(x1, y1+1, x1, y2);
}

void GScreenDC::Box(GRect *a)
{
	if (a)
	{
		Box(a->x1, a->y1, a->x2, a->y2);
	}
	else
	{
		Box(0, 0, X()-1, Y()-1);
	}
}

void GScreenDC::Rectangle(int x1, int y1, int x2, int y2)
{
	if (x2 >= x1 AND
		y2 >= y1 AND
		d->c)
	{
		d->c->drawRectangle(x1, y1, x2-x1, y2-y1);
	}
}

void GScreenDC::Rectangle(GRect *a)
{
	if (a)
	{
		if (a->X() > 0 AND
			a->Y() > 0)
		{
			d->c->drawRectangle(a->x1, a->y1, a->X(), a->Y());
		}
	}
	else
	{
		d->c->drawRectangle(0, 0, X(), Y());
	}
}

void GScreenDC::Blt(int x, int y, GSurface *Src, GRect *a)
{
	if (Src)
	{
		if (Src->IsScreen())
		{
		}
		else
		{
			// memory -> screen blt
			if (Src->GetBitmap())
			{
				GRect s;
				if (NOT a)
				{
					s.ZOff(Src->X()-1, Src->Y()-1);
					a = &s;
				}
				
				if (Src->GetBits() != GetBits())
				{
					// Do on the fly depth conversion...
					GSurface *t = NEW(GMemDC);
					if (t)
					{
						if (t->Create(a->X(), a->Y(), GetBits()))
						{
							t->GSurface::Blt(0, 0, Src, a);
							
							/*d->p.drawImage(	x + a->x1, y + a->y1,
											*t->GetBitmap(),
											0, 0, a->X(), a->Y(),
											QImage::ColorOnly);*/
						}
						
						DeleteObj(t);
					}
				}
				else
				{
					// d->p.drawImage(x, y, *Src->GetBitmap(), a->x1, a->y1, a->X(), a->Y(), QImage::ColorOnly);
				}
			}
		}
	}
}

void GScreenDC::StretchBlt(GRect *d, GSurface *Src, GRect *s)
{
}

void GScreenDC::Polygon(int Points, GdcPt2 *Data)
{
}

void GScreenDC::Bezier(int Threshold, GdcPt2 *Pt)
{
}

void GScreenDC::FloodFill(int x, int y, int Mode, COLOUR Border, GRect *Bounds)
{
}

