/*hdr
**      FILE:           Memory.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           10/4/98
**      DESCRIPTION:    Memory subsystem
**
**      Copyright (C) 1998, Matthew Allen
**              fret@memecode.com
*/

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "Lgi.h"

#ifdef MEMORY_DEBUG

char *_vmem_file = 0;
int _vmem_line = 0;

class MInfo
{
public:
	char File[16];
	int Line;
	int Count;
	
	MInfo(char *f, int l)
	{
		char *d = strrchr(f, DIR_CHAR);
		if (d) strncpy(File, d+1, sizeof(File)-1);
		else strncpy(File, f, sizeof(File)-1);
		File[sizeof(File)-1] = 0;
		Line = l;
		Count = 0;
	}
};

static _Size = 32 << 10;
static MInfo *_Table[_Size];

MInfo *_GetInfo(char *f, int l)
{
	
}

void *operator new(unsigned size, char *file, int line)
{
	return malloc(size);
}

void *operator new[](unsigned size, char *file, int line)
{
	return malloc(size);
}

void operator delete(void *p)
{
	free(p);
}

void operator delete(void *p, char *file, int line)
{
	free(p);
}

void LgiDumpHeapInfo()
{
	printf("Heap Info\n");
}

#endif


/*
#ifdef MEMORY_DEBUG

#if 0

#define _malloc(s)		VirtualAlloc(0, s, MEM_COMMIT, PAGE_READWRITE)
#define _free(p)		VirtualFree(p, 0, MEM_RELEASE);

#else

#include <malloc.h>
#define abs(a)			((a)>0 ? (a) : -(a))
#define _malloc(s)		malloc(s)
#define _free(p)		free(p);

#endif

char *_vmem_file = 0;
int _vmem_line = 0;

#include <crtdbg.h>

#define _VMEM_MAGIC		0x12345678
#define _VMEM_BUFFER	16
#define _VMEM_FILL		0xcdcdcdcd

#ifdef _METRICS
class _vmem_metrics
{
	class _metric
	{
	public:
		char *File;
		int Line;
		int Count;

		_metric()
		{
			File = 0;
			Line = 0;
			Count = 0;
		}
	};

	int Length;
	_metric *m;
	GSemaphore Lock;

public:
	_vmem_metrics()
	{
		Length = 16 << 10;
		
		int Bytes = sizeof(_metric) * Length;
		m = (_metric*) malloc(Bytes);
		if (m)
		{
			memset(m, 0, Bytes);
		}
	}

	~_vmem_metrics()
	{
		for (int i=0; i<10; i++)
		{
			_metric *Max = m;
			for (int n=0; n<Length AND m[n].File; n++)
			{
				if (m[n].Count > Max->Count)
				{
					Max = m + n;
				}
			}

			if (Max->Count)
			{
				char s[256];
				sprintf(s, "Metrics: %s,%i Count: %i\n", Max->File, Max->Line, Max->Count);
				OutputDebugString(s);
				Max->Count = 0;
			}
		}

		free(m);
	}

	void OnAlloc(char *file, int line)
	{
		if (Lock.Lock())
		{
			int i;
			for (i=0; i<Length AND m[i].File; i++)
			{
				if (stricmp(m[i].File, file) == 0 AND
				// if (m[i].File == file AND
					m[i].Line == line)
				{
					m[i].Count++;
					return;
				}
			}

			if (i<Length)
			{
				m[i].File = file;
				m[i].Line = line;
				m[i].Count = 1;
			}

			Lock.Unlock();
		}
	}

} VmemMetrics;
#endif

struct _vmem_info
{
	int Magic;
	char *File;
	int Line;
	int Size;
	_vmem_info *Prev, *Next;

	bool CheckFill()
	{
		uint32 *Pre = (uint32*)(this+1);
		uint32 *Post = (uint32*) (((uchar*)Pre) + Size + _VMEM_BUFFER);
		uint32 *e;

		for (e = Pre+(_VMEM_BUFFER/sizeof(uint32)); Pre < e; Pre++)
		{
			if (*Pre != _VMEM_FILL) return false;
		}

		for (e = Post+(_VMEM_BUFFER/sizeof(uint32)); Post < e; Post++)
		{
			if (*Post != _VMEM_FILL) return false;
		}

		return true;
	}

	void Dump()
	{
		char s[256];
		sprintf(s, "Leak: %s,%i (%i bytes, '", File, Line, Size);

		char *Data = ((char*)(this+1)) + _VMEM_BUFFER;
		char *e = s + strlen(s);
		for (int i=0; i<min(Size, 32) AND Data[i]; i++)
		{
			if (Data[i] >= ' ')
				*e++ = Data[i];
			else
				*e++ = '.';
		}
		strcpy(e, "')\n");

		OutputDebugString(s);
	}
};

int _used = 0;
int _max_used = 0;
_vmem_info *_First, *_Last;

class _VMemCleanup
{
public:
	_VMemCleanup()
	{
		_CrtSetReportMode(_CRT_WARN, _CRTDBG_MODE_WNDW);
		_CrtSetReportMode(_CRT_ERROR, _CRTDBG_MODE_WNDW);
		_CrtSetReportMode(_CRT_ASSERT, _CRTDBG_MODE_WNDW);
	}

	~_VMemCleanup()
	{
		LgiAssert(LgiCheckHeap());

		char s[256];
		if (_First)
		{
			sprintf(s, "Memory Leaks: %i (Max blocks %i)\n", _used, _max_used);
			OutputDebugString(s);

			int k=0;
			for (_vmem_info *i=_First; i; k++)
			{
				i->Dump();
				void *p = i;
				i = i->Next;
				_free(p);
			}
		}
		else
		{
			sprintf(s, "No Memory Leaks. (Max blocks %i)\n", _max_used);
			OutputDebugString(s);
		}
	}

} VMemCleanup;

int _Allocated = 0;

void *_vmem_alloc(size_t Size, char *file, int line)
{
	int TotalSize = Size + (_VMEM_BUFFER * 2);
	_vmem_info *p = (_vmem_info*)_malloc(sizeof(_vmem_info) + TotalSize);
	if (p)
	{
		#ifdef _METRICS
		VmemMetrics.OnAlloc(file, line);
		#endif

		// set the members
		p->Magic = _VMEM_MAGIC;
		p->File = file;
		p->Line = line;
		p->Size = Size;
		_used++;
		_max_used = max(_used, _max_used);

		// insert into the list
		if (NOT _First)
		{
			_First = _Last = p;
			p->Prev = p->Next = 0;
		}
		else
		{
			p->Prev = 0;
			p->Next = _First;
			_First->Prev = p;
			_First = p;
		}

		return ((uchar*)(p+1)) + _VMEM_BUFFER;
	}
	else
	{
		// out of memory
		_asm int 3
	}

	return 0;
}

void _vmem_free(void *ptr)
{
	if (ptr)
	{
		_vmem_info *p = ((_vmem_info*) ( ((uchar*)ptr) - _VMEM_BUFFER) ) - 1;

		if (p->Magic == _VMEM_MAGIC)
		{
			// Check fill
			if (NOT p->CheckFill())
			{
				_asm int 3
			}

			// Remove from the list
			if (p == _First AND p == _Last)
			{
				_First = _Last = 0;
			}
			else if (p == _First)
			{
				if (p->Prev) _asm int 3

				_First = p->Next;
				if (_First) _First->Prev = 0;
			}
			else if (p == _Last)
			{
				if (p->Next) _asm int 3

				_Last = p->Prev;
				if (p->Prev) p->Prev->Next = 0;
			}
			else
			{
				if (NOT p->Next OR NOT p->Prev) _asm int 3

				p->Next->Prev = p->Prev;
				p->Prev->Next = p->Next;
			}
			p->Prev = p->Next = 0;

			_used--;

			_free(p);
		}
		else
		{
			// this pointer wasn't allocated by us
			_asm int 3
		}
	}
	else
	{
		// no pointer
		_asm int 3
	}
}
#endif
*/

bool LgiCheckHeap()
{
	#ifdef MEMORY_DEBUG

	_vmem_info *i = 0;
	int n = 0;
	int MinD = 0x7fffffff;
	try
	{
		for (i=_First; i; i=i->Next, n++)
		{
			char c = i->File[0];

			if (i->Magic != _VMEM_MAGIC OR
				NOT i->CheckFill())
			{
				_asm int 3
			}
		}
	}
	catch (...)
	{
		_asm int 3
		return false;
	}

	#endif

	if (_heapchk() != _HEAPOK)
	{
		LgiAssert(0);
		return false;
	}

	return true;
}

bool LgiCanReadMemory(void *p, int Len)
{
	/*
	try
	{
		int i = 0;
		for (char *Ptr = (char*)p; Len > 0; Len--)
		{
			i += *Ptr++;
		}
	}
	catch (...)
	{
		// I guess not... then..
		return false;
	}
	*/

	return true;
}

void MemorySizeToStr(char *Str, quad Size)
{
	#define K				1024.0
	#define M				(K*K)
	#define G				(M*K)

	if (Size >= G)
	{
		sprintf(Str, "%.2 G", Size / G);
	}
	else if (Size >= M)
	{
		sprintf(Str, "%.2f M", Size / M);
	}
	else if (Size >= (10 * K))
	{
		sprintf(Str, "%.2f K", Size / K);
	}
	else
	{
		sprintf(Str, "%ld bytes", Size);
	}

	#undef K
	#undef M
	#undef G
}

