#include "Lgi.h"
#include "GScrollBar.h"

//////////////////////////////////////////////////////////////////////////////////////
class GScrollBarPrivate
{
public:
	int			Type;		// vertical, horzontal or control
	SCROLLINFO	Info;

	GScrollBarPrivate()
	{
		Type = SB_VERT;
		ZeroObj(Info);
		Info.cbSize = sizeof(Info);
		Info.fMask = SIF_ALL; // | SIF_DISABLENOSCROLL;
		Info.nMax = -1;
	}
};

GScrollBar::GScrollBar()
	: ResObject(Res_ScrollBar)
{
	d = NEW(GScrollBarPrivate);
	SetParent(0);
}

GScrollBar::GScrollBar(int id, int x, int y, int cx, int cy, char *name)
	: ResObject(Res_ScrollBar)
{
	d = NEW(GScrollBarPrivate);
	d->Type = SB_CTL;
	d->Info.fMask |= SIF_DISABLENOSCROLL;

	SetId(id);
	SetPos(GRect(x, y, x+cx, y+cy));
	if (name) Name(name);

	SetStyle(GetStyle() | WS_CHILD | WS_TABSTOP | WS_VISIBLE);
	SetClass(LGI_SCROLLBAR);
	if (SubClass = GWin32Class::Create(GetClass()))
	{
		SubClass->SubClass("SCROLLBAR");
	}
}

GScrollBar::~GScrollBar()
{
	if (NOT Handle())
	{
		SetParentFlag(false);
	}
	DeleteObj(d);
}

int GScrollBar::GetScrollSize()
{
	return GetSystemMetrics(SM_CXVSCROLL);
}

bool GScrollBar::SetPos(GRect &p, bool Repaint)
{
	if (d->Type == SB_CTL)
	{
		SetVertical(p.Y() > p.X());
	}

	return GView::SetPos(p, Repaint);
}

bool GScrollBar::Vertical()
{
	return d->Type == SB_VERT;
}

void GScrollBar::SetVertical(bool v)
{
	if (d->Type != SB_CTL)
	{
		if (v)
		{
			d->Type = SB_VERT;
		}
		else
		{
			d->Type = SB_HORZ;
		}
	}
	else
	{
		if (v)
		{
			SetStyle(GetStyle() | SBS_VERT);
		}
		else
		{
			SetStyle(GetStyle() & ~SBS_VERT);
		}
		if (_View)
		{
			SetWindowPos(_View, 0, 0, 0, 0, 0, SWP_NOMOVE | SWP_NOZORDER | SWP_NOREDRAW | SWP_NOSIZE);
		}
	}
}

void GScrollBar::SetParentFlag(bool Set)
{
	int Flag = (Vertical()) ? WS_VSCROLL : WS_HSCROLL;

	if (NOT Handle() AND
		GetParent())
	{
		if (GetParent()->IsAttached())
		{
			DWORD Flags = GetWindowLong(GetParent()->Handle(), GWL_STYLE);

			if (Set)
			{
				Flags |= Flag;
			}
			else
			{
				Flags &= ~Flag;

				// clear limits.
				d->Info.nMin = 0;
				d->Info.nMax = -1;
				Update();
			}

			// check that we're in thread...
			if (GetParent()->InThread())
			{
				// you can't call 'SetWindowLong' from a thread other than
				// the one that created the window.
				SetWindowLong(GetParent()->Handle(), GWL_STYLE, Flags);
				SetWindowPos(	GetParent()->Handle(),
								0, 0, 0, 0, 0,
								SWP_NOMOVE | SWP_NOZORDER | SWP_NOSIZE | SWP_FRAMECHANGED);

				// int Test = GetWindowLong(Parent->Handle(), GWL_STYLE);

				if (Set)
				{
					Update();
				}

				// int Test2 = GetWindowLong(Parent->Handle(), GWL_STYLE);
			}
			else
			{
				// out of thread... post a message to the window
				// so it can do it itself
				GetParent()->PostEvent(M_SET_WND_STYLE, 0, Flags);
			}
		}
	}
}

GView *GScrollBar::GetMyView()
{
	if (Handle())
	{
		return this;
	}
	else if (GetParent() AND
			 GetParent()->Handle())
	{
		return GetParent();
	}

	return 0;
}

void GScrollBar::Update()
{
	if (Handle())
	{
		if (InThread())
		{
			SetScrollInfo(Handle(), SB_CTL, &d->Info, true);
			EnableScrollBar(Handle(), SB_CTL, (d->Info.nMax < d->Info.nMin) ? ESB_DISABLE_BOTH : ESB_ENABLE_BOTH);
		}
		else
		{
			PostEvent(M_SCROLL_BAR_CHANGED, 0, (int)this);
		}
	}
	else if (GetParent() AND
			 GetParent()->Handle())
	{
		if (GetParent()->InThread())
		{
			d->Info.fMask = SIF_ALL;
			SetScrollInfo(GetParent()->Handle(), d->Type, &d->Info, true);
		}
		else
		{
			GetParent()->PostEvent(M_SCROLL_BAR_CHANGED, 0, (int)this);
		}
	}
}

void GScrollBar::SetParent(GView *p)
{
	GView::SetParent(p);
	SetParentFlag(true);
}

int GScrollBar::Value()
{
	return d->Info.nPos;
}

void GScrollBar::Value(int p)
{
	int n = limit(p, d->Info.nMin, d->Info.nMax);
	if (d->Info.nPos != n)
	{
		d->Info.nPos = n;
		Update();
		
		if (GetParent())
		{
			GetParent()->Invalidate();
		}
	}
}

void GScrollBar::Limits(int &Low, int &High)
{
	Low = d->Info.nMin;
	High = d->Info.nMax;
}

void GScrollBar::SetLimits(int Low, int High)
{
	d->Info.nMin = Low;
	d->Info.nMax = High;
	Update();
}

int GScrollBar::Page()
{
	return d->Info.nPage;
}

void GScrollBar::SetPage(int p)
{
	d->Info.nPage = p;
	Update();
}

bool GScrollBar::Invalidate(GRect *r, bool Repaint, bool NonClient)
{
	SetScrollInfo(GetParent()->Handle(), d->Type, &d->Info, true);
	return true;
}

int GScrollBar::OnEvent(GMessage *Msg)
{
	int Status = SubClass ? SubClass->CallParent(Handle(), Msg->Msg, Msg->a, Msg->b) : 0;

	if ((Msg->Msg == WM_HSCROLL AND NOT Vertical())
		OR
		(Msg->Msg == WM_VSCROLL AND Vertical()))
	{
		SCROLLINFO Si;
		ZeroObj(Si);
		Si.cbSize = sizeof(Si);
		Si.fMask = SIF_ALL;

		GView *Hnd = GetMyView();
		if (Hnd AND
			GetScrollInfo(Hnd->Handle(), d->Type, &Si))
		{
			int ScrollCode = (int) LOWORD(Msg->a);
			int PageMax = d->Info.nMax - d->Info.nPage + 1;

			switch (ScrollCode)
			{
				case SB_BOTTOM:
				case SB_LINEDOWN:
				{
					Si.nPos = limit(Si.nPos+1, d->Info.nMin, PageMax);
					break;
				}
				case SB_TOP:
				case SB_LINEUP:
				{
					Si.nPos = limit(Si.nPos-1, d->Info.nMin, PageMax);
					break;
				}
				case SB_PAGEDOWN:
				{
					Si.nPos = limit(Si.nPos+(int)d->Info.nPage, d->Info.nMin, PageMax);
					break;
				}
				case SB_PAGEUP:
				{
					Si.nPos = limit(Si.nPos-(int)d->Info.nPage, d->Info.nMin, PageMax);
					break;
				}
				case SB_THUMBPOSITION:
				case SB_THUMBTRACK:
				{
					Si.nPos = limit(Si.nTrackPos, d->Info.nMin, PageMax);
					break;
				}
			}

			if (d->Info.nPos != Si.nPos)
			{
				d->Info.nPos = Si.nPos;
				Update();
				OnChange(Si.nPos);

				if (GetParent())
				{
					GView *n = GetNotify() ? GetNotify() : GetParent();
					if (n)
					{
						n->OnNotify(this, Si.nPos);
					}
				}
			}
		}
	}

	switch (MsgCode(Msg))
	{
		case WM_CREATE:
		{
			Update();
			break;
		}
		case M_SCROLL_BAR_CHANGED:
		{
			if (MsgB(Msg) == (int)this)
			{
				Update();
			}
			break;
		}
	}

	return Status;
}

