#include "Lgi.h"

//////////////////////////////////////////////////////////////////////////////
#include <process.h>

uint WINAPI ThreadEntryPoint(int i)
{
#ifdef	_MT
	GThread *Thread = (GThread*) i;
	if (Thread)
	{
		// Wait for it...
		int Status = 0;
		int Start = LgiCurrentTime();

		while (Thread->State == LGITHREAD_INIT)
		{
			LgiSleep(5);
			if (LgiCurrentTime() - Start > 2000)
			{
				// If the thread object doesn't set the running state we're
				// stuffed. So timeout in that case after 2 seconds. The only
				// reason I can think that this could happen is if the ResumeThread
				// worked but returned an error.
				break;
			}
		}

		if (Thread->State == LGITHREAD_RUNNING)
		{
			// Ok now we're ready to go...
			Thread->OnBeforeMain();
			Status = Thread->ReturnValue = Thread->Main();
			Thread->OnAfterMain();
		}

		// Shutdown...
		ThreadError:
		Thread->State = LGITHREAD_EXITED;
		if (Thread->DeleteOnExit)
		{
			DeleteObj(Thread);
		}
		_endthreadex(Status);
	}
#endif

	return 0;
}

GThread::GThread()
{
	State = LGITHREAD_INIT;
	hThread = 0;
	ThreadId = 0;
	ReturnValue = 0;
	DeleteOnExit = false;

#ifdef	_MT
	hThread = (HANDLE) _beginthreadex(NULL,
									16<<10,
									(unsigned int (__stdcall *)(void *)) ThreadEntryPoint,
									(LPVOID) this,
									CREATE_SUSPENDED,
									&ThreadId);
#else
	#error "You are tryed to compile GThread support without the WIN32 multithreaded libs.";
#endif
}

GThread::~GThread()
{
	LgiAssert(State == LGITHREAD_INIT OR State == LGITHREAD_EXITED);
}

bool GThread::IsExited()
{
	if (State == LGITHREAD_EXITED)
		return true;
	
	if (ExitCode() != STILL_ACTIVE)
	{
		State = LGITHREAD_EXITED;
		return true;
	}

	return false;
}

void GThread::Run()
{
	if (State == LGITHREAD_INIT)
	{
		DWORD Status = ResumeThread(hThread);
		if (Status)
		{
			State = LGITHREAD_RUNNING;
		}
		else
		{
			State = LGITHREAD_ERROR;
		}
	}
	/*
	else
	{
		State = LGITHREAD_ASLEEP;	// set before we sleep otherwise it can't 
									// be called from the actual thread.
		Status = SuspendThread(hThread);
	}
	*/
}

void GThread::Terminate()
{
	if (hThread)
	{
		TerminateThread(hThread, 0);
		while (NOT IsExited());
	}
}

int GThread::ExitCode()
{
	DWORD d;
	GetExitCodeThread(hThread, &d);
	ReturnValue = d;
	return ReturnValue;
}

int GThread::Main()
{
	return 0;
}

int GThread::Priority()
{
	int p = 0;
	if (hThread)
	{
		p = GetThreadPriority(hThread);
	}
	return p;
}

void GThread::Priority(int p)
{
	if (hThread)
	{
		SetThreadPriority(hThread, p);
	}
}

