/*hdr
**      FILE:           GView.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           23/4/98
**      DESCRIPTION:    Win32 GView Implementation
**
**      Copyright (C) 1998-2003, Matthew Allen
**              fret@memcode.com
*/

#include <stdio.h>
#include <time.h>

#include "Lgi.h"
#include "base64.h"
#include "GCom.h"
#include "GDragAndDrop.h"
#include "GDropFiles.h"
#include "GButton.h"
#include "GdiLeak.h"
#include "GTableLayout.h"

////////////////////////////////////////////////////////////////////////////////////////////////////
#define ToLower(c) \
	( ( (c) >= 'A' AND (c) <= 'Z' ) ? (c) - 'A' + 'a' : (c) )
#define ToUpper(c) \
	( ( (c) >= 'a' AND (c) <= 'z' ) ? (c) - 'a' + 'A' : (c) )

////////////////////////////////////////////////////////////////////////////////////////////////////
class GViewPrivate
{
public:
	// Tracking handles
	static OsView	hPrevCapture;

	// Window data
	int				WndStyle;				// Win32 hWnd Style
	int				WndExStyle;				// Win32 hWnd ExStyle
	int				WndDlgCode;				// Win32 Dialog Code (WM_GETDLGCODE)
	char			*WndClass;
	UINT			TimerId;
	GDragDropTarget	*DropTarget;

	// General
	int				CtrlId;
	GView			*Parent;
	GView			*Notify;
	GdcPt2			MinimumSize;

	// Font
	GFont			*Font;
	bool			FontOwn;
	
	GViewPrivate()
	{
		Font = 0;
		FontOwn = false;
		CtrlId = -1;
		WndStyle = WS_CHILD | WS_VISIBLE | WS_CLIPCHILDREN;
		WndExStyle = 0;
		WndClass = 0;
		WndDlgCode = 0;
		TimerId = 0;
		DropTarget = 0;
		Parent = 0;
		Notify = 0;
	}
	
	~GViewPrivate()
	{
	}
};

HWND GViewPrivate::hPrevCapture = 0;

////////////////////////////////////////////////////////////////////////////////////////////////////
// Helper Stuff
#include "zmouse.h"

int MouseRollMsg = 0;

int _lgi_mouse_wheel_lines()
{
	OSVERSIONINFO Info;
	ZeroObj(Info);
	Info.dwOSVersionInfoSize = sizeof(Info);
	if (GetVersionEx(&Info) AND
		Info.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS AND
		Info.dwMajorVersion == 4 AND
		Info.dwMinorVersion == 0)
	{
       HWND hdlMSHWheel=NULL;
       UINT msgMSHWheelGetScrollLines=NULL;
       UINT uiMsh_WheelScrollLines;

       msgMSHWheelGetScrollLines = 
               RegisterWindowMessage(MSH_SCROLL_LINES);
       hdlMSHWheel = FindWindow(MSH_WHEELMODULE_CLASS, 
                                MSH_WHEELMODULE_TITLE);
       if (hdlMSHWheel && msgMSHWheelGetScrollLines)
       {
			return SendMessage(hdlMSHWheel, msgMSHWheelGetScrollLines, 0, 0);
       }
	}
	else
	{
		UINT nScrollLines;
		if (SystemParametersInfo(	SPI_GETWHEELSCROLLLINES, 
									0, 
									(PVOID) &nScrollLines, 
									0))
		{
			return nScrollLines;
		}
	}

	return 3;
}

#define SetKeyFlag(v, k, f)		if (GetKeyState(k)&0xFF00) { v |= f; }

int _lgi_get_key_flags()
{
	int Flags = 0;
	
	if (LgiGetOs() == LGI_OS_WIN9X)
	{
		SetKeyFlag(Flags, VK_MENU, LGI_EF_ALT);

		SetKeyFlag(Flags, VK_SHIFT, LGI_EF_SHIFT);

		SetKeyFlag(Flags, VK_CONTROL, LGI_EF_CTRL);
	}
	else // is NT/2K/XP
	{
		SetKeyFlag(Flags, VK_LMENU, LGI_EF_LALT);
		SetKeyFlag(Flags, VK_RMENU, LGI_EF_RALT);

		SetKeyFlag(Flags, VK_LSHIFT, LGI_EF_LSHIFT);
		SetKeyFlag(Flags, VK_RSHIFT, LGI_EF_RSHIFT);

		SetKeyFlag(Flags, VK_LCONTROL, LGI_EF_LCTRL);
		SetKeyFlag(Flags, VK_RCONTROL, LGI_EF_RCTRL);
	}

	if (GetKeyState(VK_CAPITAL))
		SetFlag(Flags, LGI_EF_CAPS_LOCK);

	return Flags;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
int GetInputACP()
{
	char Str[16];
	LCID Lcid = (LCID)GetKeyboardLayout(LgiGetCurrentThread()) & 0xffff;
	GetLocaleInfo(Lcid, LOCALE_IDEFAULTANSICODEPAGE , Str, sizeof(Str));
	return atoi(Str);
}

GKey::GKey(int vkey)
{
	char *Cp = 0;

	if (IsWin9x)
	{
		c = (uchar)vkey; 
		
		void *In = &c;
		int Len = 1;
		Cp = LgiAnsiToLgiCp(GetInputACP());
		LgiBufConvertCp(&c16, "ucs-2", sizeof(c16), In, Cp, Len);
	}
	else
	{
		c16 = vkey;
		
		void *In = &c16;
		int Len = 2;
		Cp = LgiAnsiToLgiCp();
		LgiBufConvertCp(&c, Cp, 1, In, LGI_WideCharset, Len);
	}
}

////////////////////////////////////////////////////////////////////////////////////////////////////
GMouse _lgi_adjust_click_for_window(GMouse &Info, GView *Wnd)
{
	GMouse Temp = Info;
	if (Wnd)
	{
		GdcPt2 Offset(0, 0);
		if (Wnd->WindowVirtualOffset(&Offset))
		{
			Temp.x -= Offset.x;
			Temp.y -= Offset.y;
		}
	}
	return Temp;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
LRESULT CALLBACK GWin32Class::Redir(HWND hWnd, UINT m, WPARAM a, LPARAM b)
{
	if (m == WM_NCCREATE)
	{
		LPCREATESTRUCT Info = (LPCREATESTRUCT) b;
		
		GView *NewWnd = (GView*) Info->lpCreateParams;
		NewWnd->_View = hWnd;
		SetWindowLong(hWnd, GWL_USERDATA, (int) NewWnd);
	}

	GView *Wnd = (GView*) GetWindowLong(hWnd, GWL_USERDATA);
	if (Wnd)
	{
		GMessage Msg(m, a, b);
		return Wnd->OnEvent(&Msg);
	}

	if (IsWin9x)
	{
		return DefWindowProcA(hWnd, m, a, b);
	}
	else
	{
		return DefWindowProcW(hWnd, m, a, b);
	}
}

LRESULT CALLBACK GWin32Class::SubClassRedir(HWND hWnd, UINT m, WPARAM a, LPARAM b)
{
	if (m == WM_NCCREATE)
	{
		LPCREATESTRUCT Info = (LPCREATESTRUCT) b;
		GView *NewWnd = 0;
		if (Info->lpCreateParams)
		{
			NewWnd = (GView*) Info->lpCreateParams;
			NewWnd->_View = hWnd;
		}
		SetWindowLong(hWnd, GWL_USERDATA, (int) NewWnd);
	}

	GView *Wnd = (GView*) GetWindowLong(hWnd, GWL_USERDATA);
	if (Wnd)
	{
		GMessage Msg(m, a, b);
		int Status = Wnd->OnEvent(&Msg);
		return Status;
	}

	if (IsWin9x)
	{
		return DefWindowProcA(hWnd, m, a, b);
	}
	else
	{
		return DefWindowProcW(hWnd, m, a, b);
	}
}

GWin32Class::GWin32Class(char *name)
{
	Name(name);

	Class.a.style = CS_DBLCLKS | CS_HREDRAW | CS_VREDRAW;
	Class.a.lpfnWndProc = (WNDPROC) Redir;
	Class.a.cbClsExtra = 0;
	Class.a.cbWndExtra = 8;
	Class.a.hInstance = 0;
	Class.a.hIcon = 0;
	Class.a.hCursor = 0; // LoadCursor(LgiProcessInst(), MAKEINTRESOURCE(IDC_ARROW));
	Class.a.hbrBackground = 0;
	Class.a.lpszMenuName = 0;
	Class.a.lpszClassName = 0;

	ParentProc = 0;
}

GWin32Class::~GWin32Class()
{
	if (IsWin9x)
	{
		UnregisterClassA(Name(), LgiProcessInst());
	}
	else
	{
		UnregisterClassW(NameW(), LgiProcessInst());
	}

	Class.a.lpszClassName = NULL;
}

GWin32Class *GWin32Class::Create(char *ClassName)
{
	GWin32Class *c = 0;

	if (LgiApp)
	{
		List<GWin32Class> *Classes = LgiApp->GetClasses();
		if (Classes)
		{
			for (c = Classes->First(); c; c = Classes->Next())
			{
				if (c->Name() AND
					ClassName AND
					stricmp(c->Name(), ClassName) == 0)
				{
					break;
				}
			}

			if (NOT c)
			{
				c = NEW(GWin32Class(ClassName));
				if (c)
				{
					Classes->Insert(c);
				}
			}
		}
	}

	return c;
}

bool GWin32Class::Register()
{
	if (NOT Class.a.lpszClassName)
	{
		Class.a.hInstance = LgiProcessInst();
		if (IsWin9x)
		{
			Class.a.lpszClassName = Name();
			return RegisterClassA(&Class.a) != 0;
		}
		else
		{
			Class.w.lpszClassName = NameW();
			return RegisterClassW(&Class.w);
		}
	}

	return true;
}

bool GWin32Class::SubClass(char *Parent)
{
	bool Status = false;

	if (IsWin9x)
	{
		if (NOT Class.a.lpszClassName)
		{
			HBRUSH hBr = Class.a.hbrBackground;
			if (GetClassInfoA(LgiProcessInst(), Parent, &Class.a))
			{
				ParentProc = Class.a.lpfnWndProc;
				if (hBr)
				{
					Class.a.hbrBackground = hBr;
				}

				Class.a.cbWndExtra = max(Class.a.cbWndExtra, 8);
				Class.a.hInstance = LgiProcessInst();
				Class.a.lpfnWndProc = (WNDPROC) SubClassRedir;

				Class.a.lpszClassName = Name();

				Status = RegisterClassA(&Class.a) != 0;
			}
		}
		else Status = true;
	}
	else
	{
		if (NOT Class.w.lpszClassName)
		{
			HBRUSH hBr = Class.w.hbrBackground;
			char16 *p = LgiNewUtf8To16(Parent);
			if (p)
			{
				if (GetClassInfoW(LgiProcessInst(), p, &Class.w))
				{
					ParentProc = Class.w.lpfnWndProc;
					if (hBr)
					{
						Class.w.hbrBackground = hBr;
					}

					Class.w.cbWndExtra = max(Class.w.cbWndExtra, 8);
					Class.w.hInstance = LgiProcessInst();
					Class.w.lpfnWndProc = (WNDPROC) SubClassRedir;

					Class.w.lpszClassName = NameW();
					Status = RegisterClassW(&Class.w) != 0;
				}

				DeleteArray(p);
			}
		}
		else Status = true;
	}

	return Status;
}

LRESULT CALLBACK GWin32Class::CallParent(HWND hWnd, UINT m, WPARAM a, LPARAM b)
{
	if (NOT ParentProc) return 0;

	if (IsWindowUnicode(hWnd))
	{
		return CallWindowProcW(ParentProc, hWnd, m, a, b);
	}
	else
	{
		#if _MSC_VER == 1100
		return CallWindowProcA((FARPROC) ParentProc, hWnd, m, a, b);
		#else
		return CallWindowProcA(ParentProc, hWnd, m, a, b);
		#endif
	}
}

//////////////////////////////////////////////////////////////////////////////
/*
struct WinMsg
{
	char *Name;
	int Value;
}
WndMsgs[] =
{
	{"WM_NULL", 0x0000},
	{"WM_CREATE", 0x0001},
	{"WM_DESTROY", 0x0002},
	{"WM_MOVE", 0x0003},
	{"WM_SIZE", 0x0005},
	{"WM_ACTIVATE", 0x0006},
	{"WM_SETFOCUS", 0x0007},
	{"WM_KILLFOCUS", 0x0008},
	{"WM_ENABLE", 0x000A},
	{"WM_SETREDRAW", 0x000B},
	{"WM_SETTEXT", 0x000C},
	{"WM_GETTEXT", 0x000D},
	{"WM_GETTEXTLENGTH", 0x000E},
	{"WM_PAINT", 0x000F},
	{"WM_CLOSE", 0x0010},
	{"WM_QUERYENDSESSION", 0x0011},
	{"WM_QUIT", 0x0012},
	{"WM_QUERYOPEN", 0x0013},
	{"WM_ERASEBKGND", 0x0014},
	{"WM_SYSCOLORCHANGE", 0x0015},
	{"WM_ENDSESSION", 0x0016},
	{"WM_SHOWWINDOW", 0x0018},
	{"WM_WININICHANGE", 0x001A},
	{"WM_DEVMODECHANGE", 0x001B},
	{"WM_ACTIVATEAPP", 0x001C},
	{"WM_FONTCHANGE", 0x001D},
	{"WM_TIMECHANGE", 0x001E},
	{"WM_CANCELMODE", 0x001F},
	{"WM_SETCURSOR", 0x0020},
	{"WM_MOUSEACTIVATE", 0x0021},
	{"WM_CHILDACTIVATE", 0x0022},
	{"WM_QUEUESYNC", 0x0023},
	{"WM_GETMINMAXINFO", 0x0024},
	{"WM_PAINTICON", 0x0026},
	{"WM_ICONERASEBKGND", 0x0027},
	{"WM_NEXTDLGCTL", 0x0028},
	{"WM_SPOOLERSTATUS", 0x002A},
	{"WM_DRAWITEM", 0x002B},
	{"WM_MEASUREITEM", 0x002C},
	{"WM_DELETEITEM", 0x002D},
	{"WM_VKEYTOITEM", 0x002E},
	{"WM_CHARTOITEM", 0x002F},
	{"WM_SETFONT", 0x0030},
	{"WM_GETFONT", 0x0031},
	{"WM_SETHOTKEY", 0x0032},
	{"WM_GETHOTKEY", 0x0033},
	{"WM_QUERYDRAGICON", 0x0037},
	{"WM_COMPAREITEM", 0x0039},
	{"WM_GETOBJECT", 0x003D},
	{"WM_COMPACTING", 0x0041},
	{"WM_COMMNOTIFY", 0x0044},
	{"WM_WINDOWPOSCHANGING", 0x0046},
	{"WM_WINDOWPOSCHANGED", 0x0047},
	{"WM_POWER", 0x0048},
	{"WM_COPYDATA", 0x004A},
	{"WM_CANCELJOURNAL", 0x004B},
	{"WM_NOTIFY", 0x004E},
	{"WM_INPUTLANGCHANGEREQUEST", 0x0050},
	{"WM_INPUTLANGCHANGE", 0x0051},
	{"WM_TCARD", 0x0052},
	{"WM_HELP", 0x0053},
	{"WM_USERCHANGED", 0x0054},
	{"WM_NOTIFYFORMAT", 0x0055},
	{"WM_CONTEXTMENU", 0x007B},
	{"WM_STYLECHANGING", 0x007C},
	{"WM_STYLECHANGED", 0x007D},
	{"WM_DISPLAYCHANGE", 0x007E},
	{"WM_GETICON", 0x007F},
	{"WM_SETICON", 0x0080},
	{"WM_NCCREATE", 0x0081},
	{"WM_NCDESTROY", 0x0082},
	{"WM_NCCALCSIZE", 0x0083},
	{"WM_NCHITTEST", 0x0084},
	{"WM_NCPAINT", 0x0085},
	{"WM_NCACTIVATE", 0x0086},
	{"WM_GETDLGCODE", 0x0087},
	{"WM_SYNCPAINT", 0x0088},
	{"WM_NCMOUSEMOVE", 0x00A0},
	{"WM_NCLBUTTONDOWN", 0x00A1},
	{"WM_NCLBUTTONUP", 0x00A2},
	{"WM_NCLBUTTONDBLCLK", 0x00A3},
	{"WM_NCRBUTTONDOWN", 0x00A4},
	{"WM_NCRBUTTONUP", 0x00A5},
	{"WM_NCRBUTTONDBLCLK", 0x00A6},
	{"WM_NCMBUTTONDOWN", 0x00A7},
	{"WM_NCMBUTTONUP", 0x00A8},
	{"WM_NCMBUTTONDBLCLK", 0x00A9},
	{"WM_KEYFIRST", 0x0100},
	{"WM_KEYDOWN", 0x0100},
	{"WM_KEYUP", 0x0101},
	{"WM_CHAR", 0x0102},
	{"WM_DEADCHAR", 0x0103},
	{"WM_SYSKEYDOWN", 0x0104},
	{"WM_SYSKEYUP", 0x0105},
	{"WM_SYSCHAR", 0x0106},
	{"WM_SYSDEADCHAR", 0x0107},
	{"WM_KEYLAST", 0x0108},
	{"WM_IME_STARTCOMPOSITION", 0x010D},
	{"WM_IME_ENDCOMPOSITION", 0x010E},
	{"WM_IME_COMPOSITION", 0x010F},
	{"WM_IME_KEYLAST", 0x010F},
	{"WM_INITDIALOG", 0x0110},
	{"WM_COMMAND", 0x0111},
	{"WM_SYSCOMMAND", 0x0112},
	{"WM_TIMER", 0x0113},
	{"WM_HSCROLL", 0x0114},
	{"WM_VSCROLL", 0x0115},
	{"WM_INITMENU", 0x0116},
	{"WM_INITMENUPOPUP", 0x0117},
	{"WM_MENUSELECT", 0x011F},
	{"WM_MENUCHAR", 0x0120},
	{"WM_ENTERIDLE", 0x0121},
	{"WM_MENURBUTTONUP", 0x0122},
	{"WM_MENUDRAG", 0x0123},
	{"WM_MENUGETOBJECT", 0x0124},
	{"WM_UNINITMENUPOPUP", 0x0125},
	{"WM_MENUCOMMAND", 0x0126},
	{"WM_CTLCOLORMSGBOX", 0x0132},
	{"WM_CTLCOLOREDIT", 0x0133},
	{"WM_CTLCOLORLISTBOX", 0x0134},
	{"WM_CTLCOLORBTN", 0x0135},
	{"WM_CTLCOLORDLG", 0x0136},
	{"WM_CTLCOLORSCROLLBAR", 0x0137},
	{"WM_CTLCOLORSTATIC", 0x0138},
	{"WM_MOUSEFIRST", 0x0200},
	{"WM_MOUSEMOVE", 0x0200},
	{"WM_LBUTTONDOWN", 0x0201},
	{"WM_LBUTTONUP", 0x0202},
	{"WM_LBUTTONDBLCLK", 0x0203},
	{"WM_RBUTTONDOWN", 0x0204},
	{"WM_RBUTTONUP", 0x0205},
	{"WM_RBUTTONDBLCLK", 0x0206},
	{"WM_MBUTTONDOWN", 0x0207},
	{"WM_MBUTTONUP", 0x0208},
	{"WM_MBUTTONDBLCLK", 0x0209},
	{"WM_MOUSEWHEEL", 0x020A},
	{"WM_MOUSELAST", 0x0209},
	{"WM_PARENTNOTIFY", 0x0210},
	{"WM_ENTERMENULOOP", 0x0211},
	{"WM_EXITMENULOOP", 0x0212},
	{"WM_NEXTMENU", 0x0213},
	{"WM_SIZING", 0x0214},
	{"WM_CAPTURECHANGED", 0x0215},
	{"WM_MOVING", 0x0216},
	{"WM_POWERBROADCAST", 0x0218},
	{"WM_DEVICECHANGE", 0x0219},
	{"WM_MDICREATE", 0x0220},
	{"WM_MDIDESTROY", 0x0221},
	{"WM_MDIACTIVATE", 0x0222},
	{"WM_MDIRESTORE", 0x0223},
	{"WM_MDINEXT", 0x0224},
	{"WM_MDIMAXIMIZE", 0x0225},
	{"WM_MDITILE", 0x0226},
	{"WM_MDICASCADE", 0x0227},
	{"WM_MDIICONARRANGE", 0x0228},
	{"WM_MDIGETACTIVE", 0x0229},
	{"WM_MDISETMENU", 0x0230},
	{"WM_ENTERSIZEMOVE", 0x0231},
	{"WM_EXITSIZEMOVE", 0x0232},
	{"WM_DROPFILES", 0x0233},
	{"WM_MDIREFRESHMENU", 0x0234},
	{"WM_IME_SETCONTEXT", 0x0281},
	{"WM_IME_NOTIFY", 0x0282},
	{"WM_IME_CONTROL", 0x0283},
	{"WM_IME_COMPOSITIONFULL", 0x0284},
	{"WM_IME_SELECT", 0x0285},
	{"WM_IME_CHAR", 0x0286},
	{"WM_IME_REQUEST", 0x0288},
	{"WM_IME_KEYDOWN", 0x0290},
	{"WM_IME_KEYUP", 0x0291},
	{"WM_MOUSEHOVER", 0x02A1},
	{"WM_MOUSELEAVE", 0x02A3},
	{"WM_CUT", 0x0300},
	{"WM_COPY", 0x0301},
	{"WM_PASTE", 0x0302},
	{"WM_CLEAR", 0x0303},
	{"WM_UNDO", 0x0304},
	{"WM_RENDERFORMAT", 0x0305},
	{"WM_RENDERALLFORMATS", 0x0306},
	{"WM_DESTROYCLIPBOARD", 0x0307},
	{"WM_DRAWCLIPBOARD", 0x0308},
	{"WM_PAINTCLIPBOARD", 0x0309},
	{"WM_VSCROLLCLIPBOARD", 0x030A},
	{"WM_SIZECLIPBOARD", 0x030B},
	{"WM_ASKCBFORMATNAME", 0x030C},
	{"WM_CHANGECBCHAIN", 0x030D},
	{"WM_HSCROLLCLIPBOARD", 0x030E},
	{"WM_QUERYNEWPALETTE", 0x030F},
	{"WM_PALETTEISCHANGING", 0x0310},
	{"WM_PALETTECHANGED", 0x0311},
	{"WM_HOTKEY", 0x0312},
	{"WM_PRINT", 0x0317},
	{"WM_PRINTCLIENT", 0x0318},
	{"WM_HANDHELDFIRST", 0x0358},
	{"WM_HANDHELDLAST", 0x035F},
	{"WM_AFXFIRST", 0x0360},
	{"WM_AFXLAST", 0x037F},
	{"WM_PENWINFIRST", 0x0380},
	{"WM_PENWINLAST", 0x038F},
	{0, 0}
};

class MsgLookup
{
	int Max;
	char **Lut;

public:
	MsgLookup()
	{
		Max = 0; 
		for (WinMsg *m=WndMsgs; m->Name; m++)
		{
			Max = max(Max, m->Value);
		}
		Lut = NEW(char*[++Max]);
		if (Lut)
		{
			memset(Lut, 0, sizeof(*Lut)*Max);
			for (WinMsg *m=WndMsgs; m->Name; m++)
			{
				Lut[m->Value] = m->Name;
			}
		}
	}

	char *operator [](int i)
	{
		if (Lut AND i >= 0 AND i<Max)
		{
			return Lut[i];
		}

		if (i == WM_CHANGE)
		{
			return "WM_CHANGE";
		}
		else if (i == WM_DESCRIBE)
		{
			return "WM_DESCRIBE";
		}

		static char s[64];
		if (i >= WM_USER)
		{
			sprintf(s, "WM_USER:%i", i);
		}
		else
		{
			sprintf(s, "%i", i);
		}
		return s;
	}

} WinMsgLookup;

void _mlog(GView *v, GMessage *m)
{
	static GStringPipe Log;
	static int Start = 0;
	if (NOT Start)
	{
		Start = LgiCurrentTime();
	}

	char s[256] = "";
	if (m)
	{
		switch (m->Msg)
		{
			case WM_SETCURSOR:
			case WM_NCMOUSEMOVE:
			case WM_NCHITTEST:
			{
				break;
			}
			case WM_WINDOWPOSCHANGED:
			{
				WINDOWPOS *p = (WINDOWPOS*)m->b;
				char f[256] = "";
				#define PFlag(flag) \
					if (p->flags & flag) strcat(f, #flag " ");
				PFlag(SWP_DRAWFRAME);
				PFlag(SWP_FRAMECHANGED);
				PFlag(SWP_HIDEWINDOW);
				PFlag(SWP_NOACTIVATE);
				PFlag(SWP_NOCOPYBITS);
				PFlag(SWP_NOMOVE);
				PFlag(SWP_NOOWNERZORDER);
				PFlag(SWP_NOREDRAW);
				PFlag(SWP_NOREPOSITION);
				PFlag(SWP_NOSENDCHANGING);
				PFlag(SWP_NOSIZE);
				PFlag(SWP_NOZORDER);
				PFlag(SWP_SHOWWINDOW);

				sprintf(s, "%-28s\t%s\t%i,%s\n", v->GObject::Name(), WinMsgLookup[m->Msg], m->a, f);
				break;
			}
			default:
			{
				sprintf(s, "%-28s\t%s\t%i,%i\n", v->GObject::Name(), WinMsgLookup[m->Msg], m->a, m->b);
				break;
			}
		}

		Log.Push(s);
	}
	
	if (Log.Sizeof() > (32 << 10) OR NOT m)
	{
		GFile f;
		if (f.Open("c:\\temp\\debug.txt", O_WRITE))
		{
			f.Seek(0, SEEK_END);

			char *n = Log.NewStr();
			if (n)
			{
				f.Write(n, strlen(n));
				DeleteArray(n);
			}
		}
	}
}
*/

//////////////////////////////////////////////////////////////////////////////
GView *GWindowFromHandle(HWND hWnd)
{
	if (hWnd)
	{
		int32 m = GetWindowLong(hWnd, 4);
		if (m == LGI_GViewMagic)
		{
			int Ptr = GetWindowLong(hWnd, GWL_USERDATA);
			return dynamic_cast<GView*>((GView*)Ptr);
		}
	}
	return 0;
}

//////////////////////////////////////////////////////////////////////////////
GView *GView::_Capturing = 0;
GView *GView::_Over = 0;

GView::GView(OsView Hnd)
{
	d = NEW(GViewPrivate);
	_View = 0;
	_Window = 0;
	_Lock = 0;
	_BorderSize = 0;
	Script = 0;
	WndFlags = 0;
	Pos.ZOff(0, 0);
}

GView::~GView()
{
	_Delete();
	DeleteObj(d);
}

void GView::_Delete()
{
	if (_View AND d->DropTarget)
	{
		RevokeDragDrop(_View);
	}

	// Sanity check..
	GView *c;
	for (c = Children.First(); c; c = Children.Next())
	{
		LgiAssert(c->d->Parent == this OR c->d->Parent == 0);
	}

	// Delete myself out of my parent's list
	if (d->Parent)
	{
		d->Parent->OnChildrenChanged(this, false);
		d->Parent->Children.Delete(this);
		d->Parent = 0;
	}

	// Delete all children
	while (c = Children.First())
	{
		// This will fire if a view is not properly attached in the first place.
		LgiAssert(c->d->Parent == this);

		// Delete the child view
		DeleteObj(c);
	}

	// Delete the OS representation of myself
	if (_View AND IsWindow(_View))
	{
		WndFlags |= GWF_DESTRUCTOR;
		DestroyWindow(_View);
	}
	
	// NULL my handles and flags
	_View = 0;
	WndFlags = 0;

	// Remove static references to myself
	if (_Over == this) _Over = 0;
	if (_Capturing == this) _Capturing = 0;
	if (LgiApp AND LgiApp->AppWnd == this)
	{
		LgiApp->AppWnd = 0;
	}

	// this should only exist in an ex-GWindow, due to the way
	// C++ deletes objects it needs to be here.
	DeleteObj(_Lock);
}

OsWindow GView::WindowHandle()
{
	return (_Window) ? _Window->Handle() : 0;
}

GWindow *GView::GetWindow()
{
	return dynamic_cast<GWindow*>(_Window);
}

bool GView::Lock(int TimeOut)
{
	if (_Window AND
		_Window->_Lock)
	{
		if (TimeOut < 0)
		{
			return _Window->_Lock->Lock();
		}
		else
		{
			return _Window->_Lock->LockWithTimeout(TimeOut);
		}
	}

	return true;
}

void GView::Unlock()
{
	if (_Window AND
		_Window->_Lock)
	{
		_Window->_Lock->Unlock();
	}
}

void GView::OnMouseClick(GMouse &m)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnMouseClick(m);
	}
}

void GView::OnMouseEnter(GMouse &m)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnMouseEnter(m);
	}
}

void GView::OnMouseExit(GMouse &m)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnMouseExit(m);
	}
}

void GView::OnMouseMove(GMouse &m)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnMouseMove(m);
	}
}

void GView::OnMouseWheel(double Lines)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnMouseWheel(Lines);
	}
}

bool GView::OnKey(GKey &k)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnKey(k);
	}

	return false;
}

void GView::OnCreate()
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnCreate();
	}
}

void GView::OnDestroy()
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnDestroy();
	}
}

void GView::OnFocus(bool f)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnFocus(f);
	}
}

void GView::OnPulse()
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnPulse();
	}
}

void GView::OnPosChange()
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnPosChange();
	}
}

bool GEvent::OnRequestClose(bool OsShuttingDown)
{
	return true;
}

bool GView::OnRequestClose(bool OsShuttingDown)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnRequestClose(OsShuttingDown);
	}

	return true;
}

int GView::OnHitTest(int x, int y)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnHitTest(x, y);
	}

	return -1;
}

void GView::OnChildrenChanged(GView *Wnd, bool Attaching)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnChildrenChanged(Wnd, Attaching);
	}
}

void GView::OnPaint(GSurface *pDC)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnPaint(pDC);
	}
}

int GView::OnNotify(GView *Ctrl, int Flags)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnNotify(Ctrl, Flags);
	}
	else if (Ctrl AND d->Parent)
	{
		// default behaviour is just to pass the 
		// notification up to the parent
		return d->Parent->OnNotify(Ctrl, Flags);
	}

	return 0;
}

int GView::OnCommand(int Cmd, int Event, OsView Wnd)
{
	if (Script AND Script->OnScriptEvent(this))
	{
		Script->OnCommand(Cmd, Event, Wnd);
	}

	return 0;
}

void GView::_Paint(GSurface *pDC)
{
	bool OwnDc = false;
	if (NOT pDC)
	{
		pDC = NEW(GScreenDC(_View));
		OwnDc = true;
	}

	if (pDC)
	{
		// Paint me
		OnPaint(pDC);

		// Paint any virtual children
		int x, y;
		pDC->GetOrigin(x, y);
		GRect Clip = pDC->ClipRgn();

		Iterator<GView> ChildList(&Children); // just in case the child access the child list
		for (GView *w = ChildList.First(); w; w = ChildList.Next())
		{
			if (NOT w->Handle() AND w->Visible())
			{
				GRect r = w->GetPos();
				GRect c(0, 0, r.X()-1, r.Y()-1);

				// c.Offset(-x, -y);
				// c.Intersection(&Clip);
				pDC->SetOrigin(x - r.x1, y - r.y1);
				pDC->ClipRgn(&c);

				w->_Paint(pDC);

				pDC->SetOrigin(x, y);
			}
		}
		
		if (OwnDc)
		{
			DeleteObj(pDC);
		}
	}
}

#define ADJ_LEFT			1
#define ADJ_RIGHT			2
#define ADJ_UP				3
#define ADJ_DOWN			4

int IsAdjacent(GRect &a, GRect &b)
{
	if (	(a.x1 == b.x2 + 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_LEFT;
	}

	if (	(a.x2 == b.x1 - 1) AND
		NOT (a.y1 > b.y2 OR a.y2 < b.y1))
	{
		return ADJ_RIGHT;
	}

	if (	(a.y1 == b.y2 + 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_UP;
	}

	if (	(a.y2 == b.y1 - 1) AND
		NOT (a.x1 > b.x2 OR a.x2 < b.x1))
	{
		return ADJ_DOWN;
	}

	return 0;
}

GRect JoinAdjacent(GRect &a, GRect &b, int Adj)
{
	GRect t;

	switch (Adj)
	{
		case ADJ_LEFT:
		case ADJ_RIGHT:
		{
			t.y1 = max(a.y1, b.y1);
			t.y2 = min(a.y2, b.y2);
			t.x1 = min(a.x1, b.x1);
			t.x2 = max(a.x2, b.x2);
			break;
		}
		case ADJ_UP:
		case ADJ_DOWN:
		{
			t.y1 = min(a.y1, b.y1);
			t.y2 = max(a.y2, b.y2);
			t.x1 = max(a.x1, b.x1);
			t.x2 = min(a.x2, b.x2);
			break;
		}
	}

	return t;
}

GRect *GView::FindLargest(GRegion &r)
{
	int Pixels = 0;
	GRect *Best = 0;
	static GRect Final;

	// do initial run through the list to find largest single region
	for (GRect *i = r.First(); i; i = r.Next())
	{
		int Pix = i->X() * i->Y();
		if (Pix > Pixels)
		{
			Pixels = Pix;
			Best = i;
		}
	}

	if (Best)
	{
		Final = *Best;
		Pixels = Final.X() * Final.Y();

		int LastPixels = Pixels;
		GRect LastRgn = Final;
		int ThisPixels = Pixels;
		GRect ThisRgn = Final;

		GRegion TempList;
		for (GRect *i = r.First(); i; i = r.Next())
		{
			TempList.Union(i);
		}
		TempList.Subtract(Best);

		do
		{
			LastPixels = ThisPixels;
			LastRgn = ThisRgn;

			// search for adjoining rectangles that maybe we can add
			for (GRect *i = TempList.First(); i; i = TempList.Next())
			{
				int Adj = IsAdjacent(ThisRgn, *i);
				if (Adj)
				{
					GRect t = JoinAdjacent(ThisRgn, *i, Adj);
					int Pix = t.X() * t.Y();
					if (Pix > ThisPixels)
					{
						ThisPixels = Pix;
						ThisRgn = t;

						TempList.Subtract(i);
					}
				}
			}
		} while (LastPixels < ThisPixels);

		Final = ThisRgn;
	}

	return &Final;
}

GRect *GView::FindSmallestFit(GRegion &r, int Sx, int Sy)
{
	int X = 1000000;
	int Y = 1000000;
	GRect *Best = 0;

	for (GRect *i = r.First(); i; i = r.Next())
	{
		if ((i->X() >= Sx AND i->Y() >= Sy) AND
			(i->X() < X OR i->Y() < Y))
		{
			X = i->X();
			Y = i->Y();
			Best = i;
		}
	}

	return Best;
}

GRect *GView::FindLargestEdge(GRegion &r, int Edge)
{
	GRect *Best = 0;

	for (GRect *i = r.First(); i; i = r.Next())
	{
		if (NOT Best)
		{
			Best = i;
		}

		if ((Edge & GV_EDGE_TOP) AND (i->y1 < Best->y1) OR
			(Edge & GV_EDGE_RIGHT) AND (i->x2 > Best->x2) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 > Best->y2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 < Best->x1))
		{
			Best = i;
		}

		if (((Edge & GV_EDGE_TOP) AND (i->y1 == Best->y1) OR
			(Edge & GV_EDGE_BOTTOM) AND (i->y2 == Best->y2)) AND
			(i->X() > Best->X()))
		{
			Best = i;
		}

		if (((Edge & GV_EDGE_RIGHT) AND (i->x2 == Best->x2) OR
			(Edge & GV_EDGE_LEFT) AND (i->x1 == Best->x1)) AND
			(i->Y() > Best->Y()))
		{
			Best = i;
		}
	}

	return Best;
}

void GView::Quit(bool DontDelete)
{
	if (_View)
	{
		if (NOT DontDelete)
		{
			WndFlags |= GWF_QUIT_WND;
		}

		DestroyWindow(_View);
	}
}

uint32 GView::GetDlgCode()
{
	return d->WndDlgCode;
}

void GView::SetDlgCode(uint32 i)
{
	d->WndDlgCode = i;
}

uint32 GView::GetStyle()
{
	return d->WndStyle;
}

void GView::SetStyle(uint32 i)
{
	d->WndStyle = i;
}

uint32 GView::GetExStyle()
{
	return d->WndExStyle;
}

void GView::SetExStyle(uint32 i)
{
	d->WndExStyle = i;
}

char *GView::GetClass()
{
	return d->WndClass;
}

void GView::SetClass(char *c)
{
	d->WndClass = c;
}

GWin32Class *GView::CreateClass(char *Class, HICON Icon, int AddStyles)
{
	if (Class) SetClass(Class);
	if (GetClass())
	{
		GWin32Class *c = GWin32Class::Create(GetClass());
		if (c)
		{
			if (Icon)
			{
				c->Class.a.hIcon = Icon;
			}

			if (AddStyles)
			{
				c->Class.a.style |= AddStyles;
			}

			c->Register();
			return c;
		}
	}

	return 0;
}

GView *GView::FindReal(GdcPt2 *Offset)
{
	if (Offset)
	{
		Offset->x = 0;
		Offset->y = 0;
	}

	GView *p = d->Parent;
	while (p AND NOT p->Handle())
	{
		if (Offset)
		{
			Offset->x += Pos.x1;
			Offset->y += Pos.y1;
		}

		p = p->d->Parent;
	}
	
	if (p AND p->Handle())
	{
		return p;
	}

	return NULL;
}

bool GView::Capture(bool c)
{
	return HandleCapture(this, c);
}

bool GView::IsCapturing()
{
	return _Capturing == this;
}

bool GView::HandleCapture(GView *Wnd, bool c)
{
	bool Status = false;

	if (_View)
	{
		if (c)
		{
			if (_Capturing)
			{
				_Capturing->HandleCapture(NULL, false);
			}

			d->hPrevCapture = SetCapture(_View);
			Status = true;
			_Capturing = Wnd;
		}
		else
		{
			if (_Capturing)
			{
				ReleaseCapture();
				_Capturing = 0;
			}
		}
	}
	else
	{
		if (d->Parent)
		{
			Status = d->Parent->HandleCapture(Wnd, c);
		}
	}

	return Status;
}

bool GView::Enabled()
{
	if (_View)
	{
		return IsWindowEnabled(_View);
	}
	return NOT TestFlag(d->WndStyle, WS_DISABLED);
}

void GView::Enabled(bool i)
{
	if (NOT i) SetFlag(d->WndStyle, WS_DISABLED);
	else ClearFlag(d->WndStyle, WS_DISABLED);

	if (_View)
	{
		EnableWindow(_View, i);
	}

	Invalidate();
}

bool GView::Visible()
{
	if (_View)
	{
		return IsWindowVisible(_View);
	}

	return TestFlag(d->WndStyle, WS_VISIBLE);
}

void GView::Visible(bool i)
{
	GFlags::Visible(i);
	if (i) SetFlag(d->WndStyle, WS_VISIBLE);
	else ClearFlag(d->WndStyle, WS_VISIBLE);

	if (_View)
	{
		ShowWindow(_View, (i) ? SW_SHOWNORMAL : SW_HIDE);
	}

	Invalidate();
}

bool GView::Focus()
{
	if (_View)
	{
		return GetFocus() == _View;
	}
	return TestFlag(WndFlags, GWF_FOCUS);
}

void GView::Focus(bool i)
{
	if (i) SetFlag(WndFlags, GWF_FOCUS);
	else ClearFlag(WndFlags, GWF_FOCUS);

	if (_View)
	{
		SetFocus((i) ? _View : GetDesktopWindow());
	}
	else
	{
		Invalidate();
	}
}

GDragDropTarget	*&GView::DropTargetPtr()
{
	return d->DropTarget;
}

bool GView::DropTarget()
{
	return d->DropTarget != 0;
}

bool GView::DropTarget(bool t)
{
	bool Status = false;

	if (_View)
	{
		if (t)
		{
			if (NOT d->DropTarget)
			{
				DragAcceptFiles(_View, t);
			}
			else
			{
				Status = RegisterDragDrop(_View, (IDropTarget*) d->DropTarget) == S_OK;
			}
		}
		else
		{
			if (_View AND d->DropTarget)
			{
				Status = RevokeDragDrop(_View) == S_OK;
			}
		}
	}
	else
	{
		LgiAssert(0);
	}

	return Status;
}

bool GView::Sunken()
{
	return TestFlag(d->WndExStyle, WS_EX_CLIENTEDGE);
}

void GView::Sunken(bool i)
{
	if (i) SetFlag(d->WndExStyle, WS_EX_CLIENTEDGE);
	else ClearFlag(d->WndExStyle, WS_EX_CLIENTEDGE);

	if (NOT _BorderSize AND i)
	{
		_BorderSize = 2;
	}
}

bool GView::Flat()
{
	return	NOT TestFlag(d->WndExStyle, WS_EX_CLIENTEDGE) AND
			NOT TestFlag(d->WndExStyle, WS_EX_WINDOWEDGE);
}

void GView::Flat(bool i)
{
	ClearFlag(d->WndExStyle, WS_EX_CLIENTEDGE);
	ClearFlag(d->WndExStyle, WS_EX_WINDOWEDGE);
}

bool GView::Raised()
{
	return TestFlag(d->WndExStyle, WS_EX_WINDOWEDGE);
}

void GView::Raised(bool i)
{
	if (i) SetFlag(d->WndExStyle, WS_EX_WINDOWEDGE);
	else ClearFlag(d->WndExStyle, WS_EX_WINDOWEDGE);

	if (NOT _BorderSize AND i)
	{
		_BorderSize = 2;
	}
}

int GView::GetId()
{
	return d->CtrlId;
}

void GView::SetId(int i)
{
	d->CtrlId = i;
	if (_View)
	{
		SetWindowLong(_View, GWL_ID, d->CtrlId);
	}
}

bool GView::GetTabStop()
{
	return TestFlag(d->WndStyle, WS_TABSTOP);
}

void GView::SetTabStop(bool b)
{
	if (b)
		SetFlag(d->WndStyle, WS_TABSTOP);
	else
		ClearFlag(d->WndStyle, WS_TABSTOP);
}

int GView::GetCtrlValue(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Value() : 0;
}

void GView::SetCtrlValue(int Id, int i)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Value(i);
	}
}

char *GView::GetCtrlName(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Name() : 0;
}

void GView::SetCtrlName(int Id, char *s)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Name(s);
	}
}

bool GView::GetCtrlEnabled(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Enabled() : 0;
}

void GView::SetCtrlEnabled(int Id, bool Enabled)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Enabled(Enabled);
	}
}

bool GView::GetCtrlVisible(int Id)
{
	GView *w = FindControl(Id);
	return (w) ? w->Visible() : 0;
}

void GView::SetCtrlVisible(int Id, bool v)
{
	GView *w = FindControl(Id);
	if (w)
	{
		w->Visible(v);
	}
}

bool GView::IsAttached()
{
	return _View AND IsWindow(_View);
}

GView *GView::GetParent()
{
	return d->Parent;
}

void GView::SetParent(GView *p)
{
	d->Parent = p;
}

void GView::SendNotify(int Data)
{
	GView *n = d->Notify ? d->Notify : d->Parent;
	if (n)
	{
		n->OnNotify(this, Data);
	}
}

GView *GView::GetNotify()
{
	return d->Notify;
}

void GView::SetNotify(GView *p)
{
	d->Notify = p;
}

bool GView::Attach(GView *p)
{
	bool Status = false;

	d->Parent = p;
	_Window = p ? p->_Window : this;

	if (GetClass())
	{
		// Real window with HWND
		uint Style = GetStyle();
		if (IsWin9x)
		{
			char *Text = LgiToNativeCp(GObject::Name());

			_View = CreateWindowEx(	GetExStyle() & ~WS_EX_CONTROLPARENT,
									GetClass(),
									Text,
									Style,
									Pos.x1, Pos.y1,
									Pos.X(), Pos.Y(),
									d->Parent ? d->Parent->Handle() : 0,
									NULL,
									LgiProcessInst(),
									this);

			DeleteArray(Text);
		}
		else
		{
			char16 *Text = GObject::NameW();
			char16 *Cls = LgiNewUtf8To16(GetClass());

			_View = CreateWindowExW(GetExStyle() & ~WS_EX_CONTROLPARENT,
									Cls,
									Text,
									Style,
									Pos.x1, Pos.y1,
									Pos.X(), Pos.Y(),
									d->Parent ? d->Parent->Handle() : 0,
									NULL,
									LgiProcessInst(),
									this);
			#ifdef _DEBUG
			if (NOT _View)
			{
				DWORD e = GetLastError();
				int asd=0;
			}
			#endif

			DeleteArray(Cls);
		}

		if (_View)
		{
			long lng = GetWindowLong(_View, 0);
			SetWindowLong(_View, 4, LGI_GViewMagic);
			Status = (_View != 0);

			if (d->Font)
			{
				SendMessage(_View, WM_SETFONT, (WPARAM) d->Font->Handle(), 0);
			}

			if (d->DropTarget)
			{
				RegisterDragDrop(_View, d->DropTarget);
			}
		}

		if (GFlags::Focus())
		{
			Focus(true);
		}
		
		OnAttach();
	}
	else
	{
		// Virtual window (no HWND)
		Status = true;
	}

	if (Status AND d->Parent)
	{
		if (NOT d->Parent->Children.HasItem(this))
		{
			d->Parent->Children.Insert(this);
		}
		d->Parent->OnChildrenChanged(this, true);
	}

	return Status;
}

bool GView::AttachChildren()
{
	Iterator<GView> ChildLst(&Children);
	for (GView *c = ChildLst.First(); c; c = ChildLst.Next())
	{
		if (NOT c->IsAttached())
		{
			if (NOT c->Attach(this))
			{
				return false;
			}
		}
	}

	return true;
}

bool GView::Detach()
{
	bool Status = false;
	if (d->Parent)
	{
		Visible(false);
		d->Parent->Children.Delete(this);
		d->Parent->OnChildrenChanged(this, false);
		d->Parent = 0;
		Status = true;

		if (_Capturing == this)
		{
			if (_View)
			{
				ReleaseCapture();
			}
			_Capturing = 0;
		}
		if (_View)
		{
			WndFlags &= ~GWF_QUIT_WND;
			DestroyWindow(_View);
		}
	}
	return Status;
}

GRect &GView::GetClient()
{
	RECT rc;

	if (_View)
	{
		GetClientRect(_View, &rc);
	}
	else
	{
		rc = Pos;
	}

	static GRect Client;
	Client = rc;
	Client.Offset(-Client.x1, -Client.y1);
	return Client;
}

GFont *GView::GetFont()
{
	return d->Font ? d->Font : SysFont;
}

bool GView::SetCursor(int Cursor)
{
	char *Set = 0;
	switch (Cursor)
	{
		case LCUR_UpArrow:
			Set = IDC_UPARROW;
			break;
		case LCUR_Cross:
			Set = IDC_CROSS;
			break;
		case LCUR_Wait:
			Set = IDC_WAIT;
			break;
		case LCUR_Ibeam:
			Set = IDC_IBEAM;
			break;
		case LCUR_SizeVer:
			Set = IDC_SIZENS;
			break;
		case LCUR_SizeHor:
			Set = IDC_SIZEWE;
			break;
		case LCUR_SizeBDiag:
			Set = IDC_SIZENESW;
			break;
		case LCUR_SizeFDiag:
			Set = IDC_SIZENWSE;
			break;
		case LCUR_SizeAll:
			Set = IDC_SIZEALL;
			break;
		case LCUR_PointingHand:
		{
			int Ver = 0;
			if (LgiGetOs(&Ver) == LGI_OS_WINNT AND
				Ver >= 5)
			{
				#ifndef IDC_HAND
				#define IDC_HAND MAKEINTRESOURCE(32649)
				#endif
				Set = IDC_HAND;
			}
			// else not supported
			break;
		}
		case LCUR_Forbidden:
			Set = IDC_NO;
			break;

		// Not impl
		case LCUR_SplitV:
			break;
		case LCUR_SplitH:
			break;
		case LCUR_Blank:
			break;
	}

	::SetCursor(LoadCursor(0, MAKEINTRESOURCE(Set?Set:IDC_ARROW)));

	return true;
}

void GView::SetFont(GFont *Font, bool OwnIt)
{
	bool Change = d->Font != Font;
	if (Change)
	{
		if (NOT Font)
		{
			Font = SysFont;
			OwnIt = false;
		}
		if (d->FontOwn)
		{
			DeleteObj(d->Font);
		}
		d->FontOwn = OwnIt;
		d->Font = Font;
		if (_View)
		{
			SendMessage(_View, WM_SETFONT, (WPARAM) (Font ? Font->Handle() : 0), 0);
		}
		else
		{
			Invalidate();
		}

		for (GView *p = GetParent(); p; p = p->GetParent())
		{
			GTableLayout *Tl = dynamic_cast<GTableLayout*>(p);
			if (Tl)
			{
				Tl->InvalidateLayout();
				break;
			}
		}
	}
}

bool GView::IsOver(GMouse &m)
{
	return	(m.x >= 0) AND
			(m.y >= 0) AND
			(m.x < Pos.X()) AND
			(m.y < Pos.Y());
}

void GView::PointToScreen(GdcPt2 &p)
{
	POINT pt = {p.x, p.y};
	GView *t = this;
	while (	t AND
			t->GetParent() AND
			NOT t->Handle())
	{
		pt.x += t->Pos.x1;
		pt.y += t->Pos.y1;
		t = t->GetParent();
	}
	ClientToScreen(t->Handle(), &pt);
	p.x = pt.x;
	p.y = pt.y;
}

void GView::PointToView(GdcPt2 &p)
{
	POINT pt = {p.x, p.y};
	GView *t = this;
	while (	t AND
			t->GetParent() AND
			NOT t->Handle())
	{
		pt.x -= t->Pos.x1;
		pt.y -= t->Pos.y1;
		t = t->GetParent();
	}
	ScreenToClient(t->Handle(), &pt);
	p.x = pt.x;
	p.y = pt.y;
}

bool GView::GetMouse(GMouse &m, bool ScreenCoords)
{
	// position
	POINT p;
	GetCursorPos(&p);
	if (NOT ScreenCoords)
	{
		ScreenToClient(_View, &p);
	}
	m.x = p.x;
	m.y = p.y;

	// buttons
	m.Flags =	((GetAsyncKeyState(VK_LBUTTON)&0x8000) ? LGI_EF_LEFT : 0) |
				((GetAsyncKeyState(VK_MBUTTON)&0x8000) ? LGI_EF_MIDDLE : 0) |
				((GetAsyncKeyState(VK_RBUTTON)&0x8000) ? LGI_EF_RIGHT : 0) |
				((GetAsyncKeyState(VK_CONTROL)&0x8000) ? LGI_EF_CTRL : 0) |
				((GetAsyncKeyState(VK_MENU)&0x8000) ? LGI_EF_ALT : 0) |
				((GetAsyncKeyState(VK_SHIFT)&0x8000) ? LGI_EF_SHIFT : 0);

	if (m.Flags & (LGI_EF_LEFT | LGI_EF_MIDDLE | LGI_EF_RIGHT))
	{
		m.Flags |= LGI_EF_DOWN;
	}

	return true;
}

bool GView::WindowVirtualOffset(GdcPt2 *Offset)
{
	bool Status = false;

	if (Offset)
	{
		Offset->x = 0;
		Offset->y = 0;
		GView *Wnd = this;
		while (Wnd)
		{
			GRect r = Wnd->GetPos();
			GView *p = Wnd->GetParent();
			if (NOT Wnd->Handle())
			{
				Offset->x += r.x1;
				Offset->y += r.y1;
				Wnd = p;
				Status = true;
			}
			else
			{
				Wnd = 0;
			}
		}
	}

	return Status;
}

GView *GView::WindowFromPoint(int x, int y)
{
	for (GView *c = Children.First(); c; c = Children.Next())
	{
		GRect r = c->GetPos();
		if (r.Overlap(x, y) AND
			c->Visible())
		{
			GView *Child = c->WindowFromPoint(x - r.x1, y - r.y1);
			if (Child)
			{
				return Child;
			}
		}
	}

	if (x >= 0 AND y >= 0 AND x < Pos.X() AND y < Pos.Y())
	{
		return this;
	}

	return NULL;
}

bool GView::SetPos(GRect &p, bool Repaint)
{
	bool Status = true;

	if (Pos != p)
	{
		Pos = p;
		if (_View)
		{
			Status = SetWindowPos(	_View,
									NULL,
									Pos.x1,
									Pos.y1,
									Pos.X(),
									Pos.Y(),
									// ((Repaint) ? 0 : SWP_NOREDRAW) |
									SWP_NOZORDER);
			if (Repaint)
			{
				Invalidate();
			}
		}
		else if (Repaint)
		{
			Invalidate();
		}
	}

	return Status;
}

bool GView::InThread()
{
	return GetCurrentThreadId() == GetWindowThreadProcessId(_View, NULL);
}

bool GView::PostEvent(int Cmd, int a, int b)
{
	if (_View)
	{
		return PostMessage(_View, Cmd, a, b);
	}

	GMessage e(Cmd, a, b);
	OnEvent(&e);
	return true;
}

bool GView::Invalidate(GRegion *r, bool Repaint, bool NonClient)
{
	if (r)
	{
		for (int i=0; i<r->Length(); i++)
		{
			bool Last = i == r->Length()-1;
			Invalidate((*r)[i], Last ? Repaint : false, NonClient);
		}
		
		return true;
	}
	
	return false;
}

bool GView::Invalidate(GRect *r, bool Repaint, bool Frame)
{
	if (_View)
	{
		bool Status = false;

		if (Frame)
		{
			RedrawWindow(	_View,
							NULL,
							NULL,
							RDW_FRAME |
							RDW_INVALIDATE |
							RDW_ALLCHILDREN |
							((Repaint) ? RDW_UPDATENOW : 0));
		}
		else
		{
			if (r)
			{
				Status = InvalidateRect(_View, &((RECT)*r), false);
			}
			else
			{
				RECT c = GetClient();
				Status = InvalidateRect(_View, &c, false);
			}
		}

		if (Repaint)
		{
			UpdateWindow(_View);
		}

		return Status;
	}
	else
	{
		GRect Up;
		GView *p = this;

		if (r)
		{
			Up = *r;
		}
		else
		{
			Up.Set(0, 0, Pos.X()-1, Pos.Y()-1);
		}

		while (p AND NOT p->Handle())
		{
			Up.Offset(p->Pos.x1, p->Pos.y1);
			p = p->d->Parent;
		}

		if (p AND p->Handle())
		{
			return p->Invalidate(&Up, Repaint);
		}
	}

	return false;
}

void
CALLBACK
GView::TimerProc(HWND hwnd, UINT uMsg, UINT idEvent, uint32 dwTime)
{
	GView *View = (GView*) idEvent;
	if (View)
	{
		View->OnPulse();
	}
}

void GView::SetPulse(int Length)
{
	if (_View)
	{
		if (Length > 0)
		{
			d->TimerId = SetTimer(_View, (UINT) this, Length, (TIMERPROC) TimerProc);
		}
		else
		{
			KillTimer(_View, d->TimerId);
			d->TimerId = 0;
		}
	}
}

GButton *FindDefault(GView *w)
{
	for (GView *c = w->Children.First(); c; c = w->Children.Next())
	{
		GButton *But = dynamic_cast<GButton*>(c);
		if (But AND But->Default())
		{
			return But;
		}
		
		But = FindDefault(c);
		if (But)
		{
			return But;
		}
	}

	return NULL;
}

void GetChildrenList(GView *w, List<GView> &l)
{
	if (w)
	{
		for (GView *v = w->Children.First(); v; v = w->Children.Next())
		{
			int Style = GetWindowLong(v->Handle(), GWL_STYLE);
			if (TestFlag(Style, WS_VISIBLE) AND
				NOT TestFlag(Style, WS_DISABLED))
			{
				if (TestFlag(Style, WS_TABSTOP))
				{
					l.Insert(v);
				}

				GetChildrenList(v, l);
			}
		}
	}
}

GView *GetNextTabStop(GView *v, bool Back)
{
	if (v)
	{
		GWindow *Wnd = v->GetWindow();
		if (Wnd)
		{
			List<GView> All;
			GetChildrenList(Wnd, All);

			int MyIndex = All.IndexOf(v);
			if (MyIndex >= 0)
			{
				int Inc = Back ? -1 : 1;
				int NewIndex = (MyIndex + All.GetItems() + Inc) % All.GetItems();
				return All.ItemAt(NewIndex);
			}
		}
	}

	return 0;
}

bool SysOnKey(GView *w, GMessage *m)
{
	if (m->a == VK_TAB AND
		(m->Msg == WM_KEYDOWN OR
		m->Msg == WM_SYSKEYDOWN) )
	{
		if (NOT TestFlag(w->d->WndDlgCode, DLGC_WANTTAB) AND
			NOT TestFlag(w->d->WndDlgCode, DLGC_WANTALLKEYS))
		{
			// push the focus to the next control
			bool Shifted = (GetAsyncKeyState(VK_SHIFT) & 0x8000) != 0;
			GView *Wnd = GetNextTabStop(w, Shifted);
			if (Wnd)
			{
				::SetFocus(Wnd->Handle());
				return true;
			}
		}
	}

	return false;
}


#define KEY_FLAGS		(~(MK_LBUTTON | MK_MBUTTON | MK_RBUTTON))

int GView::OnEvent(GMessage *Msg)
{
	int Status = 0;

	if (MsgCode(Msg) == MouseRollMsg)
	{
		HWND hFocus = GetFocus();
		if (_View)
		{
			int Flags = ((GetKeyState(VK_SHIFT)&0xF000) ? VK_SHIFT : 0) | 
						((GetKeyState(VK_CONTROL)&0xF000) ? VK_CONTROL : 0);
			
			PostMessage(hFocus, WM_MOUSEWHEEL, MAKELONG(Flags, (short)Msg->a), Msg->b);
		}
		return 0;
	}

	if (_View)
	{
		switch (Msg->Msg)
		{
			case 5700:
			{
				break;
			}
			case WM_ERASEBKGND:
			{
				return 1;
			}
			case WM_GETFONT:
			{
				GFont *f = GetFont();
				return (int) (f ? f->Handle() : SysFont->Handle());
				break;
			}
			case WM_MENUCHAR:
			case WM_MEASUREITEM:
			case WM_DRAWITEM:
			{
				return GMenu::_OnEvent(Msg);
				break;
			}
			case WM_ENABLE:
			{
				Invalidate(&Pos);
				break;
			}
			case WM_HSCROLL:
			case WM_VSCROLL:
			{
				GView *Wnd = FindControl((HWND) Msg->b);
				if (Wnd)
				{
					Wnd->OnEvent(Msg);
				}
				break;
			}
			case WM_GETDLGCODE:
			{
				// we handle all tab control stuff
				return d->WndDlgCode | DLGC_WANTTAB;
			}
			case WM_MOUSEWHEEL:
			{
				short fwKeys = LOWORD(Msg->a);			// key flags
				short zDelta = (short) HIWORD(Msg->a);	// wheel rotation
				short xPos = (short) LOWORD(Msg->b);	// horizontal position of pointer
				short yPos = (short) HIWORD(Msg->b);	// vertical position of pointer
				int nScrollLines = - _lgi_mouse_wheel_lines();
				OnMouseWheel( ((int) zDelta / WHEEL_DELTA) * (int) nScrollLines );
				return 0;
			}
			case M_CHANGE:
			{
				GView *Ctrl = dynamic_cast<GView*>((GView*) Msg->a);
				if (Ctrl)
				{
					return OnNotify(Ctrl, Msg->b);
				}
				break;
			}
			case WM_NOTIFY:
			{
				LPNMHDR Hdr = (LPNMHDR) Msg->b;
				GView *Ctrl = FindControl(Hdr->hwndFrom);

				int i1 = LVN_GETDISPINFO;
				if (Ctrl) 
				{
					switch (Hdr->code)
					{
						#ifdef __GLIST_H
						// List View Stuff
						//
						// this is only required when using the Win32 ListView control
						// in GList.h/GList.cpp
						// for xp compatibility use the owner draw list control in
						// GList2.h/GList2.cpp
						case LVN_GETDISPINFOW:
						case LVN_GETDISPINFOA:
						{
							bool Unicode = Hdr->code == LVN_GETDISPINFOW;
							LV_DISPINFO *Info = (LV_DISPINFO*) Msg->b;
							GListItem *Item = (GListItem*) Info->item.lParam;

							if (Item)
							{
								if (Info->item.mask & LVIF_TEXT)
								{
									if (Info->item.pszText)
									{
										char *Text = Item->GetText(Info->item.iSubItem);
										if (Text)
										{
											if (Unicode)
											{
												short *d = (short*) Info->item.pszText;
												char *s = Text;
												int Size = Info->item.cchTextMax - 2;
												while (*s AND Size > 1)
												{
													*d++ = *s++;
													Size -= 2;
												}
												*d++ = 0;
											}
											else
											{
												strcpy(Info->item.pszText, Text);
											}
										}
										else
										{
											Info->item.pszText[0] = 0;
										}
									}
									else
									{
										if (Unicode)
										{
											static short Error[] = { 'E', 'r', 'r', 'o', 'r', 0 };
											Info->item.pszText = (char*) Error;
										}
										else
										{
											// what the?
											// don't know what causes windows to call using a NULL
											// pointer. Any ideas ppl?
											Info->item.pszText = Item->GetText(Info->item.iSubItem);
										}
									}
								}

								if (Info->item.mask & LVIF_IMAGE)
								{
									Info->item.iImage = Item->GetImage();
								}
							}
							break;
						}
						case LVN_ENDLABELEDIT:
						{
							LV_DISPINFO *Info = (LV_DISPINFO*) Msg->b;
							GListItem *Item = (GListItem*) Info->item.lParam;
							if (Info AND Item)
							{
								if (Info->item.mask & LVIF_TEXT)
								{
									return Item->SetText(Info->item.pszText);
								}
							}
							break;
						}
						case LVN_BEGINDRAG:
						case LVN_BEGINRDRAG:
						{
							NM_LISTVIEW *Info = (NM_LISTVIEW*) Msg->b;
							LV_ITEM Item;

							ZeroObj(Item);
							Item.mask = LVIF_PARAM;
							Item.iItem = Info->iItem;
							if (ListView_GetItem(Ctrl->Handle(), &Item))
							{
								GMouse m;
								m.Target = this;
								m.Flags = (Hdr->code == LVN_BEGINDRAG) ? MK_LBUTTON : MK_RBUTTON;
								((GList*) Ctrl)->OnItemBeginDrag((GListItem*) Item.lParam, m);
							}
							break;
						}
						case LVN_COLUMNCLICK:
						{
							NM_LISTVIEW *Info = (NM_LISTVIEW*) Msg->b;
							((GList*) Ctrl)->OnColumnClick(Info->iSubItem);
							break;
						}
						case LVN_ITEMCHANGED:
						{
							LPNMLISTVIEW Info = (LPNMLISTVIEW) Msg->b;
							if ((Info->uNewState | Info->uOldState) & LVIS_SELECTED)
							{
								GList *Lst = (GList*) Ctrl;
								if (Info->uNewState & LVIS_SELECTED)
								{
									Lst->OnItemSelect((GListItem*) Info->lParam);
								}
								else
								{
									Lst->OnItemSelect(0);
								}
							}
							break;
						}
						#endif

						#if 0
						// Tab Control Stuff
						case TCN_SELCHANGING:
						{
							Status = Ctrl->SysOnNotify(Hdr->code);
							break;
						}
						case TCN_SELCHANGE:
						{
							Ctrl->SysOnNotify(Hdr->code);
							OnNotify(Ctrl, 0);
							break;
						}
						#endif

						#ifdef __GTREE_H
						// Tree View Control
						//
						// this is only relevent when using the Win32 Tree control
						// which resides in GTree.h and GTree.cpp
						// for xp compatibility use the owner draw tree control
						// in GTree2.h and GTree2.cpp
	 					case TVN_BEGINRDRAG:
						case TVN_BEGINDRAG:
						{
							NM_TREEVIEW *Item = (NM_TREEVIEW*) Msg->b;
							((GTree*) Ctrl)->OnItemBeginDrag((GTreeItem*) Item->itemNew.lParam, (Hdr->code == TVN_BEGINDRAG) ? MK_LBUTTON : MK_RBUTTON);
							break;
						}
						case TVN_SELCHANGED:
						{
							NM_TREEVIEW *Item = (NM_TREEVIEW*) Msg->b;
							((GTree*) Ctrl)->OnItemSelect((GTreeItem*) Item->itemNew.lParam);
							break;
						}
						case TVN_ITEMEXPANDED:
						{
							NM_TREEVIEW *Item = (NM_TREEVIEW*) Msg->b;
							((GTree*) Ctrl)->OnItemExpand((GTreeItem*) Item->itemNew.lParam,
															(Item->itemNew.state & TVIS_EXPANDED) != 0);
							break;
						}
						case TVN_GETDISPINFO:
						{
							TV_DISPINFO *Info = (TV_DISPINFO*) Msg->b;
							GTreeItem *Item = (GTreeItem*) Info->item.lParam;
							if (Item)
							{
								if (Info->item.mask & TVIF_TEXT)
								{
									char *Text = Item->GetText();
									if (NOT Text) Text = "";
									if (Info->item.pszText)
									{
										strcpy(Info->item.pszText, Text);
									}
									else
									{
										Info->item.pszText = Text;
									}								
								}

								if (Info->item.mask & TVIF_IMAGE)
								{
									Info->item.iImage = Item->GetImage();
								}
								
								if (Info->item.mask & TVIF_SELECTEDIMAGE)
								{
									Info->item.iSelectedImage = Item->GetImage(1);
								}							
							}
							break;
						}
						case TVN_DELETEITEM:
						{
							NM_TREEVIEW *pnmtv = (NM_TREEVIEW*) Msg->b;
							if (pnmtv)
							{
								GTreeItem *Item = (GTreeItem*) pnmtv->itemOld.lParam;
								if (Item)
								{
									Item->hTreeItem = 0;
								}
							}
							break;
						}
						#endif
					}
				}
				break;
			}
			case M_COMMAND:
			{
				GView *Ctrl = FindControl((HWND) Msg->b);
				if (Ctrl)
				{
					short Code = HIWORD(Msg->a);
					switch (Code)
					{
						case BN_CLICKED: // BUTTON
						{
							OnNotify(Ctrl, 0);
							break;
						}
						case EN_CHANGE: // EDIT
						{
							Ctrl->SysOnNotify(Code);
							break;
						}
						case CBN_CLOSEUP:
						{
							PostMessage(_View, WM_COMMAND, MAKELONG(Ctrl->GetId(), CBN_EDITCHANGE), Msg->b);
							break;
						}
						case CBN_EDITCHANGE: // COMBO
						{
							Ctrl->SysOnNotify(Code);
							OnNotify(Ctrl, 0);
							break;
						}
					}
				}
				break;
			}
			case WM_NCDESTROY:
			{
				SetWindowLong(_View, GWL_USERDATA, 0);
				_View = 0;
				if (WndFlags & GWF_QUIT_WND)
				{
					delete this;
				}
				break;
			}
	 		case WM_CLOSE:
			{
				if (OnRequestClose(false))
				{
					Quit();
				}
				break;
			}
			case WM_DESTROY:
			{
				OnDestroy();
				break;
			}
			case WM_CREATE:
			{
				SetId(d->CtrlId);
				OnCreate();
				break;
			}
			case WM_SETFOCUS:
			{
				OnFocus(true);
				break;
			}
			case WM_KILLFOCUS:
			{
				OnFocus(false);
				break;
			}
			case WM_WINDOWPOSCHANGED:
			{
				if (NOT IsIconic(_View))
				{
					WINDOWPOS *Info = (LPWINDOWPOS) Msg->b;
					if (Info)
					{
						if (Info->x == -32000 AND
							Info->y == -32000)
						{
							/*
							LgiTrace("WM_WINDOWPOSCHANGED %i,%i,%i,%i (icon=%i)\n",
								Info->x,
								Info->y,
								Info->cx,
								Info->cy,
								IsIconic(Handle()));
							*/
						}
						else
						{
							GRect r;
							r.ZOff(Info->cx-1, Info->cy-1);
							r.Offset(Info->x, Info->y);
							if (r.Valid())
							{
								Pos = r;
							}
						}
					}
					OnPosChange();
				}

				if (NOT (WndFlags & GWF_DIALOG))
				{
					if (IsWin9x)
					{
						return DefWindowProcA(_View, Msg->Msg, Msg->a, Msg->b);
					}
					else
					{
						return DefWindowProcW(_View, Msg->Msg, Msg->a, Msg->b);
					}
				}
				break;
			}
			case WM_CAPTURECHANGED:
			{
				_Capturing = 0;
				break;
			}
			case M_MOUSEENTER:
			{
				if (Msg->a AND NOT TestFlag(WndFlags, GWF_OVER))
				{
					GMouse Ms;
					Ms.Target = this;
					Ms.x = (short) (Msg->b&0xFFFF);
					Ms.y = (short) (Msg->b>>16);
					Ms.Flags = 0;

					GView *MouseOver = WindowFromPoint(Ms.x, Ms.y);
					if (MouseOver AND MouseOver->Handle())
					{
						MouseOver = 0;
					}

					if (_Capturing)
					{
						if (MouseOver == _Capturing)
						{
							_Capturing->OnMouseEnter(_lgi_adjust_click_for_window(Ms, _Capturing));
						}
					}
					else
					{
						if (_Over != MouseOver)
						{
							if (_Over)
							{
								_Over->OnMouseExit(_lgi_adjust_click_for_window(Ms, _Over));
							}

							_Over = MouseOver;

							if (_Over)
							{
								_Over->OnMouseEnter(_lgi_adjust_click_for_window(Ms, _Over));
							}
						}
					}

					SetFlag(WndFlags, GWF_OVER);
				}
				break;
			}
			case M_MOUSEEXIT:
			{
				if (NOT Msg->a AND TestFlag(WndFlags, GWF_OVER))
				{
					GMouse Ms;
					Ms.Target = this;
					Ms.x = (short) (Msg->b&0xFFFF);
					Ms.y = (short) (Msg->b>>16);
					Ms.Flags = 0;

					if (_Capturing)
					{
						_Capturing->OnMouseExit(_lgi_adjust_click_for_window(Ms, _Capturing));
					}
					else
					{
						if (_Over)
						{
							_Over->OnMouseExit(_lgi_adjust_click_for_window(Ms, _Over));
							_Over = 0;
						}
						else
						{
							OnMouseExit(Ms);
						}
					}

					ClearFlag(WndFlags, GWF_OVER);
				}
				break;
			}
			case WM_MOUSEMOVE:
			{
				GMouse Ms;
				Ms.Target = this;
				Ms.x = (short) (Msg->b&0xFFFF);
				Ms.y = (short) (Msg->b>>16);
				Ms.Flags = _lgi_get_key_flags();
				if (TestFlag(Msg->a, MK_LBUTTON)) SetFlag(Ms.Flags, LGI_EF_LEFT);
				if (TestFlag(Msg->a, MK_RBUTTON)) SetFlag(Ms.Flags, LGI_EF_RIGHT);
				if (TestFlag(Msg->a, MK_MBUTTON)) SetFlag(Ms.Flags, LGI_EF_MIDDLE);
				
				SetKeyFlag(Ms.Flags, VK_MENU, MK_ALT);
				Ms.Down((Msg->a & (MK_LBUTTON|MK_MBUTTON|MK_RBUTTON)) != 0);

				int Hit = OnHitTest(Ms.x, Ms.y);
				if (Hit < 0)
				{
					SetCursor(LCUR_Normal);
				}

				GView *MouseOver = WindowFromPoint(Ms.x, Ms.y);
				if (_Over != MouseOver)
				{
					if (_Over)
					{
						_Over->OnMouseExit(_lgi_adjust_click_for_window(Ms, _Over));
					}

					_Over = MouseOver;

					if (_Over)
					{
						_Over->OnMouseEnter(_lgi_adjust_click_for_window(Ms, _Over));
					}
				}

				if (_Capturing)
				{
					_Capturing->OnMouseMove(_lgi_adjust_click_for_window(Ms, _Capturing));
				}
				else if (_Over)
				{
					_Over->OnMouseMove(_lgi_adjust_click_for_window(Ms, _Over));
				}

				break;
			}
			case WM_NCHITTEST:
			{
				POINT Pt = { LOWORD(Msg->b), HIWORD(Msg->b) };
				ScreenToClient(_View, &Pt);
				int Hit = OnHitTest(Pt.x, Pt.y);
				if (Hit >= 0)
				{
					return Hit;
				}
				if (NOT (WndFlags & GWF_DIALOG))
				{
					if (IsWin9x)
					{
						return DefWindowProcA(_View, Msg->Msg, Msg->a, Msg->b);
					}
					else
					{
						return DefWindowProcW(_View, Msg->Msg, Msg->a, Msg->b);
					}
				}
				break;
			}
			case WM_LBUTTONDBLCLK:
			case WM_LBUTTONDOWN:
			case WM_LBUTTONUP:
			{
				GMouse Ms;
				Ms.x = (short) (Msg->b&0xFFFF);
				Ms.y = (short) (Msg->b>>16);
				Ms.Flags = _lgi_get_key_flags() | LGI_EF_LEFT;
				Ms.Down(Msg->Msg != WM_LBUTTONUP);
				Ms.Double(Msg->Msg == WM_LBUTTONDBLCLK);

				if (_Capturing)
				{
					Ms.Target = _Capturing;
					_Capturing->OnMouseClick(_lgi_adjust_click_for_window(Ms, _Capturing));
				}
				else if (_Over)
				{
					Ms.Target = _Over;
					_Over->OnMouseClick(_lgi_adjust_click_for_window(Ms, _Over));
				}
				else
				{
					Ms.Target = this;
					OnMouseClick(Ms);
				}
				break;
			}
			case WM_RBUTTONDBLCLK:
			case WM_RBUTTONDOWN:
			case WM_RBUTTONUP:
			{
				GMouse Ms;
				Ms.x = (short) (Msg->b&0xFFFF);
				Ms.y = (short) (Msg->b>>16);
				Ms.Flags = _lgi_get_key_flags() | LGI_EF_RIGHT;
				Ms.Down(Msg->Msg != WM_RBUTTONUP);
				Ms.Double(Msg->Msg == WM_RBUTTONDBLCLK);

				if (_Capturing)
				{
					Ms.Target = _Capturing;
					_Capturing->OnMouseClick(_lgi_adjust_click_for_window(Ms, _Capturing));
				}
				else if (_Over)
				{
					Ms.Target = _Over;
					_Over->OnMouseClick(_lgi_adjust_click_for_window(Ms, _Over));
				}
				else
				{
					Ms.Target = this;
					OnMouseClick(Ms);
				}
				break;
			}
			case WM_MBUTTONDBLCLK:
			case WM_MBUTTONDOWN:
			case WM_MBUTTONUP:
			{
				GMouse Ms;
				Ms.x = (short) (Msg->b&0xFFFF);
				Ms.y = (short) (Msg->b>>16);
				Ms.Flags = _lgi_get_key_flags() | LGI_EF_MIDDLE;
 				Ms.Down(Msg->Msg != WM_MBUTTONUP);
				Ms.Double(Msg->Msg == WM_MBUTTONDBLCLK);

				if (_Capturing)
				{
					Ms.Target = _Capturing;
					_Capturing->OnMouseClick(_lgi_adjust_click_for_window(Ms, _Capturing));
				}
				else if (_Over)
				{
					Ms.Target = _Over;
					_Over->OnMouseClick(_lgi_adjust_click_for_window(Ms, _Over));
				}
				else
				{
					Ms.Target = this;
					OnMouseClick(Ms);
				}
				break;
			}
			case WM_SYSKEYUP:
			case WM_SYSKEYDOWN:
			case WM_KEYDOWN:
			case WM_KEYUP:
			{
				bool IsDialog = TestFlag(WndFlags, GWF_DIALOG);
				HWND hwnd = _View;

				if (NOT SysOnKey(this, Msg))
				{
					GKey Key(Msg->a);

					Key.Flags = _lgi_get_key_flags();
					Key.Data = Msg->b;
					Key.Down(Msg->Msg == WM_KEYDOWN OR Msg->Msg == WM_SYSKEYDOWN);
					Key.IsChar = false;

					if (_Window)
					{
						if (Key.Alt() OR
							Key.Ctrl() OR
							(Key.c16 < 'A' OR Key.c16 > 'Z'))
						{
							_Window->OnViewKey(this, Key);
						}
					}
					else
					{
						OnKey(Key);
					}
				}

				if (IsDialog)
				{
					/*
					if (IsWin9x)
					{
						return DefDlgProcA(hwnd, Msg->Msg, Msg->a, Msg->b);
					}
					else
					{
						return DefDlgProcW(hwnd, Msg->Msg, Msg->a, Msg->b);
					}
					*/
				}
				else
				{
					// required for Alt-Key function (eg Alt-F4 closes window)
					if (IsWin9x)
					{
						return DefWindowProcA(hwnd, Msg->Msg, Msg->a, Msg->b);
					}
					else
					{
						return DefWindowProcW(hwnd, Msg->Msg, Msg->a, Msg->b);
					}
				}
				break;
			}
			case WM_CHAR:
			{
				GKey Key(Msg->a);
				Key.Flags = _lgi_get_key_flags();
				Key.Data = Msg->b;
				Key.Down(true);
				Key.IsChar = true;

				bool Shift = Key.Shift();
				bool Caps = TestFlag(Key.Flags, LGI_EF_CAPS_LOCK);
				if (NOT (Shift ^ Caps))
				{
					Key.c = ToLower(Key.c);
					Key.c16 = ToLower(Key.c16);
				}
				else
				{
					Key.c = ToUpper(Key.c);
					Key.c16 = ToUpper(Key.c16);
				}

				if (_Window)
				{
					_Window->OnViewKey(this, Key);
				}
				else
				{
					OnKey(Key);
				}
				break;
			}
			case M_SET_WND_STYLE:
			{
				SetWindowLong(Handle(), GWL_STYLE, Msg->b);
				SetWindowPos(	Handle(),
								0, 0, 0, 0, 0,
								SWP_NOMOVE | SWP_NOZORDER | SWP_NOSIZE | SWP_FRAMECHANGED);
				break;
			}
			case WM_PAINT:
			{
				// GScreenDC DeviceContext(_View); &DeviceContext
				_Paint();
				break;
			}
			case WM_NCPAINT:
			{
				bool Thin = (Sunken() OR Raised()) AND (_BorderSize == 1);
				if (Thin)
				{
					HDC hDC = GetWindowDC(_View);
					GScreenDC Dc(hDC, _View, true);
					GRect p(0, 0, Dc.X()-1, Dc.Y()-1);
					if (p.Valid())
					{
						LgiThinBorder(&Dc, p, Sunken()?SUNKEN:RAISED);
					}
				}
				else
				{
					if (IsWin9x)
					{
						return DefWindowProcA(_View, Msg->Msg, Msg->a, Msg->b);
					}
					else
					{
						return DefWindowProcW(_View, Msg->Msg, Msg->a, Msg->b);
					}
				}
				break;
			}
			case WM_NCCALCSIZE:
			{
				bool Thin = (Sunken() OR Raised()) AND _BorderSize == 1;
				if (Msg->a)
				{
					if (Thin)
					{
						NCCALCSIZE_PARAMS *p = (NCCALCSIZE_PARAMS*) Msg->b;
						p->rgrc[0].left++;
						p->rgrc[0].top++;
						p->rgrc[0].right--;
						p->rgrc[0].bottom--;
						return 0;
					}
				}
				else
				{
					if (Thin)
					{
						RECT *r = (RECT*)Msg->b;
						r->left++;
						r->top++;
						r->right--;
						r->bottom--;						
						return 0;
					}
				}
				// Fall through
			}
			default:
			{
				if (NOT (WndFlags & GWF_DIALOG))
				{
					if (IsWin9x)
					{
						return DefWindowProcA(_View, Msg->Msg, Msg->a, Msg->b);
					}
					else
					{
						return DefWindowProcW(_View, Msg->Msg, Msg->a, Msg->b);
					}
				}
				break;
			}
		}
	}

	return 0;
}

bool GView::Name(char *n)
{
	bool Status = GObject::Name(n);

	if (_View)
	{
		if (IsWin9x)
		{
			char *Temp = LgiToNativeCp(n);
			SetWindowText(_View, Temp?Temp:"");
			DeleteArray(Temp);
		}
		else
		{
			char16 *Temp = GObject::NameW();
			SetWindowTextW(_View, Temp ? Temp : L"");
		}
	}

	Invalidate();

	return true;
}

char *GView::Name()
{
	if (_View)
	{
		if (IsWin9x)
		{
			int Length = GetWindowTextLength(_View);
			if (Length > 0)
			{
				char *Buf = NEW(char[Length+1]);
				if (Buf)
				{
					Buf[0] = 0;
					int Chars = GetWindowText(_View, Buf, Length+1);
					Buf[Chars] = 0;

					char *Temp = (char*)LgiNewConvertCp("utf-8", Buf, LgiAnsiToLgiCp());
					if (Temp)
					{
						GObject::Name(Temp);
						DeleteArray(Temp);
					}
				}
				DeleteArray(Buf);
			}
			else
			{
				GObject::Name(0);
			}
		}
		else
		{
			GView::NameW();
		}
	}

	return GObject::Name();
}

bool GView::NameW(char16 *n)
{
	bool Status = GObject::NameW(n);

	if (_View AND n)
	{
		char16 *Txt = GObject::NameW();
		if (IsWin9x)
		{
			char *Temp = (char*)LgiNewConvertCp(LgiAnsiToLgiCp(), n, "utf-16");
			if (Temp)
			{
				SetWindowText(_View, Temp);
				DeleteArray(Temp);
			}
		}
		else
		{
			SetWindowTextW(_View, Txt);
		}
	}

	Invalidate();
	return true;
}

char16 *GView::NameW()
{
	if (_View)
	{
		if (IsWin9x)
		{
			GView::Name();
		}
		else
		{
			int Length = GetWindowTextLengthW(_View);
			if (Length > 0)
			{
				char16 *Buf = NEW(char16[Length+1]);
				if (Buf)
				{
					Buf[0] = 0;
					int Chars = GetWindowTextW(_View, Buf, Length+1);
					Buf[Chars] = 0;
					GObject::NameW(Buf);
				}
				DeleteArray(Buf);
			}
			else
			{
				GObject::NameW(0);
			}
		}
	}

	return GObject::NameW();
}

GView *GView::FindControl(OsView hCtrl)
{
	if (_View == hCtrl)
	{
		return this;
	}

	Iterator<GView> List(&Children);
	for (GView *c = List.First(); c; c = List.Next())
	{
		GView *Ctrl = c->FindControl(hCtrl);
		if (Ctrl)
		{
			return Ctrl;
		}
	}
	return 0;
}

GView *GView::FindControl(int Id)
{
	if (GetId() == Id)
	{
		return this;
	}

	Iterator<GView> List(&Children);
	for (GView *c = List.First(); c; c = List.Next())
	{
		GView *Ctrl = c->FindControl(Id);
		if (Ctrl)
		{
			return Ctrl;
		}
	}
	return 0;
}

GdcPt2 GView::GetMinimumSize()
{
	return d->MinimumSize;
}

void GView::SetMinimumSize(GdcPt2 Size)
{
	d->MinimumSize = Size;

	bool Change = false;
	GRect p = Pos;
	if (X() < d->MinimumSize.x)
	{
		p.x2 = p.x1 + d->MinimumSize.x - 1;
		Change = true;
	}
	if (Y() < d->MinimumSize.y)
	{
		p.y2 = p.y1 + d->MinimumSize.y - 1;
		Change = true;
	}
	if (Change)
	{
		SetPos(p);
	}
}

void GView::MoveOnScreen()
{
	GRect Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);
	GRect p = GetPos();

	if (p.x2 >= Screen.X())
		p.Offset(Screen.X() - p.x2, 0);
	if (p.y2 >= Screen.Y())
		p.Offset(0, Screen.Y() - p.y2);
	if (p.x1 < 0)
		p.Offset(-p.x1, 0);
	if (p.y1 < 0)
		p.Offset(0, -p.y1);

	SetPos(p, true);
}

void GView::MoveToCenter()
{
	GRect Screen(0, 0, GdcD->X()-1, GdcD->Y()-1);
	GRect p = GetPos();

	p.Offset(-p.x1, -p.y1);
	p.Offset((Screen.X() - p.X()) / 2, (Screen.Y() - p.Y()) / 2);

	SetPos(p, true);
}

void GView::MoveToMouse()
{
	GMouse m;
	if (GetMouse(m, true))
	{
		GRect p = GetPos();

		p.Offset(-p.x1, -p.y1);
		p.Offset(m.x-(p.X()/2), m.y-(p.Y()/2));

		SetPos(p, true);
		MoveOnScreen();
	}
}

GdcPt2 &GView::GetWindowBorderSize()
{
	static GdcPt2 s;
	ZeroObj(s);

	if (Handle())
	{
		RECT Wnd, Client;
		GetWindowRect(Handle(), &Wnd);
		GetClientRect(Handle(), &Client);
		s.x = (Wnd.right-Wnd.left) - (Client.right-Client.left);
		s.y = (Wnd.bottom-Wnd.top) - (Client.bottom-Client.top);
	}

	return s;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
GFlags::GFlags()
{
	WndFlags = GWF_VISIBLE | GWF_ENABLED;
}

bool GFlags::Enabled()
{
	return TestFlag(WndFlags, GWF_ENABLED);
}

void GFlags::Enabled(bool i)
{
	if (i) SetFlag(WndFlags, GWF_ENABLED);
	else ClearFlag(WndFlags, GWF_ENABLED);
}

bool GFlags::Visible()
{
	return TestFlag(WndFlags, GWF_VISIBLE);
}

void GFlags::Visible(bool i)
{
	if (i) SetFlag(WndFlags, GWF_VISIBLE);
	else ClearFlag(WndFlags, GWF_VISIBLE);
}

bool GFlags::Focus()
{
	return TestFlag(WndFlags, GWF_FOCUS);
}

void GFlags::Focus(bool i)
{
	if (i) SetFlag(WndFlags, GWF_FOCUS);
	else ClearFlag(WndFlags, GWF_FOCUS);
}

bool GFlags::DropTarget()
{
	return TestFlag(WndFlags, GWF_DROP_TARGET);
}

bool GFlags::DropTarget(bool i)
{
	if (i) SetFlag(WndFlags, GWF_DROP_TARGET);
	else ClearFlag(WndFlags, GWF_DROP_TARGET);
	return true;
}

#define GWF_ALL (GWF_SUNKEN | GWF_FLAT | GWF_RAISED)

bool GFlags::Sunken()
{
	return TestFlag(WndFlags, GWF_SUNKEN);
}

void GFlags::Sunken(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_SUNKEN);
}

bool GFlags::Flat()
{
	return TestFlag(WndFlags, GWF_FLAT);
}

void GFlags::Flat(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_FLAT);
}

bool GFlags::Raised()
{
	return TestFlag(WndFlags, GWF_RAISED);
}

void GFlags::Raised(bool i)
{
	WndFlags &= ~GWF_ALL;
	if (i) SetFlag(WndFlags, GWF_RAISED);
}

//////////////////////////////////////////////////////////////////////////
List<GViewFactory> *GViewFactory::Factories;

GViewFactory::GViewFactory()
{
	if (NOT Factories) Factories = NEW(List<GViewFactory>);
	if (Factories) Factories->Insert(this);
}

GViewFactory::~GViewFactory()
{
	if (Factories)
	{
		Factories->Delete(this);
		if (Factories->GetItems() == 0)
		{
			DeleteObj(Factories);
		}
	}
}

GView *GViewFactory::Create(char *Class, GRect *Pos, char *Text)
{
	if (Factories AND ValidStr(Class))
	{
		for (GViewFactory *f=Factories->First(); f; f=Factories->Next())
		{
			GView *v = f->NewView(Class, Pos, Text);
			if (v)
			{
				return v;
			}
		}
	}

	return 0;
}
