/*hdr
**      FILE:           GWidgets.cpp
**      AUTHOR:         Matthew Allen
**      DATE:           8/9/1998
**      DESCRIPTION:    Dialog components
**
**      Copyright (C) 1998-2001 Matthew Allen
**              fret@memecode.com
*/

#include <stdlib.h>
#include <stdio.h>
#include "Lgi.h"
#include <COMMCTRL.H>

///////////////////////////////////////////////////////////////////////////////////////////
short *DlgStrCopy(short *A, char *N)
{
	uchar *n = (uchar*) N;
	if (n)
	{
		while (*n)
		{
			*A++ = *n++;
		}
	}
	*A++ = 0;
	return A;
}

short *DlgStrCopy(short *A, char16 *n)
{
	if (n)
	{
		while (*n)
		{
			*A++ = *n++;
		}
	}
	*A++ = 0;
	return A;
}

short *DlgPadToDWord(short *A, bool Seek = false)
{
	char *c = (char*) A;
	while (((int) c) & 3)
	{
		if (Seek)
		{
			c++;
		}
		else
		{
			*c++ = 0;
		}
	}
	return (short*) c;
}

///////////////////////////////////////////////////////////////////////////////////////////
GDialog::GDialog()
	: ResObject(Res_Dialog)
{
	Name("Dialog");

	Mem = 0;
	WndFlags |= GWF_DIALOG;
	SetExStyle(GetExStyle() | WS_EX_CONTROLPARENT);
	IsModal = true;

	_Window = this;
}

GDialog::~GDialog()
{
	DeleteObj(Mem);
}

bool GDialog::LoadFromResource(int Resource)
{
	char n[256];
	bool Status = GLgiRes::LoadFromResource(Resource, Children, &Pos, n);
	if (Status)
	{
		Name(n);
	}
	return Status;
}

LRESULT CALLBACK DlgRedir(HWND hWnd, UINT m, WPARAM a, LPARAM b)
{
	if (m == WM_INITDIALOG)
	{
		GDialog *NewWnd = (GDialog*) b;
		NewWnd->_View = hWnd;
		SetWindowLong(hWnd, GWL_USERDATA, (int) NewWnd);
	}

	GDialog *Wnd = (GDialog*) GetWindowLong(hWnd, GWL_USERDATA);
	if (Wnd)
	{
		GMessage Msg(m, a, b);
		return Wnd->OnEvent(&Msg);
	}

	return 0;
}

bool GDialog::OnRequestClose(bool OsClose)
{
	return true;
}

int GDialog::DoModal()
{
	int Status = -1;

	IsModal = true;
	Mem = NEW(GMem(16<<10));
	if (Mem)
	{
		DLGTEMPLATE *Template = (DLGTEMPLATE*) Mem->Lock();
		if (Template)
		{
			GRect r = GetPos();
			r.x1 = (double)r.x1 / DIALOG_X;
			r.y1 = (double)r.y1 / DIALOG_Y;
			r.x2 = (double)r.x2 / DIALOG_X;
			r.y2 = (double)r.y2 / DIALOG_Y;

			Template->style =	// DS_ABSALIGN |
								DS_SETFONT |
								DS_NOFAILCREATE |
								DS_3DLOOK |
								WS_SYSMENU |
								WS_CAPTION |
								DS_SETFOREGROUND |
								DS_MODALFRAME;

			Template->dwExtendedStyle = WS_EX_DLGMODALFRAME;
			Template->cdit = 0;
			Template->x = r.x1;
			Template->y = r.y1;
			Template->cx = r.X();
			Template->cy = r.Y();

			short *A = (short*) (Template+1);
			// menu
			*A++ = 0;
			// class
			*A++ = 0;
			// title
			A = DlgStrCopy(A, NameW());
			// font
			*A++ = SysFont->PointSize();
			A = DlgStrCopy(A, SysFont->Face());
			A = DlgPadToDWord(A);

			GView *p = (GetParent()) ? ((GetParent()->Handle()) ? GetParent() : GetParent()->FindReal()) : 0;
			HWND hWindow = (p)?p->Handle():0;

			Status = DialogBoxIndirectParam(LgiProcessInst(),
											Template,
											hWindow,
											(DLGPROC) DlgRedir, 
											(LPARAM) this);
		}

		DeleteObj(Mem);
	}

	return Status;
}

int GDialog::DoModeless()
{
	int Status = -1;
	Mem = NEW(GMem(16<<10));
	IsModal = false;
	if (Mem)
	{
		DLGTEMPLATE *Template = (DLGTEMPLATE*) Mem->Lock();
		if (Template)
		{
			GRect r = Pos;
			r.x1 = (double)r.x1 / DIALOG_X;
			r.y1 = (double)r.y1 / DIALOG_Y;
			r.x2 = (double)r.x2 / DIALOG_X;
			r.y2 = (double)r.y2 / DIALOG_Y;

			Template->style =	WS_VISIBLE |
								DS_ABSALIGN |
								WS_SYSMENU |
								WS_CAPTION |
								DS_SETFONT |
								DS_NOFAILCREATE |
								DS_3DLOOK;

			Template->style |= DS_MODALFRAME;
			Template->dwExtendedStyle = WS_EX_DLGMODALFRAME;
			Template->cdit = 0;
			Template->x = r.x1;
			Template->y = r.y1;
			Template->cx = r.X();
			Template->cy = r.Y();

			IsModal = false;

			short *A = (short*) (Template+1);
			// menu
			*A++ = 0;
			// class
			*A++ = 0;
			// title
			A = DlgStrCopy(A, NameW());
			// font
			*A++ = SysFont->PointSize(); // point size
			A = DlgStrCopy(A, SysFont->Face());
			A = DlgPadToDWord(A);

			GView *p = (GetParent()) ? ((GetParent()->Handle()) ? GetParent() : GetParent()->FindReal()) : 0;
			HWND hWindow = (p)?p->Handle():0;

			CreateDialogIndirectParam(	LgiProcessInst(),
										Template,
										hWindow,
										(DLGPROC) DlgRedir, 
										(LPARAM) this);
		}
	}

	return Status;
}

extern GButton *FindDefault(GView *w);

int GDialog::OnEvent(GMessage *Msg)
{
	switch (Msg->Msg)
	{
		/*
		case DM_GETDEFID:
		{
			GView *Def = FindDefault(this);
			if (Def)
			{
				return MAKELONG(Def->GetId(), DC_HASDEFID);
			}
			break;
		}
		case WM_NCACTIVATE:
		{
			GApp::ActiveDialog = (Msg->a) ? hWnd : 0;
			break;
		}
		*/
		case WM_INITDIALOG:
		{
			SetWindowLong(_View, GWL_STYLE, GetWindowLong(_View, GWL_STYLE) | WS_CLIPCHILDREN);

			GRect r = Pos;
			Pos.ZOff(-1, -1);
			SetPos(r);	// resets the dialog to the correct
						// size when large fonts are used

			AttachChildren();

			if (NOT _Default)
			{
				_Default = FindControl(IDOK);
			}

			OnCreate();

			return true;
			break;
		}
		case WM_CLOSE:
		{
			if (IsModal)
			{
				EndModal(0);
			}
			else
			{
				EndModeless(0);
			}
			return 0;
			break;
		}
	}

	return GView::OnEvent(Msg);
}

void GDialog::OnPosChange()
{
}

void GDialog::EndModal(int Code)
{
	EndDialog(Handle(), Code);
}

void GDialog::EndModeless(int Code)
{
	Quit(Code);
}

///////////////////////////////////////////////////////////////////////////////////////////
GControl::GControl(char *SubClassName) : GView(0)
{
	SubClass = 0;
	if (SubClassName)
	{
		SetClass(SubClassName);
		SubClass = GWin32Class::Create(SubClassName); // Owned by the GWin32Class object
	}
	Pos.ZOff(10, 10);
}

GControl::~GControl()
{
}

int GControl::OnEvent(GMessage *Msg)
{
	int Status = 0;

	// Pre-OS event handler
	switch (Msg->Msg)
	{
		case WM_CREATE:
		{
			SetId(GetId());
			OnCreate();
			break;
		}
		case WM_SETTEXT:
		{
			if (IsWindowUnicode(_View))
			{
				GObject::NameW((char16*)Msg->b);
			}
			else
			{
				GObject::Name((char*)Msg->b);
			}
			break;
		}
		case WM_GETDLGCODE:
		{
			return GView::OnEvent(Msg);
		}
		case WM_NOTIFY:
		case WM_COMMAND:
		case WM_KEYDOWN:
		case WM_KEYUP:
		case WM_SYSKEYDOWN:
		case WM_SYSKEYUP:
		{
			GView::OnEvent(Msg);
			break;
		}
	}
	
	// OS event handler
	if (SubClass)
	{
		Status = SubClass->CallParent(Handle(), Msg->Msg, Msg->a, Msg->b);
	}

	// Post-OS event handler
	switch (Msg->Msg)
	{
		case WM_CREATE:
		{
			SetFont(SysFont);
			break;
		}
	}

	return Status;
}

GdcPt2 GControl::SizeOfStr(char *Str)
{
	GdcPt2 Pt(0, 0);
	if (Str)
	{
		for (char *s=Str; s AND *s; )
		{
			char *e = strchr(s, '\n');
			if (NOT e) e = s + strlen(s);

			int x = 0, y = 0;
			SysFont->Size(&x, &y, s, (int)e-(int)s);

			Pt.y += y;
			Pt.x = max(Pt.x, x);

			s = (*e=='\n') ? e + 1 : 0;
		}
	}

	return Pt;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Edit
GEdit::GEdit(int id, int x, int y, int cx, int cy, char *name) :
	GControl(LGI_EDITBOX),
	ResObject(Res_EditBox)
{
	GdcPt2 Size = SizeOfStr((name)?name:"A");
	if (cx < 0) cx = Size.x + 6;
	if (cy < 0) cy = Size.y + 4;

	IgnoreNotify = true;
	SetId(id);
	Sunken(true);
	if (name) Name(name);
	SetPos(GRect(x, y, x+max(cx, 10), y+max(cy, 10)));
	SetStyle
	(
		WS_CHILD | WS_VISIBLE | ES_AUTOHSCROLL | WS_TABSTOP
	);
	if (SubClass)
	{
		SubClass->SubClass("EDIT");
	}
}

GEdit::~GEdit()
{
}

void GEdit::Select(int Start, int Len)
{
	if (_View)
	{
		SendMessage(_View, EM_SETSEL, Start, Start+Len-1);
	}
}

bool GEdit::MultiLine()
{
	return TestFlag(GetStyle(), ES_MULTILINE);
}

void GEdit::MultiLine(bool m)
{
	if (m)
	{
		SetStyle(GetStyle() | ES_MULTILINE | ES_AUTOVSCROLL | ES_WANTRETURN | WS_VSCROLL);
	}
	else
	{
		SetStyle(GetStyle() & ~(ES_MULTILINE | ES_AUTOVSCROLL | ES_WANTRETURN | WS_VSCROLL));
	}
}

bool GEdit::Password()
{
	return TestFlag(GetStyle(), ES_PASSWORD);
}

void GEdit::Password(bool m)
{
	if (m)
	{
		SetStyle(GetStyle() | ES_PASSWORD);
	}
	else
	{
		SetStyle(GetStyle() & ~ES_PASSWORD);
	}
}

char *GEdit::Name()
{
	return GControl::Name();
}

bool GEdit::Name(char *s)
{
	bool Old = IgnoreNotify;
	IgnoreNotify = true;
	bool Status = GControl::Name(s);
	IgnoreNotify = Old;
	return Status;
}

char16 *GEdit::NameW()
{
	return GControl::NameW();
}

bool GEdit::NameW(char16 *s)
{
	bool Old = IgnoreNotify;
	IgnoreNotify = true;
	bool Status = GControl::NameW(s);
	IgnoreNotify = Old;
	return Status;
}

int GEdit::SysOnNotify(int Code)
{
	if (NOT IgnoreNotify)
	{
		int Len = SendMessage(Handle(), WM_GETTEXTLENGTH, 0, 0);
		char *Str = NEW(char[Len+1]);
		if (Str)
		{
			SendMessage(Handle(), WM_GETTEXT, (WPARAM) (Len+1), (LPARAM) Str);
			GObject::Name(Str);
			DeleteArray(Str);
		}

		if (_View)
		{
			GView *n = GetNotify() ? GetNotify() : GetParent();
			if (n)
			{
				n->OnNotify(this, 0);
			}
		}
	}

	return 0;
}

void GEdit::Value(int i)
{
	char Str[32];
	sprintf(Str, "%i", i);
	Name(Str);
}

int GEdit::Value()
{
	char *n = Name();
	return (n) ? atoi(n) : 0;
}

int GEdit::OnEvent(GMessage *Msg)
{
	switch (Msg->Msg)
	{
		case WM_DESTROY:
		{
			Name();
			_View = 0;
			break;
		}
		case WM_SETTEXT:
		{
			if (IgnoreNotify)
			{
				// return true;
			}
			break;
		}
		case WM_GETDLGCODE:
		{
			return SubClass->CallParent(Handle(), Msg->Msg, Msg->a, Msg->b) | DLGC_WANTTAB;
		}
		case WM_KEYDOWN:
		{
			if (Msg->a == VK_TAB AND MultiLine())
			{
				// this causes the multiline Edit control to
				// insert tabs into the text and not move
				// to the next field.
				return 0;
			}
			break;
		}
		case WM_CHAR:
		{
			if ((Msg->a == VK_TAB OR Msg->a == VK_RETURN) AND
				NOT MultiLine())
			{
				// single line edit controls make "click"
				// when they get a TAB char... this
				// avoids that.
				return 0;
			}
			break;
		}
	}

	int Status = GControl::OnEvent(Msg);

	switch (MsgCode(Msg))
	{
		case WM_CREATE:
		{
			IgnoreNotify = false;
			break;
		}
		case WM_LBUTTONDBLCLK:
		case WM_LBUTTONDOWN:
		case WM_LBUTTONUP:
		case WM_RBUTTONDBLCLK:
		case WM_RBUTTONDOWN:
		case WM_RBUTTONUP:
		case WM_MBUTTONDBLCLK:
		case WM_MBUTTONDOWN:
		case WM_MBUTTONUP:
		case WM_SETFOCUS:
		case WM_KILLFOCUS:
		{
			GView::OnEvent(Msg);
			break;
		}
	}

	return Status;
}

void GEdit::OnKey(GKey &k)
{
	if (k.Down() AND
		k.c == VK_RETURN)
	{
		GView *n = (GetNotify()) ? GetNotify() : GetParent();
		if (n) n->OnNotify(this, k.c);
	}
}

///////////////////////////////////////////////////////////////////////////////////////////
// Radio group
/*
int GRadioGroup::NextId = 20000;

GRadioGroup::GRadioGroup(int id, int x, int y, int cx, int cy, char *name, int Init)
	: ResObject(Res_Group)
{
	Name(name);
	SetPos(GRect(x, y, x+cx, y+cy));
	SetId(id);
	Val = Init;
}

GRadioGroup::~GRadioGroup()
{
}

void GRadioGroup::OnCreate()
{
	AttachChildren();
	Value(Val);
}

int GRadioGroup::Value()
{
	int i=0;
	for (	GView *w = Children.First();
			w;
			w = Children.Next())
	{
		GRadioButton *But = dynamic_cast<GRadioButton*>(w);
		if (But)
		{
			if (But->Value())
			{
				Val = i;
				break;
			}

			i++;
		}
	}

	return Val;
}

void GRadioGroup::Value(int Which)
{
	Val = Which;

	int i=0;
	for (	GView *w = Children.First();
			w;
			w = Children.Next())
	{
		GRadioButton *But = dynamic_cast<GRadioButton*>(w);
		if (But)
		{
			if (But->Enabled())
			{
				But->Value(Which == i);
			}
			i++;
		}
	}
}

int GRadioGroup::OnEvent(GMessage *Msg)
{
	switch (MsgCode(Msg))
	{
		case WM_ERASEBKGND:
		{
			return 1;
		}
		default:
		{
			break;
		}
	}

	return GLayout::OnEvent(Msg);
}

void GRadioGroup::OnPaint(GSurface *pDC)
{
	pDC->Colour(LC_MED, 24);
	pDC->Rectangle();

	int y = SysFont->Y("A");
	GRect b(0, y/2, X()-1, Y()-1);
	LgiWideBorder(pDC, b, CHISEL);

	char *n = Name();
	if (n)
	{
		int x;
		SysFont->Size(&x, &y, n);
		GRect t;
		t.ZOff(x, y);
		t.Offset(6, 0);
		SysFont->Colour(LC_TEXT, LC_MED);
		SysFont->Transparent(false);
		SysFont->Text(pDC, t.x1, t.y1, n, -1, &t);
	}
}

int GRadioGroup::OnNotify(GView *Ctrl, int Flags)
{
	GView *n = GetNotify() ? GetNotify() : GetParent();
	if (n)
	{
		if (dynamic_cast<GRadioButton*>(Ctrl))
		{
			return n->OnNotify(this, Flags);
		}

		return n->OnNotify(Ctrl, Flags);
	}
	return 0;
}

GRadioButton *GRadioGroup::Append(int x, int y, char *name)
{
	GRadioButton *But = NEW(GRadioButton(NextId++, x, y, -1, -1, name, Children.GetItems() == 0));
	if (But)
	{
		Children.Insert(But);
	}
	return But;
}

///////////////////////////////////////////////////////////////////////////////////////////
// Radio button
GRadioButton::GRadioButton(int id, int x, int y, int cx, int cy, char *name, bool first) :
	GControl(LGI_BUTTON),
	ResObject(Res_RadioBox)
{
	GdcPt2 Size = SizeOfStr(name);
	if (cx < 0) cx = Size.x + 26;
	if (cy < 0) cy = Size.y + 6;

	First = first;
	InitVal = false;
	Name(name);
	SetPos(GRect(x, y, x+cx, y+cy));
	SetId(id);

	SetStyle(BS_AUTORADIOBUTTON | WS_TABSTOP | GetStyle());
	SetTabStop(true);
	if (SubClass)
	{
		SubClass->SubClass("BUTTON");
	}
}

GRadioButton::~GRadioButton()
{
}

int GRadioButton::Value()
{
	if (Handle())
	{
		return SendMessage(Handle(), BM_GETCHECK, 0, 0);
	}
	return 0;
}

void GRadioButton::Value(int i)
{
	InitVal = i;
	if (Handle())
	{
		SendMessage(Handle(), BM_SETCHECK, i, 0);
	}
}

int GRadioButton::OnEvent(GMessage *Msg)
{
	int Status = GControl::OnEvent(Msg);

	switch (MsgCode(Msg))
	{
		case WM_CREATE:
		{
			Value(InitVal);

			if (GetParent() AND
				this == GetParent()->Children.First())
			{
				int s = GetWindowLong(Handle(), GWL_STYLE) | WS_GROUP;
				SetWindowLong(Handle(), GWL_STYLE, s);
			}
			break;
		}
		case WM_GETDLGCODE:
		{
			return SubClass->CallParent(Handle(), Msg->Msg, Msg->a, Msg->b);
			break;
		}
	}

	return Status;
}
*/

//////////////////////////////////////////////////////////////////////////////////
// Slider control
GSlider::GSlider(int id, int x, int y, int cx, int cy, char *name, bool vert) :
	GControl(LGI_SLIDER),
	ResObject(Res_Slider)
{
	SetId(id);
	SetPos(GRect(x, y, x+cx, y+cy));
	Name(name);
	Vertical = vert;
	Min = Max = 0;
	Val = 0;

	SetStyle(GetStyle() | WS_VISIBLE | WS_CHILD | WS_TABSTOP | ((Vertical) ? TBS_VERT : TBS_HORZ));
	if (SubClass AND
		NOT SubClass->SubClass(TRACKBAR_CLASS))
	{
		LgiAssert(0);
	}
}

GSlider::~GSlider()
{
}

int GSlider::SysOnNotify(int Code)
{
	return 0;
}

void GSlider::Value(int i)
{
	Val = i;
	if (Handle())
	{
		SendMessage(Handle(), TBM_SETPOS, TRUE, Val);
	}
}

int GSlider::Value()
{
	if (Handle())
	{
		Val = SendMessage(Handle(), TBM_GETPOS, 0, 0);
	}
	return Val;
}

void GSlider::GetLimits(int &min, int &max)
{
	if (Handle())
	{
		min = Min = SendMessage(Handle(), TBM_GETRANGEMIN, 0, 0);
		max = Max = SendMessage(Handle(), TBM_GETRANGEMAX, 0, 0);
	}
	else
	{
		min = Min;
		max = Max;
	}
}

void GSlider::SetLimits(int min, int max)
{
	Min = min;
	Max = max;
	if (Handle())
	{
		SendMessage(Handle(), TBM_SETRANGE, TRUE, (LPARAM) MAKELONG(min, max)); 
	}
}

int GSlider::OnEvent(GMessage *Msg)
{
	int Status = GControl::OnEvent(Msg);

	switch (MsgCode(Msg))
	{
		case WM_HSCROLL:
		case WM_VSCROLL:
		{
			GView *n = (GetNotify()) ? GetNotify() : GetParent();
			if (n)
			{
				n->OnNotify(this, 0);
			}
			break;
		}
		case WM_CREATE:
		{
			SetLimits(Min, Max);
			Value(Val);
			break;
		}
	}

	return Status;
}

//////////////////////////////////////////////////////////////////////////////////
#ifdef _MT
class GBitmapThread : public GThread
{
	GBitmap *Bmp;
	char *File;
	GThread **Owner;

public:
	GBitmapThread(GBitmap *bmp, char *file, GThread **owner)
	{
		Bmp = bmp;
		File = NewStr(file);
		Owner = owner;
		if (Owner)
		{
			*Owner = this;
		}
		Run();
	}

	~GBitmapThread()
	{
		if (Owner)
		{
			*Owner = 0;
		}
		DeleteArray(File);
	}

	int Main()
	{
		if (Bmp)
		{
			GSurface *pDC = LoadDC(File);
			if (pDC)
			{
				Bmp->SetDC(pDC);
				GRect r = Bmp->GetPos();
				r.Dimension(pDC->X()+4, pDC->Y()+4);

				while (NOT Bmp->Handle())
				{
					LgiSleep(10);
				}

				Bmp->SetPos(r, true);
				Bmp->PostEvent(M_CHANGE, (int)Bmp, 0);
			}
		}

		return 0;
	}
};
#endif

GBitmap::GBitmap(int id, int x, int y, char *FileName, bool Async)
	: ResObject(Res_Bitmap)
{
	pDC = 0;
	pThread = 0;

	SetClass(LGI_BITMAP);
	SetStyle(WS_CHILD | WS_VISIBLE);

	SetId(id);
	GRect r(0, 0, 4, 4);
	r.Offset(x, y);

	#ifdef _MT
	if (NOT Async)
	{
	#endif
		pDC = LoadDC(FileName);
		if (pDC)
		{
			r.Dimension(pDC->X()+4, pDC->Y()+4);
		}
	#ifdef _MT
	}
	else
	{
		NEW(GBitmapThread(this, FileName, &pThread));
	}
	#endif

	SetPos(r);
	CreateClass(LGI_BITMAP);
}

GBitmap::~GBitmap()
{
	#ifdef _MT
	if (pThread)
	{
		pThread->Terminate();
		DeleteObj(pThread);
	}
	#endif
	DeleteObj(pDC);
}

void GBitmap::SetDC(GSurface *pNewDC)
{
	DeleteObj(pDC);
	pDC = pNewDC;
	if (pDC)
	{
		GRect r = GetPos();
		r.Dimension(pDC->X() + 4, pDC->Y() + 4);
		SetPos(r);
		Invalidate();
	}
}

GSurface *GBitmap::GetDC()
{
	return pDC;
}

int GBitmap::OnEvent(GMessage *Msg)
{
	return GView::OnEvent(Msg);
}

void GBitmap::OnPaint(GSurface *pScreen)
{
	GRect a(0, 0, X()-1, Y()-1);
	LgiWideBorder(pScreen, a, SUNKEN);
	if (pDC)
	{
		pScreen->Blt(2, 2, pDC);
	}
	else
	{
		pScreen->Colour(CBit(pScreen->GetBits(), GetSysColor(COLOR_WINDOW)));
		pScreen->Rectangle(&a);
	}
}

void GBitmap::OnMouseClick(GMouse &m)
{
	if (NOT m.Down() AND GetParent())
	{
		((GDialog*) GetParent())->OnNotify(this, 0);
	}
}

///////////////////////////////////////////////////////////////////////////////////////////
GItemContainer::GItemContainer()
{
	Flags = 0;
	ImageList = 0;
}

GItemContainer::~GItemContainer()
{
	if (OwnList())
	{
		DeleteObj(ImageList);
	}
	else
	{
		ImageList = 0;
	}
}

bool GItemContainer::SetImageList(GImageList *list, bool Own)
{
	ImageList = list;
	OwnList(Own);
	AskImage(true);
	return ImageList != NULL;
}

bool GItemContainer::LoadImageList(char *File, int x, int y)
{
	GSurface *pDC = LoadDC(File);
	if (pDC)
	{
		ImageList = NEW(GImageList(x, y, pDC));
		if (ImageList)
		{
			#ifdef WIN32
			ImageList->Create(pDC->X(), pDC->Y(), pDC->GetBits());
			#endif
		}
	}
	return pDC != 0;
}

///////////////////////////////////////////////////////////////////////////////////////////
GProgress::GProgress(int id, int x, int y, int cx, int cy, char *name) :
	GControl(LGI_PROGRESS),
	ResObject(Res_Progress)
{
	SetClass(LGI_PROGRESS);
	SetId(id);
	SetPos(GRect(x, y, x+cx, y+cy));
	if (name) GControl::Name(name);

	SetStyle(GetStyle() | WS_CHILD | WS_TABSTOP | WS_VISIBLE | PBS_SMOOTH);
	if (SubClass)
	{
		SubClass->SubClass(PROGRESS_CLASS);
	}
}

GProgress::~GProgress()
{
}

void GProgress::SetLimits(int l, int h)
{
	Low = l;
	High = h;
	if (Handle())
	{
		SendMessage(Handle(), PBM_SETRANGE32, Low, High);
	}
}

void GProgress::Value(int v)
{
	Val = v;
	if (Handle())
	{
		SendMessage(Handle(), PBM_SETPOS, (WPARAM) Val, 0);
	}
}

int GProgress::OnEvent(GMessage *Msg)
{
	return GControl::OnEvent(Msg);
}

///////////////////////////////////////////////////////////////////////
