#include "Lgi.h"
#include "GSlider.h"
#include <COMMCTRL.H>

//////////////////////////////////////////////////////////////////////////////////
// Slider control
GSlider::GSlider(int id, int x, int y, int cx, int cy, char *name, bool vert) :
	GControl(LGI_SLIDER),
	ResObject(Res_Slider)
{
	SetId(id);
	SetPos(GRect(x, y, x+cx, y+cy));
	Name(name);
	Vertical = vert;
	Min = Max = 0;
	Val = 0;

	SetStyle(GetStyle() | WS_VISIBLE | WS_CHILD | WS_TABSTOP | ((Vertical) ? TBS_VERT : TBS_HORZ));
	if (SubClass AND
		NOT SubClass->SubClass(TRACKBAR_CLASS))
	{
		LgiAssert(0);
	}
}

GSlider::~GSlider()
{
}

int GSlider::SysOnNotify(int Code)
{
	return 0;
}

void GSlider::Value(int i)
{
	Val = i;
	if (Handle())
	{
		SendMessage(Handle(), TBM_SETPOS, TRUE, Val);
	}
}

int GSlider::Value()
{
	if (Handle())
	{
		Val = SendMessage(Handle(), TBM_GETPOS, 0, 0);
	}
	return Val;
}

void GSlider::GetLimits(int &min, int &max)
{
	if (Handle())
	{
		min = Min = SendMessage(Handle(), TBM_GETRANGEMIN, 0, 0);
		max = Max = SendMessage(Handle(), TBM_GETRANGEMAX, 0, 0);
	}
	else
	{
		min = Min;
		max = Max;
	}
}

void GSlider::SetLimits(int min, int max)
{
	Min = min;
	Max = max;
	if (Handle())
	{
		SendMessage(Handle(), TBM_SETRANGE, TRUE, (LPARAM) MAKELONG(min, max)); 
	}
}

int GSlider::OnEvent(GMessage *Msg)
{
	int Status = GControl::OnEvent(Msg);

	switch (MsgCode(Msg))
	{
		case WM_HSCROLL:
		case WM_VSCROLL:
		{
			GView *n = (GetNotify()) ? GetNotify() : GetParent();
			if (n)
			{
				n->OnNotify(this, 0);
			}
			break;
		}
		case WM_CREATE:
		{
			SetLimits(Min, Max);
			Value(Val);
			break;
		}
	}

	return Status;
}

